﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;

using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;
using UIPopup = Windows.UI.Xaml.Controls.Primitives.Popup;

using pdftron.PDF.Annots;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using Windows.UI.Input;
using pdftron.PDF.Tools.Controls;
using pdftron.SDF;
using pdftron.PDF.OCG;
using pdftron.PDF.Tools.Utilities;

namespace pdftron.PDF.Tools
{
    public class AnnotEditAdvancedShape : AnnotEdit
    {
        // Measurement info and feedback
        Border mBorder;
        TextBlock mTextFeedback;
        List<pdftron.PDF.Point> vertexPointsFeedback = new List<PDFPoint>();
        MeasureInfoGroup mMeasureGroup;
        bool mIsAreaAnnot = false;
        bool mIsPerimeterAnnot = false;
        int textPosXOffset = 35;

        AnnotType annotType;
        IPolyLine mPoly;
        int mVertexPoints;

        // Temp drawing objs
        Windows.UI.Xaml.Shapes.Polygon mPolygonGuide;
        Windows.UI.Xaml.Shapes.Polyline mPolylineGuide;
        List<pdftron.PDF.Point> vertexPoints = new List<pdftron.PDF.Point>();

        public AnnotEditAdvancedShape(PDFViewCtrl ctrl, ToolManager tMan)
        : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_advanced_shape_edit;
            mToolMode = ToolType.e_advanced_shape_edit;

            mSelectedColor = new Color();

            mPt1 = new Point();
            mPt2 = new Point();

            // appearance of the selection widget
            mSelectionBackgroundColor = new SolidColorBrush(Color.FromArgb(128, 128, 128, 125));
            mCtrlPointFillBrush = new SolidColorBrush(Color.FromArgb(255, 150, 180, 255));
            mCtrlPointBorderBrush = new SolidColorBrush(Color.FromArgb(255, 200, 200, 200));
            mCtrlPointBorderThickness = 2;
            mCtrlPointRadius = 7.5;

            mMinAnnotationSize = mCtrlPointRadius * 2; // Min annotation size is 2 x Control point size
        }

        #region Events

        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (!mIsManipulated || mToolManager.ReadOnlyMode || mEffectiveCtrlPoint == -1)
                return false;

            mPt2.x = e.GetCurrentPoint(mViewerCanvas).Position.X;
            mPt2.y = e.GetCurrentPoint(mViewerCanvas).Position.Y;
            CheckBounds(mPt2);

            if (!mStartMoving) // ensure we actually move a little before we start with updates and stuff
            {
                if ((Math.Abs(mPt2.x - mPt1.x) > START_MOVING_THRESHOLD)
                    || (Math.Abs(mPt2.y - mPt1.y) > START_MOVING_THRESHOLD))
                {
                    mStartMoving = true;
                }
                else
                {
                    return false;
                }
            }

            if (mEffectiveCtrlPoint == MOVE_WIDGET_CODE)
            {
                CheckBoundsOnPointerMoved();
                return false;
            }

            if (!mStartMoving) // ensure we actually move a little before we start with updates and stuff
            {
                if ((Math.Abs(mPt2.x - mPt1.x) > START_MOVING_THRESHOLD)
                    || (Math.Abs(mPt2.y - mPt1.y) > START_MOVING_THRESHOLD))
                {
                    mStartMoving = true;
                }
                else
                {
                    return false;
                }
            }

            if (mStartMoving)
            {
                // Update selected control point position
                if (annotType == AnnotType.e_Polygon)
                    UpdateShapePoints(mPolygonGuide.Points);
                else
                    UpdateShapePoints(mPolylineGuide.Points);

                UpdateSelectionAppearance();

                if (mIsAreaAnnot)
                {
                    if (mBorder.Visibility == Visibility.Collapsed)
                        mBorder.Visibility = Visibility.Visible;

                    var area = MeasureUtils.CalcArea(vertexPointsFeedback);
                    var val = area * mMeasureGroup.AreaInfo.BaseFactor;
                    mBorder.RenderTransform = new TranslateTransform { X = mPt2.x - mPageCropOnClient.x1 + textPosXOffset, Y = mPt2.y - mPageCropOnClient.y1 }; // Position feedback right beside dragged ctrl pt
                    mTextFeedback.Text = MeasureUtils.FormatDecimals(val, MeasureUtils.ConvPDFPrecisionToStr(mMeasureGroup.AreaInfo.Precision)) + " " + mMeasureGroup.AreaInfo.Unit;
                }
                else if (mIsPerimeterAnnot)
                {
                    if (mBorder.Visibility == Visibility.Collapsed)
                        mBorder.Visibility = Visibility.Visible;

                    var perimeter = MeasureUtils.CalcPerimeter(vertexPointsFeedback);
                    var val = perimeter * mMeasureGroup.AxisInfo.BaseFactor;
                    mBorder.RenderTransform = new TranslateTransform { X = mPt2.x - mPageCropOnClient.x1 + textPosXOffset, Y = mPt2.y - mPageCropOnClient.y1 }; // Position feedback right beside dragged ctrl pt
                    mTextFeedback.Text = MeasureUtils.FormatDecimals(val, MeasureUtils.ConvPDFPrecisionToStr(mMeasureGroup.DistanceInfo.Precision)) + " " + mMeasureGroup.DistanceInfo.Unit;
                }
            }

            return false;
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (mPointerID < 0 || e.Pointer.PointerId != mPointerID)
                return false;

            if (mEffectiveCtrlPoint == MOVE_WIDGET_CODE)
            {
                base.PointerReleasedHandler(sender, e);
                CreateSelectionAppearance();                
            }
            else
            {
                mPointerID = -1;
                mViewerCanvas.ReleasePointerCaptures();

                if (mStartMoving)
                {
                    try
                    {
                        mPDFView.DocLock(true);

                        CheckAndUpdateVerticesBounds();

                        // points in Screen space
                        PDFDouble x1 = new PDFDouble(mLeft);
                        PDFDouble y1 = new PDFDouble(mTop);
                        PDFDouble x2 = new PDFDouble(mRight);
                        PDFDouble y2 = new PDFDouble(mBottom);

                        // Get coordinates of two opposite points in screen space
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);
                        mPDFView.ConvScreenPtToPagePt(x1, y1, mAnnotPageNum);
                        mPDFView.ConvScreenPtToPagePt(x2, y2, mAnnotPageNum);
                        var newAnnotRect = new Rect(x1.Value, y1.Value, x2.Value, y2.Value);
                        newAnnotRect.Normalize();

                        var bb = mAnnotBBox;

                        mAnnot.SetRect(newAnnotRect);

                        // points in Screen space
                        PDFDouble x = new PDFDouble(mPt2.x);
                        PDFDouble y = new PDFDouble(mPt2.y);
                        // Get coordinates of two opposite points in screen space
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(x, y);
                        mPDFView.ConvScreenPtToPagePt(x, y, mAnnotPageNum);

                        PDFPoint newVertexPt = new PDFPoint(x.Value, y.Value);
                        mPoly.SetVertex(mEffectiveCtrlPoint, newVertexPt);

                        // Update Measure annotation if needed
                        if (mIsAreaAnnot)
                        {
                            Annots.Polygon polygon = new Annots.Polygon(mPoly.GetSDFObj());
                            MeasureUtils.UpdateAreaAnnot(polygon);
                        }
                        else if (mIsPerimeterAnnot)
                        {
                            Annots.PolyLine polyline = new Annots.PolyLine(mPoly.GetSDFObj());
                            MeasureUtils.UpdatePerimeterAnnot(polyline);
                        }

                        mAnnot.RefreshAppearance();
                        mAnnot.SetDateToNow();
                        mPDFView.UpdateWithAnnot(mAnnot, mAnnotPageNum);

                        mToolManager.RaiseAnnotationEditedEvent(mAnnot, mAnnotPageNum);
                    }

                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                            Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                            Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                    finally
                    {
                        mPDFView.DocUnlock();
                    }
                }                
            }

            mEffectiveCtrlPoint = -1;
            mStartMoving = false;
            mIsManipulated = false;
            EnableScrolling();
            return false;
        }

        #endregion

        #region Utility
        /// <summary>
        /// Update effective control point current manipulated
        /// </summary>
        /// <param name="points">Points</param>
        private void UpdateShapePoints(PointCollection points)
        {
            points.Clear();
            for (int i = 0; i < mVertexPoints; i++)
            {
                var pt = vertexPoints[i];
                if (i == mEffectiveCtrlPoint)
                {
                    vertexPoints[i] = new PDFPoint(mPt2.x, mPt2.y);
                    points.Add(new UIPoint(mPt2.x - mPageCropOnClient.x1, mPt2.y - mPageCropOnClient.y1));

                    // Update changed pt so it can give feedback
                    PDFDouble x = new PDFDouble(mPt2.x);
                    PDFDouble y = new PDFDouble(mPt2.y);
                    // Get coordinates of two opposite points in screen space
                    mPDFView.ConvAnnotationCanvasPtToScreenPt(x, y);
                    mPDFView.ConvScreenPtToPagePt(x, y, mAnnotPageNum);
                    vertexPointsFeedback[i] = new PDFPoint(x.Value, y.Value);
                }
                else
                {
                    points.Add(new UIPoint(pt.x - mPageCropOnClient.x1, pt.y - mPageCropOnClient.y1));
                }
            }
        }

        protected override void InitializeCanvasAndGeometry()
        {
            mViewerCanvas = mPDFView.GetAnnotationCanvas();

            if (mViewerCanvas != null && !mViewerCanvas.Children.Contains(this))
                mViewerCanvas.Children.Add(this);

            // Figure out selection appearance according to type of annot
            mAnnotBBoxRectangle = new Rectangle();
            mAnnotBBoxRectangle.Stroke = mCtrlPointBorderBrush;
            mAnnotBBoxRectangle.StrokeDashArray = new DoubleCollection() { 2, 2 };
            mAnnotBBoxRectangle.StrokeThickness = 2;

            // The size of the gray rectangle is accomplished by scaling it
            mAnnotBBoxRectangle.Width = 1;
            mAnnotBBoxRectangle.Height = 1;

            this.Children.Add(mAnnotBBoxRectangle);

            // Ensure the annotation is Polyline
            annotType = mAnnot.GetAnnotType();
            if ((annotType != AnnotType.e_Polygon)
                && annotType != AnnotType.e_Polyline)
                return;

            // Check if Measure and Setup Feeback info
            if (MeasureUtils.IsAreaMeasure(mAnnot))
                mIsAreaAnnot = true;
            else if (MeasureUtils.IsPerimeterMeasure(mAnnot))
                mIsPerimeterAnnot = true;

            if (mIsAreaAnnot || mIsPerimeterAnnot)
            {
                mBorder = new Border();
                mBorder.Visibility = Visibility.Collapsed;
                mBorder.BorderBrush = new SolidColorBrush(Colors.Black);
                mBorder.BorderThickness = new Thickness(1);
                mBorder.Background = new SolidColorBrush(Color.FromArgb(255, 28, 31, 34));
                mTextFeedback = new TextBlock();
                mTextFeedback.FontSize = 12;
                mTextFeedback.Foreground = new SolidColorBrush(Color.FromArgb(255, 255, 255, 255));
                mTextFeedback.FontWeight = Windows.UI.Text.FontWeights.Bold;
                mBorder.Child = mTextFeedback;
                this.Children.Add(mBorder);
                mMeasureGroup = MeasureUtils.GetAnnotMeasureInfo(mAnnot);
            }

            if (annotType == AnnotType.e_Polygon)
            {
                // Initi the temp Polygon
                mPolygonGuide = new Windows.UI.Xaml.Shapes.Polygon();
                mPolygonGuide.StrokeThickness = 2;
                mPolygonGuide.Stroke = mCtrlPointFillBrush;
                mPolygonGuide.StrokeDashArray = new DoubleCollection() { 2.0, 2.0 };
                this.Children.Add(mPolygonGuide);
            }
            else if (annotType == AnnotType.e_Polyline)
            {
                // Initi the temp Polyline
                mPolylineGuide = new Windows.UI.Xaml.Shapes.Polyline();
                mPolylineGuide.StrokeThickness = 2;
                mPolylineGuide.Stroke = new SolidColorBrush(Windows.UI.Colors.Blue);
                mPolylineGuide.Stroke = mCtrlPointFillBrush;
                mPolylineGuide.StrokeDashArray = new DoubleCollection() { 2.0, 2.0 };
                this.Children.Add(mPolylineGuide);
            }

            // Get the vertex count so we can create the ctrl points
            mPoly = mAnnot as IPolyLine;
            mNumberCtrlPts = mVertexPoints = mPoly.GetVertexCount();

            mCtrlPointTransforms = new TranslateTransform[mVertexPoints];
            mCtrlPointLocations = new UIPoint[mVertexPoints];
            mControlPointShapes = new Path[mVertexPoints];

            for (int i = 0; i < mVertexPoints; i++)
            {
                mCtrlPointTransforms[i] = new TranslateTransform();
                mCtrlPointLocations[i] = new UIPoint();
                mControlPointShapes[i] = new Path();

                if (mMaintainAspectRatio && i % 2 != 0)
                {
                    mControlPointShapes[i].Visibility = Visibility.Collapsed;
                    mCtrlPointLocations[i].X = -100;
                    continue;
                }

                EllipseGeometry geom = new EllipseGeometry();
                //geom.Center = mCtrlPointLocations[i];
                geom.RadiusX = mCtrlPointRadius;
                geom.RadiusY = mCtrlPointRadius;
                mControlPointShapes[i].Data = geom;
                mControlPointShapes[i].Stroke = mCtrlPointBorderBrush;
                mControlPointShapes[i].StrokeThickness = mCtrlPointBorderThickness;
                mControlPointShapes[i].Fill = mCtrlPointFillBrush;
                mControlPointShapes[i].RenderTransform = mCtrlPointTransforms[i];
                this.Children.Add(mControlPointShapes[i]);
            }
        }

        protected override void CreateSelectionAppearance()
        {
            // update rectangle
            mAnnotBBoxRectangle.SetValue(Canvas.LeftProperty, mLeft - mPageCropOnClient.x1);
            mAnnotBBoxRectangle.SetValue(Canvas.TopProperty, mTop - mPageCropOnClient.y1);

            mAnnotBBoxRectangle.Width = mRight - mLeft;
            mAnnotBBoxRectangle.Height = mBottom - mTop;

            if (annotType == AnnotType.e_Polygon)
                mPolygonGuide.Points.Clear();
            else
                mPolylineGuide.Points.Clear();

            vertexPoints.Clear();
            vertexPointsFeedback.Clear();
            for (int i = 0; i < mVertexPoints; i++)
            {
                var pt = mPoly.GetVertex(i);

                // Save the values of the vertex so we can provide feedback when resizing it
                vertexPointsFeedback.Add(pt);

                PDFDouble pdf_pt_x = new PDFDouble(pt.x);
                PDFDouble pdf_pt_y = new PDFDouble(pt.y);
                mPDFView.ConvPagePtToScreenPt(pdf_pt_x, pdf_pt_y, AnnotationPageNumber);
                mPDFView.ConvScreenPtToAnnotationCanvasPt(pdf_pt_x, pdf_pt_y);

                mCtrlPointLocations[i].X = pdf_pt_x.Value;
                mCtrlPointLocations[i].Y = pdf_pt_y.Value;

                vertexPoints.Add(new PDFPoint(pdf_pt_x.Value, pdf_pt_y.Value));

                if (annotType == AnnotType.e_Polygon)
                {                    
                    mPolygonGuide.Points.Add(new UIPoint(pdf_pt_x.Value - mPageCropOnClient.x1, pdf_pt_y.Value - mPageCropOnClient.y1));
                }
                else
                {
                    mPolylineGuide.Points.Add(new UIPoint(pdf_pt_x.Value - mPageCropOnClient.x1, pdf_pt_y.Value - mPageCropOnClient.y1));
                }
                // skip middle ctrl points when maintaining aspect ratio
                if (mMaintainAspectRatio && i % 2 != 0)
                    continue;

                mCtrlPointTransforms[i].X = mCtrlPointLocations[i].X - mPageCropOnClient.x1;
                mCtrlPointTransforms[i].Y = mCtrlPointLocations[i].Y - mPageCropOnClient.y1;
            }
        }

        private void UpdateSelectionAppearance()
        {
            for (int i = 0; i < mVertexPoints; i++)
            {
                mCtrlPointLocations[i].X = vertexPoints[i].x;
                mCtrlPointLocations[i].Y = vertexPoints[i].y;

                mCtrlPointTransforms[i].X = mCtrlPointLocations[i].X - mPageCropOnClient.x1;
                mCtrlPointTransforms[i].Y = mCtrlPointLocations[i].Y - mPageCropOnClient.y1;
            }
        }

        private void CheckAndUpdateVerticesBounds()
        {
            // check bounds
            if (mEffectiveCtrlPoint == -1)
            {
                mLeft = mOldLeft + mPt2.x - mPt1.x;
                mLeft = Math.Max(mLeft, mPageCropOnClient.x1);
                mLeft = Math.Min(mLeft, mPageCropOnClient.x2 + mOldLeft - mOldRight);
                mRight = mLeft + mOldRight - mOldLeft;
                mTop = mOldTop + mPt2.y - mPt1.y;
                mTop = Math.Max(mTop, mPageCropOnClient.y1);
                mTop = Math.Min(mTop, mPageCropOnClient.y2 + mOldTop - mOldBottom);
                mBottom = mTop + mOldBottom - mOldTop;
            }
            else
            {
                double x1 = vertexPoints[0].x;
                double x2 = vertexPoints[0].x;
                double y1 = vertexPoints[0].y;
                double y2 = vertexPoints[0].y;
                for (int i = 0; i < mNumberCtrlPts; i++)
                {
                    if (vertexPoints[i].x < x1)
                        x1 = vertexPoints[i].x;

                    if (vertexPoints[i].x > x2)
                        x2 = vertexPoints[i].x;

                    if (vertexPoints[i].y < y1)
                        y1 = vertexPoints[i].y;

                    if (vertexPoints[i].y > y2)
                        y2 = vertexPoints[i].y;
                }

                mTop = y1;
                mBottom = y2;
                mLeft = x1;
                mRight = x2;
            }
        }

        /// <summary>
        /// Makes sure the point is within the page boundaries.
        /// </summary>
        /// <param name="point">The point we want to check</param>
        protected void CheckBounds(PDFPoint point)
        {
            if (point.x < mPageCropOnClient.x1)
            {
                point.x = mPageCropOnClient.x1;
            }
            if (point.y < mPageCropOnClient.y1)
            {
                point.y = mPageCropOnClient.y1;
            }
            if (point.x > mPageCropOnClient.x2)
            {
                point.x = mPageCropOnClient.x2;
            }
            if (point.y > mPageCropOnClient.y2)
            {
                point.y = mPageCropOnClient.y2;
            }
        }
        #endregion
    }
}