﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;
using UIPopup = Windows.UI.Xaml.Controls.Primitives.Popup;

using pdftron.PDF;
using pdftron.Common;
using pdftron.PDF.Annots;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using Windows.UI.Input;
using pdftron.PDF.Tools.Controls;
using pdftron.SDF;
using pdftron.PDF.OCG;
using pdftron.PDF.Tools.Utilities;

namespace pdftron.PDF.Tools
{
    public class AnnotEditGroupCreate : AnnotEditRectGroup
    {
        private bool mShapeHasBeenCreated = false;
        Path mShape;

        protected int START_DRAWING_THRESHOLD = 5;

        protected bool mStartDrawing = false;
        protected bool mPointerDown = false;
        protected Point m_cnPt1 = null;
        protected Point m_cnPt2 = null;
        protected bool mMirrorX = false;
        protected bool mMirrorY = false;
        protected CompositeTransform mRenderTransform;
        protected int mDownPageNumber = -1;

        protected bool mSimpleShapeCanceledAtPress = false; // lets the derived class know that the pointer press should be disregarded.
        protected bool mCancelledDrawing = false;

        // Visuals for shapes
        protected double mStrokeThickness = 1;
        protected double mZoomLevel = 1;
        protected double mDrawThickness;
        protected SolidColorBrush mStrokeBrush = new SolidColorBrush(Color.FromArgb(100, 255, 0, 0));
        protected bool mUseFill = true;
        protected bool mUseStroke = true;
        protected SolidColorBrush mFillBrush = new SolidColorBrush(Color.FromArgb(100, 255, 0, 0));
        protected bool mSimpleShapeCreateSubclassHandlesProperties = false;

        private bool mIsUngroupAllowed = true;

        public new ToolType GetToolMode()
        {
            return mToolMode;
        }

        public AnnotEditGroupCreate(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_annot_edit_group_create;
            mToolMode = ToolType.e_annot_edit_group_create;
            mSelectedColor = new Color();

            mPt1 = new Point();
            mPt2 = new Point();

            // appearance of the selection widget
            mSelectionBackgroundColor = new SolidColorBrush(Color.FromArgb(128, 128, 128, 125));
            mCtrlPointFillBrush = new SolidColorBrush(Color.FromArgb(255, 150, 180, 255));
            mCtrlPointBorderBrush = new SolidColorBrush(Color.FromArgb(255, 200, 200, 200));
            mCtrlPointBorderThickness = 2;
            mCtrlPointRadius = 7.5;

            // register for the keyboard callback
            mKeyboardHelper = new KeyboardHelper();
            mKeyboardHelper.SubscribeToKeyboard(true);
        }

        internal override void OnCreate()
        {
            //Do nothing
        }

        /// <summary>
        /// Adds in the menu options according to the type of annotation
        /// </summary>
        protected override void PopulateMenu()
        {
            if (mAnnot != null)
            {
                mMenuTitles = null;
                try
                {
                    mMenuTitles = new Dictionary<string, string>();
                    //locks the document first as accessing annotation/doc information isn't thread safe.
                    mPDFView.DocLockRead();
                    mMenuTitles.Add("done", ResourceHandler.GetString("EditOption_Done"));
                    mMenuTitles.Add("group", ResourceHandler.GetString("EditOption_Group"));

                    if (mIsUngroupAllowed)
                    {
                        mMenuTitles.Add("ungroup", ResourceHandler.GetString("EditOption_Ungroup"));
                    }

                    mMenuTitles.Add("flatten", ResourceHandler.GetString("EditOption_Flatten"));
                    mMenuTitles.Add("opacity", ResourceHandler.GetString("EditOption_Opacity"));
                    mMenuTitles.Add("fill color", ResourceHandler.GetString("EditOption_FillColor"));
                    mMenuTitles.Add("color", ResourceHandler.GetString("EditOption_Color"));
                    mMenuTitles.Add("delete", ResourceHandler.GetString("EditOption_Delete"));

                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlockRead();
                }

                if (mMenuTitles != null && !mAnnotIsSticky) // if sticky, we don't create menu until it's needed
                {
                    mContextMenu = new ContextMenuControl(mPDFView, mMenuTitles);
                    mContextMenu.ItemClicked += OnCommandMenuClicked;
                }
            }
        }

        internal override bool PointerPressedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (!mCreated)
            {
                AddPointer(e.Pointer);
                DisableScrolling();
                mToolManager.ToolModeAfterEditing = ToolType.e_none;
                mCancelledDrawing = false;

                mSimpleShapeCanceledAtPress = false;

                if (mToolManager.ContactPoints.Count > 1)
                {
                    mToolManager.EnableOneFingerScroll = false;
                    mSimpleShapeCanceledAtPress = true;
                    mCancelledDrawing = true;
                    mStartDrawing = false;
                    mPointerID = -1;
                    if (mViewerCanvas != null && mViewerCanvas.Children.Contains(this))
                    {
                        this.Children.Clear();
                        mViewerCanvas.Children.Remove(this);
                    }
                    mToolManager.EnableScrollByManipulation = true;
                    return false; // ignore additional pointer presses
                }

                if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse && !e.GetCurrentPoint(null).Properties.IsLeftButtonPressed)
                {
                    mSimpleShapeCanceledAtPress = true;
                    return false; // if mouse, only allow left button
                }
                mPointerID = (int)e.Pointer.PointerId;
                UIPoint screenPoint = e.GetCurrentPoint(mPDFView).Position;
                PDFDoc doc = mPDFView.GetDoc();
                if (doc == null)
                {
                    mSimpleShapeCanceledAtPress = true;
                    return false;
                }

                mViewerCanvas = mPDFView.GetAnnotationCanvas();

                // Ensure we're on a page
                mToolManager.EnableOneFingerScroll = false;

                mDownPageNumber = mPDFView.GetPageNumberFromScreenPoint(screenPoint.X, screenPoint.Y);
                if (mDownPageNumber < 1)
                {
                    mDownPageNumber = mPDFView.GetCurrentPage();
                }

                if (mDownPageNumber < 1)
                {
                    mPointerID = -1;
                    mToolManager.EnableOneFingerScroll = true;
                    mToolManager.EnableScrollByManipulation = true;
                    mDownPageNumber = -1;
                    mSimpleShapeCanceledAtPress = true;
                    return false;
                }
                // Get this from mPDFViewCtrl
                mPage = doc.GetPage(mDownPageNumber);

                // Get the page's bounding box in canvas space
                mPageCropOnClient = BuildPageBoundBoxOnClient(mDownPageNumber);

                mPointerDown = true;

                //Uncomment the code below to allow users to select color
                //SettingsColor color = Settings.GetShapeColor(this.ToolMode, false);
                //byte r = color.R;
                //byte g = color.G;
                //byte b = color.B;
                //mUseStroke = true;
                //mStrokeBrush = new SolidColorBrush(Color.FromArgb(100, r, g, b));

                //color = Settings.GetShapeColor(this.ToolMode, true);
                //r = color.R;
                //g = color.G;
                //b = color.B;
                //mUseFill = color.Use;
                //mFillBrush = new SolidColorBrush(Color.FromArgb(100, r, g, b));

                mUseStroke = true;
                mStrokeThickness = Settings.GetThickness(this.ToolMode);
                mOpacity = Settings.GetOpacity(this.ToolMode);

                mZoomLevel = mPDFView.GetZoom();
                mDrawThickness = mStrokeThickness * mZoomLevel;

                UIPoint pp = e.GetCurrentPoint(mViewerCanvas).Position;
                m_cnPt1 = new Point(pp.X, pp.Y);
                if (mPageCropOnClient != null)
                {
                    if (m_cnPt1.x < mPageCropOnClient.x1)
                    {
                        m_cnPt1.x = mPageCropOnClient.x1;
                    }
                    if (m_cnPt1.x > mPageCropOnClient.x2)
                    {
                        m_cnPt1.x = mPageCropOnClient.x2;
                    }
                    if (m_cnPt1.y < mPageCropOnClient.y1)
                    {
                        m_cnPt1.y = mPageCropOnClient.y1;
                    }
                    if (m_cnPt1.y > mPageCropOnClient.y2)
                    {
                        m_cnPt1.y = mPageCropOnClient.y2;
                    }
                }
                m_cnPt2 = new Point(m_cnPt1.x, m_cnPt1.y);
                mViewerCanvas.CapturePointer(e.Pointer);
                if (mSimpleShapeCanceledAtPress)
                {
                    return false;
                }
                mShapeHasBeenCreated = false;
            }
            else
            {
                base.PointerPressedHandler(sender, e);
            }

            return false;
        }

        internal override bool PointerMovedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            if (!mCreated)
            {
                // Canvas takes care of mirroring
                if (mPointerDown)
                {
                    if (e.Pointer.PointerId != mPointerID)
                    {
                        return false;
                    }
                    UIPoint pp = e.GetCurrentPoint(mViewerCanvas).Position;
                    m_cnPt2.x = pp.X;
                    m_cnPt2.y = pp.Y;

                    // want to make sure we're within our bounding box.
                    if (mPageCropOnClient != null)
                    {
                        if (m_cnPt2.x < mPageCropOnClient.x1)
                        {
                            m_cnPt2.x = mPageCropOnClient.x1;
                        }
                        if (m_cnPt2.x > mPageCropOnClient.x2)
                        {
                            m_cnPt2.x = mPageCropOnClient.x2;
                        }
                        if (m_cnPt2.y < mPageCropOnClient.y1)
                        {
                            m_cnPt2.y = mPageCropOnClient.y1;
                        }
                        if (m_cnPt2.y > mPageCropOnClient.y2)
                        {
                            m_cnPt2.y = mPageCropOnClient.y2;
                        }
                    }

                    // We don't want to start drawing until we have a bit of a distance
                    if (!mStartDrawing)
                    {
                        // wait until you're at least START_DRAWING_THRESHOLD distance from the start
                        if ((Math.Abs(m_cnPt2.x - m_cnPt1.x) > START_DRAWING_THRESHOLD)
                            || (Math.Abs(m_cnPt2.y - m_cnPt1.y) > START_DRAWING_THRESHOLD))
                        {
                            mStartDrawing = true;
                            mViewerCanvas.Children.Add(this);
                            this.SetValue(Canvas.LeftProperty, m_cnPt1.x);
                            this.SetValue(Canvas.TopProperty, m_cnPt1.y);
                            mRenderTransform = new CompositeTransform();
                            this.RenderTransform = mRenderTransform;
                            mMirrorX = false;
                            mMirrorY = false;
                            this.Opacity = mOpacity;
                        }
                    }

                    // Size the canvas according to the two points
                    if (mStartDrawing)
                    {
                        this.Width = Math.Abs(m_cnPt2.x - m_cnPt1.x);
                        this.Height = Math.Abs(m_cnPt2.y - m_cnPt1.y);
                        if (m_cnPt2.x < m_cnPt1.x)
                        {
                            mMirrorX = true;
                            mRenderTransform.ScaleX = -1;
                        }
                        else
                        {
                            mMirrorX = false;
                            mRenderTransform.ScaleX = 1;
                        }
                        if (m_cnPt2.y < m_cnPt1.y)
                        {
                            mMirrorY = true;
                            mRenderTransform.ScaleY = -1;
                        }
                        else
                        {
                            mMirrorY = false;
                            mRenderTransform.ScaleY = 1;
                        }
                    }
                }

                if (mStartDrawing) // Simple Shape is ready to draw
                {
                    if (!mShapeHasBeenCreated)
                    {
                        mShapeHasBeenCreated = true;
                        RectangleGeometry geom = new RectangleGeometry();
                        mShape = new Path();
                        mShape.Data = geom;

                        if (mUseStroke)
                        {
                            mShape.Stroke = mStrokeBrush;
                        }
                        if (mUseFill)
                        {
                            mShape.Fill = mFillBrush;
                        }
                        this.Children.Add(mShape);
                    }

                    // Don't draw larger than canvas / 2, which gives a smoother look
                    double tempThickness = Math.Min(mDrawThickness, Math.Min(this.Width, this.Height) / 2);
                    mShape.StrokeThickness = tempThickness;

                    // place the two points on the canvas to make room for the thickness of the border.
                    (mShape.Data as RectangleGeometry).Rect = new UIRect(tempThickness / 2, tempThickness / 2,
                        Math.Max(this.Width - tempThickness, 0), Math.Max(this.Height - tempThickness, 0));

                }
            }
            else
            {
                base.PointerMovedHandler(sender, e);
            }
            return true;
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (!mCreated)
            {
                // Remove contact from dictionary.
                RemovePointer(e.Pointer);
                if (mCancelledDrawing)
                {
                    return false; // ignore any secondary pointer
                }

                mPointerDown = false;
                if (mShape == null)
                {
                    return false;
                }

                mRenderTransform = new CompositeTransform();
                RenderTransform = mRenderTransform;
                mMirrorX = false;
                mMirrorY = false;

                if (mShape.Data is RectangleGeometry rect)
                {
                    rect.Rect = new UIRect(0, 0, 0, 0);
                    mShape.Data = rect;
                }
                //(mShape.Data as RectangleGeometry).Rect = new UIRect(0, 0, 0, 0);

                mSelectedRectInPage = FindSelectedAnnotationsAndRect();

                // TODO: Ensure to proper check of other annot types
                if (mSelectedRectInPage == null
                    || mAnnot == null
                    || mAnnot.GetPage() == null
                    || mAnnot.GetAnnotType() == AnnotType.e_Link)
                {
                    RemovePointer(e.Pointer);
                    mPointerDown = false;
                    mPointerID = -1;
                    mViewerCanvas?.ReleasePointerCaptures();

                    mPt1.x = m_cnPt1.x;
                    mPt2.y = m_cnPt2.y;

                    BackToPan();
                }
                else
                {
                    mAnnotPageNum = mAnnot.GetPage().GetIndex();

                    mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);

                    PopulateMenu();
                    InitializeCanvasAndGeometry();

                    mJustSwitchedFromAnotherTool = false;
                    CreateAppearance();
                    EnableScrolling();

                    mCreated = true;

                    RemovePointer(e.Pointer);
                    mPointerDown = false;
                    mPointerID = -1;
                    mViewerCanvas?.ReleasePointerCaptures();

                    mPt1.x = m_cnPt1.x;
                    mPt2.y = m_cnPt2.y;
                }
            }

            base.PointerReleasedHandler(sender, e);
            return false;
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            return false;
        }

        internal override bool PointerCaptureLostHandler(object sender, PointerRoutedEventArgs e)
        {
            mPointerID = -1;
            return true;
        }

        internal override bool PointerCanceledHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            return true;
        }

        internal override bool OnPageNumberChanged(int current_page, int num_pages)
        {
            return false;
        }

        internal override bool HoldingHandler(object sender, HoldingRoutedEventArgs e)
        {
            return true;
        }

        internal override bool KeyDownAction(object sender, KeyRoutedEventArgs e)
        {
            return false;
        }

        /// <summary>
        /// Find all selected annotations and get the group rectangle
        /// </summary>
        /// <returns>
        ///     Selected annotations rectangle  - Rectangle with annotations
        ///     Null                            - null if no annotations are selected
        /// </returns>
        protected override PDFRect FindSelectedAnnotationsAndRect()
        {
            PDFRect pageRect = GetShapeBoundingBox();
            List<IAnnot> annotsInPage = UtilityFunctions.GetAnnotationsOnPage(mPDFView, mDownPageNumber);
            mSelectedAnnotations.Clear();
            PDFRect selectedRectInPage = null;

            mAnnotIsTextMarkup = false;
            // search for selected annotations in page
            if (pageRect != null)
            {
                foreach (IAnnot annot in annotsInPage)
                {
                    if (mAnnot == null)
                    {
                        mAnnot = annot;
                    }

                    if (annot.GetAnnotType() == AnnotType.e_Link || annot.GetAnnotType() == AnnotType.e_Widget)
                        continue;

                    PDFRect annotRect = mPDFView.GetPageRectForAnnot(annot, mDownPageNumber);
                    annotRect.Normalize();

                    PDFRect dummyPageRect = new PDFRect(pageRect.x1, pageRect.y1, pageRect.x2, pageRect.y2);

                    if (dummyPageRect.IntersectRect(dummyPageRect, annotRect))
                    {
                        if (!DoesAnnotationSupportResize(annot))
                        {
                            mAnnotIsTextMarkup = true;
                        }
                        mSelectedAnnotations.Add(annot, mDownPageNumber);
                    }
                }
            }

            FindAndAddExistingGroupAnnotations();

            // Go over all selected annotations and draw a big group rectangle
            foreach (KeyValuePair<IAnnot, int> annot in mSelectedAnnotations)
            {
                PDFRect annotRect = mPDFView.GetPageRectForAnnot(annot.Key, mDownPageNumber);
                annotRect.Normalize();

                if (selectedRectInPage == null)
                {
                    selectedRectInPage = new PDFRect(annotRect.x1, annotRect.y1, annotRect.x2, annotRect.y2);
                }
                else
                {
                    if (annotRect.x1 < selectedRectInPage.x1)
                    {
                        selectedRectInPage.x1 = annotRect.x1;
                    }

                    if (annotRect.y1 < selectedRectInPage.y1)
                    {
                        selectedRectInPage.y1 = annotRect.y1;
                    }

                    if (annotRect.x2 > selectedRectInPage.x2)
                    {
                        selectedRectInPage.x2 = annotRect.x2;
                    }

                    if (annotRect.y2 > selectedRectInPage.y2)
                    {
                        selectedRectInPage.y2 = annotRect.y2;
                    }
                }
            }
            mGroupAnnotations = mSelectedAnnotations.Keys.ToList();
            SetMAnnot();
            return selectedRectInPage;
        }

        private void FindAndAddExistingGroupAnnotations()
        {
            //find all annotations part of a group
            Dictionary<string, IAnnot> annotationsInGroup = new Dictionary<string, IAnnot>();
            foreach (KeyValuePair<IAnnot, int> annotation in mSelectedAnnotations)
            {
                string text = "";
                try
                {
                    var id = annotation.Key.GetUniqueID();
                    if (id != null)
                    {
                        text = id.GetAsPDFText();
                    }
                    else
                    {
                        mIsUngroupAllowed = false;
                        continue;
                    }
                }
                catch (NullReferenceException)
                {
                    mIsUngroupAllowed = false;
                    continue;
                }

                if (!annotationsInGroup.ContainsKey(annotation.Key.GetUniqueID().GetAsPDFText()))
                {
                    bool isRtReply = false;
                    try
                    {
                        isRtReply = UtilityFunctions.IsReplyTypeReply(annotation.Key);
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }

                    if ((mAnnot != null) && (!isRtReply))
                    {
                        KeyValuePair<ToolType, List<IAnnot>> pair = CanSelectGroupAnnotation(mPDFView, annotation.Key, annotation.Value);

                        if (pair.Value != null)
                        {
                            foreach (IAnnot item in pair.Value)
                            {
                                if (!annotationsInGroup.ContainsKey(item.GetUniqueID().GetAsPDFText()))
                                {
                                    annotationsInGroup.Add(item.GetUniqueID().GetAsPDFText(), item);
                                }
                            }
                        }
                    }
                }
            }

            //find if there is any left that is not part of a group
            foreach (KeyValuePair<IAnnot, int> annotation in mSelectedAnnotations)
            {
                string text = "";
                try
                {
                    var id = annotation.Key.GetUniqueID();
                    if (id != null)
                    {
                        text = id.GetAsPDFText();
                    }
                    else
                    {
                        mIsUngroupAllowed = false;
                        continue;
                    }
                }
                catch (NullReferenceException)
                {
                    mIsUngroupAllowed = false;
                    continue;
                }

                if (!annotationsInGroup.ContainsKey(annotation.Key.GetUniqueID().GetAsPDFText()))
                {
                    mIsUngroupAllowed = false;
                    break;
                }
            }

            //add missing annotations from groups to mSelectedAnnotations
            foreach (KeyValuePair<string, IAnnot> item in annotationsInGroup)
            {
                bool found = false;

                foreach (KeyValuePair<IAnnot, int> selectedItem in mSelectedAnnotations)
                {
                    if (selectedItem.Key.GetUniqueID().GetAsPDFText() == item.Key)
                    {
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    mSelectedAnnotations.Add(item.Value, item.Value.GetPage().GetIndex());
                }
            }
        }

        #region Handlers For Popup Menu

        internal override void OnCommandMenuClicked(string title)
        {
            if (mAnnot == null)
            {
                SetMAnnot();
            }

            HideContextMenu(mContextMenu);

            if (title.Equals("done", StringComparison.OrdinalIgnoreCase))
            {
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("group", StringComparison.OrdinalIgnoreCase))
            {
                CreateAnnotationGroup();
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("ungroup", StringComparison.OrdinalIgnoreCase))
            {
                UngroupAnnotations();
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("delete", StringComparison.OrdinalIgnoreCase))
            {
                DeleteGroupAnnotations();
                CreateNewTool(ToolType.e_pan);
            }
            else if (title.Equals("fill color", StringComparison.OrdinalIgnoreCase))
            {
                mIsFillColor = true;
                if (mAnnot.GetAnnotType() == AnnotType.e_FreeText)
                {
                    mEffectiveBrush = e_text_fill;
                    ShowColorPicker();
                }
                else
                {
                    mEffectiveBrush = e_fill;
                    ShowColorPicker();
                }
            }
            else if (title.Equals("opacity", StringComparison.OrdinalIgnoreCase))
            {
                HandleOpacity();
            }
            else if (title.Equals("color", StringComparison.OrdinalIgnoreCase))
            {
                mIsFillColor = false;
                AnnotType aType = mAnnot.GetAnnotType();
                if (aType == AnnotType.e_Line || aType == AnnotType.e_Ink || aType == AnnotType.e_Polyline)
                {
                    mEffectiveBrush = e_stroke;
                    ShowColorPicker();
                }
                else if (aType == AnnotType.e_Highlight)
                {
                    mEffectiveBrush = e_highlight;
                    ShowColorPicker();
                }
                else
                {
                    mEffectiveBrush = e_text_markup;
                    ShowColorPicker();
                }
            }
            else if (title.Equals("flatten", StringComparison.OrdinalIgnoreCase))
            {
                HandleFlatten();
            }
            else
            {
                throw new Exception("The option doesn't exist");
            }
        }

        #endregion Handlers For Popup Menu

        private PDFRect GetShapeBoundingBox()
        {
            // Computes the bounding box of the rubber band in page space.
            PDFDouble x1 = new PDFDouble(m_cnPt1.x);
            PDFDouble y1 = new PDFDouble(m_cnPt1.y);
            PDFDouble x2 = new PDFDouble(m_cnPt2.x);
            PDFDouble y2 = new PDFDouble(m_cnPt2.y);

            mPDFView.ConvAnnotationCanvasPtToScreenPt(x1, y1);
            mPDFView.ConvAnnotationCanvasPtToScreenPt(x2, y2);
            mPDFView.ConvScreenPtToPagePt(x1, y1, mDownPageNumber);
            mPDFView.ConvScreenPtToPagePt(x2, y2, mDownPageNumber);

            PDFRect rect;
            try
            {
                rect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);
                rect.Normalize();
                return rect;
            }
            catch (Exception)
            {
                return null;
            }
        }

        private void CreateAnnotationGroup()
        {
            // first find primary annot
            // here we assume the last one created is the primary
            List<IAnnot> annotations = new List<IAnnot>(mSelectedAnnotations.Keys.ToList());
            if ((annotations == null) || (annotations.Count == 0))
            {
                return;
            }

            IAnnot primaryAnnotation = annotations[annotations.Count - 1];

            try
            {
                UtilityFunctions.CreateAnnotationGroup(mPDFView, primaryAnnotation, annotations);
                mToolManager.RaiseAnnotationGroupAddedEvent(mSelectedAnnotations);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            BackToPan();
        }
    }
}

