﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using pdftron.PDF;
using pdftron.Common;

namespace pdftron.PDF.Tools
{
    public class CloudCreate : AdvancedShapeCreate
    {
        #region Members

        private bool mShapeHasBeenCreated = false;
        private Path mShape;
        private PathGeometry mPolygonGeometry = new PathGeometry();
        private PathFigure mCloudFigure = new PathFigure();
        private UIPoint mFirstPoint = new UIPoint();
        private List<PDFPoint> mPdfPointList = new List<PDFPoint>();
        public static int mBorderIntensity = 2;
        private List<UIPoint> mDrawPoints = new List<UIPoint>();
        Dictionary<int, UIPoint> mCurveSegments = new Dictionary<int, UIPoint>();
        Dictionary<int, UIPoint> mLineSegments = new Dictionary<int, UIPoint>();
        private int mSegId = 0;
        private int mPreviousDrawPointIndex = 0;
        private UIPoint mStartPoint = new UIPoint(0, 0);
        private bool mCanSelectNewPoint = false;
        private int mMinLengthToCreateOnDragging = 20; // If user is clicking and dragging the line, we only create a new point once it passes the min length value

        #endregion

        #region Constructor

        public CloudCreate(PDFViewCtrl pdfViewControl, ToolManager toolManager) 
            : base(pdfViewControl, toolManager)
        {
            mNextToolMode = ToolType.e_cloud_create;
            mToolMode = ToolType.e_cloud_create;
        }

        #endregion

        #region Event Handlers

        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            try
            {
                base.PointerPressedHandler(sender, e);

                if (mSimpleShapeCanceledAtPress)
                {
                    OnPointCreated(mPdfPointList.Count);
                    return false;
                }

                mShapeHasBeenCreated = false;

                if (mStartDrawing)
                {
                    if (mShape == null)
                    {
                        CreatePath();

                        mPreviousDrawPointIndex++;
                        mDrawPoints.Add(new UIPoint(0, 0));

                        DrawCloud();

                        TrimToPageBounds(m_cnPt1, mDrawThickness);
                        mPdfPointList.Add(new PDFPoint(m_cnPt1.x, m_cnPt1.y));
                        mFirstPoint = new UIPoint(m_cnPt1.x, m_cnPt1.y);
                    }
                    else
                    {
                        if (mCanSelectNewPoint)
                        {
                            //TrimToPageBounds already called for m_cnPt2 in PointerMovedHandler
                            double nextPointX = m_cnPt2.x - mFirstPoint.X;
                            double nextPointY = m_cnPt2.y - mFirstPoint.Y;

                            mPreviousDrawPointIndex++;
                            mPdfPointList.Add(new PDFPoint(m_cnPt2.x, m_cnPt2.y));
                            mDrawPoints.Add(new UIPoint(nextPointX, nextPointY));

                            DrawCloud();

                            mCanSelectNewPoint = false;

                            OnPointCreated(mPdfPointList.Count);
                        }
                    }

                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return false;
        }        

        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            try
            {
                if (mSimpleShapeCanceledAtPress)
                {
                    return false;
                }

                base.PointerMovedHandler(sender, e);

                if (mStartDrawing) // Shape is ready to draw
                {
                    TrimToPageBounds(m_cnPt2, mDrawThickness);

                    double nextPointX = m_cnPt2.x - mFirstPoint.X;
                    double nextPointY = m_cnPt2.y - mFirstPoint.Y;

                    if (mPointerDown && (e.Pointer.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Touch))
                    {
                        //remove previous mouse move locations from mDrawPoints
                        if (mDrawPoints.Count > (mPreviousDrawPointIndex))
                        {
                            for (int i = mDrawPoints.Count - 1; i > mPreviousDrawPointIndex; i--)
                            {
                                mDrawPoints.RemoveAt(i);
                            }
                        }

                        //add new mouse move location to mDrawPoints
                        mDrawPoints.Add(new UIPoint(nextPointX, nextPointY));

                        if (mPdfPointList.Count == (mPreviousDrawPointIndex))
                        {
                            var x_diff = Math.Abs(mPdfPointList[mPreviousDrawPointIndex - 1].x - m_cnPt2.x);
                            var y_diff = Math.Abs(mPdfPointList[mPreviousDrawPointIndex - 1].y - m_cnPt2.y);

                            if (x_diff > mMinLengthToCreateOnDragging
                                || y_diff > mMinLengthToCreateOnDragging)
                            {
                                mPdfPointList.Add(new PDFPoint(m_cnPt2.x, m_cnPt2.y));
                                OnPointCreated(mPdfPointList.Count);
                            }
                        }
                        else
                        {
                            mPdfPointList[mPdfPointList.Count - 1].x = m_cnPt2.x;
                            mPdfPointList[mPdfPointList.Count - 1].y = m_cnPt2.y;
                        }

                        DrawCloud();
                    }

                    if (mCloudFigure.Segments.Count == 3)
                    {
                        mCloudFigure.IsClosed = true;
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return false;
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            mCanSelectNewPoint = true;
            base.PointerReleasedHandler(sender, e);
            return true;
        }

        internal override bool DoubleTappedHandler(object sender, DoubleTappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Touch)
            {
                mSimpleShapeCanceledAtPress = true;
            }
            return false;
        }

        internal override bool RightTappedHandler(Object sender, RightTappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType != Windows.Devices.Input.PointerDeviceType.Touch)
            {
                mSimpleShapeCanceledAtPress = true;
            }
            return false;
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            return base.TappedHandler(sender, e);
        }

        #endregion

        #region Methods

        public void CommitAnnotation()
        {
            if (mDownPageNumber == -1)
                return;

            try
            {
                if (mStartDrawing) // don't want to add any annotations resulting from a tap
                {
                    EnableScrolling();
                    Finish();
                    mToolManager.RaiseAnnotationAddedEvent(mAnnot, mDownPageNumber);
                }

                EndCurrentTool(ToolType.e_pan);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        private void DrawCloud()
        {
            //creates LineSegments and BezierSegments based on the UIPoints calculated for Cloud annotation.
            //these points were stored in the mLineSegments and mCurveSegments dictionaries with keys indicating index/position 
            //in which they should be used for creating segments.
            try
            {
                mCloudFigure.Segments.Clear();
                mLineSegments.Clear();
                mCurveSegments.Clear();

                CalculateCloudPoints(mDrawPoints, mBorderIntensity);

                mCloudFigure.StartPoint = mStartPoint;

                //creates a segment by looking for the mSegId starting from 0; so that the segments are created in order.
                for (int i = 0; i < mSegId; i++)
                {
                    if (mLineSegments.ContainsKey(i))
                    {
                        LineSegment segment = new LineSegment();
                        segment.Point = new UIPoint(mLineSegments[i].X, mLineSegments[i].Y);
                        mCloudFigure.Segments.Add(segment);
                    }
                    else if (mCurveSegments.ContainsKey(i))
                    {
                        BezierSegment segment = new BezierSegment();
                        segment.Point1 = new UIPoint(mCurveSegments[i].X, mCurveSegments[i].Y);
                        i++;
                        segment.Point2 = new UIPoint(mCurveSegments[i].X, mCurveSegments[i].Y);
                        i++;
                        segment.Point3 = new UIPoint(mCurveSegments[i].X, mCurveSegments[i].Y);
                        mCloudFigure.Segments.Add(segment);
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        private void CreatePath()
        {
            try
            {
                if (!mShapeHasBeenCreated)
                {
                    mShapeHasBeenCreated = true;

                    if (mDrawThickness == 0)
                    {
                        mDrawThickness = 1;
                    }

                    mShape = new Path();
                    mShape.Data = mPolygonGeometry;
                    mCloudFigure.StartPoint = mStartPoint;
                    PathFigure dummyFigure = new PathFigure();
                    dummyFigure.StartPoint = new UIPoint(0, 0);
                    mPolygonGeometry.Figures.Add(dummyFigure);
                    mPolygonGeometry.Figures.Add(mCloudFigure);
                    mShape.StrokeThickness = mDrawThickness;
                    mShape.Stroke = mStrokeBrush;

                    if (mUseFill)
                        mShape.Fill = mFillBrush;

                    this.Children.Add(mShape);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        internal override void Finish()
        {
            //creates a cloud polygon annotation and adds it to the PDF page
            try
            {
                mPDFView.DocLock(true);

                PDFRect rect = GetAdvancedShapeBBox(mPdfPointList);
                if (rect != null)
                {
                    rect.Inflate(mDrawThickness);
                    Annots.Polygon pdfPolygon = Annots.Polygon.Create(mPDFView.GetDoc().GetSDFDoc(), rect);

                    pdfPolygon.SetIntentName(Annots.PolyLineIntentType.e_PolygonCloud);
                    pdfPolygon.SetBorderEffect(Annots.MarkupBorderEffect.e_Cloudy);

                    int vertexIdx = 0;
                    foreach (PDFPoint point in mPdfPointList)
                    {
                        // The Rect is normalized, so we need to recalculate the points if we don't want it to start from 0,0
                        PDFPoint pdfPoint = ConvertFromCanvasPointToPagePoint(point, mDownPageNumber);
                        pdfPolygon.SetVertex(vertexIdx++, pdfPoint);
                    }

                    SetStyle(pdfPolygon, mUseFill);
                    pdfPolygon.RefreshAppearance();

                    UtilityFunctions.SetUniqueId(pdfPolygon, Guid.NewGuid().ToString());

                    PDFPage page = mPDFView.GetDoc().GetPage(mDownPageNumber);
                    page.AnnotPushBack(pdfPolygon);

                    mAnnotPageNum = mDownPageNumber;
                    mAnnot = pdfPolygon;
                    mPDFView.UpdateWithAnnot(pdfPolygon, mAnnotPageNum);
                }               
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }
        }

        #region Cloud Points Calculations

        private void CalculateCloudPoints(List<UIPoint> drawPoints, double borderIntensity)
        {
            try
            {
                List<UIPoint> poly = GetClosedPoly(drawPoints);
                int size = poly.Count;
                if (size < 2)
                {
                    return;
                }

                double SAME_VERTEX_TH = 1.0 / 8192.0;
                if (borderIntensity < 0.1)
                {
                    borderIntensity = mBorderIntensity;
                }
                borderIntensity *= mPDFView.GetZoom();
                bool clockwise = IsPolyWrapClockwise(poly);
                double sweepDirection = clockwise ? -1.0 : 1.0;
                double maxCloudSize = 8 * borderIntensity;

                double lastCloudSize = maxCloudSize;
                double firstCloudSize = maxCloudSize;
                double edgeDegrees = 0.0;
                UIPoint firstPos = poly[0];
                UIPoint lastEdge = Subtract(poly[0], poly[size - 2]);
                bool useLargeFirstArc = true;
                bool hasFirstPoint = false;

                float startX = 0, startY = 0;

                for (int i = 0; i < size - 1; ++i)
                {
                    UIPoint pos = poly[i];
                    UIPoint edge = Subtract(poly[i + 1], pos);
                    double length = Hypotenuse(ToSingle(edge.X), ToSingle(edge.Y));
                    // avoid division by 0 from duplicated points.
                    if (length <= SAME_VERTEX_TH)
                    {
                        continue;
                    }

                    // split the edge into some integral number of clouds
                    UIPoint direction = Divide(edge, length);
                    int numClouds = (int)Math.Max(Math.Floor(length / maxCloudSize), 1);
                    double cloudSize = length / numClouds;
                    double edgeAngle = Math.Atan2(direction.Y, direction.X); // angle from x-axis

                    // back start position out to before the vertex
                    // as we're going to increment before using it
                    pos = Subtract(pos, Multiply(direction, cloudSize * .5));

                    // which direction are we turning on this vertex?
                    double cross = Cross(lastEdge, edge);

                    int c = 0;
                    if (!hasFirstPoint)
                    {
                        // skip the first iteration for the first leg (we'll complete it at the end)
                        ++c;
                        firstCloudSize = cloudSize;
                        useLargeFirstArc = (cross * sweepDirection) < 0;
                        pos = Add(pos, Multiply(direction, cloudSize));
                        firstPos = pos;

                        // start the curve
                        mLineSegments.Add(mSegId++, new UIPoint(firstPos.X, firstPos.Y));
                        mStartPoint = new UIPoint(firstPos.X, firstPos.Y);

                        startX = ToSingle(firstPos.X);
                        startY = ToSingle(firstPos.Y);
                        hasFirstPoint = true;
                    }
                    // for the first iteration, combine the radius with the previous edge
                    double radius = (lastCloudSize + cloudSize) * 0.25;
                    for (; c < numClouds; ++c)
                    {
                        if (c == 1)
                        {
                            // on the second iteration on, we can use values exclusive to this edge
                            edgeDegrees = ToDegreesMod360(edgeAngle);
                            radius = cloudSize * 0.5;
                        }
                        pos = Add(pos, Multiply(direction, cloudSize));
                        bool useLargeArc = (c == 0 && (cross * sweepDirection) < 0);
                        UIPoint point = ArcTo(startX, startY, radius, radius, edgeDegrees, useLargeArc, clockwise, pos.X, pos.Y);
                        startX = ToSingle(point.X);
                        startY = ToSingle(point.Y);
                    }
                    edgeDegrees = ToDegreesMod360(edgeAngle);
                    lastEdge = edge;
                    lastCloudSize = cloudSize;
                }
                if (!hasFirstPoint)
                {
                    mLineSegments.Add(mSegId++, new UIPoint(firstPos.X, firstPos.Y));
                    mStartPoint = new UIPoint(firstPos.X, firstPos.Y);

                    startX = ToSingle(firstPos.X);
                    startY = ToSingle(firstPos.Y);
                }
                double closingRadius = (firstCloudSize + lastCloudSize) * 0.25;
                //now we close the poly, using the values we saved on the first vertex.
                ArcTo(startX, startY, closingRadius, closingRadius, edgeDegrees, useLargeFirstArc, clockwise, firstPos.X, firstPos.Y);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        private UIPoint ArcTo(float startX, float startY,
                         double xr, double yr, double rx,
                         bool isLargeArc, bool sweep, double endX, double endY)
        {
            try
            {
                List<double> outPoints = new List<double>(30);
                List<char> outOperations = new List<char>(); // 'l':line or 'c':cube
                ArcToCurves(startX, startY, outPoints, outOperations, xr, yr, rx, isLargeArc, sweep, endX, endY);

                // write the point data gathered to the ElementBuilderImpl
                int pointIndex = 0;
                foreach (char operation in outOperations)
                {
                    if (operation == 'c')
                    {
                        mCurveSegments.Add(mSegId++, new UIPoint(outPoints[pointIndex], outPoints[pointIndex + 1]));
                        mCurveSegments.Add(mSegId++, new UIPoint(outPoints[pointIndex + 2], outPoints[pointIndex + 3]));
                        mCurveSegments.Add(mSegId++, new UIPoint(outPoints[pointIndex + 4], outPoints[pointIndex + 5]));

                        startX = System.Convert.ToSingle(outPoints[pointIndex + 4]);
                        startY = System.Convert.ToSingle(outPoints[pointIndex + 5]);
                        pointIndex += 6;
                    }
                    else if (operation == 'l')
                    {
                        mLineSegments.Add(mSegId++, new UIPoint(outPoints[pointIndex], outPoints[pointIndex + 1]));

                        startX = System.Convert.ToSingle(outPoints[pointIndex]);
                        startY = System.Convert.ToSingle(outPoints[pointIndex + 1]);
                        pointIndex += 2;
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return new UIPoint(startX, startY);
        }

        private float Hypotenuse(float x, float y)
        {
            return (float)Math.Sqrt(x * x + y * y);
        }

        /// <summary>
        /// Gets the closed path of poly
        /// </summary>
        /// <param name="input">The input poly witch might be open</param>
        /// <returns>The closed version of the input poly</returns>
        private List<UIPoint> GetClosedPoly(List<UIPoint> input)
        {
            List<UIPoint> poly = new List<UIPoint>(input);
            int size = input.Count();

            if (size >= 2)
            {
                UIPoint firstPoint = input[0];
                UIPoint lastPoint = input[size - 1];
                if (!firstPoint.Equals(lastPoint))
                {
                    poly.Add(firstPoint);
                }
            }

            return poly;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="canvasPoints">The canvas points</param>
        /// <returns>True if this polygon is clockwise oriented;
        /// False if this polygon is counterclockwise oriented
        /// </returns>
        private bool IsPolyWrapClockwise(List<UIPoint> canvasPoints)
        {
            UIPoint oldPoint;
            double accum = 0.0;
            foreach (UIPoint point in canvasPoints)
            {
                if (oldPoint != null)
                {
                    accum += (point.X - oldPoint.X) * (point.Y + oldPoint.Y);
                }
                oldPoint = point;
            }

            return accum < 0;
        }

        private UIPoint Add(UIPoint p1, UIPoint p2)
        {
            if ((p1 == null) || (p2 == null))
            {
                return new UIPoint();
            }

            return new UIPoint(p1.X + p2.X, p1.Y + p2.Y);
        }

        private UIPoint Subtract(UIPoint p1, UIPoint p2)
        {
            if ((p1 == null) || (p2 == null))
            {
                return new UIPoint();
            }

            return new UIPoint(p1.X - p2.X, p1.Y - p2.Y);
        }

        private UIPoint Multiply(UIPoint p, double n)
        {
            if (p == null)
            {
                return new UIPoint();
            }

            return new UIPoint((float)(p.X * n), (float)(p.Y * n));
        }

        private UIPoint Divide(UIPoint p, double n)
        {
            if ((p == null) || (n == 0))
            {
                return new UIPoint();
            }

            return Multiply(p, (1f / n));
        }

        private double Cross(UIPoint p1, UIPoint p2)
        {
            if ((p1 == null) || (p2 == null))
            {
                return 0;
            }

            return (p1.X * p2.Y - p1.Y * p2.X);
        }

        /// <summary>
        /// map radians to a value between 0 and 360 degrees
        /// </summary>
        private double ToDegreesMod360(double radians)
        {
            double PI_2 = Math.PI * 2;
            double invPI = 1.0 / Math.PI;
            double fracPart = ((radians + PI_2) % PI_2) * invPI;
            return fracPart * 180;
        }

        private void ArcToCurves(double startX, double startY, List<double> outPoints, List<char> outOperations, 
            double xr, double yr, double rx, bool isLargeArc, bool sweep, double endX, double endY)
        {
            if (xr < 0.0)
            {
                xr = -xr;
            }
            if (yr < 0.0)
            {
                yr = -yr;
            }

            // Calculate the middle point between
            // the current and the final points
            //------------------------
            double dx2 = (startX - endX) / 2.0;
            double dy2 = (startY - endY) / 2.0;

            double cosA = Math.Cos(rx);
            double sinA = Math.Sin(rx);

            // Calculate (x1, y1)
            //------------------------
            double x1 = cosA * dx2 + sinA * dy2;
            double y1 = -sinA * dx2 + cosA * dy2;

            // Ensure radii are large enough
            //------------------------
            double prx = xr * xr;
            double pry = yr * yr;
            double px1 = x1 * x1;
            double py1 = y1 * y1;

            // Check that radii are large enough
            //------------------------
            double radiiCheck = px1 / prx + py1 / pry;
            if (radiiCheck > 1.0)
            {
                xr = Math.Sqrt(radiiCheck) * xr;
                yr = Math.Sqrt(radiiCheck) * yr;
                prx = xr * xr;
                pry = yr * yr;
            }

            double denom = (prx * py1) + (pry * px1);
            if (denom == 0)
            {
                // we shouldn't divide by zero
                // (this is a strange case if it occurs in an actual document
                // since it seems to only occur when start=end)
                return;
            }

            // Calculate (cx1, cy1)
            //------------------------
            double sign = (isLargeArc == sweep) ? -1.0 : 1.0;
            double sq = ((prx * pry) - (prx * py1) - (pry * px1)) / (denom);
            double coef = sign * Math.Sqrt((sq < 0) ? 0 : sq);
            double cx1 = coef * ((xr * y1) / yr);
            double cy1 = coef * -((yr * x1) / xr);

            //
            // Calculate (cx, cy) from (cx1, cy1)
            //------------------------
            double sx2 = (startX + endX) / 2.0;
            double sy2 = (startY + endY) / 2.0;
            double cx = sx2 + (cosA * cx1 - sinA * cy1);
            double cy = sy2 + (sinA * cx1 + cosA * cy1);

            // Calculate the startAngle (angle1) and the sweepAngle (dangle)
            //------------------------
            double ux = (x1 - cx1) / xr;
            double uy = (y1 - cy1) / yr;
            double vx = (-x1 - cx1) / xr;
            double vy = (-y1 - cy1) / yr;
            double p, n;

            // Calculate the angle start
            //------------------------
            n = Math.Sqrt((ux * ux) + (uy * uy));
            p = ux; // (1 * ux) + (0 * uy)
            sign = (uy < 0) ? -1.0 : 1.0;
            double startAngle = sign * Math.Acos(p / n);

            // Calculate the sweep angle
            //------------------------
            n = Math.Sqrt((ux * ux + uy * uy) * (vx * vx + vy * vy));
            p = ux * vx + uy * vy;
            sign = ((ux * vy) - (uy * vx) < 0) ? -1.0 : 1.0;

            // we want to avoid taking the inverse cosine of
            // a value > 1 or < -1 since it is undefined
            // (this can occur due to rounding error)
            double acosValue = Math.Max(Math.Min(p / n, 1.0), -1.0);

            double sweepAngle = sign * Math.Acos(acosValue);
            if (!sweep && sweepAngle > 0)
            {
                sweepAngle -= Math.PI * 2.0;
            }
            else if (sweep && sweepAngle < 0)
            {
                sweepAngle += Math.PI * 2.0;
            }

            int numVertices;
            double[] vertices = new double[26];
            bool bLineTo = false;

            if (Math.Abs(sweepAngle) < 1e-10)
            {
                numVertices = 4;

                // LineTo command;
                bLineTo = true;

                vertices[0] = xr * Math.Cos(startAngle);
                vertices[1] = yr * Math.Sin(startAngle);
                vertices[2] = xr * Math.Cos(startAngle + sweepAngle);
                vertices[3] = yr * Math.Sin(startAngle + sweepAngle);
            }
            else
            {
                double totalSweep = 0.0;
                double localSweep;
                double prev_sweep;
                numVertices = 2;
                // CurveTo command;
                bool done = false;
                do
                {
                    if (sweepAngle < 0.0)
                    {
                        prev_sweep = totalSweep;
                        localSweep = -Math.PI * 0.5;
                        totalSweep -= Math.PI * 0.5;
                        if (totalSweep <= sweepAngle + 0.01) // bezier_arc_angle_epsilon = 0.01
                        {
                            localSweep = sweepAngle - prev_sweep;
                            done = true;
                        }
                    }
                    else
                    {
                        prev_sweep = totalSweep;
                        localSweep = Math.PI * 0.5;
                        totalSweep += Math.PI * 0.5;
                        if (totalSweep >= sweepAngle - 0.01) // bezier_arc_angle_epsilon = 0.01
                        {
                            localSweep = sweepAngle - prev_sweep;
                            done = true;
                        }
                    }

                    // BEGIN: Arc to cubic bezier curve

                    int OFFSET = numVertices - 2;

                    double x0 = Math.Cos(localSweep / 2.0);
                    double y0 = Math.Sin(localSweep / 2.0);
                    double tx = (1.0 - x0) * 4.0 / 3.0;
                    double ty = y0 - tx * x0 / y0;
                    double[] px = new double[4];
                    double[] py = new double[4];
                    px[0] = x0;
                    py[0] = -y0;
                    px[1] = x0 + tx;
                    py[1] = -ty;
                    px[2] = x0 + tx;
                    py[2] = ty;
                    px[3] = x0;
                    py[3] = y0;

                    double sn = Math.Sin(startAngle + localSweep / 2.0);
                    double cs = Math.Cos(startAngle + localSweep / 2.0);

                    for (int i = 0; i < 4; i++)
                    {
                        vertices[OFFSET + i * 2] = xr * (px[i] * cs - py[i] * sn);
                        vertices[OFFSET + i * 2 + 1] = yr * (px[i] * sn + py[i] * cs);
                    }

                    // END: Arc to cubic bezier curve

                    numVertices += 6;
                    startAngle += localSweep;
                }
                while (!done && numVertices < 26);
            }

            try
            {
                // We can now transform the resulting arc
                Matrix2D mtx = Matrix2D.RotationMatrix(-rx);
                mtx = mtx.Translate(cx, cy);

                for (int i = 2; i < numVertices - 2; i += 2)
                {
                    PDFDouble point1 = new PDFDouble(vertices[i]);
                    PDFDouble point2 = new PDFDouble(vertices[i + 1]);
                    mtx.Mult(point1, point2);

                    vertices[i] = point1.Value;
                    vertices[i + 1] = point2.Value;
                }
            }
            catch (Exception)
            {
                return;
            }

            // We must make sure that the starting and ending points
            // exactly coincide with the initial (x0,y0) and (x2,y2)
            vertices[0] = startX;
            vertices[1] = startY;
            if (numVertices > 2)
            {
                vertices[numVertices - 2] = endX;
                vertices[numVertices - 1] = endY;
            }

            if (bLineTo)
            {
                outOperations.Add('l');
                outOperations.Add('l');
                outPoints.Add(vertices[0]);
                outPoints.Add(vertices[1]);
                outPoints.Add(vertices[2]);
                outPoints.Add(vertices[3]);
            }
            else
            {
                outOperations.Add('l');
                outPoints.Add(vertices[0]);
                outPoints.Add(vertices[1]);

                for (int i = 2; i < numVertices; i += 6)
                {
                    outOperations.Add('c');
                    outPoints.Add(vertices[i]);
                    outPoints.Add(vertices[i + 1]);
                    outPoints.Add(vertices[i + 2]);
                    outPoints.Add(vertices[i + 3]);
                    outPoints.Add(vertices[i + 4]);
                    outPoints.Add(vertices[i + 5]);
                }
            }
        }        

        private float ToSingle(double val)
        {
            return System.Convert.ToSingle(val);
        }

        #endregion

        #endregion
    }
}
