﻿using pdftron.PDF;
using pdftron.PDF.Tools.Collections;
using pdftron.PDF.Tools.Controls.Viewer;
using System.Diagnostics.CodeAnalysis;
using System.IO;
using System.Xml.Serialization;
using Windows.Storage;

namespace pdftron.PDF.Tools.Collections
{
    /// <summary>
    /// Stores a collection of properties about the state of a recent document
    /// </summary>
    public class RecentDocumentProperties : IRecentDocumentProperties
    {
        [SuppressMessage("Minor Code Smell", "S1075:URIs should not be hardcoded", Justification = "<Pending>")]
        public const string DEFAULT_THUMB_PATH = "ms-appx:///Assets/DocumentPage/FilePlaceHolder_pdf.png";
        [SuppressMessage("Minor Code Smell", "S1075:URIs should not be hardcoded", Justification = "<Pending>")]
        public const string DEFAULT_PROTECTED_THUMB_PATH = "ms-appx:///Assets/DocumentPage/FilePlaceHolder_pdf_locked.png";
        [SuppressMessage("Minor Code Smell", "S1075:URIs should not be hardcoded", Justification = "<Pending>")]
        public const string DEFAULT_PACKAGE_THUMB_PATH = "ms-appx:///Assets/DocumentPage/FilePlaceHolder_pdf.png";
        [SuppressMessage("Minor Code Smell", "S1075:URIs should not be hardcoded", Justification = "<Pending>")]
        public const string DEFAULT_FOLDER_PATH = "ms-appx:///Assets/DocumentPage/FolderIcon.png";
        public const string ONEDRIVE_ICON_LOCATION = "ms-appx:///Assets/DocumentPage/OneDrive_rgb_Blue2728.png";

        private string _thumbnailLocation = string.Empty;

        [XmlIgnore]
        public string ThumbnailLocation
        {
            get
            {
                if (IsFolder)
                    return DEFAULT_FOLDER_PATH;

                if (IsProtected)
                    return DEFAULT_PROTECTED_THUMB_PATH;

                if (string.IsNullOrWhiteSpace(_thumbnailLocation))
                    return DEFAULT_THUMB_PATH;

                return _thumbnailLocation;
            }
            set => _thumbnailLocation = value;
        }

        [XmlElement(ElementName = "IsFolder")]
        public bool IsFolder { get; set; }

        [XmlElement(ElementName = "Path")]
        public string FilePath { get; set; }

        [XmlIgnore]
        public IStorageItem File { get; set; }

        [XmlElement(ElementName = "Pg")]
        public int PageNumber { get; set; }

        [XmlElement(ElementName = "PW")]
        public bool IsProtected { get; set; }

        [XmlElement(ElementName = "Zoom")]
        public double Zoom { get; set; }

        [XmlElement(ElementName = "HP")]
        public int HorizontalScrollPos { get; set; }

        [XmlElement(ElementName = "VP")]
        public int VerticalScrollPos { get; set; }

        [XmlElement(ElementName = "ExtraIcon")]
        public AdditionalIcons AdditonalIcon { get; set; }

        [XmlElement(ElementName = "PPM")]
        public PDFViewCtrlPagePresentationMode PagePresentationMode { get; set; } = ViewerSettings.PagePresentationMode;

        [XmlElement(ElementName = "RF")]
        public bool IsInReflowMode { get; set; }

        [XmlElement(ElementName = "Rot")]
        public PageRotate PageRotation { get; set; }

        public RecentDocumentProperties(IStorageItem file, int pageNumber, bool isProtected, bool isFolder = false)
        {
            File = file;
            PageNumber = pageNumber;
            IsProtected = isProtected;
            if (isFolder)
            {
                _thumbnailLocation = DEFAULT_FOLDER_PATH;
                IsFolder = true;
            }
            else
            {
                IsFolder = false;
            }
        }

        public RecentDocumentProperties()
        {
        }

        public static RecentDocumentProperties FromString(string data)
        {
            XmlSerializer xmlIzer = new XmlSerializer(typeof(RecentDocumentProperties));
            using (StringReader sr = new StringReader(data))
            {
                return (xmlIzer.Deserialize(sr)) as RecentDocumentProperties;
            }
        }

        public string MakeString()
        {
            XmlSerializer xmlIzer = new XmlSerializer(typeof(RecentDocumentProperties));
            StringWriter writer = new StringWriter();
            xmlIzer.Serialize(writer, this);
            return writer.ToString();
        }
    }
}
