﻿using Microsoft.UI.Xaml.Controls;
using System.Windows.Input;
using Windows.UI.Xaml;

namespace pdftron.PDF.Tools.Commands.Events
{
    public static class CloseRequestedCommand
    {
        public static readonly DependencyProperty CommandProperty = DependencyProperty.RegisterAttached(
            "Command",
            typeof(ICommand),
            typeof(CloseRequestedCommand),
            new PropertyMetadata(null, OnCommandPropertyChanged));

        public static void SetCommand(DependencyObject d, ICommand value) => d.SetValue(CommandProperty, value);

        public static ICommand GetCommand(DependencyObject d) => (ICommand)d.GetValue(CommandProperty);

        private static void OnCommandPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is TabViewItem control)
            {
                if (e.NewValue != null)
                {
                    control.CloseRequested += Control_CloseRequested;
                }
                else
                {
                    control.CloseRequested -= Control_CloseRequested;
                }
            }
        }

        private static void Control_CloseRequested(TabViewItem sender, TabViewTabCloseRequestedEventArgs args)
        {
            var command = GetCommand(sender);

            if (command?.CanExecute(args) == true)
                command.Execute(args);
        }
    }
}
