﻿using System;
using System.Windows.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;

namespace pdftron.PDF.Tools.Commands.Events
{
    public static class PointerPressedCommand
    {
        public static readonly DependencyProperty CommandProperty = DependencyProperty.RegisterAttached(
            "Command",
            typeof(ICommand),
            typeof(PointerPressedCommand),
            new PropertyMetadata(null, OnCommandPropertyChanged));

        public static void SetCommand(DependencyObject d, ICommand value) => d.SetValue(CommandProperty, value);

        public static ICommand GetCommand(DependencyObject d) => (ICommand)d.GetValue(CommandProperty);

        private static void OnCommandPropertyChanged(DependencyObject d,
            DependencyPropertyChangedEventArgs e)
        {
            if (d is FrameworkElement control)
                control.PointerPressed += Control_PointerPressed;
        }

        private static void Control_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            FrameworkElement control = sender as FrameworkElement;
            var command = GetCommand(control);

            if (command?.CanExecute(e.Pointer) == true)
                command.Execute(new Tuple<object, PointerRoutedEventArgs>(sender, e));
        }
    }
}
