﻿using pdftron.PDF.Tools.Controls.ViewModels;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class AnnotationPropertyDialog : UserControl, IDisposable
    {
        private pdftron.PDF.Tools.Controls.ViewModels.AnnotationPropertiesViewModel _ViewModel;
        public pdftron.PDF.Tools.Controls.ViewModels.AnnotationPropertiesViewModel ViewModel { get { return _ViewModel; } }

        private AnnotationPropertiesViewModel.StyleChangeRequestedHandler _StyleChangeRequested;
        /// <summary>
        /// Event for when Style properties are changed
        /// </summary>
        public event AnnotationPropertiesViewModel.StyleChangeRequestedHandler StyleChangeRequested
        {
            add
            {
                if (_StyleChangeRequested?.GetInvocationList().Contains(value) != true)
                    _StyleChangeRequested += value;
            }
            remove
            {
                _StyleChangeRequested -= value;
            }
        }

        public AnnotationPropertyDialog()
        {
            this.InitializeComponent();
            _ViewModel = new ViewModels.AnnotationPropertiesViewModel();
            this.DataContext = _ViewModel;
            _ViewModel.ToolSelected += _ViewModel_ToolSelected;
            _ViewModel.StyleChangeRequested += ViewModel_StyleChangeRequested;
        }

        /// <summary>
        /// Set the visibility of Thickness option
        /// </summary>
        public bool HasThickness
        {
            get { return ViewModel.HasThickness; }
            set { ViewModel.HasThickness = value; }
        }

        /// <summary>
        /// Get ToolType being used
        /// </summary>
        public ToolType AssociatedTool
        {
            get { return _ViewModel.AssociatedTool; }
        }

        /// <summary>
        /// Tag to handle options for a specific annotation ToolType
        /// </summary>
        public string ItemTag
        {
            get { return ViewModel.ItemTag; }
            set { ViewModel.ItemTag = value; }
        }

        private void _ViewModel_ToolSelected(ToolType type)
        {
            //For an eraser, since only thickness is needed, make the UI prettier
            if (_ViewModel.AssociatedTool == ToolType.e_ink_eraser)
            {
                //this.Height = 70;
                //CustomSubView.Margin = new Thickness(20, 0, 20, 0);
            }
        }

        private void ViewModel_StyleChangeRequested(AnnotStyleType annotStyleType, object val)
        {
            _StyleChangeRequested?.Invoke(annotStyleType, val);
        }

        public void Dispose()
        {
            if (ViewModel != null)
            {
                ViewModel.StyleChangeRequested -= ViewModel_StyleChangeRequested;
                ViewModel.ToolSelected -= _ViewModel_ToolSelected;

                ViewModel.Dispose();
            }
        }
    }
}
