﻿using System;
using System.Linq;
using System.Threading.Tasks;
using Windows.Storage;
using Windows.Storage.Pickers;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public partial class ThumbnailsViewViewModel
    {
        private async Task ExportSelectedPagesAsync(object parameter) => await SaveSelectedPagesFromPicker(parameter as ThumbnailItem);

        /// <summary>
        /// Saves all the selected PDF pages into a new PDF file
        /// </summary>
        /// <param name="thumbItem"></param>
        /// <returns></returns>
        private async Task SaveSelectedPagesFromPicker(ThumbnailItem thumbItem = null)
        {
            FileSavePicker fileSavePicker = new FileSavePicker();
            fileSavePicker.FileTypeChoices.Add("PDF", Valid_PDF_Types);
            fileSavePicker.SuggestedFileName = _pdfViewCtrl.GetDoc().GetDocInfo().GetTitle();
            fileSavePicker.SuggestedStartLocation = PickerLocationId.DocumentsLibrary;

            StorageFile pickedFile = await fileSavePicker.PickSaveFileAsync();
            if (pickedFile != null)
            {
                PDFDoc doc = new PDFDoc();
                if (thumbItem != null)
                {
                    PageIterator pageIterator = doc.GetPageIterator(thumbItem.PageNumber);
                    doc.PageInsert(pageIterator, _pdfViewCtrl.GetDoc().GetPage(thumbItem.PageNumber));
                }
                else
                {
                    foreach (ThumbnailItem item in _selectedItems.OrderBy(x => x.PageNumber))
                    {
                        PageIterator pageIterator = doc.GetPageIterator(item.PageNumber);
                        doc.PageInsert(pageIterator, _pdfViewCtrl.GetDoc().GetPage(item.PageNumber));
                    }
                }

                await doc.SaveToNewLocationAsync(pickedFile, SDF.SDFDocSaveOptions.e_compatibility);
            }
        }
    }
}
