﻿using pdftron.Common;
using System;
using System.Threading;
using System.Threading.Tasks;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Crop
{
    public partial class CropPopupViewModel
    {
        /// <summary>
        /// Iterates through all the pages of the PDF and sets the User Crop equal to the bounding box
        /// </summary>
        /// <param name="parameter"></param>
        private async Task AutomaticCroppingAsync()
        {
            // Need to close popup in order for new IsLightDismissEnabled value to take effect
            IsAutomaticCropping = true;
            _isPopupOpen = false;
            RaisePropertyChanged(nameof(IsPopupOpen));
            IsPopupOpen = true;

            _currentAutomaticPage = 1;
            RaisePropertyChanged(nameof(AutomaticProgress));
            RaisePropertyChanged(nameof(AutomaticProgressPageText));

            PDFDoc pdfDoc = _pdfViewCtrl.GetDoc();
            _automaticBoxes.Clear();

            bool hasEditedDoc = false;

            try
            {
                pdfDoc.LockRead();
                PageIterator iter = pdfDoc.GetPageIterator();
                _tokenSource = new CancellationTokenSource();
                CancellationToken token = _tokenSource.Token;

                while (iter.HasNext())
                {
                    if (_isAutomaticCancel)
                        return;

                    Page page = iter.Current();

                    Rect rect = await Task.Run(() =>
                    {
                        Task<Rect> task2 = Task.Run(() =>
                        {
                            Rect cropRect = page.GetCropBox();
                            Rect visibleRect = page.GetVisibleContentBox();
                            visibleRect.Inflate(CROP_RECT_WHITE_SPACE_MARGIN);
                            bool intersetcs = visibleRect.IntersectRect(visibleRect, cropRect);
                            if (intersetcs)
                                return visibleRect;

                            return cropRect;
                        }, token);
                        try
                        {
                            task2.Wait(token);
                        }
                        catch (OperationCanceledException)
                        {
                            return null;
                        }

                        return task2.Result;
                    }, token);

                    if (rect == null)
                        return;

                    _automaticBoxes.Add(rect);

                    _currentAutomaticPage++;
                    RaisePropertyChanged(nameof(AutomaticProgress));
                    RaisePropertyChanged(nameof(AutomaticProgressPageText));

                    iter.Next();
                }

                _currentAutomaticPage = 1;
                PageIterator iter2 = pdfDoc.GetPageIterator();
                while (iter2.HasNext())
                {
                    Page page = iter2.Current();
                    if (_automaticBoxes[_currentAutomaticPage - 1] != null)
                    {
                        Rect cropRect = page.GetCropBox();
                        Rect oldUserCropRect = page.GetBox(PageBox.e_user_crop);
                        Rect visibleRect = _automaticBoxes[_currentAutomaticPage - 1];

                        if (!CropViewViewModel.AreRectsSimilar(visibleRect, oldUserCropRect,
                            CropViewViewModel.RECT_COMPARE_THRESHOLD))
                        {
                            if (!CropViewViewModel.AreRectsSimilar(visibleRect, cropRect,
                            CropViewViewModel.RECT_COMPARE_THRESHOLD))
                            {
                                page.SetBox(PageBox.e_user_crop, _automaticBoxes[_currentAutomaticPage - 1]);
                                hasEditedDoc = true;
                            }
                            else
                            {
                                hasEditedDoc |= CropViewViewModel.RemoveUserCropFromPage(page);
                            }
                        }
                    }

                    _currentAutomaticPage++;
                    iter2.Next();
                }
            }
            catch (Exception e)
            {
                PDFNetException pdfe = new PDFNetException(e.HResult);
                if (!pdfe.IsPDFNetException)
                    throw;
            }
            finally
            {
                pdfDoc.UnlockRead();
                IsAutomaticCropping = false;
                IsPopupOpen = false;
                _tokenSource.Dispose();

                if (!_isAutomaticCancel)
                {
                    _pdfViewCtrl.UpdatePageLayout();
                    if (hasEditedDoc)
                        _documentEdited?.Invoke();
                }
                else
                {
                    _isAutomaticCancel = false;
                }
            }
        }
    }
}
