﻿using System.Collections.Generic;
using System.Threading.Tasks;

using Windows.ApplicationModel.Resources;
using Windows.UI.Xaml.Controls;

using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Utilities;


namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer
{
    public partial class ViewerViewModel
    {
        public async Task SavePermCroppedCopyAsync()
        {
            AppCenterHandlerBase.CURRENT.SendEvent("Export", new Dictionary<string, string>() { ["Command"] = "Cropped copy" });

            TabControlViewModel?.SelectedTab?.ToolManager?.CloseOpenDialog();
            IsModal = true;
            await SavePermCroppedCopyImplAsync();
        }

        private async Task SavePermCroppedCopyImplAsync()
        {
            bool isCropped = await Task.Run(() => GetUserCropBoxes());

            if (!isCropped)
            {
                var loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                string message = loader.GetString("ViewerPage_PermCropSave_Error");
                string title = loader.GetString("ViewerPage_PermCropSave_Error_Title");

                ContentDialog dialog = new ContentDialog
                {
                    Title = title,
                    Content = message,
                    PrimaryButtonText = loader.GetString("ViewerPage_PermCropSave_Error_CropNow"),
                    CloseButtonText = loader.GetString("Generic_Cancel_Text")
                    
                };

                dialog.PrimaryButtonCommand = new RelayCommand((_) =>
                {
                    IsModal = true;
                    CropPopupViewModel.IsPopupOpen = true;
                });

                dialog.CloseButtonCommand = new RelayCommand((_) =>
                {
                    IsModal = false;
                });

                await DialogHelper.ShowContentDialogAsync(dialog);
            }
            else
            {
                await SaveAsAsync(TabControlViewModel.SelectedTab, false, false, SaveAsOption.Crop);
            }
        }
    }
}
