﻿using pdftron.PDF.Tools.Controls.ControlBase;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels;
using System;
using System.Threading.Tasks;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;

namespace pdftron.PDF.Tools.Controls
{
    public enum AnnotationToolbarOrientation
    {
        Horizontal,
        Vertical,
    }

    public class AnnotationToolbarBase : UserControl, ICloseableControl
    {

        #region Tools Visibility Control Depedency Properties

        /// <summary>
        /// Allows user to externally control whether the Sticky Note Tool button is visible or collapsed
        /// </summary>
        public bool IsStickyNoteToolButtonVisible
        {
            get => ViewModel.IsStickyNoteToolButtonVisible;
            set => ViewModel.IsStickyNoteToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Highlight Text Tool button is visible or collapsed
        /// </summary>
        public bool IsHighlightToolButtonVisible
        {
            get => ViewModel.IsHighlightToolButtonVisible;
            set => ViewModel.IsHighlightToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Underline Text Tool button is visible or collapsed
        /// </summary>
        public bool IsUnderlineToolButtonVisible
        {
            get => ViewModel.IsUnderlineToolButtonVisible;
            set => ViewModel.IsUnderlineToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Strikeout Text Tool button is visible or collapsed
        /// </summary>
        public bool IsStrikeoutToolButtonVisible
        {
            get => ViewModel.IsStrikeoutToolButtonVisible;
            set => ViewModel.IsStrikeoutToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Squiggly Text Tool button is visible or collapsed
        /// </summary>
        public bool IsSquigglyToolButtonVisible
        {
            get => ViewModel.IsSquigglyToolButtonVisible;
            set => ViewModel.IsSquigglyToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Stamper Tool button is visible or collapsed
        /// </summary>
        public bool IsStamperToolButtonVisible
        {
            get => ViewModel.IsStamperToolButtonVisible;
            set => ViewModel.IsStamperToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Signature Tool button is visible or collapsed
        /// </summary>
        public bool IsSignatureToolButtonVisible
        {
            get => ViewModel.IsSignatureToolButtonVisible;
            set => ViewModel.IsSignatureToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Ink Tool button is visible or collapsed
        /// </summary>
        public bool IsInkToolButtonVisible
        {
            get => ViewModel.IsInkToolButtonVisible;
            set => ViewModel.IsInkToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Eraser Tool button is visible or collapsed
        /// </summary>
        public bool IsEraserToolButtonVisible
        {
            get => ViewModel.IsEraserToolButtonVisible;
            set => ViewModel.IsEraserToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Marker Tool button is visible or collapsed
        /// </summary>
        public bool IsMarkerToolButtonVisible
        {
            get => ViewModel.IsMarkerToolButtonVisible;
            set => ViewModel.IsMarkerToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Text Box Tool button is visible or collapsed
        /// </summary>
        public bool IsTextBoxToolButtonVisible
        {
            get => ViewModel.IsTextBoxToolButtonVisible;
            set => ViewModel.IsTextBoxToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Shape Tools button is visible or collapsed
        /// </summary>
        public bool IsShapeToolsButtonVisible
        {
            get => ViewModel.IsShapeToolsButtonVisible;
            set => ViewModel.IsShapeToolsButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Pan Tool button is visible or collapsed
        /// </summary>
        public bool IsPanToolsButtonVisible
        {
            get => ViewModel.IsPanToolsButtonVisible;
            set => ViewModel.IsPanToolsButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Multi-Select Rectangular Tool button is visible or collapsed
        /// </summary>
        public bool IsMultiSelectToolButtonVisible
        {
            get => ViewModel.IsMultiSelectToolButtonVisible;
            set => ViewModel.IsMultiSelectToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Multi-Select Lasso Tool button is visible or collapsed
        /// </summary>
        public bool IsLassoToolButtonVisible
        {
            get => ViewModel.IsLassoToolButtonVisible;
            set => ViewModel.IsLassoToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Undo and Redo buttons are visible or collapsed
        /// </summary>
        public bool IsUndoRedoToolButtonVisible
        {
            get => ViewModel.IsUndoRedoToolButtonVisible;
            set => ViewModel.IsUndoRedoToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Square Tool button is visible or collapsed
        /// </summary>
        public bool IsSquareToolButtonVisible
        {
            get => ViewModel.IsSquareToolButtonVisible;
            set => ViewModel.IsSquareToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Circle Tool button is visible or collapsed
        /// </summary>
        public bool IsCircleToolButtonVisible
        {
            get => ViewModel.IsCircleToolButtonVisible;
            set => ViewModel.IsCircleToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Line Tool button is visible or collapsed
        /// </summary>
        public bool IsLineToolButtonVisible
        {
            get => ViewModel.IsLineToolButtonVisible;
            set => ViewModel.IsLineToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Arrow Tool button is visible or collapsed
        /// </summary>
        public bool IsArrowToolButtonVisible
        {
            get => ViewModel.IsArrowToolButtonVisible;
            set => ViewModel.IsArrowToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Measurement Tools button is visible or collapsed
        /// </summary>
        public bool IsMeasureToolsButtonVisible
        {
            get => ViewModel.IsMeasureToolsButtonVisible;
            set => ViewModel.IsMeasureToolsButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Ruler Tool button is visible or collapsed
        /// </summary>
        public bool IsRulerToolButtonVisible
        {
            get => ViewModel.IsRulerToolButtonVisible;
            set => ViewModel.IsRulerToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Redaction Tools button is visible or collapsed
        /// </summary>
        public bool IsRedactionToolButtonVisible
        {
            get => ViewModel.IsRedactionToolButtonVisible;
            set => ViewModel.IsRedactionToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Text Redaction Tool button is visible or collapsed
        /// </summary>
        public bool IsRedactionTextToolButtonVisible
        {
            get => ViewModel.IsRedactionTextToolButtonVisible;
            set => ViewModel.IsRedactionTextToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Area Redaction Tool button is visible or collapsed
        /// </summary>
        public bool IsRedactionAreaToolButtonVisible
        {
            get => ViewModel.IsRedactionAreaToolButtonVisible;
            set => ViewModel.IsRedactionAreaToolButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Close Toolbar button is visible or collapsed
        /// </summary>
        public bool IsCloseToolbarButtonVisible
        {
            get => ViewModel.IsCloseToolbarButtonVisible;
            set => ViewModel.IsCloseToolbarButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Expand Toolbar button is visible or collapsed
        /// </summary>
        public bool IsExpandButtonVisible
        {
            get => ViewModel.IsExpandButtonVisible;
            set => ViewModel.IsExpandButtonVisible = value;
        }

        /// <summary>
        /// Allows user to externally control whether the Perimeter Tool button is visible or collapsed
        /// </summary>
        private static readonly DependencyProperty IsPerimeterToolButtonVisibleProperty = DependencyProperty.Register(
            "IsPerimeterToolButtonVisible",
            typeof(bool),
            typeof(AnnotationCommandBar),
            new PropertyMetadata(false));

        /// <summary>
        /// Allows user to externally control whether the Perimeter Tool button is visible or collapsed
        /// </summary>
        public bool IsPerimeterToolButtonVisible
        {
            get => (bool)GetValue(IsPerimeterToolButtonVisibleProperty);
            set
            {
                SetValue(IsPerimeterToolButtonVisibleProperty, value);
                ViewModel.IsPerimeterToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Polyline button is visible or collapsed (since view model is not exposed)
        /// </summary>
        public static readonly DependencyProperty IsPolylineToolButtonVisibleProperty = DependencyProperty.Register(
            "IsPolylineToolButtonVisible",
            typeof(bool),
            typeof(AnnotationCommandBar),
            new PropertyMetadata(false));

        /// <summary>
        /// Allows user to externally control whether the Polyline button is visible or collapsed
        /// </summary>
        public bool IsPolylineToolButtonVisible
        {
            get => (bool)GetValue(IsPolylineToolButtonVisibleProperty);
            set
            {
                SetValue(IsPolylineToolButtonVisibleProperty, value);
                ViewModel.IsPolylineToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Polygon button is visible or collapsed (since view model is not exposed)
        /// </summary>
        public static readonly DependencyProperty IsPolygonToolButtonVisibleProperty = DependencyProperty.Register(
            "IsPolygonToolButtonVisible",
            typeof(bool),
            typeof(AnnotationCommandBar),
            new PropertyMetadata(false));

        /// <summary>
        /// Allows user to externally control whether the Polygon button is visible or collapsed
        /// </summary>
        public bool IsPolygonToolButtonVisible
        {
            get => (bool)GetValue(IsPolygonToolButtonVisibleProperty);
            set
            {
                SetValue(IsPolygonToolButtonVisibleProperty, value);
                ViewModel.IsPolygonToolButtonVisible = value;
            }
        }

        /// <summary>
        /// Allows user to externally control whether the Cloud button is visible or collapsed (since view model is not exposed)
        /// </summary>
        public static readonly DependencyProperty IsCloudToolButtonVisibleProperty = DependencyProperty.Register(
            "IsCloudToolButtonVisible",
            typeof(bool),
            typeof(AnnotationCommandBar),
            new PropertyMetadata(false));

        /// <summary>
        /// Allows user to externally control whether the Cloud button is visible or collapsed
        /// </summary>
        public bool IsCloudToolButtonVisible
        {
            get => (bool)GetValue(IsCloudToolButtonVisibleProperty);
            set
            {
                SetValue(IsCloudToolButtonVisibleProperty, value);
                ViewModel.IsCloudToolButtonVisible = value;
            }
        }

        public bool IsMarqueeZoomToolButtonVisible
        {
            get => ViewModel.IsMarqueeZoomToolButtonVisible;
            set => ViewModel.IsMarqueeZoomToolButtonVisible = value;
        }

        #endregion

        #region Extra Depedency Properties

        public static readonly DependencyProperty OrientationProperty = DependencyProperty.RegisterAttached(
            "Orientation",
            typeof(AnnotationToolbarOrientation),
            typeof(AnnotationToolbarBase),
            new PropertyMetadata(null, OnOrientationPropertyChanged));

        public static void SetOrientation(DependencyObject d, AnnotationToolbarOrientation value) => d.SetValue(OrientationProperty, value);

        public static AnnotationToolbarOrientation GetOrientation(DependencyObject d) => (AnnotationToolbarOrientation)d.GetValue(OrientationProperty);

        private static void OnOrientationPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            AnnotationToolbarOrientation oldValue = (AnnotationToolbarOrientation)e.OldValue;
            AnnotationToolbarOrientation newValue = (AnnotationToolbarOrientation)e.NewValue;
            if (newValue != oldValue)
            {

            }
        }

        public static readonly DependencyProperty SmallScreenLayoutProperty = DependencyProperty.RegisterAttached(
            "SmallScreenLayout",
            typeof(bool),
            typeof(AnnotationToolbarBase),
            new PropertyMetadata(false));

        public bool SmallScreenLayout
        {
            get => (bool)GetValue(SmallScreenLayoutProperty);
            set => SetValue(SmallScreenLayoutProperty, value);
        }

        #endregion

        #region Public Properties
        public AnnotationToolbarBase(AnnotationToolbarViewModel viewModel)
        {
            if (viewModel != null)
                ViewModel = viewModel;

            if (ViewModel == null)
                ViewModel =  new AnnotationToolbarViewModel(this);

            DataContext = ViewModel;
            SizeChanged += AnnotationCommandBar_SizeChanged;
        }

        public AnnotationToolbarViewModel ViewModel
        {
            get => (AnnotationToolbarViewModel)DataContext;
            set => DataContext = value;
        }

        /// <summary>
        /// Gets or sets whether or not the buttons stay down once pressed.
        /// When set to true, the toolbar will not switch back to the pan tool after each
        /// annotation is drawn.
        /// </summary>
        public ToolManager ToolManager
        {
            get => ViewModel.ToolManager;
            set => ViewModel.ToolManager = value;
        }

        /// <summary>
        /// Gets or sets whether or not the buttons stay down once pressed.
        /// When set to true, the toolbar will not switch back to the pan tool after each
        /// annotation is drawn.
        /// </summary>
        public bool ButtonsStayDown
        {
            get => ViewModel.ButtonsStayDown;
            set => ViewModel.ButtonsStayDown = value;
        }

        /// <summary>
        /// Gets or sets wether the Ink Muilti-Stroke option is enabled.
        /// It will lock the PDF Viewer to the selected page and enable undo, redo and save options
        /// </summary>
        public bool InkMultiStroke
        {
            get => ViewModel.InkMultiStroke;
            set => ViewModel.InkMultiStroke = value;
        }

        #endregion

        /// <summary>
        /// Occurs when the control has been requested to close usually by pressing the close button
        /// </summary>
        public event ControlClosedDelegate ControlClosed = delegate { };

        /// <summary>
        /// Perform the closing of the annotation toolbar
        /// </summary>
        public void CloseControl()
        {
            ViewModel.IsMainAppBarOpen = false;
            ViewModel.ShowInkToolbar = false;
            ViewModel.UpdateCurrentToolTo(ToolType.e_pan);
            ControlClosed();
        }

        protected bool IsButtonHidden(ButtonBase currButtonBase)
        {
            return (currButtonBase.Name == "AnnotStickyNoteButton" && !ViewModel.IsStickyNoteToolButtonVisible)
                || (currButtonBase.Name == "AnnotHighlightButton" && !ViewModel.IsHighlightToolButtonVisible)
                || (currButtonBase.Name == "AnnotUnderlineButton" && !ViewModel.IsUnderlineToolButtonVisible)
                || (currButtonBase.Name == "AnnotStrikeoutButton" && !ViewModel.IsStrikeoutToolButtonVisible)
                || (currButtonBase.Name == "AnnotSquigglyButton" && !ViewModel.IsSquigglyToolButtonVisible)
                || (currButtonBase.Name == "AnnotStamperButton" && !ViewModel.IsStamperToolButtonVisible)
                || (currButtonBase.Name == "AnnotSignatureButton" && !ViewModel.IsSignatureToolButtonVisible)
                || (currButtonBase.Name == "AnnotFreeHandButton" && !ViewModel.IsInkToolButtonVisible)
                || (currButtonBase.Name == "AnnotEraserButton" && !ViewModel.IsEraserToolButtonVisible)
                || (currButtonBase.Name == "AnnotMarkerButton" && !ViewModel.IsMarkerToolButtonVisible)
                || (currButtonBase.Name == "AnnotFreeTextButton" && !ViewModel.IsTextBoxToolButtonVisible)
                || (currButtonBase.Name == "AnnotPolygonButton" && !ViewModel.IsShapeToolsButtonVisible)
                || (currButtonBase.Name == "AnnotGroupCreateButton" && !ViewModel.IsMultiSelectToolButtonVisible)
                || (currButtonBase.Name == "AnnotLassoCreateButton" && !ViewModel.IsLassoToolButtonVisible)
                || (currButtonBase.Name == "AnnotPanButton" && !ViewModel.IsPanToolsButtonVisible)
                || (currButtonBase.Name == "UndoButton" && !ViewModel.IsUndoRedoToolButtonVisible)
                || (currButtonBase.Name == "RedoButton" && !ViewModel.IsUndoRedoToolButtonVisible)
                || (currButtonBase.Name == "AnnotSquareButton" && !ViewModel.IsSquareToolButtonVisible)
                || (currButtonBase.Name == "AnnotCircleButton" && !ViewModel.IsCircleToolButtonVisible)
                || (currButtonBase.Name == "AnnotLineButton" && !ViewModel.IsLineToolButtonVisible)
                || (currButtonBase.Name == "AnnotArrowButton" && !ViewModel.IsArrowToolButtonVisible)
                || (currButtonBase.Name == "AnnotPolylineButton" && !ViewModel.IsPolylineToolButtonVisible)
                || (currButtonBase.Name == "AnnotPolygonButton2" && !ViewModel.IsPolygonToolButtonVisible)
                || (currButtonBase.Name == "AnnotCloudButton" && !ViewModel.IsCloudToolButtonVisible)
                || (currButtonBase.Name == "AnnotZoomButton" && !ViewModel.IsMarqueeZoomToolButtonVisible)
                || (currButtonBase.Name == "AnnotMeasureButton" && !ViewModel.IsMeasureToolsButtonVisible)
                || (currButtonBase.Name == "MeasureRulerTool" && !ViewModel.IsRulerToolButtonVisible)
                || (currButtonBase.Name == "MeasurePerimeterTool" && !ViewModel.IsPerimeterToolButtonVisible)
                || (currButtonBase.Name == "MeasureAreaTool" && !ViewModel.IsAreaToolButtonVisible)
                || (currButtonBase.Name == "AnnotRedactionButton" && !ViewModel.IsRedactionToolButtonVisible)
                || (currButtonBase.Name == "TextRedactionTool" && !ViewModel.IsRedactionTextToolButtonVisible)
                || (currButtonBase.Name == "RectRedactionTool" && !ViewModel.IsRedactionAreaToolButtonVisible)
                || (currButtonBase.Name == "ToolbarCloseButton" && !ViewModel.IsCloseToolbarButtonVisible)
                || (currButtonBase.Name == "ExpandButton" && !ViewModel.IsExpandButtonVisible);
        }

        public event ControlClosedDelegate ControlAppBarOpened = delegate { };

        #region resizing

        protected double MINIMUM_BUTTON_WIDTH = 60;
        protected double MINIMUM_BUTTON_WIDTH_SMALL_SCREEN = 50;
        protected ButtonBase[] ToolbarButtonRanking;
        protected System.Collections.Generic.List<ButtonBase> ResizableInkButtons;
        protected System.Collections.Generic.List<ButtonBase> ResizablePolygonSaveButtons;
        protected double ExtraPaddingForInkToolbar;
        protected double ExtraPaddingForPolygonSaveToolbar;
        protected virtual void AnnotationCommandBar_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            ResizeButtons(e.NewSize);
        }

        protected virtual void ResizeButtons(Windows.Foundation.Size newSize)
        {
            if (ToolbarButtonRanking == null)
                return;

            int numButtons = ToolbarButtonRanking.Length;
            double minButtonWidth = MINIMUM_BUTTON_WIDTH;
            if (SmallScreenLayout)
                minButtonWidth = MINIMUM_BUTTON_WIDTH_SMALL_SCREEN;

            int maxButtons = Math.Min((int)((newSize.Width) / minButtonWidth), numButtons);

            double buttonwidth = Math.Round(newSize.Width / maxButtons); // Note: avoid infinite loop cycle on new size
            double remainder = 0;

            int vis_buttons = 0;
            for (int i = 0; i < numButtons; i++)
            {
                double trunkWidth = (int)buttonwidth;
                remainder += buttonwidth - trunkWidth;
                if (remainder >= 1)
                {
                    ++trunkWidth;
                    --remainder;
                }
                var currButtonBase = ToolbarButtonRanking[i];
                if (currButtonBase != null)
                {
                    var tg = currButtonBase.Tag as string;
                    if (!IsButtonHidden(currButtonBase))
                    {
                        if (vis_buttons < maxButtons)
                        {
                            currButtonBase.Visibility = Visibility.Visible;
                            vis_buttons++;
                        }
                        else
                        {
                            currButtonBase.Visibility = Visibility.Collapsed;
                        }
                    }
                    currButtonBase.Width = (int)trunkWidth;
                }
            }
            if (ResizableInkButtons != null)
            {
                int numInkButtons = 6;
                if (!SmallScreenLayout)
                    numInkButtons = 8;

                double totalAvailableWidth = Math.Max(100, newSize.Width - ExtraPaddingForInkToolbar);
                double maxButtonWidth = totalAvailableWidth / numInkButtons;
                if (buttonwidth > maxButtonWidth)
                    buttonwidth = maxButtonWidth;

                foreach (ButtonBase inkButton in ResizableInkButtons)
                {
                    inkButton.Width = buttonwidth;
                }
            }
            if (ResizablePolygonSaveButtons != null)
            {
                int numPolygonSaveButtons = 5;
                double totalAvailableWidth = Math.Max(100, newSize.Width - ExtraPaddingForPolygonSaveToolbar);
                double maxButtonWidth = totalAvailableWidth / numPolygonSaveButtons;
                if (buttonwidth > maxButtonWidth)
                    buttonwidth = maxButtonWidth;

                foreach (ButtonBase polygonSaveButton in ResizablePolygonSaveButtons)
                {
                    polygonSaveButton.Width = buttonwidth;
                }
            }
        }

        #endregion resizing
    }
}
