﻿using pdftron.PDF.Tools.Controls.ViewModels;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;


namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class CreateDocControl : ContentDialog, IDisposable
    {
        private CreateDocControlViewModel _ViewModel;

        #region Constructors
        public CreateDocControl()
        {
            this.InitializeComponent();

            _ViewModel = new CreateDocControlViewModel();
            DataContext = _ViewModel;
            SetRefViewModels(ref _ViewModel);

            _ViewModel.NewDocumentCreated += RaiseNewDocumentCreated;
            _ViewModel.DocumentCreationCancelled += RaiseDocumentCreationCancelled;
            _ViewModel.DocumentCreationFailed += RaiseDocumentCreationFailed;
            _ViewModel.ImageLoaded += _ViewModel_ImageLoaded;
            _ViewModel.ImageLoadingFailed += _ViewModel_ImageLoadingFailed;

            CloseButtonClick += CreateDocControl_CloseButtonClick;
        }
        #endregion

        #region Public Methods
        public void Dispose()
        {
            if (_ViewModel != null)
            {
                _ViewModel.NewDocumentCreated -= RaiseNewDocumentCreated;
                _ViewModel.DocumentCreationCancelled -= RaiseDocumentCreationCancelled;
                _ViewModel.DocumentCreationFailed -= RaiseDocumentCreationFailed;
                CloseButtonClick -= CreateDocControl_CloseButtonClick;
                _ViewModel = null;
            }
        }
        #endregion

        #region Private Methods
        private void NavigationView_SelectionChanged(Microsoft.UI.Xaml.Controls.NavigationView sender, Microsoft.UI.Xaml.Controls.NavigationViewSelectionChangedEventArgs args)
        {
            Microsoft.UI.Xaml.Controls.NavigationViewItem item = args.SelectedItem as Microsoft.UI.Xaml.Controls.NavigationViewItem;
            if (item == null || _ViewModel == null)
                return;

            string tag = item.Tag.ToString();
            if (tag == string.Empty)
                return;

            if (tag == "Blank")
            {
                _ViewModel.DocumentType = CreateDocControlViewModel.CreateDocFrom.Blank;
            }
            else if (tag == "Image")
            {
                _ViewModel.DocumentType = CreateDocControlViewModel.CreateDocFrom.Image;
            }
        }

        private void SetRefViewModels(ref CreateDocControlViewModel viewModel)
        {
            if (BlankDocControl != null)
            {
                BlankDocControl.ViewModel = viewModel;
            }
            if (ImageDocControl != null)
            {
                ImageDocControl.ViewModel = viewModel;
            }
        }

        private void CreateDocControl_CloseButtonClick(ContentDialog sender, ContentDialogButtonClickEventArgs args)
        {
            RaiseDocumentCreationCancelled();
        }

        private void RaiseDocumentCreationFailed()
        {
            if (_DocumentCreationFailed != null)
            {
                _DocumentCreationFailed.Invoke();
            }
        }

        private void RaiseNewDocumentCreated(Windows.Storage.StorageFile file)
        {
            if (_NewDocumentCreated != null)
            {
                _NewDocumentCreated.Invoke(file);
            }
        }

        private void RaiseDocumentCreationCancelled()
        {
            if (_DocumentCreationCancelled != null)
                _DocumentCreationCancelled.Invoke();
        }

        private void _ViewModel_ImageLoadingFailed()
        {
            _imageLoadingFailed?.Invoke();
        }

        private void _ViewModel_ImageLoaded(Windows.Storage.StorageFile file)
        {
            _imageLoaded?.Invoke(file);
        }
        #endregion

        #region Events

        private CreateDocControlViewModel.NewDocumentCreatedDelegate _NewDocumentCreated;
        /// <summary>
        /// Raised when the document is successfully created
        /// </summary>
        public event CreateDocControlViewModel.NewDocumentCreatedDelegate NewDocumentCreated
        {
            add
            {
                if (_NewDocumentCreated?.GetInvocationList().Contains(value) != true)
                    _NewDocumentCreated += value;
            }
            remove => _NewDocumentCreated -= value;
        }

        private CreateDocControlViewModel.DocumentCreationFailedDelegate _DocumentCreationFailed;
        /// <summary>
        /// Raised when new document creation fails
        /// </summary>
        public event CreateDocControlViewModel.DocumentCreationFailedDelegate DocumentCreationFailed
        {
            add
            {
                if (_DocumentCreationFailed?.GetInvocationList().Contains(value) != true)
                    _DocumentCreationFailed += value;
            }
            remove => _DocumentCreationFailed -= value;
        }

        private CreateDocControlViewModel.DocumentCreationCancelledDelegate _DocumentCreationCancelled;
        /// <summary>
        /// Raised when new document creation is cancelled
        /// </summary>
        public event CreateDocControlViewModel.DocumentCreationCancelledDelegate DocumentCreationCancelled
        {
            add
            {
                if (_DocumentCreationCancelled?.GetInvocationList().Contains(value) != true)
                    _DocumentCreationCancelled += value;
            }
            remove => _DocumentCreationCancelled -= value;
        }

        private CreateDocControlViewModel.ImageLoadedDelegate _imageLoaded;
        /// <summary>
        /// Raised when an image has been loaded in the document creation dialog
        /// </summary>
        public event CreateDocControlViewModel.ImageLoadedDelegate ImageLoaded
        {
            add
            {
                if (_imageLoaded?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _imageLoaded += value;
            }
            remove => _imageLoaded -= value;
        }

        private CreateDocControlViewModel.ImageLoadingFailedDelegate _imageLoadingFailed;
        /// <summary>
        /// Raised when an image failed to during loading in the document creation dialog
        /// </summary>
        public event CreateDocControlViewModel.ImageLoadingFailedDelegate ImageLoadingFailed
        {
            add
            {
                if (_imageLoadingFailed?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _imageLoadingFailed += value;
            }
            remove => _imageLoadingFailed -= value;
        }

        #endregion
    }
}