﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;

namespace pdftron.PDF.Tools.Controls
{
    public class FlipViewCustom : FlipView
    {

        #region Private Properties
        private static List<string> ButtonNameList = new List<string>
        {
            "PreviousButtonHorizontal",
            "NextButtonHorizontal",
            "PreviousButtonVertical",
            "NextButtonVertical",
        };
        #endregion

        #region Constructors
        public FlipViewCustom()
        {
        }
        #endregion

        #region Public Properties
        public static DependencyProperty ShowFlipViewButtonsProperty = DependencyProperty.Register(
            "ShowFlipViewButtons",
            typeof(bool),
            typeof(FlipViewCustom),
            new PropertyMetadata(true, OnShowFlipViewButtonsPropertyChanged)
        );

        public bool ShowFlipViewButtons
        {
            get { return (bool)GetValue(ShowFlipViewButtonsProperty); }
            set { SetValue(ShowFlipViewButtonsProperty, value); }
        }
        #endregion

        #region Private Methods
        private static void OnShowFlipViewButtonsPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is FlipViewCustom flipView)
            {
                if (flipView != null)
                    flipView.AdjustButtonVisibility();
            }
        }

        private void AdjustButtonVisibility()
        {
            if (!IsLoaded)
            {
                //Can't get the children of this element if they havent been loaded.
                return;
            }

            foreach (var item in ButtonNameList)
            {
                Button b = GetTemplateChild(item) as Button;
                if (b != null)
                {
                    b.IsHitTestVisible = ShowFlipViewButtons;
                    if (ShowFlipViewButtons)
                    {
                        b.Opacity = 1;
                    }
                    else
                    {
                        b.Opacity = 0;
                    }
                }
            }
        }

        protected override void OnApplyTemplate()
        {
            //Need to call OnApplyTemplate on the base class first
            //otherwise the buttons within this class will not fire click events
            base.OnApplyTemplate();
            AdjustButtonVisibility();
        }
        #endregion
    }
}
