﻿using pdftron.PDF.Tools.Controls.ControlBase;
using pdftron.PDF.Tools.Controls.ViewModels.View;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;
using Windows.ApplicationModel.DataTransfer;
using Windows.ApplicationModel.Resources;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Storage;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Navigation;
using pdftron.PDF.Tools.Utilities;

// The User Control item template is documented at https://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls
{
    public sealed partial class ViewPdfControl : UserControl, ControlBase.ICloseableControl, IMultiWindowUserControl
    {
        ResourceLoader _loader;
        string mDraggingFeedback; // cached text for dragging feedback

        public ViewPdfControl()
        {
            this.InitializeComponent();
            this.DataContext = ViewModel = new ViewPdfViewModel();

            Init();
        }

        /// <summary>
        /// Initialize some additional resources
        /// </summary>
        private void Init()
        {
            _loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
            browsePdf.Text = mDraggingFeedback = _loader.GetString("PDFCompareControl_browse"); // NOTE: reusing PDF Compare strings
            viewPdfDropDesc.Text = _loader.GetString("PDFCompareControl_dropFileDescription"); // NOTE: reusing PDF Compare strings
            SaveDocToolTip.Content = _loader.GetString("Viewerpage_DocumentModifiedDialog_Created_Title");
            SaveAsDocToolTip.Content = _loader.GetString("SaveBeforeClosingDialog_SaveAs_Option");
        }

        public ViewPdfViewModel ViewModel { get; set; }

        public event ControlBase.ControlClosedDelegate ControlClosed = delegate { };

        public void CloseControl()
        {
            ViewModel.CleanUp();
            ControlClosed();
        }

        private async void Pdf_OpenFile_DragOver(object sender, DragEventArgs e)
        {
            await HandleDragAndDropOperation(e, false, mDraggingFeedback);
        }

        private async void Pdf_OpenFile_Drop(object sender, DragEventArgs e)
        {
            await HandleDragAndDropOperation(e, true);
        }

        /// <summary>
        /// Handle drag and drop operations
        /// </summary>
        /// <param name="e"></param>
        /// <param name="isDrop">True if dropping, otherwise dragging</param>
        /// <param name="isPdfA">True if Pdf A / left side, otherwise Pdf B / right side</param>
        /// <param name="drag_caption">Feedback when dragging PDF</param>
        /// <returns></returns>
        private async Task HandleDragAndDropOperation(DragEventArgs e, bool isDrop, string drag_caption = null)
        {
            DragOperationDeferral deferral = e.GetDeferral();

            if (e.DataView.Contains(StandardDataFormats.StorageItems))
            {
                IReadOnlyList<IStorageItem> items = await e.DataView.GetStorageItemsAsync();
                if (items?.Count == 1)
                {
                    if (items[0] is StorageFile storageFile && storageFile.FileType.Equals(".pdf", StringComparison.OrdinalIgnoreCase))
                    {
                        if (isDrop)
                        {
                            if (ViewModel != null)
                            {
                                 await ViewModel.OpenFileAsync(storageFile);
                            }
                        }
                        else
                        {
                            e.AcceptedOperation = DataPackageOperation.Copy;

                            if (string.IsNullOrEmpty(drag_caption) == false)
                                e.DragUIOverride.Caption = drag_caption;
                        }
                    }
                    else
                    {
                        e.AcceptedOperation = DataPackageOperation.None;
                    }
                }
            }

            deferral.Complete();
        }
    }
}
