﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using static pdftron.PDF.Tools.UtilityFunctions;
using UIRect = Windows.Foundation.Rect;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    /// <summary>
    /// This is for tracking what Style Properties are updated
    /// </summary>
    public enum AnnotStyleType
    {
        MainColor,
        FillColor,
        Thickness,
        Opacity,
        FontSize,
        FontName,
        Icon,
        FontFormat,
        FormatText
    }

    #region Preset Colors

    public class AnnotationPropertiesPresetsColorGridViewModel : ViewModelBase
    {
        public static string SegoeRemoveString = new string((char)0xE108, 1);
        public static string SegoeAddString = new string((char)0xE109, 1);
        public static string SegoeAcceptString = new string((char)0xE10B, 1);

        private static int MAX_PRESETS_ROWS = 5;

        #region Preset Color Item

        public class PresetItem : ViewModelBase
        {
            public bool IsColorOption { get; private set; }
            public bool IsEmptyOption { get; private set; }
            public bool IsAddRemoveButton { get { return IsAddButton || IsRemoveButton; } }
            public SolidColorBrush ColorBrush { get; private set; }
            public Grid CheckPatternGrid { get; private set; }

            public PresetItem MySelf { get; private set; }

            private string _AddOrRemoveContent;
            public string AddOrRemoveContent
            {
                get { return _AddOrRemoveContent; }
                set
                {
                    if (value != null || _AddOrRemoveContent != null)
                    {
                        if (!string.Equals(_AddOrRemoveContent, value))
                        {
                            _AddOrRemoveContent = value;
                            RaisePropertyChanged();
                        }
                    }
                }
            }

            public bool IsAddButton { get; private set; }
            public bool IsRemoveButton { get; private set; }

            private bool _ShowDelete = false;
            public bool ShowDelete
            {
                get { return _ShowDelete; }
                set
                {
                    if (_ShowDelete != value)
                    {
                        _ShowDelete = value;
                        RaisePropertyChanged();
                        RaisePropertyChanged("ShowSelected");
                    }
                }
            }

            private bool _IsSelected = false;
            public bool IsSelected
            {
                get { return _IsSelected; }
                set
                {
                    if (_IsSelected != value)
                    {
                        _IsSelected = value;
                        RaisePropertyChanged();
                        RaisePropertyChanged("ShowSelected");
                    }
                }
            }

            public bool ShowSelected
            {
                get
                {
                    return IsSelected && !ShowDelete;
                }
            }

            public PresetItem(Color color)
            {
                Init(color, false, false, false);
            }

            public PresetItem(Color color, bool isEmpty, bool isAdd, bool isRemove)
            {
                Init(color, isEmpty, isAdd, isRemove);
            }

            private void Init(Color color, bool isEmpty, bool isAdd, bool isRemove)
            {
                MySelf = this;
                IsColorOption = false;
                IsEmptyOption = false;
                IsAddButton = false;
                IsRemoveButton = false;
                ColorBrush = new SolidColorBrush(Colors.Transparent);
                CheckPatternGrid = new Grid();
                AddOrRemoveContent = "";

                if (isAdd)
                {
                    IsAddButton = true;
                    AddOrRemoveContent = AnnotationPropertiesPresetsColorGridViewModel.SegoeAddString;
                }
                else if (isRemove)
                {
                    IsRemoveButton = true;
                    AddOrRemoveContent = AnnotationPropertiesPresetsColorGridViewModel.SegoeRemoveString;
                }
                else if (isEmpty)
                {
                    IsEmptyOption = true;
                    //CheckPatternGrid.Children.Add(CreateEmptyGrid());
                }
                else
                {
                    IsColorOption = true;
                    ColorBrush = new SolidColorBrush(color);
                }
            }
        }

        #endregion Preset Color Item

        public delegate void ColorSelectedDelegate(Color color, bool isEmpty);
        public event ColorSelectedDelegate ColorSelectedEvent = delegate { };
        public delegate void RequestNewColorDelegate(IList<Color> existingColors);
        public event RequestNewColorDelegate RequestNewColor = delegate { };

        public AnnotationPropertiesPresetsColorGridViewModel(ToolType toolType, bool canBeEmpty, int maxColumns) : base()
        {
            _ColorItems = new ObservableCollection<PresetItem>();
            Init(toolType, canBeEmpty, maxColumns);
        }

        private ObservableCollection<PresetItem> _ColorItems = null;
        public ObservableCollection<PresetItem> PresetsColors
        {
            get { return _ColorItems; }
            private set
            {
                _ColorItems = value;
                RaisePropertyChanged();
            }
        }

        private int _PresetsMaxColumns = 5;
        public int PresetsMaxColumns
        {
            get { return _PresetsMaxColumns; }
            set
            {
                if (value != _PresetsMaxColumns)
                {
                    _PresetsMaxColumns = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _ShowAddRemoveButtons = true;
        /// <summary>
        /// Toggle displaying Show and remove buttons
        /// </summary>
        public bool ShowAddRemoveButtons
        {
            get { return _ShowAddRemoveButtons; }
            private set
            {
                if (ShowAddRemoveButtons != value)
                {
                    _ShowAddRemoveButtons = value;
                    RaisePropertyChanged();
                    ResolveAddRemoveButtons(_PresetsMaxColumns * MAX_PRESETS_ROWS);
                }
            }
        }

        private bool _CanBeEmpty = false;
        private ToolType _ToolType = ToolType.e_none;

        private void Init(ToolType toolType, bool canBeEmpty, int maxColumns)
        {
            PresetsMaxColumns = maxColumns;
            _CanBeEmpty = canBeEmpty;
            _ToolType = toolType;
            IList<PresetItem> items = GetItems();
            foreach (PresetItem item in items)
            {
                PresetsColors.Add(item);
            }
            if (canBeEmpty)
            {
                PresetsColors.Add(new PresetItem(Colors.Transparent, true, false, false));
            }
            ResolveAddRemoveButtons(_PresetsMaxColumns * MAX_PRESETS_ROWS);      
            //PresetsColors.Add(new PresetItem(Colors.Transparent, false, true, false));
            //PresetsColors.Add(new PresetItem(Colors.Transparent, false, false, true));

            ItemClickCommand = new RelayCommand(ItemClickCommandImpl);
        }

        private IList<PresetItem> GetItems()
        {
            IList<PresetItem> items = new List<PresetItem>();

            IList<Color> colors = Settings.GetPresetColors(_ToolType, !_CanBeEmpty);
            foreach (Color color in colors)
            {
                PresetItem item = new PresetItem(color);
                items.Add(item);
            }
            return items;
        }

        /// <summary>
        /// Use Transparent to select empty color
        /// </summary>
        /// <param name="color"></param>
        public bool Select(Color color)
        {
            bool selected = false;
            foreach (PresetItem item in PresetsColors)
            {
                if ((color.A == 0 && item.IsEmptyOption)
                    || (item.IsColorOption && item.ColorBrush.Color == color))
                {
                    item.IsSelected = true;
                    selected = true;
                }
                else
                {
                    item.IsSelected = false;
                }
            }
            return selected;
        }

        public void AddNewColor(Color color)
        {
            if (!Select(color))
            {
                // check if we have room for add button.
                ResolveAddRemoveButtons(PresetsMaxColumns * MAX_PRESETS_ROWS - 1);
                int index = PresetsColors.Count - 1;
                while (index >= 0 && !PresetsColors[index].IsColorOption)
                {
                    --index;
                }
                PresetItem item = new PresetItem(color, false, false, false);
                PresetsColors.Insert(index + 1, item);
                item.IsSelected = true;
                IList<Color> colors = GetColorsFromItems();
                Settings.SetPresetColors(_ToolType, !_CanBeEmpty, colors);
            }
            ColorSelectedEvent(color, false);
        }


        /// <summary>
        /// Set ShowAddRemoveButtons property and resolve changes to buttons collection
        /// </summary>
        /// <param name="showAddRemoveButtons"></param>
        public void SetShowAddRemoveButtons(bool showAddRemoveButtons)
        {
            ShowAddRemoveButtons = showAddRemoveButtons;
            ResolveAddRemoveButtons(_PresetsMaxColumns * MAX_PRESETS_ROWS);
        }

        #region Commands

        public RelayCommand ItemClickCommand { get; private set; }

        private void ItemClickCommandImpl(object sender)
        {
            PresetItem item = sender as PresetItem;
            if (item != null)
            {
                if (_CurrentMode == ButtonMode.normal)
                {
                    HandleClickInNormalMode(item);
                }
                else if (_CurrentMode == ButtonMode.remove)
                {
                    HandleClickInRemoveMode(item);
                }
            }
        }

        #endregion Commands

        #region Behaviour

        private enum ButtonMode
        {
            normal,
            remove
        }

        private ButtonMode _CurrentMode = ButtonMode.normal;
        private bool _WasItemDeleted = false;


        private void HandleClickInNormalMode(PresetItem item)
        {
            if (item.IsColorOption)
            {
                ColorSelectedEvent(item.ColorBrush.Color, false);
                Select(item.ColorBrush.Color);
            }
            if (item.IsEmptyOption)
            {
                ColorSelectedEvent(Colors.Transparent, true);
                Select(Colors.Transparent);
            }
            if (item.IsRemoveButton)
            {
                _CurrentMode = ButtonMode.remove;
                _WasItemDeleted = false;
                item.AddOrRemoveContent = AnnotationPropertiesPresetsColorGridViewModel.SegoeAcceptString;
                foreach (PresetItem preset in PresetsColors)
                {
                    if (preset.IsColorOption)
                    {
                        preset.ShowDelete = true;
                    }
                }
            }
            if (item.IsAddButton)
            {
                IList<Color> colors = new List<Color>();
                foreach (PresetItem itm in _ColorItems)
                {
                    if (itm.IsColorOption)
                    {
                        colors.Add(itm.ColorBrush.Color);
                    }
                }
                RequestNewColor(colors);
            }
        }

        private void HandleClickInRemoveMode(PresetItem item)
        {
            if (item.IsColorOption)
            {
                if (PresetsColors.Contains(item))
                {
                    _WasItemDeleted = true;
                    PresetsColors.Remove(item);
                }
            }
            else if (item.IsRemoveButton)
            {
                _CurrentMode = ButtonMode.normal;
                item.AddOrRemoveContent = AnnotationPropertiesPresetsColorGridViewModel.SegoeRemoveString;
                foreach (PresetItem preset in PresetsColors)
                {
                    if (preset.IsColorOption)
                    {
                        preset.ShowDelete = false;
                    }
                }
                if (_WasItemDeleted)
                {
                    IList<Color> colors = GetColorsFromItems();
                    Settings.SetPresetColors(_ToolType, !_CanBeEmpty, colors);
                    ResolveAddRemoveButtons(PresetsMaxColumns * MAX_PRESETS_ROWS);
                }
            }
        }

        private IList<Color> GetColorsFromItems()
        {
            IList<Color> colors = new List<Color>();
            foreach (PresetItem item in PresetsColors)
            {
                if (item.IsColorOption)
                {
                    colors.Add(item.ColorBrush.Color);
                }
            }
            return colors;
        }

        private void ResolveAddRemoveButtons(int maxButtons)
        {
            bool hasEmpty = false;
            PresetItem addButton = null;
            PresetItem removeButton = null;
            foreach (PresetItem item in PresetsColors)
            {
                if (item.IsEmptyOption)
                {
                    hasEmpty = true;
                }
                if (item.IsAddButton)
                {
                    addButton = item;
                }
                if (item.IsRemoveButton)
                {
                    removeButton = item;
                }
            }

            if (!ShowAddRemoveButtons)
            {
                if (addButton != null && _ShowAddRemoveButtons == false)
                {
                    PresetsColors.Remove(addButton);
                }

                if (removeButton != null && _ShowAddRemoveButtons == false)
                {
                    PresetsColors.Remove(removeButton);
                }
            }
            else
            {
                int numColors = PresetsColors.Count;
                if (addButton != null)
                {
                    --numColors;
                }
                if (removeButton != null)
                {
                    --numColors;
                }

                // for the purpose of the add button, we consider the empty button a color
                if (addButton == null && numColors < maxButtons - 1)
                {
                    if (removeButton != null)
                    {
                        PresetsColors.Insert(PresetsColors.Count - 1, new PresetItem(Colors.Transparent, false, true, false));
                    }
                    else
                    {
                        PresetsColors.Add(new PresetItem(Colors.Transparent, false, true, false));
                    }
                }
                if (addButton != null && numColors >= maxButtons - 1)
                {
                    PresetsColors.Remove(addButton);
                }

                // for the purpose of the remove button, empty is not considered a color
                if (hasEmpty)
                {
                    --numColors;
                }
                if (removeButton == null && numColors > 0)
                {
                    PresetsColors.Add(new PresetItem(Colors.Transparent, false, false, true));
                }

                if (removeButton != null && numColors == 0)
                {
                    PresetsColors.Remove(removeButton);
                }
            }
        }

        #endregion Behaviour
    }

    #endregion Preset Colors

    #region Preset Buttons

    public class PresetButtonViewModel : ViewModelBase
    {
        #region Preset Button Item

        public class PresetButtonItem : ViewModelBase
        {
            public enum SelectionState
            {
                selected,
                notselected,
            }

            private SelectionState _SelectedStatus = SelectionState.notselected;
            public SelectionState SelectedStatus
            {
                get { return _SelectedStatus; }
                set
                {
                    if (value != _SelectedStatus)
                    {
                        _SelectedStatus = value;
                        RaisePropertyChanged();
                    }
                }
            }
            private SolidColorBrush _SelectedForegroundColor;
            public SolidColorBrush SelectedForegroundColor
            {
                get { return _SelectedForegroundColor; }
                set
                {
                    if (_SelectedForegroundColor != value)
                    {
                        _SelectedForegroundColor = value;
                        RaisePropertyChanged();
                    }
                }
            }
            private bool _IsOption = true;
            public bool IsOption { get { return _IsOption; } }
            public bool IsSeparator { get { return !_IsOption; } }
            public string Option { get; set; }

            private double _ItemWidth = 0;
            public double ItemWidth
            {
                get { return _ItemWidth; }
                set
                {
                    if (_ItemWidth != value && value > 0)
                    {
                        _ItemWidth = value;
                        RaisePropertyChanged();
                    }
                }
            }

            public PresetButtonItem(string content)
            {
                if (!string.IsNullOrEmpty(content))
                {
                    _IsOption = true;
                    Option = content;
                }
                else
                {
                    _IsOption = false;
                }
            }
        }

        #endregion Preset Button Item

        private const double SeparatorWidth = 1;

        public delegate void OptionSelectedDelegate(string option);
        public event OptionSelectedDelegate OptionSelected = delegate { };

        public ObservableCollection<PresetButtonItem> Items { get; private set; }

        public RelayCommand ItemClickCommand { get; private set; }
        public RelayCommand SizeChangedCommand { get; private set; }

        public ColumnDefinitionCollection ColumnDefs { get; private set; }

        private PresetButtonItem _SelectedButton = null;

        public PresetButtonViewModel(IList<string> options, bool addNull = true)
        {
            this.ItemClickCommand = new RelayCommand(ItemClickCommandImpl);
            this.SizeChangedCommand = new RelayCommand(SizeChangedCommandImpl);

            Items = new ObservableCollection<PresetButtonItem>();

            foreach (string option in options)
            {
                // Not sure what this part is for, so setting a boolean for me to disable for use case
                if (Items.Count > 0 && addNull)
                {
                    Items.Add(new PresetButtonItem(null));
                }
                Items.Add(new PresetButtonItem(option));
            }
        }

        public void Select(string val)
        {
            if (_SelectedButton != null && !val.Equals(_SelectedButton.Option))
            {
                _SelectedButton.SelectedStatus = PresetButtonItem.SelectionState.notselected;
                _SelectedButton = null;
            }

            foreach (PresetButtonItem item in Items)
            {
                if (item.IsOption)
                {
                    string content = item.Option;
                    if (val.Equals(content))
                    {
                        _SelectedButton = item;
                        _SelectedButton.SelectedStatus = PresetButtonItem.SelectionState.selected;
                    }
                }
            }
        }


        private void ItemClickCommandImpl(object content)
        {
            string val = content as string;
            if (!string.IsNullOrWhiteSpace(val))
            {
                OptionSelected(val);
            }

        }

        private void SizeChangedCommandImpl(object sizeArgs)
        {
            SizeChangedEventArgs args = sizeArgs as SizeChangedEventArgs;

            int numItems = (Items.Count + 1) / 2;
            double buttonWidth = (args.NewSize.Width - numItems) / numItems;
            foreach (PresetButtonItem item in Items)
            {
                if (item.IsOption)
                {
                    item.ItemWidth = buttonWidth;
                }
                else
                {
                    item.ItemWidth = SeparatorWidth;
                }
            }
        }

    }

    #endregion Preset Buttons

    #region Custom Colors

    public class AnnotationPropertiesCustomColorGridViewModel : ViewModelBase
    {
        #region Custom Color Item

        public class CustomItem : ViewModelBase
        {
            public SolidColorBrush ColorBrush { get; private set; }
            public bool IsColorOption { get; private set; }
            public bool IsEmptyOption { get; private set; }

            private bool _IsSelectable = true;
            public bool IsSelectable
            {
                get { return _IsSelectable; }
                set
                {
                    if (value != _IsSelectable)
                    {
                        _IsSelectable = value;
                        RaisePropertyChanged();
                        RaisePropertyChanged("Opacity");
                    }
                }
            }
            public double Opacity
            {
                get
                {
                    if (_IsSelectable)
                    {
                        return 1;
                    }
                    return 0.5;
                }

            }

            private bool _IsSelected = false;
            public bool IsSelected
            {
                get { return _IsSelected; }
                set
                {
                    if (_IsSelected != value)
                    {
                        _IsSelected = value;
                        RaisePropertyChanged();
                    }
                }
            }

            public CustomItem(Color color, bool isEmpty)
            {
                IsEmptyOption = isEmpty;
                IsColorOption = !isEmpty;
                if (IsColorOption)
                {
                    ColorBrush = new SolidColorBrush(color);
                }
                else
                {
                    ColorBrush = new SolidColorBrush(Colors.Transparent);
                }
            }
        }

        #endregion Custom Color Item

        public delegate void ColorSelectedDelegate(Color color, bool isEmpty);
        public event ColorSelectedDelegate ColorSelectedEvent = delegate { };

        public AnnotationPropertiesCustomColorGridViewModel(bool canBeEmpty, int maxColumns, IList<Color> colorList)
        {
            Init(canBeEmpty, maxColumns, colorList);
        }

        private void Init(bool canBeEmpty, int maxColumns, IList<Color> colorList)
        {
            PointerContactCommand = new RelayCommand(PointerContactCommandImpl);

            CustomMaxColumns = maxColumns;

            CustomColors = new ObservableCollection<CustomItem>();
            foreach (Color color in colorList)
            {
                CustomItem item = new CustomItem(color, false);
                _CustomColors.Add(item);
            }
            if (canBeEmpty)
            {
                _CustomColors.Add(new CustomItem(Colors.Transparent, true));
            }
        }

        public int CustomMaxColumns { get; private set; }

        private ObservableCollection<CustomItem> _CustomColors;
        public ObservableCollection<CustomItem> CustomColors
        {
            get { return _CustomColors; }
            set
            {
                _CustomColors = value;
                RaisePropertyChanged();
            }
        }

        public void Select(Color color)
        {
            if (color.A == 0)
            {
                SelectEmpty();
                return;
            }
            foreach (CustomItem item in CustomColors)
            {
                item.IsSelected = (!item.IsEmptyOption && item.ColorBrush.Color == color);
            }
        }

        public void SelectEmpty()
        {
            foreach (CustomItem item in CustomColors)
            {
                item.IsSelected = item.IsEmptyOption;
            }
        }

        public void DisableColors(IList<Color> colors)
        {
            foreach (CustomItem item in _CustomColors)
            {
                if (item.IsColorOption && colors.Contains(item.ColorBrush.Color))
                {
                    item.IsSelectable = false;
                }
                else
                {
                    item.IsSelectable = true;
                }
            }
        }

        #region Commands


        public RelayCommand PointerContactCommand { get; private set; }

        private void PointerContactCommandImpl(object senderAndArgs)
        {
            Tuple<FrameworkElement, Windows.UI.Xaml.Input.PointerRoutedEventArgs> args =
                senderAndArgs as Tuple<FrameworkElement, Windows.UI.Xaml.Input.PointerRoutedEventArgs>;
            if (args != null)
            {
                FrameworkElement control = args.Item1;
                Windows.Foundation.Point point = args.Item2.GetCurrentPoint(control).Position;

                if (!double.IsNaN(control.ActualWidth) && control.ActualWidth > 0 && !double.IsNaN(control.ActualHeight) && control.ActualHeight > 0)
                {
                    double xPercentage = point.X / control.ActualWidth;
                    double yPercentage = point.Y / control.ActualHeight;
                    if (xPercentage >= 0 && xPercentage <= 1 && yPercentage >= 0 && yPercentage <= 1)
                    {
                        int numcols = Math.Min(_CustomColors.Count, CustomMaxColumns);
                        int numrows = (CustomMaxColumns - 1 + _CustomColors.Count) / CustomMaxColumns;
                        int col = (int)(xPercentage * numcols);
                        if (col >= numcols) // in case the percentage was right on 1
                        {
                            col = numcols - 1;
                        }
                        int row = (int)(yPercentage * numrows);
                        if (row >= numrows)
                        {
                            row = numrows - 1;
                        }
                        int index = row * CustomMaxColumns + col;
                        if (index < _CustomColors.Count)
                        {
                            ColorSelectedEvent(_CustomColors[index].ColorBrush.Color, _CustomColors[index].IsEmptyOption);
                        }
                    }
                }
            }
        }

        #endregion Commands

    }

    #endregion Custom Colors

    #region Font Names

    public class AnnotationPropertiesFontNameSelectorViewModel : ViewModelBase
    {
        public delegate void FontNameSelectedDelegate(FontBundle selectedFont);
        public event FontNameSelectedDelegate FontNameSelected;

        private static Windows.UI.Xaml.Controls.Primitives.Popup _CurrentPopup;
        public static void ClosePopups()
        {
            if (_CurrentPopup != null)
            {
                _CurrentPopup.IsOpen = false;
            }
        }

        public ObservableCollection<FontBundle> FontNameItems { get; private set; }

        private bool _IsSelectionInternal = false;
        private FontBundle _PreviouslySelectedFontNameItem = null;
        private FontBundle _SelectedFontNameItem = null;
        public FontBundle SelectedFontNameItem
        {
            get { return _SelectedFontNameItem; }
            set
            {
                _PreviouslySelectedFontNameItem = _SelectedFontNameItem;
                if (Set(ref _SelectedFontNameItem, value))
                {
                    if (value != null)
                    {
                        if (_MoreFontString.Equals(value.DisplayName))
                        {
                            ShowAllFonts();
                        }
                        else if (!_IsSelectionInternal)
                        {
                            FontNameSelected?.Invoke(_SelectedFontNameItem);
                        }
                    }
                }
            }
        }

        private string _MoreFontString = null;
        private HashSet<string> _FontWhiteList;

        public AnnotationPropertiesFontNameSelectorViewModel()
        {
            FontNameItems = new ObservableCollection<FontBundle>();
            _FontWhiteList = Settings.FontWhiteList;
            _MoreFontString = ResourceHandler.GetString("AnnotationProperties_FontCombo_MoreItem");
            PopulateList();
        }

        private void PopulateList()
        {
            try
            {
                List<FontBundle> fullFontList = UtilityFunctions.GetListOfSystemFonts();
                foreach (FontBundle bundle in fullFontList)
                {
                    if (_FontWhiteList.Contains(bundle.DisplayName))
                    {
                        FontNameItems.Add(bundle);
                    }
                }
                FontBundle extra = new FontBundle(_MoreFontString, _MoreFontString, _MoreFontString, _MoreFontString,
                    Windows.UI.Text.FontWeights.Normal, Windows.UI.Text.FontStyle.Normal);
                FontNameItems.Add(extra);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        public void Select(FontBundle value)
        {
            if (value != null && !UtilityFunctions.FontBundle.IsNullFont(value))
            {
                try
                {
                    _IsSelectionInternal = true;
                    int index = 0;
                    foreach (FontBundle item in FontNameItems)
                    {
                        if (item.Equals(value))
                        {
                            SelectedFontNameItem = item;
                            return;
                        }
                        if (item.DisplayName.CompareTo(value.DisplayName) < 0 && !item.DisplayName.Equals(_MoreFontString))
                        {
                            index++;
                        }
                    }
                    FontNameItems.Insert(index, value);
                    SelectedFontNameItem = value;
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    _IsSelectionInternal = false;
                }
            }
        }

        Windows.UI.Xaml.Controls.Primitives.Popup _FontPickerPopup = null;
        Grid _FontPickerGrid = null;
        private void ShowAllFonts()
        {
            _FontPickerPopup = new Windows.UI.Xaml.Controls.Primitives.Popup();
            _FontPickerPopup.Width = Window.Current.Bounds.Width;
            _FontPickerPopup.Height = Window.Current.Bounds.Height;

            _CurrentPopup = _FontPickerPopup;

            _FontPickerGrid = new Grid();
            _FontPickerGrid.Width = _FontPickerPopup.Width;
            _FontPickerGrid.Height = _FontPickerPopup.Height;
            _FontPickerGrid.Background = new SolidColorBrush(Color.FromArgb(44, 44, 44, 44));
            _FontPickerPopup.Child = _FontPickerGrid;

            _FontPickerGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });
            _FontPickerGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(500, GridUnitType.Star), MaxWidth = 400 });
            _FontPickerGrid.ColumnDefinitions.Add(new ColumnDefinition() { Width = new GridLength(1, GridUnitType.Star) });

            Grid g = new Grid();
            g.SetValue(Grid.ColumnProperty, 1);
            g.HorizontalAlignment = HorizontalAlignment.Stretch;
            g.VerticalAlignment = VerticalAlignment.Stretch;
            g.Background = ResourceHandler.ViewerResources["ApplicationPageBackgroundThemeBrush"] as SolidColorBrush;
            _FontPickerGrid.Children.Add(g);

            ToolSettings.FreeTextFontControl fontControl = new ToolSettings.FreeTextFontControl();
            fontControl.HorizontalAlignment = HorizontalAlignment.Stretch;
            fontControl.VerticalAlignment = VerticalAlignment.Stretch;
            g.Children.Add(fontControl);

            (fontControl.DataContext as ToolSettings.FreeTextFontViewModel).ExitRequested += FontListSelector_ExitRequested;

            Window.Current.SizeChanged += Current_SizeChanged;

            if (AnnotationPropertiesDisplay.CurrentPopup != null)
            {
                AnnotationPropertiesDisplay.CurrentPopup.IsLightDismissEnabled = false;
            }

            _FontPickerPopup.IsOpen = true;
        }

        private void Current_SizeChanged(object sender, Windows.UI.Core.WindowSizeChangedEventArgs e)
        {
            _FontPickerPopup.Width = e.Size.Width;
            _FontPickerPopup.Height = e.Size.Height;
            _FontPickerGrid.Width = _FontPickerPopup.Width;
            _FontPickerGrid.Height = _FontPickerPopup.Height;
        }

        private void FontListSelector_ExitRequested(object sender, RoutedEventArgs e)
        {
            _CurrentPopup = null;
            _FontPickerPopup.IsOpen = false;
            try
            {
                _IsSelectionInternal = true;
                FontBundle previousSelection = _PreviouslySelectedFontNameItem;
                _FontWhiteList = Settings.FontWhiteList;
                FontNameItems.Clear();
                PopulateList();
                Select(previousSelection);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _IsSelectionInternal = false;
            }
            if (AnnotationPropertiesDisplay.CurrentPopup != null)
            {
                AnnotationPropertiesDisplay.CurrentPopup.IsLightDismissEnabled = true;
            }
        }
    }

    #endregion Font Names

    #region Measure

    public class AnnotationMeasurePropertiesViewModel : ViewModelBase
    {
        public AnnotationMeasurePropertiesViewModel(ToolType toolType)
        {
            BaseUnits = new ObservableCollection<string>();
            BaseUnits.Add("pt");
            BaseUnits.Add("in");
            BaseUnits.Add("mm");
            BaseUnits.Add("cm");

            TranslateUnits = new ObservableCollection<string>();
            TranslateUnits.Add("pt");
            TranslateUnits.Add("in");
            TranslateUnits.Add("mm");
            TranslateUnits.Add("cm");
            TranslateUnits.Add("m");
            TranslateUnits.Add("km");
            TranslateUnits.Add("ft");
            TranslateUnits.Add("yd");
            TranslateUnits.Add("mi");

            Precisions = new ObservableCollection<string>();
            Precisions.Add("1");
            Precisions.Add("0.1");
            Precisions.Add("0.01");
            Precisions.Add("0.001");
            Precisions.Add("0.0001");

            SelectedPrecision = Settings.GetMeasure(toolType, MeasureInfoType.Precision, MeasureUtils.DEFAULT_PRECISION);
            BaseScale = Settings.GetMeasure(toolType, MeasureInfoType.BaseScale, MeasureUtils.DEFAULT_BASE_SCALE);
            SelectedBaseUnit = Settings.GetMeasure(toolType, MeasureInfoType.BaseUnit, MeasureUtils.DEFAULT_BASE_UNIT);
            TranslateScale = Settings.GetMeasure(toolType, MeasureInfoType.TranslateScale, MeasureUtils.DEFAULT_TRANSLATE_SCALE);
            SelectedTranslateUnit = Settings.GetMeasure(toolType, MeasureInfoType.TranslateUnit, MeasureUtils.DEFAULT_TRANSLATE_UNIT);
        }

        #region Events
        public delegate void PrecisionSelectedDelegate(string precision, bool isEmpty);
        public event PrecisionSelectedDelegate PrecisionSelected = delegate { };

        public delegate void BaseUnitSelectedDelegate(string baseUnit, bool isEmpty);
        public event BaseUnitSelectedDelegate BaseUnitSelected = delegate { };

        public delegate void TranslateUnitSelectedDelegate(string translateUnit, bool isEmpty);
        public event TranslateUnitSelectedDelegate TranslateUnitSelected = delegate { };

        public delegate void BaseScaleSelectedDelegate(string baseUnit, bool isEmpty);
        public event BaseScaleSelectedDelegate BaseScaleSelected = delegate { };

        public delegate void TranslateScaleSelectedDelegate(string translateUnit, bool isEmpty);
        public event TranslateScaleSelectedDelegate TranslateScaleSelected = delegate { };
        #endregion

        #region Public Properties

        public ObservableCollection<string> BaseUnits { get; set; }

        public ObservableCollection<string> TranslateUnits { get; set; }

        public ObservableCollection<string> Precisions { get; set; }

        string _selectedBaseUnit;
        public string SelectedBaseUnit
        {
            get { return _selectedBaseUnit; }
            set 
            {
                _selectedBaseUnit = value;
                BaseUnitSelected(_selectedBaseUnit, false);
                RaisePropertyChanged();
            }
        }

        string _selectedTranslateUnit;
        public string SelectedTranslateUnit
        {
            get { return _selectedTranslateUnit; }
            set 
            { 
                _selectedTranslateUnit = value;
                TranslateUnitSelected(_selectedTranslateUnit, false);
                RaisePropertyChanged();
            }
        }

        string _selectedPrecision;
        public string SelectedPrecision
        {
            get { return _selectedPrecision; }
            set 
            {
                _selectedPrecision = value;
                PrecisionSelected(_selectedPrecision, false);
                RaisePropertyChanged();
            }
        }

        string _baseScale;
        public string BaseScale
        {
            get { return _baseScale; }
            set
            { 
                _baseScale = value;
                BaseScaleSelected(_baseScale, false);
                RaisePropertyChanged();
            }
        }

        string _translateScale;
        public string TranslateScale 
        {
            get { return _translateScale; }
            set 
            { 
                _translateScale = value;
                TranslateScaleSelected(_translateScale, false);
                RaisePropertyChanged();
            }
        }

        #endregion
    }

    #endregion

    public class AnnotationPropertiesViewModel : ViewModelBase, IDisposable
    {
        private static int MAX_PRESETS_COLUMNS = 6;
        private static int MAX_CUSTOM_COLUMNS = 9;
        private static double MIN_THICKNESS = 0.5;
        private static double MAX_THICKNESS = 12;
        private static double MIN_FONT_SIZE = 4;
        private static double MAX_FONT_SIZE = 72;
        private static double MIN_OPACITY = 0.05;
        private static double MAX_OPACITY = 1;
        public double MinThickness { get { return MIN_THICKNESS; } }
        public double MaxThickness { get { return MAX_THICKNESS; } }
        public double MinFontSize { get { return MIN_FONT_SIZE; } }
        public double MaxFontSize { get { return MAX_FONT_SIZE; } }
        public double MinOpacity { get { return MIN_OPACITY; } }
        public double MaxOpacity { get { return MAX_OPACITY; } }

        public delegate void ToolSelectedDelegate(ToolType type);
        public event ToolSelectedDelegate ToolSelected = delegate { };


        public delegate void StyleChangeRequestedHandler(AnnotStyleType annotStyleType, object val);
        private StyleChangeRequestedHandler _StyleChangeRequested;
        /// <summary>
        /// Event for when Style properties are changed
        /// </summary>
        public event StyleChangeRequestedHandler StyleChangeRequested
        {
            add
            {
                if (_StyleChangeRequested?.GetInvocationList().Contains(value) != true)
                    _StyleChangeRequested += value;
            }
            remove => _StyleChangeRequested -= value;
        }

        public static string StrokeBrushIcon = new string((char)0xE235, 1);
        public static string TextBrushIcon = new string((char)0x0050, 1);
        public static string FillBrushIcon = new string((char)0x005D, 1);

        private static IList<string> _ThicknessOptions;
        private static IList<string> _OpacityOptions;
        private static IList<string> _FontSizeOptions;
        private static IList<string> _IconOptions;
        private static IList<string> _FormatTextIconOptions;

        private static IList<string> ThicknessOptions
        {
            get
            {
                if (_ThicknessOptions == null)
                {
                    _ThicknessOptions = new List<string>();
                    _ThicknessOptions.Add("1");
                    _ThicknessOptions.Add("3");
                    _ThicknessOptions.Add("7");
                    _ThicknessOptions.Add("12");
                }
                return _ThicknessOptions;
            }
        }
        private static IList<string> OpacityOptions
        {
            get
            {
                if (_OpacityOptions == null)
                {
                    _OpacityOptions = new List<string>();
                    _OpacityOptions.Add("25");
                    _OpacityOptions.Add("50");
                    _OpacityOptions.Add("75");
                    _OpacityOptions.Add("100");
                }
                return _OpacityOptions;
            }
        }
        private static IList<string> FontSizeOptions
        {
            get
            {
                if (_FontSizeOptions == null)
                {
                    _FontSizeOptions = new List<string>();
                    _FontSizeOptions.Add("8");
                    _FontSizeOptions.Add("12");
                    _FontSizeOptions.Add("24");
                    _FontSizeOptions.Add("36");
                }
                return _FontSizeOptions;
            }
        }
        private static IList<string> IconOptions
        {
            get
            {
                if (_IconOptions == null)
                {
                    _IconOptions = new List<string>();
                    _IconOptions.Add("Comment");
                    _IconOptions.Add("Help");
                    _IconOptions.Add("InsertCaret");
                    _IconOptions.Add("Star");
                    _IconOptions.Add("LightBulb");
                    _IconOptions.Add("NewParagraph");
                    _IconOptions.Add("CrossMark");
                    _IconOptions.Add("CheckMark");
                    _IconOptions.Add("LocationPin");
                    _IconOptions.Add("ThumbsUp");

                }
                return _IconOptions;
            }
        }
        private static IList<string> FormatTextIconOptions
        {
            get
            {
                if (_FormatTextIconOptions == null)
                {
                    _FormatTextIconOptions = new List<string>();
                    _FormatTextIconOptions.Add("Bold");
                    _FormatTextIconOptions.Add("Italic");
                    _FormatTextIconOptions.Add("Underline");
                    _FormatTextIconOptions.Add("Strikethrough");
                }
                return _FormatTextIconOptions;
            }
        }

        private string _ItemTag = "";
        public string ItemTag
        {
            get { return _ItemTag; }
            set
            {
                _ItemTag = value;
            }
        }

        private ToolType _AssociatedTool = ToolType.e_none;
        /// <summary>
        /// Please set the AnnotationProperties before you set this.
        /// </summary>
        public ToolType AssociatedTool
        {
            get { return _AssociatedTool;}
            set
            {
                _AssociatedTool = value;
                RaisePropertyChanged(nameof(AreSubViewSelectionButtonsVisible));
                ToolSelected(_AssociatedTool);
                Init();
            }
        }

        private string _SettingString = "";

        public AnnotationPropertiesViewModel()
        {
            CreateCommands();
        }

        // things that need to be initialized before control is created.
        private void CreateCommands()
        {
            PresetOrCustomCommand = new RelayCommand(PresetOrCustomCommandImpl);
            PresetColorTargetCommand = new RelayCommand(PresetColorTargetCommandImpl);
            CustomColorTargetCommand = new RelayCommand(CustomColorTargetCommandImpl);
            NewColorAcceptCommand = new RelayCommand(NewColorAccpetCommandImpl);
            NewColorCancelCommand = new RelayCommand(NewColorCancelCommandImpl);
        }

        // things that need to be initialized once we have a tool type
        private void Init()
        {
            _SettingString = AssociatedTool.ToString() + ItemTag;
            if (Sequencenumber >= 0)
            {
                ActiveSubView = (SubviewStates)Settings.GetView(AssociatedTool, (int)(SubviewStates.presets), Sequencenumber);
            }
            else
            {
                ActiveSubView = (SubviewStates)Settings.GetView(AssociatedTool, (int)(SubviewStates.presets));
            }
            ResolveOptions(AssociatedTool);
            ResolvePreview(AssociatedTool);
        }

        private void ResolveOptions(ToolType toolType)
        {
            switch (toolType)
            {
                case ToolType.e_annot_edit_rect_group:
                case ToolType.e_annot_edit_group_create:
                case ToolType.e_lasso_create:
                case ToolType.e_rect_create:
                case ToolType.e_oval_create:
                case ToolType.e_cloud_create:
                case ToolType.e_polygon_create:
                    SecondaryColorType = ColorTypes.fill;
                    break;
                case ToolType.e_text_highlight:
                    HasThickness = false;
                    break;
                case ToolType.e_text_annot_create:
                    HasThickness = false;
                    PrimaryColorType = ColorTypes.text;
                    SecondaryColorType = ColorTypes.fill;
                    HasFontSize = true;
                    HasFontName = true;
                    HasFormatTextOptions = true;
                    if (ItemTag == "EditText")
                    {
                        ActiveSubView = SubviewStates.editrichtext;
                    }
                    break;
                case ToolType.e_sticky_note_create:
                    HasThickness = false;
                    HasOpacity = false;
                    HasFontSize = false;
                    HasIcons = true;
                    HasAnnotPreview = false;
                    break;
                case ToolType.e_ink_eraser:
                    HasOpacity = false;
                    break;
                case ToolType.e_signature:
                    HasOpacity = false;
                    break;
                case ToolType.e_area_create:
                    SecondaryColorType = ColorTypes.fill;
                    HasMeasure = true;
                    break;
                case ToolType.e_perimeter_create:
                case ToolType.e_ruler_create:
                    HasThickness = true;
                    HasOpacity = true;
                    HasMeasure = true;
                    break;
                case ToolType.e_rect_redaction_create:
                    HasThickness = false;
                    HasOpacity = false;
                    HasAnnotPreview = false;
                    HasRedaction = true;                    
                    SecondaryColorType = ColorTypes.fill;
                    RedactionOverlayText = Settings.GetRedactionText();
                    break;
                case ToolType.e_text_redaction:
                    HasThickness = false;
                    HasOpacity = false;
                    HasAnnotPreview = false;
                    break;
            }

            if (HasMeasure)
            {
                // Measure Options Setup
                MeasureOption = new AnnotationMeasurePropertiesViewModel(toolType);
                MeasureOption.PrecisionSelected += MeasureOption_PrecisionSelected;
                MeasureOption.BaseScaleSelected += MeasureOption_BaseScaleSelected;
                MeasureOption.BaseUnitSelected += MeasureOption_BaseUnitSelected;
                MeasureOption.TranslateScaleSelected += MeasureOption_TranslateScaleSelected;
                MeasureOption.TranslateUnitSelected += MeasureOption_TranslateUnitSelected;
            }

            PrimaryColorOption = new AnnotationPropertiesPresetsColorGridViewModel(toolType, false, MAX_PRESETS_COLUMNS);
            if (ActiveSubView == SubviewStates.editrichtext)
            {
                PrimaryColorOption.SetShowAddRemoveButtons(false);
            }
            PrimaryColorOption.ColorSelectedEvent += PrimaryColorOption_ColorSelectedEvent;
            PrimaryColorOption.RequestNewColor += ColorOption_RequestNewColor;
            RaisePropertyChanged(nameof(PrimaryColorOption));
            PrimaryCustomColorOption = new AnnotationPropertiesCustomColorGridViewModel(false, MAX_CUSTOM_COLUMNS, Settings.GetCustomColors());
            PrimaryCustomColorOption.ColorSelectedEvent += PrimaryColorOption_ColorSelectedEvent;
            RaisePropertyChanged(nameof(PrimaryCustomColorOption));
            SelectMainColor();
            if (SecondaryColorType != ColorTypes.none)
            {
                SecondaryColorOption = new AnnotationPropertiesPresetsColorGridViewModel(toolType, true, MAX_PRESETS_COLUMNS);
                if (ActiveSubView == SubviewStates.editrichtext)
                {
                    SecondaryColorOption.SetShowAddRemoveButtons(false);
                }
                SecondaryColorOption.ColorSelectedEvent += SecondaryColorOption_ColorSelectedEvent;
                SecondaryColorOption.RequestNewColor += ColorOption_RequestNewColor;
                RaisePropertyChanged(nameof(SecondaryColorOption));
                SecondaryCustomColorOption = new AnnotationPropertiesCustomColorGridViewModel(true, MAX_CUSTOM_COLUMNS, Settings.GetCustomColors());
                SecondaryCustomColorOption.ColorSelectedEvent += SecondaryColorOption_ColorSelectedEvent;
                RaisePropertyChanged(nameof(SecondaryCustomColorOption));
                SelectFillColor();
            }

            if (HasThickness)
            {
                // Configure thickness options depending if it is a marker or not
                if (toolType == ToolType.e_ink_marker || toolType == ToolType.e_ink_eraser)
                {
                    ThicknessOptions.Clear();
                    ThicknessOptions.Add("10");
                    ThicknessOptions.Add("12");
                    ThicknessOptions.Add("15");
                    ThicknessOptions.Add("20");
                    if (toolType == ToolType.e_ink_eraser)
                    {
                        MIN_THICKNESS = 1;
                    }
                    else
                    {
                        MIN_THICKNESS = 4;
                    }
                    MAX_THICKNESS = 24;
                }
                else
                {
                    ThicknessOptions.Clear();
                    ThicknessOptions.Add("1");
                    ThicknessOptions.Add("3");
                    ThicknessOptions.Add("7");
                    ThicknessOptions.Add("12");
                    MIN_THICKNESS = 0.5;
                    MAX_THICKNESS = 12;
                }
                ThicknessButtonViewModel = new PresetButtonViewModel(ThicknessOptions);
                SelectThickness();
                ThicknessButtonViewModel.OptionSelected += ThicknessButtonViewModel_OptionSelected;
                RaisePropertyChanged(nameof(ThicknessButtonViewModel));
            }
            else
            {
                Thickness = 1;
            }

            if (HasOpacity)
            {
                OpacityButtonViewModel = new PresetButtonViewModel(OpacityOptions);
                SelectOpacity();
                OpacityButtonViewModel.OptionSelected += OpacityButtonViewModel_OptionSelected;
                RaisePropertyChanged(nameof(OpacityButtonViewModel));
            }
            if (HasFontSize)
            {
                FontSizeButtonViewModel = new PresetButtonViewModel(FontSizeOptions);
                SelectFontSize();
                FontSizeButtonViewModel.OptionSelected += FontSizeButtonViewModel_OptionSelected;
                RaisePropertyChanged(nameof(FontSizeButtonViewModel));
            }
            if (HasFontName)
            {
                FontNameViewModel = new AnnotationPropertiesFontNameSelectorViewModel();
                SelectFontName();
                FontNameViewModel.FontNameSelected += FontNameViewModel_FontNameSelected;
                RaisePropertyChanged(nameof(FontNameViewModel));
            }
            if (HasIcons)
            {
                IconButtonViewModel = new PresetButtonViewModel(IconOptions, false);
                SettingsColor color = Settings.GetShapeColor(ToolType.e_sticky_note_create, false);
                SetSelectedForegroundColor(new SolidColorBrush(Color.FromArgb(255, color.R, color.G, color.B)), IconButtonViewModel);
                SelectIcon();
                IconButtonViewModel.OptionSelected += IconButtonViewModel_OptionSelected;
                RaisePropertyChanged(nameof(IconButtonViewModel));
            }
            if (HasFormatTextOptions)
            {
                FormatTextViewModel = new PresetButtonViewModel(FormatTextIconOptions, false);
                FormatTextViewModel.OptionSelected += FormatTextViewModel_OptionSelected;
                RaisePropertyChanged(nameof(FormatTextViewModel));
            }
        }

        private void SetSelectedForegroundColor(SolidColorBrush brush, PresetButtonViewModel viewModel)
        {
            foreach(PresetButtonViewModel.PresetButtonItem button in viewModel.Items)
            {
                button.SelectedForegroundColor = brush;
            }
        }

        private void ResolvePreview(ToolType toolType)
        {
            switch (toolType)
            {
                case ToolType.e_line_create:
                case ToolType.e_arrow_create:
                case ToolType.e_polyline_create:
                    PreviewType = PreviewTypes.DiagonalLine;
                    break;
                case ToolType.e_ink_create:
                case ToolType.e_ink_marker:
                case ToolType.e_ink_eraser:
                case ToolType.e_signature:
                    PreviewType = PreviewTypes.CurvedLine;
                    break;
                case ToolType.e_annot_edit_rect_group:
                case ToolType.e_annot_edit_group_create:
                case ToolType.e_lasso_create:
                case ToolType.e_rect_create:
                case ToolType.e_cloud_create:
                case ToolType.e_polygon_create:
                case ToolType.e_area_create:
                    PreviewType = PreviewTypes.Rectangle;
                    break;
                case ToolType.e_oval_create:
                    PreviewType = PreviewTypes.Ellipse;
                    break;
                case ToolType.e_text_underline:
                case ToolType.e_text_squiggly:
                case ToolType.e_text_strikeout:
                    PreviewType = PreviewTypes.HorizontalLine;
                    break;
                case ToolType.e_text_highlight:
                    PreviewType = PreviewTypes.HighlightedText;
                    break;
                case ToolType.e_text_annot_create:
                    PreviewType = PreviewTypes.Text;
                    break;
                default:
                    PreviewType = PreviewTypes.DiagonalLine;
                    break;
            }
        }

        void ThicknessButtonViewModel_OptionSelected(string option)
        {
            double result = 0;
            bool parsed = double.TryParse(option, out result);
            if (parsed)
            {
                ThicknessInternal = result;
            }
        }

        void OpacityButtonViewModel_OptionSelected(string option)
        {
            if (option.Length < 1)
            {
                return;
            }
            double result = 0;
            bool parsed = double.TryParse(option, out result);
            if (parsed)
            {
                OpacityInternal = result / 100.0;
            }
        }

        void FontSizeButtonViewModel_OptionSelected(string option)
        {
            double result = 0;
            bool parsed = double.TryParse(option, out result);
            if (parsed)
            {
                FontSizeInternal = result;
            }
        }

        private void FontNameViewModel_FontNameSelected(FontBundle selectedFont)
        {
            FontNameInternal = selectedFont;
        }

        void IconButtonViewModel_OptionSelected(string option)
        {
            if (!string.IsNullOrEmpty(option))
                IconInternal = option;
        }

        void FormatTextViewModel_OptionSelected(string option)
        {
            if (!string.IsNullOrEmpty(option))
            {
                FormatText = option;
            }
        }

        private void SelectMainColor()
        {
            SettingsColor settingsColor;
            if (_MainColorSet)
            {
                settingsColor = MainColor;
            }
            else
            {
                MainColor = settingsColor = Settings.GetShapeColor(AssociatedTool, false);
            }
            PrimaryColorOption.Select(UtilityFunctions.ConvertToColor(settingsColor));
            PrimaryCustomColorOption.Select(UtilityFunctions.ConvertToColor(settingsColor));
        }

        private void SelectFillColor()
        {
            SettingsColor settingsColor;
            if (_FillColorSet)
            {
                settingsColor = FillColor;
            }
            else
            {
                FillColor = settingsColor = Settings.GetShapeColor(AssociatedTool, true);
            }
            SecondaryColorOption.Select(UtilityFunctions.ConvertToColor(settingsColor));
            SecondaryCustomColorOption.Select(UtilityFunctions.ConvertToColor(settingsColor));
        }

        private void SelectThickness()
        {
            if (Thickness < 0) // means it's not been set, so grab default
            {
                Thickness = Settings.GetThickness(AssociatedTool, Sequencenumber);
            }
            if (ThicknessButtonViewModel != null)
            {
                ThicknessButtonViewModel.Select("" + Thickness);
            }
        }

        private void SelectOpacity()
        {
            if (Opacity < 0) // means it's not been set, so grab default
            {
                Opacity = Settings.GetOpacity(AssociatedTool, Sequencenumber);
            }
            if (OpacityButtonViewModel != null)
            {
                OpacityButtonViewModel.Select("" + (int)((Opacity * 100) + 0.5));
            }
        }

        private void SelectFontSize()
        {
            if (FontSize < 0) // means it's not been set, so grab default
            {
                FontSize = Settings.GetFontSize(AssociatedTool, Sequencenumber);
            }
            if (FontSizeButtonViewModel != null)
            {
                FontSizeButtonViewModel.Select("" + FontSize);
            }
        }

        private void SelectFontName()
        {
            if (FontName == null) // means it's not been set, so grab default
            {
                FontName = FontBundle.CreateFromDisplayName(Settings.GetFontName(AssociatedTool, Sequencenumber));
                if (FontName == null)
                {
                    // This means that the font is no longer available, so we're setting a new default font.
                    FontBundle defaultFont = FontBundle.CreateDefaultFontBundle();
                    FontNameInternal = defaultFont;
                }
            }
        }

        private void SelectIcon()
        {
            if (Icon == "") // means it's not been set, so grab default
            {
                Icon = Settings.GetIcon(AssociatedTool, Sequencenumber);
            }
            if (IconButtonViewModel != null)
            {
                IconButtonViewModel.Select("" + Icon);
            }
        }

        #region Annotation Properties Views
        public AnnotationMeasurePropertiesViewModel MeasureOption { get; private set; }

        private void MeasureOption_TranslateUnitSelected(string translateUnit, bool isEmpty)
        {
            Settings.SetMeasure(AssociatedTool, MeasureInfoType.TranslateUnit, translateUnit);
        }

        private void MeasureOption_BaseUnitSelected(string baseUnit, bool isEmpty)
        {
            Settings.SetMeasure(AssociatedTool, MeasureInfoType.BaseUnit, baseUnit);
        }

        private void MeasureOption_TranslateScaleSelected(string translateScale, bool isEmpty)
        {
            Settings.SetMeasure(AssociatedTool, MeasureInfoType.TranslateScale, translateScale);
        }

        private void MeasureOption_BaseScaleSelected(string baseUnit, bool isEmpty)
        {
            Settings.SetMeasure(AssociatedTool, MeasureInfoType.BaseScale, baseUnit);
        }

        private void MeasureOption_PrecisionSelected(string precision, bool isEmpty)
        {
            Settings.SetMeasure(AssociatedTool, MeasureInfoType.Precision, precision);
        }

        public AnnotationPropertiesPresetsColorGridViewModel PrimaryColorOption { get; private set; }
        public AnnotationPropertiesPresetsColorGridViewModel SecondaryColorOption { get; private set; }
        public AnnotationPropertiesCustomColorGridViewModel PrimaryCustomColorOption { get; private set; }
        public AnnotationPropertiesCustomColorGridViewModel SecondaryCustomColorOption { get; private set; }
        public AnnotationPropertiesCustomColorGridViewModel _AddNewColorOptions = null;
        public AnnotationPropertiesCustomColorGridViewModel AddNewColorOptions
        {
            get { return _AddNewColorOptions; }
            set
            {
                if (value != _AddNewColorOptions)
                {
                    _AddNewColorOptions = value;
                    RaisePropertyChanged();
                }
            }
        }

        void PrimaryColorOption_ColorSelectedEvent(Color color, bool isEmpty)
        {
            MainColorInternal = new SettingsColor(color.R, color.G, color.B, !isEmpty);
            SelectMainColor();
        }

        void SecondaryColorOption_ColorSelectedEvent(Color color, bool isEmpty)
        {
            if (isEmpty)
            {
                FillColorInternal = new SettingsColor(0, 0, 0, false);
            }
            else
            {
                FillColorInternal = new SettingsColor(color.R, color.G, color.B, true);
            }
            SelectFillColor();
        }

        public PresetButtonViewModel ThicknessButtonViewModel { get; private set; }
        public PresetButtonViewModel OpacityButtonViewModel { get; private set; }
        public PresetButtonViewModel FontSizeButtonViewModel { get; set; }
        public PresetButtonViewModel IconButtonViewModel { get; private set; }
        public PresetButtonViewModel FormatTextViewModel { get; private set; }

        public AnnotationPropertiesFontNameSelectorViewModel FontNameViewModel { get; private set; }


        #endregion Annotation Properties Views

        #region Add a new Color

        private bool _NewColorSelected = false;
        public bool NewColorSelected
        {
            get { return _NewColorSelected; }
            set
            {
                if (value != _NewColorSelected)
                {
                    _NewColorSelected = value;
                    RaisePropertyChanged();
                }
            }
        }

        private Color _NewColor = Colors.Transparent;

        void ColorOption_RequestNewColor(IList<Color> existingColors)
        {
            AddNewColorOptions = new AnnotationPropertiesCustomColorGridViewModel(false, MAX_CUSTOM_COLUMNS, Settings.GetCustomColors());
            AddNewColorOptions.ColorSelectedEvent += AddNewColorOptions_ColorSelectedEvent;
            NewColorSelected = false;
            _NewColor = Colors.Transparent;
            //AddNewColorOptions.DisableColors(existingColors);
            ActiveSubView = SubviewStates.addnew;
        }

        private void AddNewColorOptions_ColorSelectedEvent(Color color, bool isEmpty)
        {
            AddNewColorOptions.Select(color);
            NewColorSelected = true;
            _NewColor = color;
        }

        public RelayCommand NewColorAcceptCommand { get; private set; }
        public RelayCommand NewColorCancelCommand { get; private set; }

        private void NewColorAccpetCommandImpl(object sender)
        {
            ActiveSubView = SubviewStates.presets;
            if (PresetColorTarget == ColorTarget.primary)
            {
                PrimaryColorOption.AddNewColor(_NewColor);
            }
            else
            {
                SecondaryColorOption.AddNewColor(_NewColor);
            }
        }

        private void NewColorCancelCommandImpl(object sender)
        {
            ActiveSubView = SubviewStates.presets;
        }

        #endregion Add a new Color

        #region AnnotationProperties

        private bool _PropertyWasUpdated = false;
        public bool PropertyWasUpdated
        {
            get { return _PropertyWasUpdated; }
            private set { _PropertyWasUpdated = value; }
        }

        private bool _MainColorSet = false;
        private SettingsColor _MainColor;
        public SettingsColor MainColor
        {
            get { return _MainColor; }
            set
            {
                PropertyWasUpdated = true;
                _MainColorSet = true;
                _MainColor = value;
                RaisePropertyChanged("PrimaryBrush");
                if (IconButtonViewModel != null)
                    SetSelectedForegroundColor(PrimaryBrush, IconButtonViewModel);
                _StyleChangeRequested?.Invoke(AnnotStyleType.MainColor, _MainColor);
            }
        }

        private SettingsColor MainColorInternal
        {
            set
            {
                MainColor = value;
                PropertyWasUpdated = true;
                if (_SequenceNumber >= 0)
                {
                    Settings.SetShapeColor(AssociatedTool, false, MainColor, _SequenceNumber);
                }
                else
                {
                    Settings.SetShapeColor(AssociatedTool, false, MainColor);
                }
                _StyleChangeRequested?.Invoke(AnnotStyleType.MainColor, _MainColor);
            }
        }

        private bool _FillColorSet = false;
        private SettingsColor _FillColor;
        /// <summary>
        /// Doubles as background color for FreeTextAnnotations
        /// </summary>
        public SettingsColor FillColor
        {
            get { return _FillColor; }
            set
            {
                _FillColorSet = true;
                _FillColor = value;
                RaisePropertyChanged("SecondaryBrush");
                _StyleChangeRequested?.Invoke(AnnotStyleType.FillColor, _FillColor);
            }
        }

        private SettingsColor FillColorInternal
        {
            set
            {
                FillColor = value;
                PropertyWasUpdated = true;
                if (_SequenceNumber >= 0)
                {
                    Settings.SetShapeColor(AssociatedTool, true, FillColor, _SequenceNumber);
                }
                else
                {
                    Settings.SetShapeColor(AssociatedTool, true, FillColor);
                }
            }
        }

        private double _Opacity = -1;
        public double Opacity
        {
            get { return _Opacity; }
            set
            {
                if (value != _Opacity)
                {
                    _Opacity = value;
                    RaisePropertyChanged();
                    SelectOpacity();
                    _StyleChangeRequested?.Invoke(AnnotStyleType.Opacity, _Opacity);
                }
            }
        }

        public double OpacityInternal
        {
            get { return _Opacity; }
            set
            {
                Opacity = value;
                PropertyWasUpdated = true;
                if (_SequenceNumber >= 0)
                {
                    Settings.SetOpacity(AssociatedTool, Opacity, _SequenceNumber);
                }
                else
                {
                    Settings.SetOpacity(AssociatedTool, Opacity);
                }
                RaisePropertyChanged();
            }
        }

        private double _Thickness = -1;
        public double Thickness
        {
            get { return _Thickness; }
            set
            {
                if (value != _Thickness)
                {
                    _Thickness = value;
                    RaisePropertyChanged();
                    SelectThickness();
                    _StyleChangeRequested?.Invoke(AnnotStyleType.Thickness, _Thickness);
                }
            }
        }

        public double ThicknessInternal
        {
            get { return _Thickness; }
            set
            {
                Thickness = value;
                PropertyWasUpdated = true;
                if (_SequenceNumber >= 0)
                {
                    Settings.SetThickness(AssociatedTool, Thickness, _SequenceNumber);
                }
                else
                {
                    Settings.SetThickness(AssociatedTool, Thickness);
                }
                RaisePropertyChanged();
            }
        }

        private double _FontSize = -1;
        public double FontSize
        {
            get { return _FontSize; }
            set
            {
                if (value != _FontSize)
                {
                    _FontSize = value;
                    RaisePropertyChanged();
                    SelectFontSize();
                    _StyleChangeRequested?.Invoke(AnnotStyleType.FontSize, _FontSize);
                }
            }
        }


        public double FontSizeInternal
        {
            get { return _FontSize; }
            set
            {
                FontSize = value;
                PropertyWasUpdated = true;
                if (_SequenceNumber >= 0)
                {
                    Settings.SetFontSize(AssociatedTool, FontSize, _SequenceNumber);
                }
                else
                {
                    Settings.SetFontSize(AssociatedTool, FontSize);
                }
                RaisePropertyChanged();
            }
        }

        private FontBundle _FontName = null;
        public FontBundle FontName
        {
            get { return _FontName; }
            set
            {
                if ((value != null || _FontName != null) && !(value != null && value.Equals(_FontName)))
                {
                    _FontName = value;
                    RaisePropertyChanged();
                    _StyleChangeRequested?.Invoke(AnnotStyleType.FontName, _FontName);
                }
            }
        }

        public FontBundle FontNameInternal
        {
            get { return _FontName; }
            set
            {
                FontName = value;
                PropertyWasUpdated = true;
                string displayName = Settings.DEFAULT_FONT_NAME;
                if (FontName != null)
                {
                    displayName = FontName.DisplayName;
                }
                if (_SequenceNumber >= 0)
                {
                    Settings.SetFontName(AssociatedTool, displayName, _SequenceNumber);
                }
                else
                {
                    Settings.SetFontName(AssociatedTool, displayName);
                }
                RaisePropertyChanged();
            }
        }

        private string _Icon = "";
        public string Icon
        {
            get { return _Icon; }
            set
            {
                if (value != _Icon)
                {
                    _Icon = value;
                    RaisePropertyChanged();
                    _StyleChangeRequested?.Invoke(AnnotStyleType.Icon, _Icon);
                    SelectIcon();
                }
            }
        }

        public string IconInternal
        {
            get { return _Icon; }
            set
            {
                Icon = value;
                PropertyWasUpdated = true;
                if (_SequenceNumber >= 0)
                {
                    Settings.SetIcon(AssociatedTool, Icon, _SequenceNumber);
                }
                else
                {
                    Settings.SetIcon(AssociatedTool, Icon);
                }
                _StyleChangeRequested?.Invoke(AnnotStyleType.Icon, _Icon);
                RaisePropertyChanged();
            }
        }

        private string _FormatText = "";
        /// <summary>
        /// Checking what format button was pressed
        /// </summary>
        public string FormatText
        {
            get { return _FormatText; }
            set
            {
                _FormatText = value;
                PropertyWasUpdated = true;
                RaisePropertyChanged();
                _StyleChangeRequested?.Invoke(AnnotStyleType.FormatText, _FormatText);
            }
        }

        private bool _HasFormatTextOptions = false;
        /// <summary>
        /// Toggle text format options
        /// </summary>
        public bool HasFormatTextOptions
        {
            get { return _HasFormatTextOptions; }
            set
            {
                _HasFormatTextOptions = value;
                RaisePropertyChanged();
            }
        }

        private int _SequenceNumber = -1;
        /// <summary>
        /// Sets a unique identifier for the purpose of the settings, making it possible to store multiple
        /// settings for the same tooltype. Anything above 0 will be identified by this value.
        /// </summary>
        public int Sequencenumber
        {
            get { return _SequenceNumber; }
            set { _SequenceNumber = value; }
        }

        string _RedactionOverlayText = string.Empty;
        public string RedactionOverlayText
        {
            get { return _RedactionOverlayText; }
            set
            {
                if (value != _RedactionOverlayText)
                {
                    _RedactionOverlayText = value;
                    RaisePropertyChanged();
                    Settings.SetRedactionText(_RedactionOverlayText);
                }
            }
        }

        #endregion Annotation Properties

        #region Commands

        public RelayCommand PresetOrCustomCommand { get; private set; }
        public RelayCommand PresetColorTargetCommand { get; private set; }
        public RelayCommand CustomColorTargetCommand { get; private set; }


        private void PresetOrCustomCommandImpl(object desiredView)
        {
            string desired = desiredView as string;
            if (!string.IsNullOrWhiteSpace(desired))
            {
                if (desired.Equals("presets", StringComparison.OrdinalIgnoreCase))
                {
                    ActiveSubView = SubviewStates.presets;
                }
                else if (desired.Equals("custom", StringComparison.OrdinalIgnoreCase))
                {
                    ActiveSubView = SubviewStates.custom;
                }
                else if (desired.Equals("measure", StringComparison.OrdinalIgnoreCase))
                {
                    ActiveSubView = SubviewStates.measure;
                }
                else if (desired.Equals("EditRichText", StringComparison.OrdinalIgnoreCase))
                {
                    ActiveSubView = SubviewStates.editrichtext;
                }

                if (Sequencenumber >= 0)
                {
                    Settings.SetView(AssociatedTool, (int)ActiveSubView, Sequencenumber);
                }
                else
                {
                    Settings.SetView(AssociatedTool, (int)ActiveSubView);
                }
            }
        }

        private void PresetColorTargetCommandImpl(object target)
        {
            string targetString = target as string;
            if (!string.IsNullOrWhiteSpace(targetString))
            {
                if (targetString.Equals("primary", StringComparison.OrdinalIgnoreCase))
                {
                    PresetColorTarget = ColorTarget.primary;
                }
                else if (targetString.Equals("secondary", StringComparison.OrdinalIgnoreCase))
                {
                    PresetColorTarget = ColorTarget.secondary;
                }
                else if (targetString.Equals("fontName", StringComparison.OrdinalIgnoreCase))
                {
                    PresetColorTarget = ColorTarget.fontName;
                }
            }
        }

        private void CustomColorTargetCommandImpl(object target)
        {
            string targetString = target as string;
            if (!string.IsNullOrWhiteSpace(targetString))
            {
                if (targetString.Equals("primary", StringComparison.OrdinalIgnoreCase))
                {
                    CustomColorTarget = ColorTarget.primary;
                }
                else if (targetString.Equals("secondary", StringComparison.OrdinalIgnoreCase))
                {
                    CustomColorTarget = ColorTarget.secondary;
                }
            }
        }

        public void Dispose()
        {
            if (MeasureOption != null)
            {
                MeasureOption.PrecisionSelected -= MeasureOption_PrecisionSelected;
                MeasureOption.BaseScaleSelected -= MeasureOption_BaseScaleSelected;
                MeasureOption.BaseUnitSelected -= MeasureOption_BaseUnitSelected;
                MeasureOption.TranslateScaleSelected -= MeasureOption_TranslateScaleSelected;
                MeasureOption.TranslateUnitSelected -= MeasureOption_TranslateUnitSelected;
            }

            if (PrimaryColorOption != null)
            {
                PrimaryColorOption.ColorSelectedEvent -= PrimaryColorOption_ColorSelectedEvent;
                PrimaryColorOption.RequestNewColor -= ColorOption_RequestNewColor;
            }

            if (PrimaryCustomColorOption != null)
            {
                PrimaryCustomColorOption.ColorSelectedEvent -= PrimaryColorOption_ColorSelectedEvent;
            }

            if (SecondaryColorOption != null)
            {
                SecondaryColorOption.ColorSelectedEvent -= SecondaryColorOption_ColorSelectedEvent;
                SecondaryColorOption.RequestNewColor -= ColorOption_RequestNewColor;
            }

            if (SecondaryCustomColorOption != null)
            {
                SecondaryCustomColorOption.ColorSelectedEvent -= SecondaryColorOption_ColorSelectedEvent;
            }

            if (ThicknessButtonViewModel != null)
            {
                ThicknessButtonViewModel.OptionSelected -= ThicknessButtonViewModel_OptionSelected;
            }

            if (OpacityButtonViewModel != null)
            {
                OpacityButtonViewModel.OptionSelected -= OpacityButtonViewModel_OptionSelected;
            }

            if (FontSizeButtonViewModel != null)
            {
                FontSizeButtonViewModel.OptionSelected -= FontSizeButtonViewModel_OptionSelected;
            }

            if (FontNameViewModel != null)
            {
                FontNameViewModel.FontNameSelected -= FontNameViewModel_FontNameSelected;
            }

            if (IconButtonViewModel != null)
            {
                IconButtonViewModel.OptionSelected -= IconButtonViewModel_OptionSelected;
            }

            if (FormatTextViewModel != null)
            {
                FormatTextViewModel.OptionSelected -= FormatTextViewModel_OptionSelected;
            }

            if (AddNewColorOptions != null)
            {
                AddNewColorOptions.ColorSelectedEvent -= AddNewColorOptions_ColorSelectedEvent;
            }
        }

        #endregion Commands

        #region Visual Properties

        public enum SubviewStates
        {
            presets,
            custom,
            addnew,
            measure,
            editrichtext
        }

        public enum ColorTypes
        {
            stroke,
            fill,
            text,
            none,
        }

        public enum ColorTarget
        {
            primary,
            secondary,
            fontName,
        }

        public enum PreviewTypes
        {
            HorizontalLine,
            DiagonalLine,
            CurvedLine,
            Rectangle,
            Ellipse,
            Text,
            HighlightedText,
        }

        private SubviewStates _ActiveSubView = SubviewStates.presets;
        public SubviewStates ActiveSubView
        {
            get
            {
                // For eraser, only show custom (which simply shows thickness slider)
                if (AssociatedTool == ToolType.e_ink_eraser)
                {
                    return SubviewStates.custom;
                }

                return _ActiveSubView;
            }
            set
            {
                if (_ActiveSubView != value)
                {
                    _ActiveSubView = value;
                    RaisePropertyChanged();
                    RaisePropertyChanged("AreSubViewSelectionButtonsVisible");
                }
            }
        }

        public bool AreSubViewSelectionButtonsVisible
        {
            get { return _ActiveSubView != SubviewStates.addnew && _ActiveSubView != SubviewStates.editrichtext && AssociatedTool != ToolType.e_ink_eraser; }
        }

        public bool HasSecondaryColor
        {
            get { return _SecondaryColorType != ColorTypes.none; }
        }

        public bool HasMultipleTabs
        {
            get { return HasSecondaryColor || HasFontName; }
        }

        private ColorTypes _PrimaryColorType = ColorTypes.stroke;
        public ColorTypes PrimaryColorType
        {
            get { return _PrimaryColorType; }
            set
            {
                if (_PrimaryColorType != value)
                {
                    _PrimaryColorType = value;
                    RaisePropertyChanged();
                }
            }
        }

        public string PrimaryColorTypeIcon
        {
            get
            {
                if (_PrimaryColorType == ColorTypes.text)
                {
                    return TextBrushIcon;
                }
                return StrokeBrushIcon;
            }
        }

        private ColorTypes _SecondaryColorType = ColorTypes.none;
        public ColorTypes SecondaryColorType
        {
            get { return _SecondaryColorType; }
            set
            {
                if (_SecondaryColorType != value)
                {
                    _SecondaryColorType = value;
                    RaisePropertyChanged();
                    RaisePropertyChanged("HasSecondaryColor");
                    RaisePropertyChanged("HasMultipleTabs");
                }
            }
        }

        private ColorTarget _PresetColorTarget = ColorTarget.primary;
        public ColorTarget PresetColorTarget
        {
            get { return _PresetColorTarget; }
            set
            {
                if (_PresetColorTarget != value)
                {
                    _PresetColorTarget = value;
                    RaisePropertyChanged();
                    if (_PresetColorTarget == ColorTarget.fontName && FontNameViewModel != null)
                    {
                        FontNameViewModel.Select(FontName);
                    }
                }
            }
        }

        private ColorTarget _CustomColorTarget = ColorTarget.primary;
        public ColorTarget CustomColorTarget
        {
            get { return _CustomColorTarget; }
            set
            {
                if (_CustomColorTarget != value)
                {
                    _CustomColorTarget = value;
                    RaisePropertyChanged();
                }
            }
        }

        private PreviewTypes _PreviewType = PreviewTypes.HorizontalLine;
        public PreviewTypes PreviewType
        {
            get { return _PreviewType; }
            set
            {
                if (value != _PreviewType)
                {
                    _PreviewType = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _HasThickness = true;
        public bool HasThickness
        {
            get { return _HasThickness; }
            set
            {
                if (_HasThickness != value)
                {
                    _HasThickness = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _HasOpacity = true;
        public bool HasOpacity
        {
            get { return _HasOpacity; }
            set
            {
                if (_HasOpacity != value)
                {
                    _HasOpacity = value;
                    if (!_HasOpacity)
                    {
                        Opacity = 1;
                    }
                    RaisePropertyChanged();
                }
            }
        }

        private bool _HasFontSize = false;
        public bool HasFontSize
        {
            get { return _HasFontSize; }
            set
            {
                if (_HasFontSize != value)
                {
                    _HasFontSize = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _HasFontName = false;
        public bool HasFontName
        {
            get { return _HasFontName; }
            set
            {
                if (_HasFontName != value)
                {
                    _HasFontName = value;
                    RaisePropertyChanged();
                    RaisePropertyChanged("HasMultipleTabs");
                }
            }
        }

        private bool _HasIcons = false;
        public bool HasIcons
        {
            get { return _HasIcons; }
            set
            {
                if (_HasIcons != value)
                {
                    _HasIcons = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _HasMeasure = false;
        public bool HasMeasure
        {
            get { return _HasMeasure; }
            set
            {
                if (_HasMeasure != value)
                {
                    _HasMeasure = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _HasRedaction = false;
        public bool HasRedaction
        {
            get { return _HasRedaction; }
            set
            {
                if (_HasRedaction != value)
                {
                    _HasRedaction = value;
                    RaisePropertyChanged();
                }
            }
        }

        private bool _HasAnnotPreview = true;
        public bool HasAnnotPreview
        {
            get { return _HasAnnotPreview; }
            set { _HasAnnotPreview = value; RaisePropertyChanged(); }
        }

        public SolidColorBrush PrimaryBrush
        {
            get
            {
                return new SolidColorBrush(Color.FromArgb(255, MainColor.R, MainColor.G, MainColor.B));
            }
        }

        public SolidColorBrush SecondaryBrush
        {
            get
            {
                byte alpha = 0;
                if (FillColor.Use)
                {
                    alpha = 255;
                }
                return new SolidColorBrush(Color.FromArgb(alpha, FillColor.R, FillColor.G, FillColor.B));
            }
        }

        bool _arrowHeadEnabled = false;
        public bool ArrowHeadEnabled
        { 
            get { return _arrowHeadEnabled; }
            set
            {
                _arrowHeadEnabled = value;
                RaisePropertyChanged();
            }
        }

        static double INITIAL_ARROW_HEAD_POSITION = 119; // Initial position is always in the center
        double _arrowHeadPosition = INITIAL_ARROW_HEAD_POSITION;
        /// <summary>
        /// Arrow head is always centered in the popup so this property can be used to adjust it left ot right
        /// </summary>
        public double ArrowHeadPositionOffset
        {
            get { return _arrowHeadPosition; }
            set
            {
                _arrowHeadPosition = INITIAL_ARROW_HEAD_POSITION + value;
                RaisePropertyChanged();
            }
        }

        #endregion Visual Properties
    }

    public class AnnotationPropertiesDisplay
    {
        public delegate void AnnotationPropertiesDisplayClosedHandler(AnnotationPropertyDialog dialog);
        public event AnnotationPropertiesDisplayClosedHandler AnnotationPropertyDialogClosed = delegate { };

        private AnnotationPropertiesViewModel.StyleChangeRequestedHandler _StyleChangeRequested;
        /// <summary>
        /// Event for when Style properties are changed
        /// </summary>
        public event AnnotationPropertiesViewModel.StyleChangeRequestedHandler StyleChangeRequested
        {
            add
            {
                if (_StyleChangeRequested?.GetInvocationList().Contains(value) != true)
                    _StyleChangeRequested += value;
            }
            remove
            {
                _StyleChangeRequested -= value;
            }
        }

        private static Windows.UI.Xaml.Controls.Primitives.Popup _CurrentPopup;
        public static Windows.UI.Xaml.Controls.Primitives.Popup CurrentPopup
        {
            get { return _CurrentPopup; }
        }

        /// <summary>
        /// Set current Popup.IsOpen property to false
        /// </summary>
        public static void CloseCurrentPopup()
        {
            if (CurrentPopup != null)
            {
                CurrentPopup.IsOpen = false;
            }
        }

        private AnnotationPropertyDialog _Dialog;
        /// <summary>
        /// Get the UserControl Dialog
        /// </summary>
        public AnnotationPropertyDialog Dialog { get { return _Dialog; } }

        private Windows.UI.Xaml.Controls.Primitives.Popup _Popup;
        /// <summary>
        /// Get the Windows Popup UI
        /// </summary>
        public Windows.UI.Xaml.Controls.Primitives.Popup PopupUI { get { return _Popup; } }

        /// <summary>
        /// Popup IsOpen property opens/closes popup
        /// </summary>
        public bool IsPopupOpen
        {
            get
            {
                if (_Popup != null)
                {
                    return _Popup.IsOpen;
                }
                else
                {
                    return false;
                }
            }
            set
            {
                if (_Popup != null)
                {
                    _Popup.IsOpen = value;
                }
            }
        }

        /// <summary>
        /// Popup IsLightDismissed property
        /// </summary>
        public bool IsPopupLightDismissed
        {
            get
            {
                if (_Popup != null)
                {
                    return _Popup.IsLightDismissEnabled;
                }
                else
                {
                    return false;
                }
            }
            set
            {
                if (_Popup != null)
                {
                    _Popup.IsLightDismissEnabled = value;
                }
            }
        }

        private UIRect _Target;
        private UIRect _TargetNormalizedToWindow;
        private Placement _Placement = Placement.Below;
        private CommitedPositions _CommitedPosition = CommitedPositions.None;
        private SizeChangedEventHandler _SizeChangedHandler;
        private EventHandler<object> _PopupClosedHandler;

        public enum Placement
        {
            Below,
            AlignTop,
            PreferBeside,
            PreferAboveOrBelow,
        }

        private enum CommitedPositions
        {
            None,
            Below,
            Above,
            Side,
        }

        private AnnotationPropertiesDisplay()
        {

        }

        public static AnnotationPropertiesDisplay CreateAnnotationPropertiesDisplay(UIRect position, AnnotationPropertyDialog dialog,
            Placement placement = Placement.Below)
        {
            AnnotationPropertiesDisplay display = new AnnotationPropertiesDisplay();
            display._Target = position;
            display._TargetNormalizedToWindow = position;
            //display._TargetNormalizedToWindow = Window.Current.Bounds.Left; // need to push it by the window bounds
            display._Dialog = dialog;
            display._Placement = placement;
            display._Popup = new Windows.UI.Xaml.Controls.Primitives.Popup();

            return display;
        }

        /// <summary>
        /// Place popup and init events. Note: Set properties before placing the popup.
        /// </summary>
        public void PlacePopup()
        {
            _Dialog.HorizontalAlignment = HorizontalAlignment.Left;
            _Dialog.VerticalAlignment = VerticalAlignment.Top;

            _Popup.Width = Window.Current.Bounds.Width;
            _Popup.Height = Window.Current.Bounds.Height;

            Grid g = new Grid();
            g.Width = _Popup.Width;
            g.Height = _Popup.Height;

            _Popup.Child = g;
            g.Children.Add(_Dialog);

            _Popup.IsOpen = true;

            _PopupClosedHandler = new EventHandler<object>(Popup_Closed);
            _Popup.Closed += _PopupClosedHandler;

            _SizeChangedHandler = new SizeChangedEventHandler(PropertiesDialog_SizeChanged);
            _Dialog.SizeChanged += _SizeChangedHandler;

            _Dialog.StyleChangeRequested += PropertiesDialog_StyleChangeRequested;

            _Dialog.MaxWidth = _Popup.Width;
            _Dialog.MaxHeight = _Popup.Height;
        }

        void Popup_Closed(object sender, object e)
        {
            _Dialog.SizeChanged -= _SizeChangedHandler;
            _Popup.Closed -= _PopupClosedHandler;
            _Dialog.StyleChangeRequested -= PropertiesDialog_StyleChangeRequested;
            AnnotationPropertyDialogClosed(_Dialog);
            _Dialog.Dispose();
            _Popup = null;
            _Dialog = null;
        }

        private void PropertiesDialog_StyleChangeRequested(AnnotStyleType annotStyleType, object val)
        {
            _StyleChangeRequested?.Invoke(annotStyleType, val);
        }

        void PropertiesDialog_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            UIRect bounds = GetNormalizedWindowsBounds();

            if (_CommitedPosition == CommitedPositions.None)
            {
                if (_Placement == Placement.Below)
                {
                    PositionDialogBelow(e.NewSize);
                }
                else if (_Placement == Placement.AlignTop)
                {
                    PositionDialogAlignedTop(e.NewSize);
                }
                else if (_Placement == Placement.PreferBeside)
                {
                    PositionDialogPreferSide(e.NewSize, true);
                }
                else if (_Placement == Placement.PreferAboveOrBelow)
                {
                    PositionDialogPreferAboveOrBelow(e.NewSize, true);
                }
            }
            else if (_CommitedPosition == CommitedPositions.Side)
            {
                if (_Dialog.VerticalAlignment == VerticalAlignment.Top && _Dialog.Margin.Top + e.NewSize.Height + 1 > bounds.Height)
                {
                    // this means our position is determined by the bottom of the target, we want to instead use the bottom of the screen as the dialog is bigger than that gap.
                    _Dialog.VerticalAlignment = VerticalAlignment.Bottom;
                    _Dialog.Margin = new Thickness(_Dialog.Margin.Left, 0, 0, 0);
                }
                else if (_Dialog.VerticalAlignment == VerticalAlignment.Bottom && e.NewSize.Height + 1 < bounds.Bottom - _Target.Top)
                {
                    // this means our position is determined by the bottom of the screen, we want to instead use the bottom of the target as the dialog is smaller than that gap.
                    _Dialog.VerticalAlignment = VerticalAlignment.Top;
                    _Dialog.Margin = new Thickness(_Dialog.Margin.Left, _Target.Top, 0, 0);
                }
            }
            else if (_CommitedPosition == CommitedPositions.Below)
            {
                if (_Dialog.VerticalAlignment == VerticalAlignment.Top && _Dialog.Margin.Top + e.NewSize.Height + 1 > bounds.Height)
                {
                    // this means our position is determined by the bottom of the target, we want to instead use the bottom of the screen as the dialog is bigger than that gap.
                    _Dialog.VerticalAlignment = VerticalAlignment.Bottom;
                    _Dialog.Margin = new Thickness(_Dialog.Margin.Left, 0, 0, 0);
                }
                else if (_Dialog.VerticalAlignment == VerticalAlignment.Bottom && e.NewSize.Height + 1 < bounds.Bottom - _Target.Bottom)
                {
                    // this means our position is determined by the bottom of the screen, we want to instead use the bottom of the target as the dialog is smaller than that gap.
                    _Dialog.VerticalAlignment = VerticalAlignment.Top;
                    _Dialog.Margin = new Thickness(_Dialog.Margin.Left, _Target.Bottom, 0, 0);
                }
            }
            else // above
            {
                if (_Dialog.VerticalAlignment == VerticalAlignment.Bottom && _Dialog.Margin.Bottom + e.NewSize.Height + 1 > bounds.Height)
                {
                    // this means our position is determined by the top of the target, we want to instead use the top of the screen as the dialog is bigger than that gap.
                    _Dialog.VerticalAlignment = VerticalAlignment.Top;
                    _Dialog.Margin = new Thickness(_Dialog.Margin.Left, 0, 0, 0);
                }
                else if (_Dialog.VerticalAlignment == VerticalAlignment.Top && e.NewSize.Height + 1 < _Target.Top - bounds.Top)
                {
                    // this means our position is determined by the top of the screen, we want to instead use the top of the target as the dialog is smaller than that gap.
                    _Dialog.VerticalAlignment = VerticalAlignment.Bottom;
                    _Dialog.Margin = new Thickness(_Dialog.Margin.Left, 0, 0, bounds.Bottom - _Target.Top);
                }
            }
        }

        private static int VERTICAL_BELOW_EXTRA_OFFSET = -10;

        /// <summary>
        /// Will place the dialog below the target, or as far down as possible if it does not fit otherwise.
        /// </summary>
        /// <param name="dialogSize"></param>
        private void PositionDialogBelow(Windows.Foundation.Size dialogSize)
        {
            UIRect bounds = GetNormalizedWindowsBounds();

            // position vertically
            double verticalOffset = 0;
            if (dialogSize.Height + 1 >= bounds.Bottom - _Target.Bottom)
            {
                _Dialog.VerticalAlignment = VerticalAlignment.Bottom;
            }
            else
            {
                _Dialog.VerticalAlignment = VerticalAlignment.Top;
                verticalOffset = _Target.Bottom;
            }

            // position horizontally
            double center = (_Target.Left + _Target.Right) / 2;
            double halfWidth = dialogSize.Width / 2;
            double horizontalOffset = 0;
            if (center + halfWidth > bounds.Right)
            {
                _Dialog.HorizontalAlignment = HorizontalAlignment.Right;
            }
            else
            {
                _Dialog.HorizontalAlignment = HorizontalAlignment.Left;
                horizontalOffset = center - halfWidth;
                if (horizontalOffset < 0)
                {
                    horizontalOffset = 0;
                }
            }

            if (_Dialog.ViewModel.ArrowHeadEnabled)
            {
                /* When ArrowHead is enabled the dialog must be alligned properly
                 * and also ensure the arrow head follows the target */
                verticalOffset += VERTICAL_BELOW_EXTRA_OFFSET;

                if (_Dialog.HorizontalAlignment == HorizontalAlignment.Left)
                {
                    if ((horizontalOffset + halfWidth) != center)
                    {
                        var arrow_head_offset = horizontalOffset + halfWidth - center;
                        _Dialog.ViewModel.ArrowHeadPositionOffset = -arrow_head_offset;
                    }
                }else
                {
                    _Dialog.ViewModel.ArrowHeadPositionOffset = halfWidth - (bounds.Right - center);
                }
            }

            _Dialog.Margin = new Thickness(horizontalOffset, verticalOffset, 0, 0);
        }

        /// <summary>
        /// Will place the dialog so that it aligns with the top of the target. It will place it
        /// to the side of the target, with preference to the right side, but will use the side with the most space.
        /// </summary>
        /// <param name="dialogSize"></param>
        private void PositionDialogAlignedTop(Windows.Foundation.Size dialogSize)
        {
            UIRect bounds = GetNormalizedWindowsBounds();
            // position vertically
            double verticalOffset = _Target.Top;
            _Dialog.VerticalAlignment = VerticalAlignment.Top;

            // position horizontally
            double horizontalOffset = 0;
            _Dialog.HorizontalAlignment = HorizontalAlignment.Left;
            if (bounds.Right - _Target.Right >= dialogSize.Width)
            {
                horizontalOffset = _Target.Right;
            }
            else if (_Target.Left - bounds.Left >= dialogSize.Width)
            {
                horizontalOffset = _Target.Left - dialogSize.Width;
            }
            else
            {
                if (bounds.Right - _Target.Right >= _Target.Left - bounds.Left)
                {
                    _Dialog.HorizontalAlignment = HorizontalAlignment.Right;
                }
                else
                {
                    _Dialog.HorizontalAlignment = HorizontalAlignment.Left;
                }
            }

            _Dialog.Margin = new Thickness(horizontalOffset, verticalOffset, 0, 0);
        }

        /// <summary>
        /// Will place the dialog to the side of the target if it fits completely, otherwise, it will put it above or below.
        /// If it doesn't fit anywhere, it will place it to the side with the most room.
        /// </summary>
        /// <returns>True if the dialog was placed</returns>
        /// <param name="dialogSize"></param>
        /// <param name="force">Forces the dialog to be placed. It will first try above or below though</param>
        private bool PositionDialogPreferSide(Windows.Foundation.Size dialogSize, bool force = false)
        {
            UIRect bounds = GetNormalizedWindowsBounds();

            bool placed = false;
            bool fitsHorizontally = false;
            // position vertically
            double verticalOffset = _Target.Top;
            _Dialog.VerticalAlignment = VerticalAlignment.Top;

            // position horizontally
            double horizontalOffset = 0;
            _Dialog.HorizontalAlignment = HorizontalAlignment.Left;
            if (bounds.Right - _Target.Right >= dialogSize.Width)
            {
                horizontalOffset = _Target.Right;
                fitsHorizontally = true;
                _CommitedPosition = CommitedPositions.Side;
            }
            else if (_Target.Left - bounds.Left >= dialogSize.Width)
            {
                horizontalOffset = _Target.Left - dialogSize.Width;
                fitsHorizontally = true;
                _CommitedPosition = CommitedPositions.Side;
            }
            else if (force)
            {
                placed = PositionDialogPreferAboveOrBelow(dialogSize);

                if (!placed)
                {
                    _CommitedPosition = CommitedPositions.Side;
                    if (bounds.Right - _Target.Right >= _Target.Left - bounds.Left)
                    {
                        _Dialog.HorizontalAlignment = HorizontalAlignment.Right;
                    }
                    else
                    {
                        _Dialog.HorizontalAlignment = HorizontalAlignment.Left;
                    }
                }
                fitsHorizontally = true;
            }

            if (fitsHorizontally && !placed)
            {
                if (dialogSize.Height > bounds.Bottom - _Target.Top)
                {
                    verticalOffset = 0;
                    _Dialog.VerticalAlignment = VerticalAlignment.Bottom;
                    //verticalOffset = bounds.Bottom - dialogSize.Height;
                    //if (verticalOffset < bounds.Top)
                    //{
                    //    verticalOffset = bounds.Top;
                    //}
                }
                placed = true;
                _Dialog.Margin = new Thickness(horizontalOffset, verticalOffset, 0, 0);
            }

            return placed;
        }

        /// <summary>
        /// Tries to position the dialog below the target, but if it doesn't fit will go above.
        /// If it can't fit the rectangle either above or below, it will return false.
        ///
        /// If force is set to true, then it will see if it can place it to the side, and if that doesn't work, it will place it above or below, wherever there is more room.
        /// </summary>
        /// <param name="dialogSize"></param>
        /// <param name="force"></param>
        /// <returns></returns>
        private bool PositionDialogPreferAboveOrBelow(Windows.Foundation.Size dialogSize, bool force = false)
        {
            UIRect bounds = GetNormalizedWindowsBounds();
            bool placed = false;
            bool fitsVertically = false;

            // position horizontally
            double horizontalOffset = _Target.Left;
            _Dialog.HorizontalAlignment = HorizontalAlignment.Left;

            // position vertically
            double verticalOffset = 0;
            _Dialog.VerticalAlignment = VerticalAlignment.Top;

            if (bounds.Bottom - _Target.Bottom >= dialogSize.Height)
            {
                verticalOffset = _Target.Bottom;
                fitsVertically = true;
                _CommitedPosition = CommitedPositions.Below;
            }
            else if (_Target.Top - bounds.Top >= dialogSize.Height)
            {
                _Dialog.VerticalAlignment = VerticalAlignment.Bottom;
                verticalOffset = bounds.Bottom - _Target.Top;
                fitsVertically = true;
                _CommitedPosition = CommitedPositions.Above;
            }
            else if (force)
            {
                placed = PositionDialogPreferSide(dialogSize);

                if (!placed)
                {
                    if (bounds.Bottom - _Target.Bottom >= _Target.Top - bounds.Top)
                    {
                        _CommitedPosition = CommitedPositions.Below;
                        _Dialog.VerticalAlignment = VerticalAlignment.Bottom;
                    }
                    else
                    {
                        _CommitedPosition = CommitedPositions.Above;
                        _Dialog.VerticalAlignment = VerticalAlignment.Top;
                    }
                }
                fitsVertically = true;
            }

            if (fitsVertically && !placed)
            {
                if (dialogSize.Width > bounds.Right - _Target.Left)
                {
                    horizontalOffset = bounds.Right - dialogSize.Width;
                    if (horizontalOffset < bounds.Left)
                    {
                        horizontalOffset = bounds.Left;
                    }
                }

                placed = true;
                if (_Dialog.VerticalAlignment == VerticalAlignment.Bottom)
                {
                    _Dialog.Margin = new Thickness(horizontalOffset, 0, 0, verticalOffset);
                }
                else
                {
                    _Dialog.Margin = new Thickness(horizontalOffset, verticalOffset, 0, 0);
                }

            }

            return placed;
        }

        private UIRect GetNormalizedWindowsBounds()
        {
            UIRect bounds = Window.Current.Bounds;
            bounds.X = 0;
            bounds.Y = 0;
            return bounds;
        }

    }
}
