﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using Windows.System;
using Windows.UI.Xaml.Input;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class AuthorDialogViewModel : ViewModelBase
    {
        public delegate void AuthorDialogFinishedDelegate(string authorName);
        public event AuthorDialogFinishedDelegate AuthorDialogFinished;

        private string _authorName = string.Empty;

        public AuthorDialogViewModel()
        {
            InitCommands();
        }

        private void InitCommands()
        {
            AuthorNameKeyUpCommand = new RelayCommand(AuthorNameKeyUpCommandImpl);
            AuthorNameChangedCommand = new RelayCommand(AuthorNameChangedCommandImpl);
            AuthorNameOkPressedCommand = new RelayCommand(AuthorNameOkPressedCommandImpl);
            AuthorNameCancelPressedCommand = new RelayCommand(AuthorNameCancelPressedCommandImpl);
        }

        public RelayCommand AuthorNameKeyUpCommand { get; private set; }

        public RelayCommand AuthorNameChangedCommand { get; private set; }

        public RelayCommand AuthorNameOkPressedCommand { get; private set; }

        public RelayCommand AuthorNameCancelPressedCommand { get; private set; }

        private void AuthorNameKeyUpCommandImpl(object keyArgs)
        {
            KeyRoutedEventArgs args = keyArgs as KeyRoutedEventArgs;
            if (args.Key == VirtualKey.Enter)
            {
                AuthorEntered();
            }
            else if (args.Key == VirtualKey.Escape)
            {
                AuthorCanceled();
            }
        }

        private void AuthorNameChangedCommandImpl(object newAuthorName)
        {
            if (newAuthorName is string authorName)
                AuthorName = authorName;
        }

        private void AuthorNameOkPressedCommandImpl(object sender) => AuthorEntered();

        private void AuthorNameCancelPressedCommandImpl(object sender) => AuthorCanceled();

        public string AuthorName
        {
            get => _authorName;
            set
            {
                if (value?.Equals(_authorName) == false)
                {
                    _authorName = value;
                    RaisePropertyChanged();
                    RaisePropertyChanged(nameof(DoesAuthorNameHaveContent));
                }
            }
        }

        public bool DoesAuthorNameHaveContent => !string.IsNullOrWhiteSpace(AuthorName);

        private void AuthorEntered() => RaiseAuthorFinishedEvent(AuthorName);

        private void AuthorCanceled() => RaiseAuthorFinishedEvent(null);

        private void RaiseAuthorFinishedEvent(string password) => AuthorDialogFinished?.Invoke(password);

        public void GoBack() => AuthorCanceled();
    }
}
