﻿using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Utilities;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public class CalibrationViewModel : ViewModelBase
    {
        MeasureInfoGroup mMeasureInfoGroup;
        IAnnot mAnnot;
        PDFViewCtrl mPDFView;

        public CalibrationViewModel(IAnnot annot, PDFViewCtrl pDFViewCtrl)
        {
            TranslateUnits = new ObservableCollection<string>();
            TranslateUnits.Add("pt");
            TranslateUnits.Add("in");
            TranslateUnits.Add("mm");
            TranslateUnits.Add("cm");
            TranslateUnits.Add("m");
            TranslateUnits.Add("km");
            TranslateUnits.Add("ft");
            TranslateUnits.Add("yd");
            TranslateUnits.Add("mi");

            mAnnot = annot;
            mPDFView = pDFViewCtrl;

            mMeasureInfoGroup = MeasureUtils.GetAnnotMeasureInfo(annot);
            SelectedTranslateUnit = mMeasureInfoGroup.DistanceInfo.Unit;
        }

        #region Properties

        string _selectedTranslateUnit;
        public string SelectedTranslateUnit
        {
            get { return _selectedTranslateUnit; }
            set
            {
                _selectedTranslateUnit = value;
                RaisePropertyChanged();
            }
        }

        string _calibrationValue;
        public string CalibrationValue
        {
            get { return _calibrationValue; }
            set
            {
                _calibrationValue = value;
                RaisePropertyChanged();
            }
        }

        public ObservableCollection<string> TranslateUnits { get; set; }

        #endregion

        public bool PerformCalibration()
        {
            if (mAnnot == null)
                return false;

            var mLine = (pdftron.PDF.Annots.Line)mAnnot;

            double dValue = 0;
            if (double.TryParse(CalibrationValue, out dValue) == false)
                return false;

            // Prepare annotation data
            MeasureInfo measureInfo = new MeasureInfo();
            measureInfo.Precision = MeasureUtils.ConvStrToPDFPrecision(CalibrationValue);
            measureInfo.TranslateFactor = mMeasureInfoGroup.AxisInfo.TranslateFactor;
            measureInfo.BaseUnit = mMeasureInfoGroup.AxisInfo.BaseUnit;
            measureInfo.Unit = SelectedTranslateUnit;

            var startPt = mLine.GetStartPoint();
            var endPt = mLine.GetEndPoint();

            var Width = Math.Abs(endPt.x - startPt.x);
            var Height = Math.Abs(endPt.y - startPt.y);

            var line_length = Math.Sqrt(Width * Width + Height * Height);
            // Calculate new Factor based on the calibrate value
            measureInfo.BaseFactor = (dValue / line_length);

            try
            {
                mPDFView.DocLock(true);               

                //var dVal = dValue * mMeasureInfoGroup.AxisInfo.BaseFactor * mMeasureInfoGroup.DistanceInfo.BaseFactor;
                var strVal = MeasureUtils.FormatDecimals(dValue, MeasureUtils.ConvPDFPrecisionToStr(measureInfo.Precision)) + " " + measureInfo.Unit;

                MeasureUtils.SetMeasurementInfo(mLine, measureInfo);

                mLine.SetContents(strVal);
                mLine.RefreshAppearance();
                mPDFView.UpdateWithAnnot(mLine, mPDFView.GetCurrentPage());
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }

            return true;
        }

    }
}
