﻿using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Runtime.CompilerServices;

namespace pdftron.PDF.Tools.Controls.ViewModels.Common
{
    public abstract class ViewModelBase : INotifyPropertyChanged
    {
        PropertyChangedEventHandler _propertyChangedHandler;
        public event PropertyChangedEventHandler PropertyChanged
        {
            add
            {
                if (_propertyChangedHandler?.GetInvocationList().Contains(value) != true)
                    _propertyChangedHandler += value;
            }
            remove => _propertyChangedHandler -= value;
        }

        protected bool Set<T>(ref T field, T newValue = default, [CallerMemberName] string propertyName = null)
        {
            if (EqualityComparer<T>.Default.Equals(field, newValue))
                return false;

            field = newValue;
            RaisePropertyChanged(propertyName);

            return true;
        }

        protected virtual void RaisePropertyChanged([CallerMemberName] string propertyName = null)
        {
            _propertyChangedHandler?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        /// <summary>
        /// This will make the view model go back if applicable. Override this on ViewModels.
        /// </summary>
        /// <returns>True if the view model performed a backward navigation step, false otherwise</returns>
        public virtual bool GoBack()
        {
            return false;
        }
    }
}
