﻿namespace pdftron.PDF.Tools.Controls.ViewModels.FileOpening
{
    public enum OpenedDocumentStates
    {
        /// <summary>
        /// The default state. Should never be in this state when a document is passed to the Viewer
        /// </summary>
        None,
        /// <summary>
        /// Normal document, can read/write.
        /// </summary>
        Normal,
        /// <summary>
        /// This document can be saved to, but requires a full save (mostly for dropbox).
        /// </summary>
        NeedsFullSave,
        /// <summary>
        /// This is a corrupt document, the user will have to save it manually
        /// </summary>
        Corrupted,
        /// <summary>
        /// The user has edited a corrupt document. This means we need to notify them when they try to navigate away.
        /// </summary>
        CorruptedAndModified,
        /// <summary>
        /// This is a read only document. If it is modified, we should prompt the user to save it.
        /// </summary>
        ReadOnly,
        /// <summary>
        /// This is a document created by the app, we should always prompt the user to save it.
        /// </summary>
        Created,
        /// <summary>
        /// This is a non-pdf document, so we can't save back to it.
        /// </summary>
        NonePDF,
        /// <summary>
        /// This is a state in which we do not allow the document to be edited at all.
        /// This because we could not make a copy of it.
        /// </summary>
        Uneditable,
        /// <summary>
        /// This state signifies that we are opening a universal document.
        /// </summary>
        Universal
    }
}
