﻿using System;
using System.Threading.Tasks;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ControlBase;
using Windows.UI.Xaml.Media;
using System.Collections.ObjectModel;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml;
using System.Collections.Generic;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Linq;

using UIRect = Windows.Foundation.Rect;
using System.IO;
using Windows.UI.Xaml.Media.Imaging;
using Windows.UI;
using Windows.Storage;
using Windows.Storage.Pickers;
using Windows.Storage.Streams;
using pdftron.Common;
using Windows.UI.Popups;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using System.Diagnostics.CodeAnalysis;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Extensions;
using System.Collections.Specialized;
using pdftron.PDF.Tools.Controls.Resources;
using Windows.Foundation;
using pdftron.PDF.Tools.Utilities;
using Windows.System.Profile;
using System.Threading;
using Windows.UI.Core;
using pdftron.PDF.Tools.Controls.ViewModels.Password;
using pdftron.PDF.Tools.Controls.Password;
using Windows.ApplicationModel.Resources;

namespace pdftron.PDF.Tools.Controls.ViewModels
{
    public partial class ThumbnailsViewViewModel : ViewModelBase
    {
        private const int MAX_SPACE_PC = 200000000;
        private const int MAX_SPACE_PHONE = 50000000;

        protected readonly ResourceLoader _loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");

        private readonly IList<string> Valid_Image_Types = new List<string>() { ".jpg", ".bmp", ".jpeg", ".tif", ".tiff", ".png", ".gif", ".ico", ".jpeg-xr" };
        private readonly IList<string> Valid_PDF_Types = new List<string>() { ".pdf" };

        private bool _reorderItems = true;
        private readonly ICloseableControl _view;
        private bool _hasModifiedPageLayout;
        private readonly List<int> _thumbnailRequestQueue;
        private bool _clearingSelection;
        private bool _gotBookmarksFromDocument;
        private bool _filePickingInProgress;
        private bool _enableThumbnailOverlay = true;
        private bool _windowsEditingStyle = true;
        private ThumbnailItem _currentItem;
        private OnThumbnailRequestCompletedEventHandler _thumbnailGeneratedHandler;
        private OnPageNumberChangedEventHandler _pageNumberChangedHanlder;
        private OnConversionEventHandler _conversionChangedHandler;
        private DateTime _lastConversionEvent = DateTime.Now;
        private NotifyCollectionChangedEventHandler _collectionChangedHandler;
        private ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo> _thumbnailsList;
        private ListViewSelectionMode _selectionMode = ListViewSelectionMode.Multiple;
        private bool _isSelectButtonVisible = true;
        private int _numberOfSelectedItems;
        private Color _pageDefaultColor = Colors.White;
        private bool _readOnly;
        private Size _imageMaxDimensions = new Size(0, 0);
        private readonly PDFViewCtrl _pdfViewCtrl;
        private PDFViewCtrlThumbnailRequestHandler _thumbHandler;
        private readonly string _documentTag;
        private bool _pause;
        private bool _unloadingThumbs;
        private SemaphoreSlim _unloadingMutex = new SemaphoreSlim(1);
        private SortedDictionary<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>> _unloadingJobsPending =
            new SortedDictionary<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>>();
        private List<ThumbnailItem> _selectedItems = new List<ThumbnailItem>();
        private object _reorderItem;
        private int _reorderIndexFrom;
        private IAsyncOperation<IUICommand> _deleteErrorMessage;

        public delegate void PageDeletedDelegate(int pageNumber);
        private PageDeletedDelegate _pageDeleted;
        public event PageDeletedDelegate PageDeleted
        {
            add
            {
                if (_pageDeleted?.GetInvocationList().Contains(value) != true)
                    _pageDeleted += value;
            }
            remove => _pageDeleted -= value;
        }

        public delegate void PageAddedDelegate(int pageNumber, ThumbnailItem item);
        private PageAddedDelegate _pageAdded;
        public event PageAddedDelegate PageAdded
        {
            add
            {
                if (_pageAdded?.GetInvocationList().Contains(value) != true)
                    _pageAdded += value;
            }
            remove => _pageAdded -= value;
        }

        public delegate void PageMovedDelegate(int pageNumber, int newLocation);
        private PageMovedDelegate _pageMoved;
        public event PageMovedDelegate PageMoved
        {
            add
            {
                if (_pageMoved?.GetInvocationList().Contains(value) != true)
                    _pageMoved += value;
            }
            remove => _pageMoved -= value;
        }

        public delegate void PageClickedDelegate(int prevPage, int newPage);
        private PageClickedDelegate _pageClicked;
        public event PageClickedDelegate PageClicked
        {
            add
            {
                if (_pageClicked?.GetInvocationList().Contains(value) != true)
                    _pageClicked += value;
            }
            remove => _pageClicked -= value;
        }

        public delegate void PageRotatedDelegate(int pageNumber);
        private PageRotatedDelegate _pageRotated;
        public event PageRotatedDelegate PageRotated
        {
            add
            {
                if (_pageRotated?.GetInvocationList().Contains(value) != true)
                    _pageRotated += value;
            }
            remove => _pageRotated -= value;
        }

        public delegate void ThumbnailReceivedDelegate(int pageNumber);
        private ThumbnailReceivedDelegate _thumbnailReceived;
        public event ThumbnailReceivedDelegate ThumbnailReceived
        {
            add
            {
                if (_thumbnailReceived?.GetInvocationList().Contains(value) != true)
                    _thumbnailReceived += value;
            }
            remove => _thumbnailReceived -= value;
        }

        public RelayCommand RotateCommand { get; }

        public RelayCommand DeleteCommand { get; }

        public RelayCommand CloseCommand { get; }

        public RelayCommand SelectionChangedCommand { get; }

        public RelayCommand ClearSelectionCommand { get; }

        public RelayCommand ItemClickCommand { get; }

        public RelayCommand EditCommand { get; }

        public RelayCommand DoneEditingCommand { get; }

        public RelayCommand ExportSelectedPagesCommand { get; }

        public RelayCommand DuplicateCommand { get; }

        public RelayCommand AddPageCommand { get; }

        public RelayCommand AddImageCommand { get; }

        public RelayCommand AddPDFCommand { get; }

        public RelayCommand InsertPageCommand { get; }

        public RelayCommand InsertImageCommand { get; }

        public RelayCommand InsertPDFCommand { get; }

        public ThumbnailsViewViewModel(PDFViewCtrl ctrl, string docTag, ICloseableControl view)
        {
            _pdfViewCtrl = ctrl;
            _thumbHandler = ctrl.ThumbnailResponseHandler;
            _view = view;

            _documentTag = docTag;
            PDFDoc doc = _pdfViewCtrl.GetDoc();
            if (doc == null && !_pdfViewCtrl.IsConverting)
                return;

            SelectionChangedCommand = new RelayCommand(SelectionChangedCommandImpl);
            DeleteCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(DeleteSelectedPagesAsync);
                command.Execute(parameter);
            });
            RotateCommand = new RelayCommand(RotateSelectedPages);
            CloseCommand = new RelayCommand(Close);
            ClearSelectionCommand = new RelayCommand(ClearSelection);
            ItemClickCommand = new RelayCommand(ItemClick);

            EditCommand = new RelayCommand(Edit);
            DoneEditingCommand = new RelayCommand(DoneEditing);

            ExportSelectedPagesCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(ExportSelectedPagesAsync);
                command.Execute(parameter);
            });
            DuplicateCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(DuplicateSelectedPagesAsync);
                command.Execute(parameter);
            });
            AddPageCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(AddPageAsync);
                command.Execute(parameter);
            });
            AddImageCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(AddImageAsync);
                command.Execute(parameter);
            });
            AddPDFCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(AddPdfAsync);
                command.Execute(parameter);
            });
            InsertPageCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(InsertPageAsync);
                command.Execute(parameter);
            });
            InsertImageCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(InsertImageAsync);
                command.Execute(parameter);
            });
            InsertPDFCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.CreateWithParameter(InsertPdfAsync);
                command.Execute(parameter);
            });

            if (_pdfViewCtrl == null)
                return;

            _thumbnailRequestQueue = new List<int>();
            _thumbnailGeneratedHandler = new OnThumbnailRequestCompletedEventHandler(ThumbHandler_OnThumbnailGenerated);
            _thumbHandler.OnThumbnailGenerated += _thumbnailGeneratedHandler;
            _pageNumberChangedHanlder = new OnPageNumberChangedEventHandler(PDFViewCtrl_OnPageNumberChanged);
            _pdfViewCtrl.OnPageNumberChanged += _pageNumberChangedHanlder;
            if (_pdfViewCtrl.IsConverting)
            {
                _conversionChangedHandler = new OnConversionEventHandler(PDFViewCtrl_OnConversionChanged);
                _pdfViewCtrl.OnConversionChanged += _conversionChangedHandler;
            }

            CreateCollection();

            if (!_pdfViewCtrl.IsConverting)
                GetBookmarks(doc);

            RaisePropertyChanged(nameof(IsConverting));

            SelectionMode = ListViewSelectionMode.None;
        }

        public bool WindowsEditingStyle
        {
            get => _windowsEditingStyle;
            set
            {
                if (value != _windowsEditingStyle)
                {
                    _windowsEditingStyle = value;
                    RaisePropertyChanged();
                    if (_windowsEditingStyle)
                    {
                        SelectionMode = ListViewSelectionMode.Single;
                    }
                    else
                    {
                        SelectionMode = ListViewSelectionMode.None;
                    }
                }
            }
        }

        public ThumbnailItem CurrentItem
        {
            get => _currentItem;
            private set
            {
                if (_currentItem != value)
                {
                    if (_currentItem != null)
                        _currentItem.IsCurrentPage = false;

                    _currentItem = value;
                    if (_currentItem != null)
                        _currentItem.IsCurrentPage = true;

                    RaisePropertyChanged();
                }
            }
        }

        public ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo> ThumbnailsList
        {
            get => _thumbnailsList;
            private set
            {
                if (value != _thumbnailsList)
                {
                    _thumbnailsList = value;
                    RaisePropertyChanged();
                }
            }
        }

        public ListViewSelectionMode SelectionMode
        {
            get => _selectionMode;
            set
            {
                if (value != _selectionMode)
                {
                    _selectionMode = value;
                    RaisePropertyChanged();
                    RaisePropertyChanged(nameof(IsInEditMode));
                }
            }
        }

        public bool IsInEditMode => _selectionMode != ListViewSelectionMode.None;

        public bool IsSelectButtonVisible
        {
            get => _isSelectButtonVisible;
            set => Set(ref _isSelectButtonVisible, value);
        }

        public bool HasSelectedItems => _numberOfSelectedItems > 0;

        public int NumberOfSelectedItems
        {
            get => _numberOfSelectedItems;
            set
            {
                if (value != _numberOfSelectedItems)
                {
                    _numberOfSelectedItems = value;
                    RaisePropertyChanged();
                    RaisePropertyChanged(nameof(HasSelectedItems));
                }
            }
        }

        public Color PageDefaultColor
        {
            get => _pageDefaultColor;
            set
            {
                _pageDefaultColor = value;
                CreateCollection();
            }
        }

        public bool ReadOnly
        {
            get => _readOnly;
            set
            {
                if (value != _readOnly)
                {
                    _readOnly = value;
                    RaisePropertyChanged();
                    if (_readOnly)
                        SelectionMode = ListViewSelectionMode.None;
                }
            }
        }

        public bool IsConverting => _pdfViewCtrl?.IsConverting == true;

        public Size ImageMaxDimensions
        {
            get => _imageMaxDimensions;
            set
            {
                if (Set(ref _imageMaxDimensions, value) && ThumbnailsList?.ItemsWithResource?.Keys != null)
                {
                    foreach (int key in ThumbnailsList.ItemsWithResource.Keys)
                    {
                        if (key >= 0 && key < ThumbnailsList.Count)
                            UpdateScaleFromRotation(ThumbnailsList[key]);
                    }
                }
            }
        }

        public NavigationStack NavigationStack { get; set; }

        public ToolManager ToolManager { get; set; }

        public bool EnabledThumbnailOverlay
        {
            get => _enableThumbnailOverlay;
            set => Set(ref _enableThumbnailOverlay, value);
        }

        public IPasswordUserControl PasswordUserControl => new PdfPasswordUserControl();

        private async void GetBookmarks(PDFDoc doc)
        {
            if (doc != null)
            {
                try
                {
                    if (UserBookmarkManager.DoesDocHaveBookmarks(doc))
                    {
                        await UserBookmarkManager.LoadBookmarksListAsync(doc);
                        _gotBookmarksFromDocument = true;
                    }
                    else
                    {
                        await UserBookmarkManager.LoadBookmarksListAsync(_documentTag);
                    }
                }
                catch (Exception ex)
                {
                    AnalyticsHandlerBase.CURRENT.SendEvent(
                        AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
        }

        private void CreateCollection()
        {
            if (ThumbnailsList != null && _collectionChangedHandler != null)
            {
                ThumbnailsList.CollectionChanged -= _collectionChangedHandler;
                _collectionChangedHandler = null;
            }

            ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo> items = new ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo>
            {
                RequestBehaviour = ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo>.RequestBehaviours.FILO
            };

            int pageCount = _pdfViewCtrl.GetPageCount();
            if (_pdfViewCtrl.IsConverting)
                pageCount = _pdfViewCtrl.ConvertedPages;

            for (int i = 1; i <= pageCount; ++i)
            {
                items.Add(
                    new ThumbnailItem(i, _pageDefaultColor)
                    {
                        MainViewModel = this
                    });
            }

            if (pageCount > 0)
            {
                int currentIndex = _pdfViewCtrl.GetCurrentPage() - 1;
                if (currentIndex >= 0 && currentIndex < pageCount)
                {
                    CurrentItem = items[currentIndex];
                }
                else
                {
                    CurrentItem = items[0];
                }
            }

            ThumbnailsList = items;

            if (!AnalyticsInfo.VersionInfo.DeviceFamily.Equals("Windows.Mobile", StringComparison.OrdinalIgnoreCase))
            {
                ThumbnailsList.MaxAllowedSize = MAX_SPACE_PC;
            }
            else
            {
                ThumbnailsList.MaxAllowedSize = MAX_SPACE_PHONE;
            }

            _collectionChangedHandler = new NotifyCollectionChangedEventHandler(ThumbnailsList_CollectionChanged);
            _thumbnailsList.CollectionChanged += _collectionChangedHandler;
            _thumbnailsList.ResourceNeededForIndex += ThumbnailsList_ResourceNeededForIndex;
            _thumbnailsList.DiscardResourceForIndex += ThumbnailsList_DiscardResourceForIndex;
            _thumbnailsList.AllRequestsCancelled += ThumbnailsList_AllRequestsCancelled;
        }

        public void CleanUp()
        {
            if (_thumbnailGeneratedHandler != null)
            {
                _pdfViewCtrl.CancelAllThumbRequests();
                _thumbHandler.OnThumbnailGenerated -= _thumbnailGeneratedHandler;
                _pdfViewCtrl.OnPageNumberChanged -= _pageNumberChangedHanlder;
                _thumbnailGeneratedHandler = null;
                _pageNumberChangedHanlder = null;

                if (_conversionChangedHandler != null)
                {
                    _pdfViewCtrl.OnConversionChanged -= _conversionChangedHandler;
                    _conversionChangedHandler = null;
                }
            }

            _thumbHandler = null;
            if (_collectionChangedHandler != null)
            {
                _thumbnailsList.CollectionChanged -= _collectionChangedHandler;
                _collectionChangedHandler = null;
            }
            try
            {
                _unloadingMutex.Wait();
                _unloadingJobsPending.Clear();
            }
            finally
            {
                _unloadingMutex.Release();
            }
        }

        /// <summary>
        /// This method handles the XAML rotation of a thumbnailItem to increase responsiveness
        /// when rotating a thumbnail, rather than relying on the result of the Thumbnail_Requested event
        /// </summary>
        /// <param name="thumbItem"></param>
        private void HandleRotation(ThumbnailItem thumbItem)
        {
            Page pageToRotate = _pdfViewCtrl.GetDoc().GetPage(thumbItem.PageNumber);
            int pageRotation = ((int)pageToRotate.GetRotation() + 1) % 4;

            thumbItem.Angle = thumbItem.Angle == 270 ? 0 : thumbItem.Angle + 90;
            pageToRotate.SetRotation((PageRotate)pageRotation);

            UpdateScaleFromRotation(thumbItem);
        }

        private void SelectionChangedCommandImpl(object changeArgs)
        {
            if (changeArgs is SelectionChangedEventArgs args)
                ResolveSelection(args);

            if (_clearingSelection)
                return;
        }

        private void ClearSelection(object parameter) => SelectionMode = ListViewSelectionMode.None;

        private void OpenPage(int index)
        {
            if (!ReadOnly)
            {
                _pdfViewCtrl.CancelAllThumbRequests();
                try
                {
                    _pdfViewCtrl.UpdatePageLayout();
                }
                catch (Exception ex)
                {
                    AnalyticsHandlerBase.CURRENT.SendEvent(
                        AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }

            _pageClicked?.Invoke(_pdfViewCtrl.GetCurrentPage(), index + 1);
            if (NavigationStack != null)
            {
                NavigationStack.SetCurrentPage(index + 1);
            }
            else
            {
                _pdfViewCtrl.SetCurrentPage(index + 1);
            }

            _pdfViewCtrl.RequestRendering();

            if (!ReadOnly)
            {
                CleanUp();
                _view.CloseControl();
            }
        }

        [SuppressMessage("Redundancy", "RCS1163:Unused parameter.", Justification = "<Pending>")]
        [SuppressMessage("Style", "IDE0060:Remove unused parameter", Justification = "<Pending>")]
        public void ThumbnailViewerDoubleTapped(object sender, DoubleTappedRoutedEventArgs e)
        {
            if (e.OriginalSource is FrameworkElement element && element.DataContext is ThumbnailItem item)
            {
                int index = ThumbnailsList.IndexOf(item);
                OpenPage(index);
            }
        }

        private void Edit(object parameter)
        {
            SelectionMode = ListViewSelectionMode.Multiple;
            IsSelectButtonVisible = false;
        }

        private void DoneEditing(object parameter)
        {
            SelectionMode = ListViewSelectionMode.None;
            IsSelectButtonVisible = true;
        }

        private async Task DuplicateSelectedPagesAsync(object parameter) => await AddNewPageAsync(NewPageType.Duplicate);

        private async Task AddPageAsync(object sender) => await AddNewPageAsync(NewPageType.Blank);

        private async Task AddImageAsync(object sender) => await AddNewPageAsync(NewPageType.Image);

        private async Task AddPdfAsync(object sender) => await AddNewPageAsync(NewPageType.ExistingPDF);

        private async Task InsertPageAsync(object sender) => await AddNewPageAsync(NewPageType.Blank, false);

        private async Task InsertImageAsync(object sender) => await AddNewPageAsync(NewPageType.Image, false);

        private async Task InsertPdfAsync(object sender) => await AddNewPageAsync(NewPageType.ExistingPDF, false);

        /// <summary>
        /// Gets or sets whether this ViewModel is paused.
        /// Settings the ViewModel to paused will not stop any current requests, but
        /// will prevent it from generating new ones.zd
        /// </summary>
        public bool Pause
        {
            get => _pause;
            set
            {
                if (Set(ref _pause, value) && !_pause)
                    ThumbnailsList.Refresh();
            }
        }

        /// <summary>
        /// Enable double tap to open/navigate selected thumbnail to a page. Default is single tap to open
        /// </summary>
        public bool DoubleTapToNavigate { get; set; }

        public override bool GoBack()
        {
            if (IsInEditMode)
            {
                SelectionMode = ListViewSelectionMode.None;
                return true;
            }

            _view.CloseControl();
            return true;
        }

        public void Refresh()
        {
            foreach (int index in ThumbnailsList.ItemsWithResource.Keys)
            {
                ThumbnailsList[index].Thumbnail = null;
            }

            ThumbnailsList.Refresh();
        }

        public void PageModified(int pageNumber)
        {
            int index = pageNumber - 1;
            if (index >= 0 && ThumbnailsList.Count > index && ThumbnailsList[index].HasThumbnail)
                ThumbnailsList.RefreshResource(index);
        }

        private void UpdatePageCount()
        {
            for (int i = 0; i < ThumbnailsList.Count; i++)
            {
                ThumbnailsList[i].PageNumber = i + 1;
            }

            List<ThumbInfo> newInfoList = new List<ThumbInfo>();
            foreach (ThumbInfo info in ThumbnailsList.ItemsWithResource.Values)
            {
                newInfoList.Add(info);
            }

            ThumbnailsList.ItemsWithResource.Clear();
            foreach (ThumbInfo info in newInfoList)
            {
                if (!info.SourceItem.IsDeleted)
                    ThumbnailsList.ItemsWithResource.TryAdd(info.SourceItem.PageNumber, info);
            }
        }

        private void ThumbnailsList_ResourceNeededForIndex(object sender, int index)
        {
            if (_thumbHandler == null || _pdfViewCtrl.GetDoc() == null)
                return;

            if (_pdfViewCtrl.IsConverting && _pdfViewCtrl.ConvertedPages <= 0)
                return;

            try
            {
                _thumbHandler.GetThumbAsync(index + 1);
            }
            catch
            {
                //swallow exception
            }
        }

        private void ThumbnailsList_DiscardResourceForIndex(object sender, int index)
        {
            ThumbnailsList[index].Thumbnail = null;
            ThumbnailsList[index].Angle = 0;
        }

        private void ThumbnailsList_AllRequestsCancelled(object sender, object e)
        {
            if (_thumbHandler == null)
                return;

            if (_pdfViewCtrl.IsConverting && _pdfViewCtrl.ConvertedPages <= 0)
                return;

            _pdfViewCtrl.CancelAllThumbRequests();
            try
            {
                _unloadingMutex.Wait();
                _unloadingJobsPending.Clear();
            }
            finally
            {
                _unloadingMutex.Release();
            }
        }

        private KeyValuePair<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>> FindFirstVisibleJob()
        {
            KeyValuePair<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>> job = _unloadingJobsPending.Last();
            if (ThumbnailsList.Direction == ObservableCollectionWithAsyncResource<ThumbnailItem, ThumbInfo>.Directions.Forwards)
            {
                int firstVisiblePageNumber = ThumbnailsList.FirstVisibleIndex + 1;
                if (job.Key > firstVisiblePageNumber)
                    job = _unloadingJobsPending.First(kvp => kvp.Key >= firstVisiblePageNumber);
            }
            else
            {
                job = _unloadingJobsPending.First();
                int lastVisiblePageNumber = ThumbnailsList.LastVisibleIndex + 1;
                if (job.Key < lastVisiblePageNumber)
                    job = _unloadingJobsPending.Last(kvp => kvp.Key <= lastVisiblePageNumber);
            }

            return job;
        }

        private Tuple<ThumbInfo, InMemoryRandomAccessStream> GetNextJob()
        {
            Tuple<ThumbInfo, InMemoryRandomAccessStream> currentJob = null;
            try
            {
                _unloadingMutex.Wait();
                if (_unloadingJobsPending.Count > 0 && _thumbHandler != null)
                {
                    KeyValuePair<int, Tuple<ThumbInfo, InMemoryRandomAccessStream>> job = FindFirstVisibleJob();
                    currentJob = job.Value;
                    _unloadingJobsPending.Remove(job.Key);
                }
                else
                {
                    _unloadingThumbs = false;
                }
            }
            finally
            {
                _unloadingMutex.Release();
            }

            return currentJob;
        }

        private async Task StartUnloadingTaskAsync()
        {
            Tuple<ThumbInfo, InMemoryRandomAccessStream> currentJob = GetNextJob();

            while (currentJob != null && _thumbHandler != null)
            {
                int index = currentJob.Item1.PageNumber - 1;
                if (index >= 0 && index < ThumbnailsList.Count)
                {
                    currentJob.Item1.SourceItem = ThumbnailsList[index];
                    BitmapImage bmi = new BitmapImage();
                    await bmi.SetSourceAsync(currentJob.Item2);
                    if (_thumbHandler != null)
                    {
                        bool hadThumbnail = ThumbnailsList[index].Thumbnail != null;
                        ThumbnailsList[index].Thumbnail = bmi;
                        ThumbnailsList[index].BitmapSize = new Size(bmi.PixelWidth, bmi.PixelHeight);
                        UpdateScaleFromRotation(ThumbnailsList[index]);
                        bool couldAdd = ThumbnailsList.ItemsWithResource.TryAdd(index, currentJob.Item1);
                        if (!hadThumbnail && couldAdd)
                            _thumbnailReceived?.Invoke(currentJob.Item1.PageNumber);

                        int visibleItems = ThumbnailsList.LastVisibleIndex - ThumbnailsList.FirstVisibleIndex + 1;

                        await Task.Delay(1000 / visibleItems);
                    }

                    currentJob = GetNextJob();
                }
            }

            if (_thumbHandler != null)
                ThumbnailsList.RequestResources();
        }

        private async void ThumbHandler_OnThumbnailGenerated(object sender, int pageNumber, bool wasThumbFound, byte[] thumb, int w, int h)
        {
            InMemoryRandomAccessStream iras = null;
            if (wasThumbFound)
            {
                try
                {
                    iras = await UtilityFunctions.ConvertToPngStreamAsync(thumb, w, h).ConfigureAwait(false);
                }
                catch (Exception ex)
                {
                    AnalyticsHandlerBase.CURRENT.SendEvent(
                        AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        AnalyticsHandlerBase.GetExceptionMessage(ex));
                    wasThumbFound = false;
                }
            }

            if (wasThumbFound)
            {
                wasThumbFound = false;
                try
                {
                    _unloadingMutex.Wait();
                    if (!_unloadingJobsPending.ContainsKey(pageNumber))
                    {
                        ThumbInfo thumbInfo = new ThumbInfo(pageNumber, w, h, null);
                        _unloadingJobsPending.Add(pageNumber, new Tuple<ThumbInfo, InMemoryRandomAccessStream>(thumbInfo, iras));
                    }
                    if (!_unloadingThumbs)
                    {
                        _unloadingThumbs = true;
                        wasThumbFound = true;
                    }
                }
                finally
                {
                    _unloadingMutex.Release();
                }
            }

            if (wasThumbFound)
                await _pdfViewCtrl.Dispatcher.RunAsync(CoreDispatcherPriority.Normal, async () => await StartUnloadingTaskAsync());
        }

        private void PDFViewCtrl_OnPageNumberChanged(int current_page, int num_pages)
        {
            if (ThumbnailsList != null && current_page > 0 && current_page <= ThumbnailsList.Count)
                CurrentItem = ThumbnailsList[current_page - 1];
        }

        private void PDFViewCtrl_OnConversionChanged(PDFViewCtrlConversionType type, int totalPagesConverted)
        {
            bool update = false;
            if (type == PDFViewCtrlConversionType.e_conversion_finished)
            {
                update = true;
                RaisePropertyChanged(nameof(IsConverting));
            }
            else if (type == PDFViewCtrlConversionType.e_conversion_progress && DateTime.Now - _lastConversionEvent > TimeSpan.FromSeconds(1))
            {
                update = true;
                GetBookmarks(_pdfViewCtrl.GetDoc());
            }

            if (update)
            {
                _lastConversionEvent = DateTime.Now;
                int totalPages = _pdfViewCtrl.GetPageCount();
                for (int i = ThumbnailsList.Count; i < totalPages; ++i)
                {
                    ThumbnailItem item = new ThumbnailItem(i + 1, _pageDefaultColor)
                    {
                        MainViewModel = this
                    };

                    ThumbnailsList.Add(item);
                }
            }
        }

        private void UpdateScaleFromRotation(ThumbnailItem item)
        {
            if (!item.HasThumbnail
                || ImageMaxDimensions.IsEmpty
                || ImageMaxDimensions.Width == 0
                || ImageMaxDimensions.Height == 0
                || item.BitmapSize.IsEmpty
                || item.BitmapSize.Width == 0
                || item.BitmapSize.Height == 0)
            {
                item.UniformScale = 1.0;
                return;
            }

            if (item.Angle == 0 || item.Angle == 180)
            {
                item.UniformScale = 1.0;
                return;
            }

            double widthRatio = ImageMaxDimensions.Width / item.BitmapSize.Width;
            double heightRatio = ImageMaxDimensions.Height / item.BitmapSize.Height;

            double defaultScale = Math.Min(widthRatio, heightRatio);
            double visualImageWidth = defaultScale * item.BitmapSize.Width;
            double visualImageHeight = defaultScale * item.BitmapSize.Height;

            double rotatedWidthGrowth = ImageMaxDimensions.Width / visualImageHeight;
            double rotatedHeightGrowth = ImageMaxDimensions.Height / visualImageWidth;

            item.UniformScale = Math.Min(rotatedWidthGrowth, rotatedHeightGrowth);
        }

        private void ResolveSelection(SelectionChangedEventArgs args)
        {
            foreach (ThumbnailItem item in args.RemovedItems)
            {
                if (_selectedItems.Contains(item))
                    _selectedItems.Remove(item);
            }
            foreach (ThumbnailItem item in args.AddedItems)
            {
                if (!_selectedItems.Contains(item))
                    _selectedItems.Add(item);
            }

            NumberOfSelectedItems = _selectedItems.Count;
            if (NumberOfSelectedItems == 0)
                DoneEditing(null);
        }

        private void ThumbnailsList_CollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            switch (e.Action)
            {
                case NotifyCollectionChangedAction.Remove:
                    _reorderItem = e.OldItems[0];
                    _reorderIndexFrom = e.OldStartingIndex;
                    break;
                case NotifyCollectionChangedAction.Add:
                    if (_reorderItems)
                    {
                        if (_reorderItem == null)
                            return;

                        HandleReorder(_reorderItem, _reorderIndexFrom, e.NewStartingIndex);
                        _reorderItem = null;
                    }
                    break;
            }
        }

        private void HandleReorder(object item, int indexFrom, int indexTo)
        {
            _hasModifiedPageLayout = true;
            _pdfViewCtrl.CancelAllThumbRequests();
            PDFDoc doc = _pdfViewCtrl.GetDoc();

            try
            {
                _pdfViewCtrl.CancelRendering();
                doc.Lock();

                Page pageToMove = doc.GetPage(indexFrom + 1);
                int oldNum = pageToMove.GetSDFObj().GetObjNum();

                PageIterator removeIter = doc.GetPageIterator(indexFrom + 1);
                doc.PageRemove(removeIter);

                PageIterator moveTo = doc.GetPageIterator(indexTo + 1);
                doc.PageInsert(moveTo, pageToMove);
                
                if (_gotBookmarksFromDocument)
                {
                    PageIterator createdPage = doc.GetPageIterator(indexTo + 1);
                    UserBookmarkManager.MoveBookmarkDestination(doc, pageToMove, createdPage.Current());
                }
                else
                {
                    Page newPage = doc.GetPage(indexTo + 1);
                    int newNum = newPage.GetSDFObj().GetObjNum();
                    UserBookmarkManager.PageMoved(_documentTag, oldNum, newNum, indexFrom + 1, indexTo + 1);
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.Unlock();
            }

            ThumbnailsList.RequestResources();
            _pageMoved?.Invoke(indexFrom + 1, indexTo + 1);
            UpdatePageCount();
        }

        public void DuplicatePage(int pageNumber)
        {
            PDFDoc doc = _pdfViewCtrl.GetDoc();
            _pdfViewCtrl.CancelAllThumbRequests();

            // Calculate where to place new pages
            int offset = pageNumber + 1;

            int pages = 1;
            bool needsUnlock = false;

            try
            {
                _pdfViewCtrl.DocLock(true);
                needsUnlock = true;

                PageIterator pageIterator = doc.GetPageIterator(pageNumber);
                Page page = doc.GetPage(pageNumber);
                page.SetRotation(pageIterator.Current().GetRotation());

                if (pageIterator.HasNext())
                {
                    pageIterator.Next();
                    doc.PageInsert(pageIterator, page);
                }
                else
                {
                    doc.PagePushBack(page);
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (needsUnlock)
                    doc.Unlock();
            }

            _reorderItems = false;
            _hasModifiedPageLayout = true;
            // Add all the needed pages to thumbnail items
            ThumbnailItem item = null;
            for (int i = 0; i < pages; i++)
            {
                item = new ThumbnailItem(offset + i, _pageDefaultColor)
                {
                    MainViewModel = this
                };

                ThumbnailsList.Insert(offset + i - 1, item);
            }

            ThumbnailsList.Refresh();

            if (doc.GetPageCount() != pageNumber)
                UpdatePageCount();

            _pdfViewCtrl.UpdatePageLayout();
            ThumbnailsList.RequestResources();

            _pageAdded?.Invoke(offset, item);

            if (ToolManager?.UndoRedoAction != null)
                ToolManager.UndoRedoAction.TakeSnapshot("Duplicate Page", true);

            _reorderItems = true;
        }

        private async Task DeleteItemsAsync(ThumbnailItem thumbItem = null)
        {
            if (NumberOfSelectedItems == ThumbnailsList.Count && thumbItem == null)
            {
                if (_deleteErrorMessage == null)
                {
                    ContentDialog contentDialog = new ContentDialog
                    {
                        Title = ResourceHandler.GetString("ThumbnailsView_Delete_Error_Info"),
                        Content = ResourceHandler.GetString("ThumbnailsView_Delete_Error_Title"),
                        CloseButtonText = ResourceHandler.GetString("Button_OK_Text"),
                    };

                    try
                    {
                        await DialogHelper.ShowContentDialogAsync(contentDialog);
                    }
                    catch (Exception ex)
                    {
                        AnalyticsHandlerBase.CURRENT.SendEvent(
                            AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                            AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }
                }

                return;
            }

            _hasModifiedPageLayout = true;

            _pdfViewCtrl.CancelAllThumbRequests();
            List<ThumbnailItem> itemsToDelete = new List<ThumbnailItem>();
            List<int> pagesToDelete = new List<int>();
            if (thumbItem != null)
            {
                itemsToDelete.Add(thumbItem);
                pagesToDelete.Add(ThumbnailsList.IndexOf(thumbItem) + 1);
            }
            else
            {
                foreach (ThumbnailItem item in _selectedItems)
                {
                    itemsToDelete.Add(item);
                    pagesToDelete.Add(ThumbnailsList.IndexOf(item) + 1);
                }

                SelectionMode = ListViewSelectionMode.None;
            }

            ThumbnailItem currentItem = CurrentItem;
            foreach (ThumbnailItem item in itemsToDelete)
            {
                if (item == currentItem)
                    currentItem = null;

                item.IsDeleted = true;
                ThumbnailsList.Remove(item);
            }

            PDFDoc doc = _pdfViewCtrl.GetDoc();
            try
            {
                doc.Lock();
                pagesToDelete.Sort();
                for (int i = pagesToDelete.Count - 1; i >= 0; --i)
                {
                    PageIterator itr = doc.GetPageIterator(pagesToDelete[i]);
                    if (_gotBookmarksFromDocument)
                    {
                        UserBookmarkManager.PageDeleted(doc, itr.Current());
                    }
                    else
                    {
                        UserBookmarkManager.PageDeleted(_documentTag, itr.Current().GetSDFObj().GetObjNum(), pagesToDelete[i], _pdfViewCtrl.GetPageCount());
                    }
                    doc.PageRemove(itr);
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                doc.Unlock();
            }

            _pdfViewCtrl.UpdatePageLayout();
            if (currentItem != null && ThumbnailsList.Contains(currentItem))
            {
                int index = ThumbnailsList.IndexOf(currentItem);
                _pdfViewCtrl.SetCurrentPage(index + 1);
                CurrentItem = currentItem;
            }

            ThumbnailsList.Refresh();
            foreach (int pageNum in pagesToDelete)
            {
                _pageDeleted?.Invoke(pageNum);
            }

            // Update all page numbers after removing pages
            UpdatePageCount();
        }

        private async Task<StorageFile> GetOpenFileFromPicker(IList<string> fileTypes)
        {
            FileOpenPicker fileOpenPicker = new FileOpenPicker
            {
                ViewMode = PickerViewMode.Thumbnail
            };

            foreach (string filetype in fileTypes)
            {
                fileOpenPicker.FileTypeFilter.Add(filetype);
            }

            return await fileOpenPicker.PickSingleFileAsync();
        }

        public void AddBlankPage(bool beforePage, int pageNumber)
        {
            PDFDoc doc = _pdfViewCtrl.GetDoc();
            _pdfViewCtrl.CancelAllThumbRequests();

            // Calculate where to place new pages
            int offset;
            if (beforePage)
            {
                offset = pageNumber;
            }
            else
            {
                offset = pageNumber + 1;
            }

            int pages = 1;
            bool needsUnlock = false;

            try
            {
                _pdfViewCtrl.DocLock(true);
                needsUnlock = true;

                PageIterator pageIterator = doc.GetPageIterator(pageNumber);
                Rect mediaBox = pageIterator.Current().GetBox(PageBox.e_media);
                Rect cropBox = pageIterator.Current().GetBox(PageBox.e_crop);
                Rect userCrop = pageIterator.Current().GetBox(PageBox.e_user_crop);

                Page page = doc.PageCreate(mediaBox);
                page.SetBox(PageBox.e_crop, cropBox);
                page.SetBox(PageBox.e_user_crop, userCrop);
                page.SetRotation(pageIterator.Current().GetRotation());

                if (beforePage && pageNumber == 1)
                {
                    //Insert page before the first page
                    doc.PagePushFront(page);
                }
                else if (beforePage)
                {
                    doc.PageInsert(pageIterator, page);
                }
                else if (!beforePage && pageIterator.HasNext())
                {
                    pageIterator.Next();
                    doc.PageInsert(pageIterator, page);
                }
                else
                {
                    doc.PagePushBack(page);
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (needsUnlock)
                    doc.Unlock();
            }

            _reorderItems = false;
            _hasModifiedPageLayout = true;
            // Add all the needed pages to thumbnail items
            ThumbnailItem item = null;
            for (int i = 0; i < pages; i++)
            {
                item = new ThumbnailItem(offset + i, _pageDefaultColor)
                {
                    MainViewModel = this
                };

                ThumbnailsList.Insert(offset + i - 1, item);
            }

            ThumbnailsList.Refresh();

            if (doc.GetPageCount() != pageNumber)
                UpdatePageCount();

            _pdfViewCtrl.UpdatePageLayout();
            ThumbnailsList.RequestResources();

            _pageAdded?.Invoke(offset, item);

            if (ToolManager?.UndoRedoAction != null)
                ToolManager.UndoRedoAction.TakeSnapshot("Add new Page", true);

            _reorderItems = true;
        }

        /// <summary>
        /// This method adds a new page (blank, image, PDF) into the current PDF
        /// </summary>
        /// <param name="PageType"></param>
        /// <param name="insertAtEnd"> Whether to insert at end. If false, insert immediately after last selected item.</param>
        private async Task AddNewPageAsync(NewPageType PageType, bool insertAtEnd = true)
        {
            PDFDoc doc = _pdfViewCtrl.GetDoc();
            _pdfViewCtrl.CancelAllThumbRequests();

            // Calculate where to place new pages
            int offset = insertAtEnd ? doc.GetPageCount() + 1 : _selectedItems.Max(x => x.PageNumber) + 1;
            int pages = 1;

            switch (PageType)
            {
                case NewPageType.Blank:
                    InsertBlankPage(insertAtEnd);
                    break;
                case NewPageType.Duplicate:
                    InsertDuplicatePage();
                    break;
                case NewPageType.Image:
                    bool pickedImage = await InsertImagePage(insertAtEnd);
                    if (!pickedImage)
                        return;

                    break;
                case NewPageType.ExistingPDF:
                    pages = await InsertPDF(insertAtEnd);
                    if (pages == 0)
                        return;

                    break;
            }

            _hasModifiedPageLayout = true;
            // Add all the needed pages to thumbnail items
            ThumbnailItem item = null;
            for (int i = 0; i < pages; i++)
            {
                item = new ThumbnailItem(offset + i, _pageDefaultColor)
                {
                    MainViewModel = this
                };

                ThumbnailsList.Insert(offset + i - 1, item);
            }

            ThumbnailsList.Refresh();

            if (!insertAtEnd)
                UpdatePageCount();

            _pdfViewCtrl.UpdatePageLayout();
            ThumbnailsList.RequestResources();

            _pageAdded?.Invoke(offset, item);

            if (ToolManager?.UndoRedoAction != null)
                ToolManager.UndoRedoAction.TakeSnapshot("Add new Page", true);
        }

        private void InsertBlankPage(bool insertAtEnd)
        {
            bool needsUnlock = false;
            PDFDoc doc = _pdfViewCtrl.GetDoc();

            try
            {
                _pdfViewCtrl.DocLock(true);
                needsUnlock = true;
                int pageNumberToFolow = 0;
                if (insertAtEnd)
                {
                    pageNumberToFolow = doc.GetPageCount();
                }
                else
                {
                    // Insert right after the max selected page number
                    pageNumberToFolow = _selectedItems.Max(x => x.PageNumber);
                }

                PageIterator pageIterator = doc.GetPageIterator(pageNumberToFolow);
                Rect mediaBox = pageIterator.Current().GetBox(PageBox.e_media);
                Rect cropBox = pageIterator.Current().GetBox(PageBox.e_crop);
                Rect userCrop = pageIterator.Current().GetBox(PageBox.e_user_crop);

                Page page = doc.PageCreate(mediaBox);
                page.SetBox(PageBox.e_crop, cropBox);
                page.SetBox(PageBox.e_user_crop, userCrop);
                page.SetRotation(pageIterator.Current().GetRotation());

                if (!insertAtEnd && pageIterator.HasNext())
                {
                    pageIterator.Next();
                    doc.PageInsert(pageIterator, page);
                }
                else
                {
                    doc.PagePushBack(page);
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (needsUnlock)
                    doc.Unlock();
            }
        }

        private void InsertDuplicatePage()
        {
            bool needsUnlock = false;
            PDFDoc doc = _pdfViewCtrl.GetDoc();

            try
            {
                _pdfViewCtrl.DocLock(true);
                needsUnlock = true;
                int pageNumberToFolow = _selectedItems.Max(x => x.PageNumber);

                PageIterator pageIterator = doc.GetPageIterator(pageNumberToFolow);
                Page page = doc.GetPage(pageNumberToFolow);
                page.SetRotation(pageIterator.Current().GetRotation());

                if (pageIterator.HasNext())
                {
                    pageIterator.Next();
                    doc.PageInsert(pageIterator, page);
                }
                else
                {
                    doc.PagePushBack(page);
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (needsUnlock)
                    doc.Unlock();
            }
        }

        private async Task<bool> InsertImagePage(bool insertAtEnd)
        {
            if (_filePickingInProgress)
                return false;

            bool needsUnlock = false;
            PDFDoc doc = _pdfViewCtrl.GetDoc();
            Page page = doc.PageCreate();
            try
            {
                _pdfViewCtrl.DocLock(true);
                needsUnlock = true;
                _filePickingInProgress = true;

                StorageFile file = await GetOpenFileFromPicker(Valid_Image_Types);
                if (file == null)
                    return false;

                Stream stream = await file.OpenStreamForReadAsync();
                byte[] bytes = new byte[(int)stream.Length];
                stream.Read(bytes, 0, (int)stream.Length);

                Image img = Image.Create(doc.GetSDFDoc(), bytes);
                ElementBuilder f = new ElementBuilder();
                Element element = f.CreateImage(img, new Matrix2D(img.GetImageWidth(), 0, 0, img.GetImageHeight(), 0, 0));

                page.SetMediaBox(new Rect(0, 0, img.GetImageWidth(), img.GetImageHeight()));
                page.SetCropBox(new Rect(0, 0, img.GetImageWidth(), img.GetImageHeight()));

                ElementWriter writer = new ElementWriter();
                writer.Begin(page);
                writer.WritePlacedElement(element);
                writer.End();

                if (!insertAtEnd)
                {
                    // Insert right after the max selected page number
                    int offset = _selectedItems.Max(x => x.PageNumber) + 1;
                    PageIterator pageIterator = doc.GetPageIterator(offset);
                    doc.PageInsert(pageIterator, page);
                }
                else
                {
                    doc.PagePushBack(page);
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _filePickingInProgress = false;
                if (needsUnlock)
                    doc.Unlock();
            }

            return true;
        }

        private async Task<int> InsertPDF(bool insertAtEnd)
        {
            if (_filePickingInProgress)
                return 0;

            bool needsUnlock = false;
            PDFDoc doc = _pdfViewCtrl.GetDoc();
            int pagesAdded = 0;

            try
            {
                _pdfViewCtrl.DocLock(true);
                needsUnlock = true;
                _filePickingInProgress = true;

                StorageFile file = await GetOpenFileFromPicker(Valid_PDF_Types);
                if (file == null)
                    return 0;

                PDFDoc docTemp = new PDFDoc(file);
                //Check for password #160
                if (await OpenWithPasswordAsync(docTemp, file))
                {

                    pagesAdded = docTemp.GetPageCount();

                    // Check where to place new pages
                    int offset = insertAtEnd ? doc.GetPageCount() + 1 : _selectedItems.Max(x => x.PageNumber) + 1;
                    doc.InsertPages(offset, docTemp, 1, pagesAdded, PDFDocInsertFlag.e_none);
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(
                    AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                _filePickingInProgress = false;
                if (needsUnlock)
                    doc.Unlock();
            }

            return pagesAdded;
        }

        private async Task<bool> OpenWithPasswordAsync(PDFDoc doc, IStorageFile file)
        {
            bool valid = false;

            //check for PDF password
            PdfPasswordViewModel viewModel = new PdfPasswordViewModel
            {
                FileToCheck = file
            };

            PasswordUserControl.SetViewModel(viewModel);

            if (viewModel.GetPasswordRequired())
            {
                ContentDialog contentDialog = new ContentDialog
                {
                    DataContext = viewModel,
                    Content = PasswordUserControl,
                    Background = (SolidColorBrush)Application.Current.Resources["ContentDialogBackground"],
                    PrimaryButtonText = _loader.GetString("Generic_Unlock_Text"),
                    SecondaryButtonText = _loader.GetString("Generic_Cancel_Text"),
                    PrimaryButtonCommand = viewModel.PasswordEnteredCommand,
                    PrimaryButtonCommandParameter = viewModel.UserPassword,
                    SecondaryButtonCommand = viewModel.PasswordCancelCommand,
                    DefaultButton = ContentDialogButton.Primary
                };

                while (viewModel.AttemptsAtPassword > 0)
                {
                    ContentDialogResult result = await DialogHelper.ShowContentDialogAsync(contentDialog);
                    if (result == ContentDialogResult.Primary && viewModel.ValidPassword)
                    {
                        doc.InitStdSecurityHandler(viewModel.UserPassword);
                        valid = true;
                        break;
                    }
                    else if (result == ContentDialogResult.Secondary || result == ContentDialogResult.None)
                    {
                        //Cancel
                        valid = false;
                        break;
                    }
                }

                contentDialog.Content = null;
            }
            else
            {
                valid = true;
            }

            return valid;
        }
    }
}