﻿using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using Windows.ApplicationModel.Resources;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Crop
{
    public partial class CropPopupViewModel : ViewModelBase
    {
        private const int CROP_RECT_WHITE_SPACE_MARGIN = 2;

        private readonly PDFViewCtrl _pdfViewCtrl;
        private bool _isPopupOpen;
        private readonly List<Rect> _automaticBoxes = new List<Rect>();
        private bool _isAutomaticCropping;
        private int _currentAutomaticPage = 1;
        private bool _isAutomaticCancel;
        private CancellationTokenSource _tokenSource;

        private PopupClosedDelegate _popupClosed;
        public delegate void PopupClosedDelegate();
        public event PopupClosedDelegate PopupClosed
        {
            add
            {
                if (_popupClosed?.GetInvocationList().Contains(value) != true)
                    _popupClosed += value;
            }
            remove => _popupClosed -= value;
        }

        private ManualCropRequestedDelegate _manualCropRequested;
        public delegate void ManualCropRequestedDelegate();
        public event ManualCropRequestedDelegate ManualCropRequested
        {
            add
            {
                if (_manualCropRequested?.GetInvocationList().Contains(value) != true)
                    _manualCropRequested += value;
            }
            remove => _manualCropRequested -= value;
        }

        private DocumentEditedDelegate _documentEdited;
        public delegate void DocumentEditedDelegate();
        public event DocumentEditedDelegate DocumentEdited
        {
            add
            {
                if (_documentEdited?.GetInvocationList().Contains(value) != true)
                    _documentEdited += value;
            }
            remove => _documentEdited -= value;
        }

        public RelayCommand AutomaticCroppingCommand { get; }

        public RelayCommand ManualCroppingCommand { get; }

        public RelayCommand RemoveCroppingCommand { get; }

        public RelayCommand OkCommand { get; }

        public RelayCommand CancelAutomaticCommand { get; }

        public CropPopupViewModel(PDFViewCtrl ctrl)
        {
            _pdfViewCtrl = ctrl;

            AutomaticCroppingCommand = new RelayCommand(parameter =>
            {
                var command = AsyncCommand.Create(AutomaticCroppingAsync);
                command.Execute(parameter);
            });
            ManualCroppingCommand = new RelayCommand(ManualCropping);
            RemoveCroppingCommand = new RelayCommand(RemoveCropping);
            OkCommand = new RelayCommand(ClosePopup);
            CancelAutomaticCommand = new RelayCommand(CancelAutomatic);
        }

        public bool IsPopupOpen
        {
            get => _isPopupOpen;
            set
            {
                if (Set(ref _isPopupOpen, value) && !value)
                    _popupClosed?.Invoke();
            }
        }

        public bool IsAutomaticCropping
        {
            get => _isAutomaticCropping;
            set => Set(ref _isAutomaticCropping, value);
        }

        public int AutomaticProgress => (int)(100.0 * (_currentAutomaticPage - 1) / _pdfViewCtrl.GetPageCount());

        public string AutomaticProgressPageText
        {
            get
            {
                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                return string.Format(loader.GetString("CropPopup_Automatic_PageText"), _currentAutomaticPage, _pdfViewCtrl.GetPageCount());
            }
        }

        /// <summary>
        /// This command raises an event for the CropView to be used
        /// </summary>
        /// <param name="parameter"></param>
        private void ManualCropping(object parameter)
        {
            _manualCropRequested?.Invoke();
            IsPopupOpen = false;
        }

        private void ClosePopup(object parameter) => IsPopupOpen = false;

        private void CancelAutomatic(object parameter)
        {
            _isAutomaticCancel = true;
            _tokenSource?.Cancel();
        }
    }
}
