﻿using pdftron.Filters;
using pdftron.PDF.Tools.Commands;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Common;
using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Outline;
using pdftron.PDF.Tools.Utilities;
using pdftron.SDF;
using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Threading;
using System.Threading.Tasks;
using System.Xml.Serialization;
using Windows.ApplicationModel;
using Windows.Foundation;
using Windows.Graphics.Imaging;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.UI.Core;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;

namespace pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab
{
    public class PdfViewCtrlTabInfo : ViewModelBase
    {
        private TabSelectedHandler _tabSelected;
        public delegate void TabSelectedHandler(PdfViewCtrlTabInfo item);

        /// <summary>
        /// Raised when a tab has been selected
        /// </summary>
        public event TabSelectedHandler TabSelected
        {
            add
            {
                if (_tabSelected?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _tabSelected += value;
            }
            remove
            {
                if (_tabSelected != null)
                {
                    TabSelectedHandler eventMethod = (TabSelectedHandler)Array.Find(
                        _tabSelected.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _tabSelected -= eventMethod;
                }
            }
        }

        /// <summary>
        /// Request tab closed event
        /// </summary>
        public void RaiseTabClosedEvent()
        {
            if (_tabClosed != null)
                _tabClosed.Invoke(this);
        }

        private TabClosedHandler _tabClosed;
        public delegate void TabClosedHandler(PdfViewCtrlTabInfo item);

        /// <summary>
        /// Raised when a tab has been requested to closed
        /// </summary>
        public event TabClosedHandler TabClosed
        {
            add
            {
                if (_tabClosed?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _tabClosed += value;
            }
            remove
            {
                if (_tabClosed != null)
                {
                    TabClosedHandler eventMethod = (TabClosedHandler)Array.Find(
                        _tabClosed.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _tabClosed -= eventMethod;
                }
            }
        }

        private SetUpPDFViewCtrlAndToolsHandler _setUpPDFViewCtrlAndTools;
        public delegate void SetUpPDFViewCtrlAndToolsHandler(PDFViewCtrl ctrl, ToolManager toolManager, Deferral deferral);
        public event SetUpPDFViewCtrlAndToolsHandler SetUpPDFViewCtrlAndTools
        {
            add
            {
                if (_setUpPDFViewCtrlAndTools?.GetInvocationList().Contains(value) != true)
                    _setUpPDFViewCtrlAndTools += value;
            }
            remove => _setUpPDFViewCtrlAndTools -= value;
        }

        private EventHandler<RoutedEventArgs> _pdfUpdated;
        public event EventHandler<RoutedEventArgs> PdfUpdated
        {
            add
            {
                if (_pdfUpdated?.GetInvocationList().Contains(value) != true)
                    _pdfUpdated += value;
            }
            remove => _pdfUpdated -= value;
        }

        private EventHandler<RoutedEventArgs> _pdfIsReady;
        public event EventHandler<RoutedEventArgs> PdfIsReady
        {
            add
            {
                if (_pdfIsReady?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _pdfIsReady += value;
            }
            remove
            {
                if (_pdfIsReady != null)
                {
                    EventHandler<RoutedEventArgs> eventMethod = (EventHandler<RoutedEventArgs>)Array.Find(
                        _pdfIsReady.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _pdfIsReady -= eventMethod;
                }
            }
        }

        private EventHandler<TappedRoutedEventArgs> _reflowTapped;
        public event EventHandler<TappedRoutedEventArgs> ReflowTapped
        {
            add
            {
                if (_reflowTapped?.GetInvocationList().Contains(value) != true)
                    _reflowTapped += value;
            }
            remove => _reflowTapped -= value;
        }

        private EventHandler<TappedRoutedEventArgs> _tappedBeforeTools;
        public event EventHandler<TappedRoutedEventArgs> TappedBeforeTools
        {
            add
            {
                if (_tappedBeforeTools?.GetInvocationList().Contains(value) != true)
                    _tappedBeforeTools += value;
            }
            remove => _tappedBeforeTools -= value;
        }

        private EventHandler<TappedRoutedEventArgs> _tappedAfterTools;
        public event EventHandler<TappedRoutedEventArgs> TappedAfterTools
        {
            add
            {
                if (_tappedAfterTools?.GetInvocationList().Contains(value) != true)
                    _tappedAfterTools += value;
            }
            remove => _tappedAfterTools -= value;
        }

        private bool _active;
        private bool _showPreview;
        private bool _first;
        private bool _last;
        private bool _isSelected;
        private PDFViewCtrl _pdfViewCtrl;
        private OutlineDialogStateBundle _outlineState;
        private OutlineDialogViewModel _outlineDialogViewModel;
        private ReflowView _reflowView;
        private ToolManager _toolManager;
        private StorageFile _tempFile;
        private bool _fileLoadingError;
        private bool _hasUnsavedChanged;
        private ShowIfModifiedStates _showIfDocumentModified;
        private bool _updatingFile;
        private bool _pdfReady;
        private bool _isWaitingForConversionToStart;
        private bool _showConversionProgress;
        private bool _activatingConversion;
        private SemaphoreSlim _docSavingSemaphore;
        private bool _isPaused;
        private bool _settingDoc;
        private OnSetDocEventHandler _setDocHandler;
        private OnConversionEventHandler _conversionChangedHandler;
        private OnThumbnailGeneratedEventHandler _thumbnailHandler;
        private ViewModelBase _parentViewModel;
        private int _reflowCurrentPage = 1;
        private bool _isCloseTabButtonVisible = true;
        private bool _isCloseTabButtonEnabled = true;

        /// <summary>
        /// Choose whether or not the close button within tabs are Visible or Collapsed
        /// </summary>
        public bool IsCloseTabButtonVisible
        {
            get
            {
                if (!_isCloseTabButtonEnabled)
                    return false;

                return _isCloseTabButtonVisible;
            }
            set => Set(ref _isCloseTabButtonVisible, value);
        }

        /// <summary>
        /// Choose whether or not the close button within tabs are enabled or disabled
        /// </summary>
        public bool IsCloseTabButtonEnabled
        {
            get { return _isCloseTabButtonEnabled; }
            set => Set(ref _isCloseTabButtonEnabled, value);
        }

        public PdfViewCtrlTabInfo(StorageFile orig, StorageFile tmp, PDFDoc doc, string title)
        {
            MetaData = new PdfViewCtrlTabMetaData();

            Title = title ?? string.Empty;
            TempFile = tmp;
            OriginalFile = orig;
            Doc = doc;
        }

        /// <summary>
        /// Helper method to restore opened tabs
        /// </summary>
        /// <param name="orig"> Original File </param>
        /// <param name="tmp"> Temperory File </param>
        /// <param name="restoredData"></param>
        public PdfViewCtrlTabInfo(StorageFile orig, StorageFile tmp, PdfViewCtrlTabMetaData restoredData)
        {
            MetaData = restoredData;
            TempFile = tmp;
            OriginalFile = orig;
            if (tmp == null)
                FileLoadingError = true;
        }

        public string ThumbnailFolder { get; set; }

        public string ThumbnailFallback { get; set; }

        public string PreviewSource
        {
            get
            {
                if (!string.IsNullOrEmpty(MetaData.ThumbnailLocation))
                    return Path.Combine(ApplicationData.Current.LocalFolder.Path, DocumentManager.TabPreviewFolderName, MetaData.ThumbnailLocation);

                return ThumbnailFallback;
            }
        }

        public bool ShowPreview
        {
            get => _showPreview;
            set => Set(ref _showPreview, value);
        }

        public string Title
        {
            get => MetaData.TabTitle;
            set
            {
                if (value?.Equals(MetaData.TabTitle) == false)
                {
                    MetaData.TabTitle = value;
                    RaisePropertyChanged();
                }
            }
        }

        public bool First
        {
            get => _first;
            set => Set(ref _first, value);
        }

        public bool Last
        {
            get => _last;
            set => Set(ref _last, value);
        }

        public bool IsSelected
        {
            get => _isSelected;
            set => Set(ref _isSelected, value);
        }

        public void Deactivate()
        {
            Pause();
            ToolManager?.CreateDefaultTool();
        }

        public PDFViewCtrl PDFViewCtrl
        {
            get => _pdfViewCtrl;
            set => Set(ref _pdfViewCtrl, value);
        }

        public OutlineDialogViewModel OutlineDialogViewModel
        {
            get
            {
                if (_outlineDialogViewModel == null)
                {
                    _outlineDialogViewModel = new OutlineDialogViewModel(PDFViewCtrl)
                    {
                        Isconverting = PDFViewCtrl.IsConverting,
                        NavigationStack = NavigationStack
                    };

                    if (_outlineState != null)
                    {
                        _outlineDialogViewModel.RestoreState(_outlineState);
                    }
                    else
                    {
                        OutlineDialogStateBundle outlineState = new OutlineDialogStateBundle((SubViews)ViewerSettings.OutlineDefaultView, 0, null, 0);
                        _outlineDialogViewModel.RestoreState(outlineState);
                    }

                    _outlineDialogViewModel.AnnotationList.ViewModel.ToolManager = ToolManager;
                    _outlineDialogViewModel.Outline.DocumentTitle = Title;
                }

                return _outlineDialogViewModel;
            }
            internal set => _outlineDialogViewModel = value;
        }

        public ReflowView ReflowView
        {
            get => _reflowView;
            set => Set(ref _reflowView, value);
        }

        public bool IsReflow
        {
            get
            {
                if (MetaData == null)
                    return false;

                return MetaData.IsReflow;
            }
            set
            {
                if (MetaData == null)
                    return;

                if (MetaData.IsReflow != value)
                {
                    MetaData.IsReflow = value;

                    if (NavigationStack != null)
                        NavigationStack.CleanPositionData();

                    if (!MetaData.IsReflow)
                    {
                        if (PDFViewCtrl != null && ReflowView != null)
                        {
                            PDFViewCtrl.SetCurrentPage(ReflowView.CurrentPage);
                            PDFViewCtrl.RequestRendering();
                        }
                    }

                    UpdateReflowState();
                    RaisePropertyChanged();
                }
            }
        }

        public bool IsRightToLeftEnabled
        {
            get => MetaData.IsRightToLeft;
            set
            {
                if (MetaData.IsRightToLeft != value)
                {
                    MetaData.IsRightToLeft = value;

                    PDFViewCtrl.SetRightToLeftLanguage(MetaData.IsRightToLeft);

                    // Note: Update page layout if any of these will make use of RTL
                    var presentationMode = PDFViewCtrl.GetPagePresentationMode();
                    if (presentationMode == PDFViewCtrlPagePresentationMode.e_single_page
                        || presentationMode == PDFViewCtrlPagePresentationMode.e_facing
                        || presentationMode == PDFViewCtrlPagePresentationMode.e_facing_cover)
                    {
                        _pdfViewCtrl.UpdatePageLayout();

                    }

                    RaisePropertyChanged();
                }
            }
        }

        public ToolManager ToolManager
        {
            get => _toolManager;
            set
            {
                _toolManager = value;
                RaisePropertyChanged();
            }
        }

        /// <summary>
        /// The temporary file created in the app's temp folder to be used for Viewing/Editing by the user
        /// </summary>
        /// <remarks>
        /// All changes made to the temporary file will be transfered to the <see cref="OriginalFile"/> once Save (Manual) or AutoSave is performed
        /// </remarks>
        public StorageFile TempFile
        {
            get => _tempFile;
            set => Set(ref _tempFile, value);
        }

        /// <summary>
        /// The original file opened by the user
        /// </summary>
        /// <remarks>
        /// Changes to this file will be transfered from the <see cref="TempFile"/> on AutoSave or Manual Save is issued
        /// </remarks>
        public StorageFile OriginalFile { get; set; }

        public StorageFile FileSourceIfNotSaveable { get; set; }

        public IRandomAccessStream StreamForConversion { get; set; }

        /// <summary>
        /// It holds a reference to the current opened PDFDoc, which is related to a temporary StorageFile
        /// </summary>
        public PDFDoc Doc { get; set; }

        public NavigationStack NavigationStack { get; set; }

        public bool FileLoadingError
        {
            get => _fileLoadingError;
            set => Set(ref _fileLoadingError, value);
        }

        public bool ShowHasUnsavedChanged
        {
            get => _hasUnsavedChanged;
            set => Set(ref _hasUnsavedChanged, value);
        }

        [XmlElement(ElementName = "ModSinceSave")]
        public bool IsDocumentModifiedSinceLastSave
        {
            get => MetaData.IsDocumentModifiedSinceLastSave;
            set
            {
                if (value != MetaData.IsDocumentModifiedSinceLastSave)
                {
                    MetaData.IsDocumentModifiedSinceLastSave = value;
                    ShowHasUnsavedChanged = value && ShouldShowIfDocumentModified;
                }
            }
        }

        /// <summary>
        /// If false, auto-saving is implied
        /// </summary>
        [XmlIgnore]
        public ShowIfModifiedStates ShowIfDocumentModified
        {
            get => _showIfDocumentModified;
            set
            {
                if (value != _showIfDocumentModified)
                {
                    _showIfDocumentModified = value;
                    if (ShouldShowIfDocumentModified && !ShowHasUnsavedChanged)
                    {
                        ShowHasUnsavedChanged = IsDocumentModifiedSinceLastSave;
                    }
                    else
                    {
                        ShowHasUnsavedChanged = false;
                    }
                }
            }
        }

        private bool ShouldShowIfDocumentModified
        {
            get
            {
                if (_showIfDocumentModified == ShowIfModifiedStates.ShowIfModified)
                {
                    return true;
                }
                else if (_showIfDocumentModified == ShowIfModifiedStates.ShowIfModidiedAndCantSave)
                {
                    return DocumentState != OpenedDocumentStates.Normal && DocumentState != OpenedDocumentStates.NeedsFullSave;
                }

                return false;
            }
        }

        public OpenedDocumentStates DocumentState
        {
            get => MetaData.DocumentState;
            set
            {
                MetaData.DocumentState = value;
                RaisePropertyChanged(nameof(IsConverting));
            }
        }

        public bool HasDocumentBeenModifiedSinceOpening
        {
            get => MetaData.HasDocumentBeenModifiedSinceOpening;
            set => MetaData.HasDocumentBeenModifiedSinceOpening = value;
        }

        public bool HasUserBeenWarnedAboutSaving
        {
            get => MetaData.HasUserBeenWarnedAboutSaving;
            set => MetaData.HasUserBeenWarnedAboutSaving = value;
        }

        public DateTimeOffset SystemLastModifiedDate { get; set; } = DateTimeOffset.MinValue;

        public ulong SystemFileSize { get; set; } = 0;

        public bool IsUpToDate
        {
            get
            {
                if (DocumentState == OpenedDocumentStates.Corrupted || DocumentState == OpenedDocumentStates.CorruptedAndModified
                    || DocumentState == OpenedDocumentStates.Normal || DocumentState == OpenedDocumentStates.NeedsFullSave)
                {
                    return SystemLastModifiedDate == MetaData.LastModifiedDate || SystemFileSize == MetaData.FileSize;
                }

                return true;
            }
        }

        /// <summary>
        /// True if the original file is out of date and we have to load a new file
        /// </summary>
        public bool UpdatingFile
        {
            get => _updatingFile;
            set => Set(ref _updatingFile, value);
        }

        public bool PdfReady
        {
            get => _pdfReady;
            private set
            {
                if (Set(ref _pdfReady, value) && _pdfReady && _pdfIsReady != null)
                    _pdfIsReady?.Invoke(this, new RoutedEventArgs());
            }
        }

        public bool IsWaitingForConversionToStart
        {
            get => _isWaitingForConversionToStart;
            set => Set(ref _isWaitingForConversionToStart, value);
        }

        public bool CanBeSaved => DocumentState != OpenedDocumentStates.Universal && DocumentState != OpenedDocumentStates.Uneditable;

        public bool IsConverting => MetaData.DocumentState == OpenedDocumentStates.Universal;

        public bool ShowConversionProgress
        {
            get => _showConversionProgress;
            set => Set(ref _showConversionProgress, value);
        }

        public ViewModelBase ParentViewModel
        {
            get => _parentViewModel;
            set => Set(ref _parentViewModel, value);
        }

        public int ReflowCurrentPage
        {
            get => _reflowCurrentPage;
            set => Set(ref _reflowCurrentPage, value);
        }

        public static PdfViewCtrlTabInfo CreateUniversalTabInfo(PdfViewCtrlTabMetaData restoredData, StorageFile sourceFile) =>
            new PdfViewCtrlTabInfo(null, null, restoredData)
            {
                FileLoadingError = false,
                FileSourceIfNotSaveable = sourceFile,
                OriginalFile = sourceFile,
                DocumentState = OpenedDocumentStates.Universal
            };

        public PdfViewCtrlTabMetaData MetaData { get; set; }

        /// <summary>
        /// Update the metadata (saved in disk) for a particular document tab
        /// </summary>
        public void RefreshState()
        {
            if (_active && PDFViewCtrl != null && !_settingDoc)
                MetaData.UpdateSettings(PDFViewCtrl);
        }

        private enum OpeningType
        {
            DocExists = 0,
            PDFTempFileExists,
            OpenUniversalFromStorageFile,
            OpenUniveralFromPath,
        }

        /// <summary>
        /// Activate the PDF tab by handling the initialization and creation of the document with the PDFViewCtrl
        /// </summary>
        public async Task ActivateAsync()
        {
            if (!IsUpToDate)
            {
                if (!IsDocumentModifiedSinceLastSave)
                    await UpdateFileAsync();

                return;
            }

            FileOpeningType openingType = ViewerSettings.GetFileOpeningType(OriginalFile);

            if (Doc == null && !FileLoadingError)
            {
                if (TempFile != null)
                {
                    bool success = false;
                    try
                    {
                        Doc = new PDFDoc(TempFile);
                        bool hasOC = Doc.HasOC();
                        if (!string.IsNullOrEmpty(MetaData.Password))
                        {
                            success = Doc.InitStdSecurityHandler(MetaData.Password);
                        }
                        else
                        {
                            Doc.InitStdSecurityHandler("");
                        }

                        success = true;
                    }
                    catch
                    {
                        //Swallow exception
                    }

                    FileLoadingError = !success;
                }
            }

            if (PDFViewCtrl == null)
            {
                // NOTE: high risk of exception and causing layout cycle issues due resizing of the XMAL
                PDFViewCtrl = new PDFViewCtrl
                {
                    Opacity = 0
                };

                PDFViewCtrl.SetPageSpacing(2, 2, 2, 0);

                _thumbnailHandler = new OnThumbnailGeneratedEventHandler(PDFViewCtrl_OnThumbnailGenerated);
                PDFViewCtrl.OnThumbnailGenerated += _thumbnailHandler;

                if (ViewerSettings.RememberLastPage)
                {
                    PDFViewCtrl.SetPagePresentationMode(MetaData.PagePresentationMode);
                    switch (MetaData.PageRotation)
                    {
                        case PageRotate.e_90:
                            PDFViewCtrl.RotateClockwise();
                            break;
                        case PageRotate.e_180:
                            PDFViewCtrl.RotateClockwise();
                            PDFViewCtrl.RotateClockwise();
                            break;
                        case PageRotate.e_270:
                            PDFViewCtrl.RotateCounterClockwise();
                            break;
                    }
                }
                else
                {
                    PDFViewCtrl.SetPagePresentationMode(ViewerSettings.PagePresentationMode);
                }
            }

            if (ToolManager == null)
            {
                PDFViewCtrl.Tapped += PDFViewCtrlBeforeTools_Tapped;
                ToolManager = new ToolManager(PDFViewCtrl);
                ToolManager.IsRichTextEnabled = ViewerSettings.IsRichTextEnabled;
                ToolManager.SetUndoRedoManager(new UndoRedoManager());

                PDFViewCtrl.Tapped += PDFViewCtrlAfterTools_Tapped;
            }

            if (NavigationStack == null)
            {
                NavigationStack = new NavigationStack(PDFViewCtrl);
                ToolManager.NavigationStack = NavigationStack;
            }

            if (!_active && _setUpPDFViewCtrlAndTools != null)
            {
                if (_setUpPDFViewCtrlAndTools != null)
                {
                    var tcs = new TaskCompletionSource<bool>();
                    using (var deferral = new Deferral(() => tcs.SetResult(true)))
                    {
                        _setUpPDFViewCtrlAndTools.Invoke(PDFViewCtrl, ToolManager, deferral);
                        await tcs.Task;
                    }
                }

                ToolManager.DocumentName = Title;
                ToolManager.ReadOnlyMode = (DocumentState == OpenedDocumentStates.Uneditable);
                ToolManager.CopyAnnotatedTextToNote = ViewerSettings.CopyAnnotatedTextToNote;
            }

            if (Doc == null && openingType == FileOpeningType.ImageConversion)
            {
                Doc = await HandleImageConversionAsync();
            }

            if (!_active && Doc != null)
            {
                _active = true;

                if (_setDocHandler == null)
                {
                    _setDocHandler = new OnSetDocEventHandler(PDFViewCtrl_OnSetDoc);
                    PDFViewCtrl.OnSetDoc += _setDocHandler;
                    _settingDoc = true;
                }

                try
                {
                    PDFViewCtrl.SetDoc(Doc);
                }
                catch (Exception)
                {
                    FileLoadingError = true;
                }

                if (FileLoadingError)
                    return;

                PdfReady = true;
                if (!MetaData.IsThumbnailUpToDate && ShowPreview)
                    UpdatePreview();

                StreamForConversion?.Dispose();
            }
            else if (!_active && (openingType == FileOpeningType.UniversalConversion || openingType == FileOpeningType.UniversalConversionFromString))
            {
                await HandleUniversalConversionAsync(openingType);
            }

            UpdateReflowState();

            MetaData.TabLastViewedTimeStamp = DateTime.Now;
        }

        /// <summary>
        /// Convert Image to PDF
        /// </summary>
        private async Task<PDFDoc> HandleImageConversionAsync()
        {
            return await PDFDocCreator.CreateFromImageAsync(FileSourceIfNotSaveable);
        }

        /// <summary>
        /// Handle the conversion from a specific file format to PDF (image, office, xps, etc)
        /// </summary>
        /// <param name="openingType">The file opening type</param>
        private async Task HandleUniversalConversionAsync(FileOpeningType openingType)
        {
            _active = true;
            try
            {               
                ShowConversionProgress = true;
                IsWaitingForConversionToStart = true;
                _conversionChangedHandler = new OnConversionEventHandler(PDFViewCtrl_OnConversionChanged);
                PDFViewCtrl.OnConversionChanged += _conversionChangedHandler;
                DocumentConversion conversion = null;

                WordToPDFOptions opts = new WordToPDFOptions();
                opts.SetLayoutResourcesPluginPath(Path.Combine(Package.Current.InstalledLocation.Path, "Resources"));
                opts.SetSmartSubstitutionPluginPath(Path.Combine(Package.Current.InstalledLocation.Path, "Resources"));
                if (openingType == FileOpeningType.UniversalConversion && StreamForConversion != null)
                {
                    IFilter filter = new RandomAccessStreamFilter(StreamForConversion);
                    conversion = Convert.UniversalConversion(filter, null);
                    StreamForConversion = null;
                }
                else if (openingType == FileOpeningType.UniversalConversion && FileSourceIfNotSaveable != null)
                {
                    await ActivateConversionAsync(FileSourceIfNotSaveable);
                    return;
                }
                else
                {
                    conversion = Convert.UniversalConversion(OriginalFile.Path, opts);
                }

                if (conversion != null)
                {
                    PDFViewCtrl.OpenUniversalDocument(conversion);
                }
                else
                {
                    FileLoadingError = true;
                }
            }
            catch (Exception ex)
            {
                var exception = new pdftron.Common.PDFNetException(ex.HResult);

                if (exception.IsPDFNetException)
                {
                    var loader = Windows.ApplicationModel.Resources.ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                    var dialog = new Windows.UI.Xaml.Controls.ContentDialog
                    {
                        Title = loader.GetString("DocumentsPage_FileOpeningError_Title"),
                        Content = loader.GetString("pdftron_Tool_Office_Conversion_Failed") + Environment.NewLine + Environment.NewLine + "Error: " + exception.Message,
                        CloseButtonText = loader.GetString("pdftron_Controls_Content_Dialog_OK")
                    };

                    await DialogHelper.ShowContentDialogAsync(dialog);
                }

                FileLoadingError = true;
            }
            finally 
            {
                ShowConversionProgress = false;
            }

            if (FileLoadingError)
                return;
        }

        private async Task ActivateConversionAsync(StorageFile file)
        {
            if (_activatingConversion)
            {
                return;
            }
            _activatingConversion = true;
            DocumentConversion conversion = null;
            try
            {
                WordToPDFOptions opts = new WordToPDFOptions();
                opts.SetLayoutResourcesPluginPath(Path.Combine(Package.Current.InstalledLocation.Path, "Resources"));
                opts.SetSmartSubstitutionPluginPath(Path.Combine(Package.Current.InstalledLocation.Path, "Resources"));
                IRandomAccessStream iras = await file.OpenReadAsync();
                IFilter filter = new RandomAccessStreamFilter(iras);
                conversion = pdftron.PDF.Convert.UniversalConversion(filter, null);
            }
            finally
            {
                _activatingConversion = false;
            }

            if (conversion != null)
            {
                PDFViewCtrl.OpenUniversalDocument(conversion);
            }
            else
            {
                FileLoadingError = true;
            }

            UpdateReflowState();
            MetaData.TabLastViewedTimeStamp = DateTime.Now;
        }

        public void CloseOutline()
        {
            if (_outlineDialogViewModel != null)
            {
                _outlineState = _outlineDialogViewModel.GetState();
                _outlineDialogViewModel.SaveBookmarks();
                _outlineDialogViewModel.NavigationStack = null;
                _outlineDialogViewModel.PDFViewCtrl = null;
                _outlineDialogViewModel.AnnotationList?.ViewModel?.ClearViewModel();
                _outlineDialogViewModel.Thumbnails?.ViewModel?.CleanUp();
                _outlineDialogViewModel.Outline?.Dispose();
                _outlineDialogViewModel.Dispose();
                _outlineDialogViewModel = null;
            }
        }

        public async Task CloseOutlineAsync()
        {
            if (_outlineDialogViewModel != null)
            {
                OutlineDialogViewModel vmToClose = _outlineDialogViewModel;
                _outlineState = _outlineDialogViewModel.GetState();
                _outlineDialogViewModel.SaveBookmarks();
                _outlineDialogViewModel.NavigationStack = null;
                _outlineDialogViewModel.PDFViewCtrl = null;
                _outlineDialogViewModel.AnnotationList?.ViewModel?.ClearViewModel();
                _outlineDialogViewModel.Thumbnails?.ViewModel?.CleanUp();
                _outlineDialogViewModel.Outline?.Dispose();

                if (vmToClose.UserBookmarks != null)
                    await vmToClose.UserBookmarks.WaitForBookmarkSavingAsync();

                if (vmToClose.AnnotationList != null)
                    await vmToClose.AnnotationList.WaitForAnnotationListToFinish();

                _outlineDialogViewModel.Dispose();
                _outlineDialogViewModel = null;
            }
        }

        /// <summary>
        /// Update Reflow control
        /// </summary>
        /// <param name="currentPage">True to update with current PDFViewCtrl page, otherwise MetaData.LasPage</param>
        public void UpdateReflowState(bool currentPage = true)
        {
            if (ReflowView == null && IsReflow && !_isPaused && !IsWaitingForConversionToStart)
            {
                ReflowView = new ReflowView(Doc);
                ReflowView.IsEditingEnabled = ViewerSettings.IsReflowEditingEnabled;
                ReflowView.ToolManager = ToolManager;

                int startPage = 1;
                if (PDFViewCtrl != null)
                    startPage = PDFViewCtrl.GetCurrentPage();

                ReflowView.CurrentPage = currentPage ? startPage : MetaData.LastPage;
                ReflowView.ReflowTapped += ReflowView_ReflowTapped;
                ReflowView.PageChanged += ReflowView_PageChanged;
                ReflowView.DocumentChanged += ReflowView_DocEditedChanged;
            }
            else if (ReflowView != null && (!IsReflow || _isPaused))
            {
                ReflowCurrentPage = ReflowView.CurrentPage;
                ReflowView.Dispose();
                ReflowView.ReflowTapped -= ReflowView_ReflowTapped;
                ReflowView.PageChanged -= ReflowView_PageChanged;
                ReflowView = null;
            }
        }

        private void ReflowView_DocEditedChanged()
        {
            PDFViewCtrl.Update();
        }

        private void ReflowView_PageChanged(int page, int totalPages)
        {
            ReflowCurrentPage = page;
            _pdfViewCtrl.SetCurrentPage(page);
        }
        private void ReflowView_ReflowTapped(object sender, TappedRoutedEventArgs e) => _reflowTapped?.Invoke(sender, e);

        private void PDFViewCtrlBeforeTools_Tapped(object sender, TappedRoutedEventArgs e) => _tappedBeforeTools?.Invoke(sender, e);

        private void PDFViewCtrlAfterTools_Tapped(object sender, TappedRoutedEventArgs e) => _tappedAfterTools?.Invoke(sender, e);

        public async Task ReActivateAsync()
        {
            CleanUp();
            Doc = null;
            DisposeWhenSaved = false;
            _active = false;
            await ActivateAsync();
        }

        public void Pause()
        {
            _isPaused = true;
            UpdateReflowState();
        }

        public void Resume()
        {
            _isPaused = false;
        }

        public SemaphoreSlim DocSavingSemaphore => _docSavingSemaphore ?? (_docSavingSemaphore = new SemaphoreSlim(1));

        public void DocSemWait(
            [CallerMemberName] string member = "",
            [CallerLineNumber] int line = 0)
        {
#if DEBUG
            if (TabControlViewModel._showSemaphoreLogging)
            {
                Debug.WriteLine("Function {0}, line {1} waiting for doc Semaphore for {2}", member, line, Title);
            }
#endif
            SemaphoreSlim tmpSema = DocSavingSemaphore;
            tmpSema.Wait();
#if DEBUG
            if (TabControlViewModel._showSemaphoreLogging)
            {
                Debug.WriteLine("Function {0}, line {1} got doc Semaphore for {2}", member, line, Title);
            }
#endif
        }

        public async Task DocSemWaitAsync(
            [CallerMemberName] string member = "",
            [CallerLineNumber] int line = 0)
        {
#if DEBUG
            if (TabControlViewModel._showSemaphoreLogging)
            {
                Debug.WriteLine("Function {0}, line {1} waiting async for doc Semaphore for {2}", member, line, Title);
            }
#endif
            SemaphoreSlim tmpSema = DocSavingSemaphore;
            await tmpSema.WaitAsync().ConfigureAwait(false);
#if DEBUG
            if (TabControlViewModel._showSemaphoreLogging)
            {
                Debug.WriteLine("Function {0}, line {1} got async doc Semaphore for {2}", member, line, Title);
            }
#endif
        }

        public void DocSemRelease(
            [CallerMemberName] string member = "",
            [CallerLineNumber] int line = 0)
        {
#if DEBUG
            if (TabControlViewModel._showSemaphoreLogging)
            {
                Debug.WriteLine("Function {0}, line {1} released doc Semaphore for {2}", member, line, Title);
            }
#endif
            SemaphoreSlim tmpSema = DocSavingSemaphore;
            tmpSema.Release();
        }

        public bool DisposeWhenSaved { get; private set; }

        public void CleanUp()
        {
            if (PDFViewCtrl != null)
            {
                try
                {
                    PDFViewCtrl.Deactivate();

                    if (_conversionChangedHandler != null)
                    {
                        _pdfViewCtrl.OnConversionChanged -= _conversionChangedHandler;
                        _conversionChangedHandler = null;
                    }

                    PDFViewCtrl.Tapped -= PDFViewCtrlBeforeTools_Tapped;
                    PDFViewCtrl.Tapped -= PDFViewCtrlAfterTools_Tapped;
                    PDFViewCtrl.CloseDoc();
                    ShowConversionProgress = false;
                }
                catch (Exception ex)
                {
                    AppCenterHandlerBase.CURRENT.TrackError(ex);
                }
            }
            if (ReflowView != null)
            {
                ReflowView.ReflowTapped -= ReflowView_ReflowTapped;
            }
            if (Doc != null)
            {
                try
                {
                    Doc.Destroy();
                    Doc = null;
                }
                catch (Exception ex)
                {
                    AppCenterHandlerBase.CURRENT.TrackError(ex);
                }

            }

            if (_setDocHandler != null)
            {
                if (PDFViewCtrl != null)
                    PDFViewCtrl.OnSetDoc -= _setDocHandler;

                _setDocHandler = null;
            }

            if (_thumbnailHandler != null)
            {
                if (PDFViewCtrl != null)
                {
                    PDFViewCtrl.OnThumbnailGenerated -= _thumbnailHandler;
                }

                _thumbnailHandler = null;
            }

            if (ToolManager != null)
            {
                ToolManager.CreateDefaultTool();
                ToolManager.Dispose();
                ToolManager = null;
            }

            if (NavigationStack != null)
                NavigationStack = null;
            
            PDFViewCtrl = null;
            _active = false;
        }

        public async Task SaveDocumentAsync() => await SaveDocumentImplAsync();

        public async Task SaveDocumentImplAsync()
        {
            if (DocumentState == OpenedDocumentStates.Uneditable || DocumentState == OpenedDocumentStates.Universal)
                return;

            await DocSemWaitAsync().ConfigureAwait(false);
            bool locked = false;
            try
            {
                if (_outlineDialogViewModel?.HasUnsavedUserbookmarks == true)
                {
                    _outlineDialogViewModel.UserBookmarks.SaveBookmarks();
                    await _outlineDialogViewModel.CleanUpSubViewsAsync();
                }

                if (Doc != null)
                {
                    Doc.LockRead();
                    locked = true;
                    if (Doc.IsModified())
                    {
                        Doc.UnlockRead();
                        locked = false;
                        if ((MetaData.DocumentState == OpenedDocumentStates.Corrupted || MetaData.DocumentState == OpenedDocumentStates.CorruptedAndModified)
                            && !MetaData.IsBrokenDocumentRestored)
                        {
                            await Doc.SaveAsync(SDFDocSaveOptions.e_remove_unused).AsTask().ConfigureAwait(false);
                        }
                        else
                        {
                            await Doc.SaveAsync().AsTask().ConfigureAwait(false);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (locked)
                    Doc.UnlockRead();

                DocSemRelease();
            }
        }

        public void UpdateIsReflow()
        {
            RaisePropertyChanged(nameof(IsReflow));
            UpdateReflowState();
        }

        public void UpdatePreview()
        {
            MetaData.IsThumbnailUpToDate = false;
            if (PDFViewCtrl?.GetColorPostProcessMode() == PDFRasterizerColorPostProcessMode.e_postprocess_none)
            {
                try
                {
                    PDFViewCtrl.GetThumbAsync(1);
                }
                catch(Exception ex)
                {
                    AppCenterHandlerBase.CURRENT.TrackError(ex);
                }
            }
        }

        public async Task UpdateFileAsync()
        {
            await DocSemWaitAsync();
            bool didUpdateDoc = false;
            try
            {
                if (!IsUpToDate)
                {
                    PdfReady = false;
                    UpdatingFile = true;
                    if (Doc != null) // need to set it to null so that CleanUp doesn't try to lock.
                    {
                        PDFViewCtrl?.CloseDoc();
                        Doc.Dispose();
                        Doc = null;
                    }

                    if (_active)
                        CleanUp();

                    await OriginalFile.CopyAndReplaceAsync(TempFile);
                    MetaData.LastModifiedDate = SystemLastModifiedDate;
                    MetaData.FileSize = SystemFileSize;
                    ShowHasUnsavedChanged = false;
                    IsDocumentModifiedSinceLastSave = false;
                    didUpdateDoc = true;
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error updating tab: " + e.ToString());
            }
            finally
            {
                UpdatingFile = false;
                DocSemRelease();
            }

            if (didUpdateDoc)
            {
                _pdfUpdated?.Invoke(this, new RoutedEventArgs());
                await ActivateAsync();
            }
        }

        private void PDFViewCtrl_OnSetDoc()
        {
            _settingDoc = false;

            if (ViewerSettings.RememberLastPage)
            {
                PDFViewCtrl.SetCurrentPage(MetaData.LastPage);
                double desiredHPos = MetaData.HScrollPos;
                double desiredVPos = MetaData.VScrollPos;

                if (MetaData.Zoom <= 0)
                {
                    PDFViewCtrl.SetZoom(ViewerSettings.DefaultZoomLevel);
                }
                else
                {
                    PDFViewCtrl.SetZoom(MetaData.Zoom);
                    if (Math.Abs(PDFViewCtrl.GetZoom() - MetaData.Zoom) > 0.01)
                    {
                        double zoomDifference = PDFViewCtrl.GetZoom() / MetaData.Zoom;
                        desiredHPos *= zoomDifference;
                        desiredVPos *= zoomDifference;
                    }
                }

                PDFViewCtrl.SetHScrollPos(desiredHPos);
                PDFViewCtrl.SetVScrollPos(desiredVPos);

                if (IsReflow && ReflowView != null)
                {
                    ReflowView.CurrentPage = MetaData.LastPage;
                    PDFViewCtrl.CancelRendering();
                }
            }

            PDFViewCtrl.Opacity = 1;
        }

        private async void PDFViewCtrl_OnConversionChanged(PDFViewCtrlConversionType type, int totalPagesConverted)
        {
            bool wasWaitingForConversion = IsWaitingForConversionToStart;
            IsWaitingForConversionToStart = false;
            if (type == PDFViewCtrlConversionType.e_conversion_failed || type == PDFViewCtrlConversionType.e_conversion_finished)
            {
                PDFViewCtrl.OnConversionChanged -= _conversionChangedHandler;
                _conversionChangedHandler = null;
                ShowConversionProgress = false;
            }

            PDFViewCtrl.Opacity = 1;
            switch (type)
            {
                case PDFViewCtrlConversionType.e_conversion_progress:
                    Doc = PDFViewCtrl.GetDoc();
                    if (wasWaitingForConversion)
                    {
                        UpdateReflowState();
                        PDFViewCtrl.GetThumbAsync(1);
                    }
                    else if (IsReflow && ReflowView != null)
                    {
                        ReflowView?.ViewModel.UpdatePageNumberProperties();
                    }
                    // update page numbers
                    break;

                case PDFViewCtrlConversionType.e_conversion_finished:
                    await HandleOpenFileOnConversionFinishedAsync();
                    break;

                case PDFViewCtrlConversionType.e_conversion_failed:
                    // notify
                    break;
            }
        }

        private async Task HandleOpenFileOnConversionFinishedAsync()
        {
            PDFViewCtrl.SetZoom(ViewerSettings.DefaultZoomLevel);
            // save and update status
            PDFDoc doc = PDFViewCtrl.GetDoc();
            StorageFile tempPDFFile = null;
            try
            {
                if (!DocumentManager.IsReady)
                {
                    AnalyticsHandlerBase.CURRENT.SendEvent(AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        "DocumentOpener invoked without DocumentManager ready");
                    Debug.WriteLine("DocumentOpener invoked without DocumentManager ready");
                    await DocumentManager.Instance.InitializeInstanceAsync().ConfigureAwait(false);
                }

                Debug.WriteLine("About to ask for temp file");
                tempPDFFile = await DocumentManager.Instance.OpenTemporaryCopyAsync(FileSourceIfNotSaveable, false).ConfigureAwait(false);

                if (tempPDFFile == null)
                {
                    tempPDFFile = await ApplicationData.Current.TemporaryFolder.CreateFileAsync(
                        "FromConversion.pdf",
                        CreationCollisionOption.GenerateUniqueName).AsTask().ConfigureAwait(false);
                }
            }
            catch
            {
                //Swallow exception
            }

            if (tempPDFFile != null)
            {
                await DocSemWaitAsync().ConfigureAwait(false);
                try
                {
                    Debug.WriteLine("Trying to create temp doc");
                    Debug.WriteLine("Got a temp file");
                    if (tempPDFFile != null)
                    {
                        Debug.WriteLine("About to save");
                        await doc.SaveToNewLocationAsync(tempPDFFile, SDFDocSaveOptions.e_remove_unused).AsTask().ConfigureAwait(false);
                        Debug.WriteLine("Save completed");
                    }
                }
                catch (Exception ex)
                {
                    AnalyticsHandlerBase.CURRENT.SendEvent(
                        AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                        AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    DocSemRelease();
                }

                if (PDFViewCtrl != null)
                {
                    await PDFViewCtrl.Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () =>
                    {
                        if (PDFViewCtrl != null) // can happen if someone closes a tab that just finished converting
                        {
                            TempFile = tempPDFFile;
                            Doc = doc;
                            DocumentState = OpenedDocumentStates.ReadOnly;

                            _pdfUpdated?.Invoke(this, new RoutedEventArgs());

                            PDFViewCtrl.GetThumbAsync(1);
                            PdfReady = true;

                            if (_outlineDialogViewModel != null)
                                _outlineDialogViewModel.Isconverting = false;
                        }
                    });
                }
            }
            else
            {
                DocumentState = OpenedDocumentStates.Uneditable;
            }
        }

        async void PDFViewCtrl_OnThumbnailGenerated(int pageNumber, bool wasThumbGenerated, byte[] thumb, int w, int h)
        {
            if (pageNumber != 1 || !ShowPreview || MetaData.IsThumbnailUpToDate ||
                PDFViewCtrl.GetColorPostProcessMode() != PDFRasterizerColorPostProcessMode.e_postprocess_none)
            {
                return;
            }

            if (!wasThumbGenerated)
            {
                UpdatePreview();
                return;
            }

            try
            {
                StorageFolder folder = await ApplicationData.Current.LocalFolder.CreateFolderAsync(DocumentManager.TabPreviewFolderName, CreationCollisionOption.OpenIfExists);
                if (folder == null)
                    return;

                string filename = Guid.NewGuid().ToString() + ".bmp";
                StorageFile file = null;
                IRandomAccessStream fileStream = null;

                file = await folder.CreateFileAsync(filename, CreationCollisionOption.ReplaceExisting);
                if (file == null)
                    return;

                fileStream = await file.OpenAsync(FileAccessMode.ReadWrite);
                if (fileStream == null)
                    return;

                BitmapEncoder enco = await BitmapEncoder.CreateAsync(BitmapEncoder.BmpEncoderId, fileStream);

                const uint maxSide = 500;
                uint pixW = maxSide;
                uint pixH = maxSide;

                if (w > h)
                {
                    pixH = (uint)(h * (maxSide / (double)w));
                }
                else
                {
                    pixW = (uint)(w * (maxSide / (double)h));
                }

                enco.BitmapTransform.ScaledHeight = pixH;
                enco.BitmapTransform.ScaledWidth = pixW;
                enco.SetPixelData(BitmapPixelFormat.Bgra8, BitmapAlphaMode.Straight, (uint)w, (uint)h, 72, 72, thumb);

                await enco.FlushAsync();
                fileStream.Dispose();
                try
                {
                    if (!string.IsNullOrEmpty(MetaData.ThumbnailLocation))
                    {
                        IStorageItem oldFile = await folder.TryGetItemAsync(MetaData.ThumbnailLocation);
                        if (oldFile != null)
                            await oldFile.DeleteAsync();
                    }
                }
                catch (Exception ex)
                {
                    AppCenterHandlerBase.CURRENT.TrackError(ex);
                }

                MetaData.ThumbnailLocation = file.Name;
                MetaData.IsThumbnailUpToDate = true;
                RaisePropertyChanged(nameof(PreviewSource));
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }
    }
}
