﻿using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Outline;
using pdftron.PDF.Tools.Converters.Outline;
using System;
using System.Collections.Generic;
using System.Linq;
using Windows.ApplicationModel.Resources;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;

// The User Control item template is documented at http://go.microsoft.com/fwlink/?LinkId=234236

namespace pdftron.PDF.Tools.Controls.Viewer.Dialogs
{
    public sealed partial class OutlineDialog : UserControl
    {
        private static Dictionary<WidthLayouts, double> _LayoutMinWidths;
        public Dictionary<WidthLayouts, double> LayoutMinWidths
        {
            get
            {
                if (_LayoutMinWidths == null)
                {
                    _LayoutMinWidths = new Dictionary<WidthLayouts, double>();
                    _LayoutMinWidths.Add(WidthLayouts.Narrow, 4 * (double)this.Resources["OutlineTabButtonSize"]);
                    _LayoutMinWidths.Add(WidthLayouts.Full, 6 * (double)this.Resources["OutlineTabButtonSize"]);
                }
                return _LayoutMinWidths;
            }
        }

        public OutlineDialog()
        {
            this.InitializeComponent();
            this.SizeChanged += OutlineDialog_SizeChanged;
            this.DataContextChanged += OutlineDialog_DataContextChanged;
            SetupColumnButtons();
        }

        OutlineDialogViewModel _OldDataContext = null;
        private void OutlineDialog_DataContextChanged(FrameworkElement sender, DataContextChangedEventArgs args)
        {
            if (_OldDataContext != null)
            {
                _OldDataContext.PropertyChanged -= OutlineDialog_PropertyChanged;
                _OldDataContext = null;
            }
            if (this.DataContext is OutlineDialogViewModel)
            {
                _OldDataContext = this.DataContext as OutlineDialogViewModel;
                _OldDataContext.PropertyChanged += OutlineDialog_PropertyChanged;
                if (_OldDataContext.Isconverting)
                {
                    if (BookmarksOptions.Flyout == null)
                    {
                        BookmarksOptions.Flyout = BookmarksOptions.Resources["BookmarkButtonFlyout"] as Flyout;
                        BookmarksOptions.Foreground = Resources["SystemControlDisabledBaseMediumLowBrush"] as SolidColorBrush;
                    }
                }
                else
                {
                    BookmarksOptions.Flyout = null;
                    BookmarksOptions.Foreground = ResourceHandler.ViewerResources["ThemeBrushHighlightDifferentBrightness"] as SolidColorBrush;
                }               
            }

            UpdatePivotItemsVisibility();
        }

        private void OutlineDialog_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            ResolveWidthButtons();
            if (e.NewSize.Width < LayoutMinWidths[WidthLayouts.Full])
            {
                WidthLayout = WidthLayouts.Narrow;
            }
            else
            {
                WidthLayout = WidthLayouts.Full;
            }
        }

        private void SetupColumnButtons()
        {
            try
            {
                StackPanel stack = NumColumnsFlyout.Content as StackPanel;
                if (stack != null)
                {
                    NumberToColumnNameConverter conv = new NumberToColumnNameConverter();
                    for (int i = 0; i < stack.Children.Count; ++i)
                    {
                        if (stack.Children[i] is Button button)
                        {
                            button.Content = conv.Convert(i + 1, typeof(object), null, "");
                        }
                    }
                }
            }
            catch
            {
                //swallow exception
            }
        }

        private void ColumnNumberButton_Click(object sender, RoutedEventArgs e)
        {
            NumColumnsFlyout.Hide();
        }

        private void DockAndWidthFlyoutButton_Click(object sender, RoutedEventArgs e)
        {
            Button button = sender as Button;
            string tagText = button.Tag as string;
            if (!string.IsNullOrWhiteSpace(tagText))
            {
                int width = 0;
                try
                {
                    width = System.Convert.ToInt32(tagText);
                    ViewerSettings.OutlineDialogWidth = width;
                    ResolveWidthButtons();
                }
                catch
                {
                    //swallow exception
                }

            }

            DockAndWidthFlyout.Hide();
        }

        private void ResolveWidthButtons()
        {
            Brush SelectedBrush = ResourceHandler.ViewerResources["ThemeBrushHighlightDifferentBrightness"] as Brush;
            Brush UnselectedBrush = Resources["SystemControlForegroundBaseHighBrush"] as Brush;
            if (ViewerSettings.OutlineDialogWidth == 0)
            {
                WidthButton0.Foreground = SelectedBrush;
            }
            else
            {
                WidthButton0.Foreground = UnselectedBrush;
            }
            if (ViewerSettings.OutlineDialogWidth == 1)
            {
                WidthButton1.Foreground = SelectedBrush;
            }
            else
            {
                WidthButton1.Foreground = UnselectedBrush;
            }
            if (ViewerSettings.OutlineDialogWidth == 2)
            {
                WidthButton2.Foreground = SelectedBrush;
            }
            else
            {
                WidthButton2.Foreground = UnselectedBrush;
            }
        }

        void OutlineDialog_Loaded(object sender, RoutedEventArgs e)
        {
            if (this.DataContext is OutlineDialogViewModel)
            {
                (this.DataContext as OutlineDialogViewModel).FadeAnimation = FadeOutStoryBoard;
                (this.DataContext as OutlineDialogViewModel).PropertyChanged += OutlineDialog_PropertyChanged;
            }
        }

        private void OutlineDialog_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            if (e.PropertyName.Equals("Isconverting", StringComparison.OrdinalIgnoreCase))
            {
                if ((this.DataContext as OutlineDialogViewModel).Isconverting)
                {
                    if (BookmarksOptions.Flyout == null)
                    {
                        BookmarksOptions.Flyout = BookmarksOptions.Resources["BookmarkButtonFlyout"] as Flyout;
                    }
                }
                else
                {
                    BookmarksOptions.Flyout = null;
                }
            }

            if (e.PropertyName.Equals(nameof(_OldDataContext.IsPivotItemAnnotationsEnabled), StringComparison.OrdinalIgnoreCase) ||
                    e.PropertyName.Equals(nameof(_OldDataContext.IsPivotItemBookmarksEnabled), StringComparison.OrdinalIgnoreCase) ||
                    e.PropertyName.Equals(nameof(_OldDataContext.IsPivotItemOutlineEnabled), StringComparison.OrdinalIgnoreCase) ||
                    e.PropertyName.Equals(nameof(_OldDataContext.IsPivotItemThumbnailsEnabled), StringComparison.OrdinalIgnoreCase))
            {
                UpdatePivotItemsVisibility();
            }
        }

        public static readonly DependencyProperty AnchorSideProperty = DependencyProperty.Register(
          "AnchorSide",
          typeof(AnchorSides),
          typeof(OutlineDialog),
          new PropertyMetadata(AnchorSides.None, new PropertyChangedCallback(OnAnchorSideChanged))
        );

        public AnchorSides AnchorSide
        {
            get
            {
                return (AnchorSides)GetValue(AnchorSideProperty);
            }
            set
            {
                SetValue(AnchorSideProperty, value);
            }
        }

        private static void OnAnchorSideChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            OutlineDialog ctrl = d as OutlineDialog; //null checks omitted
            AnchorSides s = (AnchorSides)e.NewValue; //null checks omitted
            if (ctrl != null)
            {
                if (s == AnchorSides.None)
                {
                    ctrl.DockButton.Visibility = Visibility.Collapsed;
                }
                else
                {
                    ResourceLoader _ResourceLoader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                    ToolTip toolTip = new ToolTip();
                    ctrl.DockButton.Visibility = Visibility.Visible;
                    if (s == AnchorSides.Left)
                    {
                        ctrl.DockButtonFontIcon.Glyph = new string((char)0xE146, 1);
                        toolTip.Content = _ResourceLoader.GetString("OutlineDialog_DockButton_Right_ToolTip");
                        ctrl.SwitchDockSideButton.Content = _ResourceLoader.GetString("OutlineDialog_DockButton_Right_ToolTip");
                    }
                    else if (s == AnchorSides.Right)
                    {
                        ctrl.DockButtonFontIcon.Glyph = new string((char)0xE145, 1);
                        toolTip.Content = _ResourceLoader.GetString("OutlineDialog_DockButton_Left_ToolTip");
                        ctrl.SwitchDockSideButton.Content = _ResourceLoader.GetString("OutlineDialog_DockButton_Left_ToolTip");
                    }
                    ToolTipService.SetToolTip(ctrl.DockButton, toolTip);
                }
            }
        }

        private WidthLayouts _WidthLayout = WidthLayouts.Full;
        private WidthLayouts WidthLayout
        {
            get { return _WidthLayout; }
            set
            {
                if (_WidthLayout != value)
                {
                    _WidthLayout = value;
                    UpdateWidthLayout();
                }
            }
        }

        private void UpdateWidthLayout()
        {
            if (WidthLayout == WidthLayouts.Narrow)
            {
                VisualStateManager.GoToState(this, "Thin", false);
                if (ThumbnailsOptions.Flyout != null)
                {
                    ThumbnailsOptions.Flyout = null;
                }
                SecondaryColumnsButton.Flyout = NumColumnsFlyout;
            }
            else
            {
                VisualStateManager.GoToState(this, "Middle", false);
                if (SecondaryColumnsButton.Flyout != null)
                {
                    SecondaryColumnsButton.Flyout = null;
                }
                ThumbnailsOptions.Flyout = NumColumnsFlyout;
            }
        }

        /// <summary>
        /// Update the visibility of the Side Panel / Outline pivot items
        /// according to the ViewModel properties
        /// </summary>
        private void UpdatePivotItemsVisibility()
        {
            if (_OldDataContext != null && MainPivot.Items != null)
            {
                if ((!_OldDataContext.IsPivotItemThumbnailsEnabled) && MainPivot.Items.Contains(PivotItem_Thumbnails))
                {
                    MainPivot.Items.Remove(PivotItem_Thumbnails);
                }

                if ((!_OldDataContext.IsPivotItemBookmarksEnabled) && MainPivot.Items.Contains(PivotItem_Bookmarks))
                {
                    MainPivot.Items.Remove(PivotItem_Bookmarks);
                }

                if ((!_OldDataContext.IsPivotItemOutlineEnabled) && MainPivot.Items.Contains(PivotItem_Outline))
                {
                    MainPivot.Items.Remove(PivotItem_Outline);
                }

                if ((!_OldDataContext.IsPivotItemAnnotationsEnabled) && MainPivot.Items.Contains(PivotItem_Annotations))
                {
                    MainPivot.Items.Remove(PivotItem_Annotations);
                }
            }
        }
    }
}
