﻿using pdftron.PDF.Tools.Controls.ViewModels.FileOpening;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Outline;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Settings;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using Windows.ApplicationModel;
using Windows.ApplicationModel.Resources;
using Windows.Storage;
using Windows.UI;

namespace pdftron.PDF.Tools.Controls.Viewer
{
    public static class ViewerSettings
    {
        private static SettingUpdatedDelegate _settingUpdated;
        public delegate void SettingUpdatedDelegate(string settingName);
        public static event SettingUpdatedDelegate SettingUpdated
        {
            add
            {
                if (_settingUpdated?.GetInvocationList().Select(x => x.Method).Contains(value.Method) != true)
                    _settingUpdated += value;
            }
            remove
            {
                if (_settingUpdated != null)
                {
                    SettingUpdatedDelegate eventMethod = (SettingUpdatedDelegate)Array.Find(
                        _settingUpdated.GetInvocationList(),
                        x => x.Method == value.Method);
                    if (eventMethod != null)
                        _settingUpdated -= eventMethod;
                }
            }
        }

        private const int NUM_CUSTOM_COLORS = 15;

        private static string CUSTOM_COLOR_OPTIONS_STRING =
            "241,244,245,0,0,0:54,59,61,0,0,0:191,216,194,64,75,64:208,217,227,41,60,88:222,206,172,91,67,46:233,205,213,170,69,92:153,255,204,0,0,0:153,204,255,0,0,0:52,48,46,137,135,132:52,45,44,252,220,220:90,95,55,252,220,220:64,47,33,222,206,172:8,48,69,219,225,227:53,59,61,135,150,153:0,51,0,255,255,255";
        private static string CUSTOM_COLOR_LIST_NAME = "CompleteReader_CustomColorListName";
        private static string CUSTOM_COLOR_LIST_INDEX_NAME = "CompleteReader_CustomColorListIndexName";
        private static string COLOR_MODE_SETTING_NAME = "CompleteReader_ColorMode";
        private static string OUTLINE_DEFAULT_VIEW_NAME = "CompleteReader_OutlineDefaultView";
        private static string OUTLINE_DIALOG_THUMBNAIL_NUM_COLUMNS_NAME = "CompleteReader_OutlineDialogNumThumbnailColumns";
        private static string OUTLINE_DIALOG_WIDTH_NAME = "CompleteReader_OutlineDialogWidth";
        private static string OUTLINE_DIALOG_ANCHOR_SIDE_NAME = "CompleteReader_OutlineDialogAnchorSide";
        private static string PAGE_PRESENTATION_MODE_SETTING_NAME = "CompleteReader_PagePresentationMode2";
        private static string REMEMBER_LAST_PAGE_SETTING_NAME = "CompleteReader_RememberLastPage";
        private static string COPY_ANNOTATED_TEXT_TO_NOTE_NAME = "CompleteReader_CopyAnnotatedText";
        private static string MAINTAIN_ZOOM_SETTING_NAME = "CompleteReader_MaintainZoom";
        private static string DEFAULT_ZOOM_SETTING_NAME = "CompleteReader_DefaultZoom";
        private static string MAX_ZOOM_SETTING_NAME = "CompleteReader_MaxZoomLevel";
        private static string MIN_ZOOM_SETTING_NAME = "CompleteReader_MinZoomLevel";
        private static string RIGHT_TO_LEFT_VISIBILE_NAME = "CompleteReader_RightToLeftVisible";
        private static string PHONE_FULL_SCREEN_NAME = "CompleteReader_PhoneFullScreen";
        private static string EMAIL_SIGNATURE_SETTING_NAME = "CompleteReader_EmailSignature";
        private static string PIN_COMMAND_BAR_SETTING_NAME = "CompleteReader_PinCommandBar";
        private static string ENABLEJAVASCRIPT_SETTING_NAME = "CompleteReader_EnableJavaScript";
        private static string SCREEN_SLEEP_LOCK = "CompleteReader_ScreenSleepLock";
        private static string BUTTON_STAY_DOWN_SETTING_NAME = "CompleteReader_ButtonsStayDown";
        private static string OUTLINE_DIALOG_OPEN_NAME = "CompleteReader_OutlineDialogOpen";
        private static string AUTO_SAVE_ON_SETTING_NAME = "CompleteReader_AutoSaveOn";
        private static string AUTO_SAVE_ON_EVERY_CHANGE_NAME = "CompleteReader_AutoSaveOnEveryChange";
        private static string AUTO_SAVE_INTERVAL = "CompleteReader_AutoSaveInterval";
        private static string INK_SMOOTHING_SETTING_NAME = "CompleteReader_InkSmmothing";
        private static string INK_MULTI_STROKE_NAME = "CompleteReader_InkMultiStroke";
        private static string SIGNATURE_DIALOG_FULL_SCREEN_NAME = "CompleteReader_SignatureFullScreen";
        private static string STYLUS_AS_PEN_NAME = "CompleteReader_StylusAsPen";
        private static string TAB_SCROLLING_SETTING_NAME = "CompleteReader_TabScrolling";
        private static string ALWAYS_SHOW_PAGE_NAVIGATION_TOOLS_SETTING_NAME = "CompleteReader_AlwaysShowPageNavigationTools";
        private static string SAVE_COPY_TIP_SETTING_NAME = "CompleteReader_SaveCopy";
        private static string REDACTION_TIP_SETTING_NAME = "CompleteReader_RedactionTip";
        private static string EXPORT_OPTIONS_TIP_SETTING_NAME = "CompleteReader_ExportOptsTip";
        private static string CONTEXT_MENU_FLYOUT_OPENED = "CompleteReader_ContextMenuIsFlyoutOpened";
        private static string REFLOW_EDITING_NAME = "CompleteReader_ReflowEditing";
        private static string RICH_TEXT_SUPPORT_NAME = "CompleteReader_RichTextSupport";

        private static IList<Tuple<Color, Color>> CUSTOM_COLOR_OPTIONS;
        private const int CUSTOM_COLOR_ICON_DEFAULT = 0;
        private const CustomColorModeType COLOR_MODE_DEFAULT = CustomColorModeType.none;
        private const int OUTLINE_DEFAULT_VIEW_DEFAULT = 0;
        private const int OUTLINE_DIALOG_THUMBNAIL_NUM_COLUMNS_DEFAULT = 2;
        private const int OUTLINE_DIALOG_WIDTH_DEFAULT = 2;
        private const AnchorSides OUTLINE_DIALOG_ANCHOR_SIDE_DEFAULT = AnchorSides.Left;
        private const bool REMEMBER_LAST_PAGE_DEFAULT = true;
        private const bool COPY_ANNOTATED_TEXT_TO_NOTE_DEFAULT = false;
        private const bool MAINTAIN_ZOOM_DEFAULT = true;
        private const bool RIGHT_TO_LEFT_VISIBLE_DEFAULT = false;
        private const double DEFAULT_ZOOM_LEVEL = 1.0;
        private const double DEFAULT_MAX_ZOOM_LEVEL = 20;
        private const double DEFAULT_MIN_ZOOM_LEVEL = 0.1;
        private const int PHONE_FULL_SCREEN_DEFAULT = 0;
        private const bool PIN_COMMAND_BAR_DEFAULT = true;
        private const bool ENABLEJAVASCRIPT_DEFAULT = true;
        private const bool SCREEN_SLEEP_LOCK_DEFAULT = false;
        private const bool BUTTONS_STAY_DOWN_DEFAULT = true;
        private const bool OUTLINE_DIALOG_OPEN_DEFAULT = false;
        private const bool AUTO_SAVE_ON_DEFAULT = false;
        private const bool AUTO_SAVE_ON_EVERY_CHANGE_DEFAULT = true;
        private const int AUTO_SAVE_INTERVAL_DEFAULT = 30;
        private static ToolManager.InkSmoothingOptions INK_SMOOTHING_DEFAULT;
        private static bool INK_MULTI_STROKE_DEFAULT = true;
        private static bool SIGNATURE_DIALOG_FULL_SCREEN = true;
        private static bool STYLUS_AS_PEN_DEFAULT = true;
        private static bool SAVE_COPY_TIP_DEFAULT;
        private static bool REDACTION_TIP_DEFAULT; 
        private static bool EXPORT_OPTIONS_TIP_DEFAULT;
        private static bool TAB_SCROLLING_DEFAULT = false;
        private static bool ALWAYS_SHOW_PAGE_NAVIGATION_TOOLS_DEFAULT = false;
        private static bool CONTEXT_MENU_FLYOUT_OPENED_DEFAULT = false;
        private static bool REFLOW_EDITING_DEFAULT = true;
        private static bool RICH_TEXT_SUPPORT_DEFAULT = false;

        private static readonly List<string> _ImageTypesForConversion = new List<string> { ".jpg", ".bmp", ".jpeg", ".tif", ".tiff", ".png", ".gif", ".ico", ".jpeg-xr" };
        private static readonly List<string> _imageTypes = new List<string> { ".jpeg", ".jpg", ".png", ".bmp", ".cbz" };
        private static readonly List<string> _staticConversionTypes = new List<string> { ".xps", ".oxps" };
        private static readonly List<string> _officeFileTypes = new List<string> { ".docx", ".doc", ".ppt", ".pptx", ".xls", ".xlsx" };
        private static readonly List<string> _pdfFileTypes = new List<string> { ".pdf" };
        private static readonly List<string> _universalConversionFromString = new List<string>();

        private static List<string> _associatedFileTypes;
        private static List<string> _filterPdfFileTypes;
        private static List<string> _filterOfficeFileTypes;
        private static List<string> _filterImageFileTypes;

        private static ApplicationDataContainer _localSettings;
        private static ApplicationDataContainer LocalSettings => _localSettings ?? (_localSettings = ApplicationData.Current.LocalSettings);

        private static ApplicationDataContainer _roamingSettings;
        private static ApplicationDataContainer RoamingSettings => _roamingSettings ?? (_roamingSettings = ApplicationData.Current.RoamingSettings);

        private static void RaiseSettingChangedDelegate([CallerMemberName] string settingName = "") => _settingUpdated?.Invoke(settingName);

        public static string DisplayName { get; set; } = Package.Current.DisplayName;

        public static string SupportName { get; set; } = "support@xodo.com";

        public static void SetCustomColorIcon(Color background, Color foreground, int index)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(background.R + "," + background.G + "," + background.B + "," + foreground.R + "," + foreground.G + "," + foreground.B);
            LocalSettings.Values[CUSTOM_COLOR_LIST_NAME + index] = sb.ToString();
            if (CUSTOM_COLOR_OPTIONS != null)
                CUSTOM_COLOR_OPTIONS[index] = new Tuple<Color, Color>(background, foreground);
        }

        public static int CurrentCustomColorIcon
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(CUSTOM_COLOR_LIST_INDEX_NAME))
                    {
                        int index = (int)LocalSettings.Values[CUSTOM_COLOR_LIST_INDEX_NAME];
                        if (index < 0 || index >= CUSTOM_COLOR_OPTIONS.Count)
                        {
                            index = 0;
                        }
                        return index;
                    }
                }
                catch
                {
                    //swallow exception
                }

                return CUSTOM_COLOR_ICON_DEFAULT;
            }
            set
            {
                LocalSettings.Values[CUSTOM_COLOR_LIST_INDEX_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static CustomColorModeType ColorMode
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(COLOR_MODE_SETTING_NAME))
                        return (CustomColorModeType)LocalSettings.Values[COLOR_MODE_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return COLOR_MODE_DEFAULT;
            }
            set
            {
                LocalSettings.Values[COLOR_MODE_SETTING_NAME] = (int)value;
                RaiseSettingChangedDelegate();
            }
        }

        public static void SetDefaultCustomColors()
        {
            char[] separateColors = { ':' };
            string[] colorStrings = CUSTOM_COLOR_OPTIONS_STRING.Split(separateColors);
            for (int i = 0; i < NUM_CUSTOM_COLORS; ++i)
            {
                LocalSettings.Values[CUSTOM_COLOR_LIST_NAME + i] = colorStrings[i];
            }

            CUSTOM_COLOR_OPTIONS = ParseColorString(CUSTOM_COLOR_OPTIONS_STRING);
        }

        public static IList<Tuple<Color, Color>> GetCustomColors() => CUSTOM_COLOR_OPTIONS ?? (CUSTOM_COLOR_OPTIONS = ParseColorString(CUSTOM_COLOR_OPTIONS_STRING));

        public static IList<Tuple<Color, Color>> ParseColorString(string colorString)
        {
            try
            {
                List<Tuple<Color, Color>> colors = new List<Tuple<Color, Color>>();
                char[] separateColors = { ':' };
                char[] separateBytes = { ',' };
                string[] colorStrings = colorString.Split(separateColors);
                int index = 0;
                foreach (string color in colorStrings)
                {
                    string[] byteStrings;
                    if (LocalSettings.Values.ContainsKey(CUSTOM_COLOR_LIST_NAME + index))
                    {
                        string str = (string)LocalSettings.Values[CUSTOM_COLOR_LIST_NAME + index];
                        byteStrings = str.Split(separateBytes);
                    }
                    else
                    {
                        byteStrings = color.Split(separateBytes);
                    }

                    Color background = Color.FromArgb(255, byte.Parse(byteStrings[0]), byte.Parse(byteStrings[1]), byte.Parse(byteStrings[2]));
                    Color foreground = Color.FromArgb(255, byte.Parse(byteStrings[3]), byte.Parse(byteStrings[4]), byte.Parse(byteStrings[5]));
                    Tuple<Color, Color> pair = new Tuple<Color, Color>(background, foreground);
                    colors.Add(pair);

                    index++;
                }

                return colors;
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.StackTrace);
            }

            return null;
        }

        public static int OutlineDefaultView
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(OUTLINE_DEFAULT_VIEW_NAME))
                    {
                        return (int)LocalSettings.Values[OUTLINE_DEFAULT_VIEW_NAME];
                    }
                }
                catch
                {
                    //swallow
                }

                return OUTLINE_DEFAULT_VIEW_DEFAULT;
            }
            set
            {
                LocalSettings.Values[OUTLINE_DEFAULT_VIEW_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static int OutlineDialogNumThumbnailColumns
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(OUTLINE_DIALOG_THUMBNAIL_NUM_COLUMNS_NAME))
                    {
                        return (int)LocalSettings.Values[OUTLINE_DIALOG_THUMBNAIL_NUM_COLUMNS_NAME];
                    }
                }
                catch
                {
                    //swallow exception
                }

                return OUTLINE_DIALOG_THUMBNAIL_NUM_COLUMNS_DEFAULT;
            }
            set
            {
                LocalSettings.Values[OUTLINE_DIALOG_THUMBNAIL_NUM_COLUMNS_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static int OutlineDialogWidth
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(OUTLINE_DIALOG_WIDTH_NAME))
                    {
                        return (int)LocalSettings.Values[OUTLINE_DIALOG_WIDTH_NAME];
                    }
                }
                catch
                {
                    //swallow exception
                }

                return OUTLINE_DIALOG_WIDTH_DEFAULT;
            }
            set
            {
                LocalSettings.Values[OUTLINE_DIALOG_WIDTH_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static AnchorSides OutlineDialogAnchorSide
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(OUTLINE_DIALOG_ANCHOR_SIDE_NAME))
                    {
                        return (AnchorSides)LocalSettings.Values[OUTLINE_DIALOG_ANCHOR_SIDE_NAME];
                    }
                }
                catch
                {
                    //swallow exception
                }

                return OUTLINE_DIALOG_ANCHOR_SIDE_DEFAULT;
            }
            set
            {
                LocalSettings.Values[OUTLINE_DIALOG_ANCHOR_SIDE_NAME] = (int)value;
                RaiseSettingChangedDelegate();
            }
        }

        public static PDFViewCtrlPagePresentationMode PagePresentationMode
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(PAGE_PRESENTATION_MODE_SETTING_NAME))
                    {
                        if (RoamingSettings.Values[PAGE_PRESENTATION_MODE_SETTING_NAME] is bool isContinuous)
                        {
                            if (isContinuous)
                            {
                                return PDFViewCtrlPagePresentationMode.e_single_continuous;
                            }
                            else
                            {
                                return PDFViewCtrlPagePresentationMode.e_single_page;
                            }
                        }
                        else // it's an int
                        {
                            int presMode = (int)RoamingSettings.Values[PAGE_PRESENTATION_MODE_SETTING_NAME];
                            return (PDFViewCtrlPagePresentationMode)presMode;
                        }
                    }
                }
                catch
                {
                    //Swallow exception
                }

                return PDFViewCtrlPagePresentationMode.e_single_continuous;
            }
            set
            {
                RoamingSettings.Values[PAGE_PRESENTATION_MODE_SETTING_NAME] = (int)value;
                RaiseSettingChangedDelegate();
            }
        }

        public static FileOpeningType GetFileOpeningType(StorageFile file)
        {
            if (file == null)
                return FileOpeningType.PDF;

            var fileTypeWithDot = file.FileType;

            if (_staticConversionTypes.Contains(fileTypeWithDot, StringComparer.OrdinalIgnoreCase))
                return FileOpeningType.StaticConversion;

            if (_imageTypes.Contains(fileTypeWithDot, StringComparer.OrdinalIgnoreCase))
                return FileOpeningType.ImageConversion;

            if (_officeFileTypes.Contains(fileTypeWithDot, StringComparer.OrdinalIgnoreCase) || _imageTypes.Contains(fileTypeWithDot, StringComparer.OrdinalIgnoreCase))
                return FileOpeningType.UniversalConversion;

            if (_universalConversionFromString.Contains(fileTypeWithDot, StringComparer.OrdinalIgnoreCase))
                return FileOpeningType.UniversalConversionFromString;

            return FileOpeningType.PDF;
        }

        public static List<string> AssociatedFileTypes
        {
            get
            {
                if (_associatedFileTypes == null)
                {
                    _associatedFileTypes = new List<string>();
                    _associatedFileTypes.AddRange(_pdfFileTypes);
                    _associatedFileTypes.AddRange(_staticConversionTypes);
                    _associatedFileTypes.AddRange(_officeFileTypes);
                    _associatedFileTypes.AddRange(_imageTypes);
                    _associatedFileTypes.AddRange(_universalConversionFromString);
                }

                return _associatedFileTypes;
            }
        }

        public static List<string> FilterPdfFileTypes
        {
            get
            {
                if (_filterPdfFileTypes == null)
                {
                    _filterPdfFileTypes = new List<string>();
                    _filterPdfFileTypes.AddRange(_pdfFileTypes);
                }

                return _filterPdfFileTypes;
            }
        }

        public static List<string> FilterOfficeFileTypes
        {
            get
            {
                if (_filterOfficeFileTypes == null)
                {
                    _filterOfficeFileTypes = new List<string>();
                    _filterOfficeFileTypes.AddRange(_staticConversionTypes);
                    _filterOfficeFileTypes.AddRange(_officeFileTypes);
                }

                return _filterOfficeFileTypes;
            }
        }

        public static List<string> FilterImageFileTypes
        {
            get
            {
                if (_filterImageFileTypes == null)
                {
                    _filterImageFileTypes = new List<string>();
                    _filterImageFileTypes.AddRange(_imageTypes);
                }

                return _filterImageFileTypes;
            }
        }

        /// <summary>
        /// A list of all image format types that are supported for conversion (Image to PDF)
        /// </summary>
        public static List<string> ImagesTypesForConversion { get { return _ImageTypesForConversion; }}

        public static bool RememberLastPage
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(REMEMBER_LAST_PAGE_SETTING_NAME))
                        return (bool)RoamingSettings.Values[REMEMBER_LAST_PAGE_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return REMEMBER_LAST_PAGE_DEFAULT;
            }
            set
            {
                RoamingSettings.Values[REMEMBER_LAST_PAGE_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool CopyAnnotatedTextToNote
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(COPY_ANNOTATED_TEXT_TO_NOTE_NAME))
                    {
                        return System.Convert.ToBoolean(LocalSettings.Values[COPY_ANNOTATED_TEXT_TO_NOTE_NAME]);
                    }
                }
                catch
                {
                    //swallow exception
                }

                return COPY_ANNOTATED_TEXT_TO_NOTE_DEFAULT;
            }
            set
            {
                LocalSettings.Values[COPY_ANNOTATED_TEXT_TO_NOTE_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        /// <summary>
        /// Choose weather or not to maintain the same zoom level when changing pages in the viewer
        /// </summary>
        public static bool MaintainZoom
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(MAINTAIN_ZOOM_SETTING_NAME))
                        return (bool)LocalSettings.Values[MAINTAIN_ZOOM_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return MAINTAIN_ZOOM_DEFAULT;
            }
            set
            {
                LocalSettings.Values[MAINTAIN_ZOOM_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        /// <summary>
        /// The deafult zoom level when opening a new PDF document
        /// </summary>
        public static double DefaultZoomLevel 
        { 
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(DEFAULT_ZOOM_SETTING_NAME))
                        return (double)LocalSettings.Values[DEFAULT_ZOOM_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return DEFAULT_ZOOM_LEVEL;
            }
            set 
            {
                LocalSettings.Values[DEFAULT_ZOOM_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        /// <summary>
        /// The max zoom level allowed
        /// </summary>
        public static double MaxZoomLevel
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(MAX_ZOOM_SETTING_NAME))
                        return (double)LocalSettings.Values[MAX_ZOOM_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return DEFAULT_MAX_ZOOM_LEVEL;
            }
            set
            {
                LocalSettings.Values[MAX_ZOOM_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        /// <summary>
        /// The min zoom level allowed
        /// </summary>
        public static double MinZoomLevel
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(MIN_ZOOM_SETTING_NAME))
                        return (double)LocalSettings.Values[MIN_ZOOM_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return DEFAULT_MIN_ZOOM_LEVEL;
            }
            set
            {
                LocalSettings.Values[MIN_ZOOM_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static int PhoneFullScreen
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(PHONE_FULL_SCREEN_NAME))
                    {
                        return (int)LocalSettings.Values[PHONE_FULL_SCREEN_NAME];
                    }
                }
                catch
                {
                    //swallow exception
                }
                return PHONE_FULL_SCREEN_DEFAULT;
            }
            set
            {
                LocalSettings.Values[PHONE_FULL_SCREEN_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static string EmailSignature
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(EMAIL_SIGNATURE_SETTING_NAME))
                        return RoamingSettings.Values[EMAIL_SIGNATURE_SETTING_NAME] as string;
                }
                catch
                {
                    //Swallow exception
                }

                ResourceLoader loader = ResourceLoader.GetForCurrentView("pdftron.PDF.Tools/Resources");
                return string.Format(loader.GetString("Settings_Options_EmailSignature_Default"), DisplayName);
            }
            set
            {
                RoamingSettings.Values[EMAIL_SIGNATURE_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool PinCommandBar
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(PIN_COMMAND_BAR_SETTING_NAME))
                    {
                        return (bool)LocalSettings.Values[PIN_COMMAND_BAR_SETTING_NAME];
                    }
                }
                catch
                {
                    //swallow
                }

                return PIN_COMMAND_BAR_DEFAULT;
            }
            set
            {
                LocalSettings.Values[PIN_COMMAND_BAR_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool EnableJavaScript
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(ENABLEJAVASCRIPT_SETTING_NAME))
                        return (bool)_roamingSettings.Values[ENABLEJAVASCRIPT_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return ENABLEJAVASCRIPT_DEFAULT;
            }
            set
            {
                RoamingSettings.Values[ENABLEJAVASCRIPT_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool ScreenSleepLock
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(SCREEN_SLEEP_LOCK))
                    {
                        return System.Convert.ToBoolean((LocalSettings.Values[SCREEN_SLEEP_LOCK]));
                    }
                }
                catch
                {
                    //swallow exception
                }

                return SCREEN_SLEEP_LOCK_DEFAULT;
            }
            set
            {
                LocalSettings.Values[SCREEN_SLEEP_LOCK] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool ButtonsStayDown
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(BUTTON_STAY_DOWN_SETTING_NAME))
                        return (bool)RoamingSettings.Values[BUTTON_STAY_DOWN_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return BUTTONS_STAY_DOWN_DEFAULT;
            }
            set
            {
                RoamingSettings.Values[BUTTON_STAY_DOWN_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool OutlineDialogOpen
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(OUTLINE_DIALOG_OPEN_NAME))
                    {
                        return (bool)LocalSettings.Values[OUTLINE_DIALOG_OPEN_NAME];
                    }
                }
                catch
                {
                    //swallow exception
                }

                return OUTLINE_DIALOG_OPEN_DEFAULT;
            }
            set
            {
                LocalSettings.Values[OUTLINE_DIALOG_OPEN_NAME] = value;
            }
        }

        public static bool AutoSaveOn
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(AUTO_SAVE_ON_SETTING_NAME))
                        return (bool)RoamingSettings.Values[AUTO_SAVE_ON_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return AUTO_SAVE_ON_DEFAULT;
            }
            set
            {
                RoamingSettings.Values[AUTO_SAVE_ON_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        /// <summary>
        /// If true the save will be performed on every change made to the document. If False will save on a timer event.
        /// </summary>
        public static bool AutoSaveOnEveryChange
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(AUTO_SAVE_ON_EVERY_CHANGE_NAME))
                        return (bool)RoamingSettings.Values[AUTO_SAVE_ON_EVERY_CHANGE_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return AUTO_SAVE_ON_EVERY_CHANGE_DEFAULT;
            }
            set
            {
                RoamingSettings.Values[AUTO_SAVE_ON_EVERY_CHANGE_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static int AutoSaveInterval
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(AUTO_SAVE_INTERVAL))
                        return (int)RoamingSettings.Values[AUTO_SAVE_INTERVAL];
                }
                catch
                {
                    //Swallow exception
                }

                return AUTO_SAVE_INTERVAL_DEFAULT;
            }
            set
            {
                RoamingSettings.Values[AUTO_SAVE_INTERVAL] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static ToolManager.InkSmoothingOptions InkSmoothingOption
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(INK_SMOOTHING_SETTING_NAME))
                    {
                        return (ToolManager.InkSmoothingOptions)LocalSettings.Values[INK_SMOOTHING_SETTING_NAME];
                    }
                }
                catch
                {
                    //swallow exception
                }

                return INK_SMOOTHING_DEFAULT;
            }
            set
            {
                LocalSettings.Values[INK_SMOOTHING_SETTING_NAME] = (int)value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool InkMultiStrokeOption
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(INK_MULTI_STROKE_NAME))
                    {
                        return System.Convert.ToBoolean((LocalSettings.Values[INK_MULTI_STROKE_NAME]));
                    }
                }
                catch
                {
                    //swallow exception
                }

                return INK_MULTI_STROKE_DEFAULT;
            }
            set
            {
                LocalSettings.Values[INK_MULTI_STROKE_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool SignatureFullScreenOption
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(SIGNATURE_DIALOG_FULL_SCREEN_NAME))
                    {
                        return System.Convert.ToBoolean((LocalSettings.Values[SIGNATURE_DIALOG_FULL_SCREEN_NAME]));
                    }
                }
                catch
                {
                    //swallow exception
                }

                return SIGNATURE_DIALOG_FULL_SCREEN;
            }
            set
            {
                LocalSettings.Values[SIGNATURE_DIALOG_FULL_SCREEN_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool StylusAsPen
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(STYLUS_AS_PEN_NAME))
                    {
                        return System.Convert.ToBoolean((LocalSettings.Values[STYLUS_AS_PEN_NAME]));
                    }
                }
                catch
                {
                    //swallow exception
                }

                return STYLUS_AS_PEN_DEFAULT;
            }
            set
            {
                LocalSettings.Values[STYLUS_AS_PEN_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool SaveCopyExplained
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(SAVE_COPY_TIP_SETTING_NAME))
                    {
                        return System.Convert.ToBoolean(LocalSettings.Values[SAVE_COPY_TIP_SETTING_NAME]);
                    }
                }
                catch
                {
                    //swallow exception
                }

                return SAVE_COPY_TIP_DEFAULT;
            }
            set
            {
                LocalSettings.Values[SAVE_COPY_TIP_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool RedactionExplained
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(REDACTION_TIP_SETTING_NAME))
                    {
                        return System.Convert.ToBoolean(LocalSettings.Values[REDACTION_TIP_SETTING_NAME]);
                    }
                }
                catch
                {
                    //swallow exception
                }

                return REDACTION_TIP_DEFAULT;
            }
            set
            {
                LocalSettings.Values[REDACTION_TIP_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool ExportOptsExplained
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(EXPORT_OPTIONS_TIP_SETTING_NAME))
                    {
                        return System.Convert.ToBoolean(LocalSettings.Values[EXPORT_OPTIONS_TIP_SETTING_NAME]);
                    }
                }
                catch
                {
                    //swallow exception
                }

                return EXPORT_OPTIONS_TIP_DEFAULT;
            }
            set
            {
                LocalSettings.Values[EXPORT_OPTIONS_TIP_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool TabScrolling
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(TAB_SCROLLING_SETTING_NAME))
                        return (bool)RoamingSettings.Values[TAB_SCROLLING_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return TAB_SCROLLING_DEFAULT;
            }
            set
            {
                RoamingSettings.Values[TAB_SCROLLING_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        public static bool AlwaysShowPageNavigationTools
        {
            get
            {
                try
                {
                    if (RoamingSettings.Values.ContainsKey(ALWAYS_SHOW_PAGE_NAVIGATION_TOOLS_SETTING_NAME))
                        return (bool)RoamingSettings.Values[ALWAYS_SHOW_PAGE_NAVIGATION_TOOLS_SETTING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return ALWAYS_SHOW_PAGE_NAVIGATION_TOOLS_DEFAULT;
            }
            set
            {
                RoamingSettings.Values[ALWAYS_SHOW_PAGE_NAVIGATION_TOOLS_SETTING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        /// <summary>
        /// Hide or Show Right-To-Left option in the Viewer
        /// </summary>
        public static bool IsRightToLeftVisible
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(RIGHT_TO_LEFT_VISIBILE_NAME))
                        return (bool)LocalSettings.Values[RIGHT_TO_LEFT_VISIBILE_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return RIGHT_TO_LEFT_VISIBLE_DEFAULT;
            }
            set
            {
                LocalSettings.Values[RIGHT_TO_LEFT_VISIBILE_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        /// <summary>
        /// The last state of the context menu flyout
        /// </summary>
        public static bool IsContextMenuFlyoutOpened
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(CONTEXT_MENU_FLYOUT_OPENED))
                        return (bool)LocalSettings.Values[CONTEXT_MENU_FLYOUT_OPENED];
                }
                catch
                {
                    //Swallow exception
                }

                return CONTEXT_MENU_FLYOUT_OPENED_DEFAULT;
            }
            set
            {
                LocalSettings.Values[CONTEXT_MENU_FLYOUT_OPENED] = value;
                RaiseSettingChangedDelegate();
            }
        }

        /// <summary>
        /// The last state of the context menu flyout
        /// </summary>
        public static bool IsReflowEditingEnabled
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(REFLOW_EDITING_NAME))
                        return (bool)LocalSettings.Values[REFLOW_EDITING_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return REFLOW_EDITING_DEFAULT;
            }
            set
            {
                LocalSettings.Values[REFLOW_EDITING_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }

        /// <summary>
        /// If enabled it will allow Rich Text support during FreeText annotation creation
        /// </summary>
        public static bool IsRichTextEnabled
        {
            get
            {
                try
                {
                    if (LocalSettings.Values.ContainsKey(RICH_TEXT_SUPPORT_NAME))
                        return (bool)LocalSettings.Values[RICH_TEXT_SUPPORT_NAME];
                }
                catch
                {
                    //Swallow exception
                }

                return RICH_TEXT_SUPPORT_DEFAULT;
            }
            set
            {
                LocalSettings.Values[RICH_TEXT_SUPPORT_NAME] = value;
                RaiseSettingChangedDelegate();
            }
        }
#if SUBSCRIPTIONS
        /*
         * This entire section is meant for subscription calls for products that rely on activate or deactivate features and
         * displays a lock or not allowed feature
         */

        /// <summary>
        /// Cache the current subscription status for Pro features
        /// </summary>
        public static bool IsSubscriptionActivated { get; set; }

#endif
    }
}
