﻿using System;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Data;

namespace pdftron.PDF.Tools.Converters
{
    /// <summary>
    /// This lets you go from an enum value to one of two styles, based on whether the converter parameter matches the enum's tostring.
    /// You need to provide it with the styles. This can be done in XAML, using the attached dependency properties.
    /// </summary>
    public class EnumToStyleConverter : IValueConverter
    {
        public static readonly DependencyProperty MatchedStyleProperty = DependencyProperty.RegisterAttached(
            "MatchedStyle",
            typeof(Style),
            typeof(EnumToStyleConverter),
            new PropertyMetadata(null));

        public Style MatchedStyle { get; set; }

        public static readonly DependencyProperty NotMatchedStyleProperty = DependencyProperty.RegisterAttached(
            "NotMatchedStyle",
            typeof(Style),
            typeof(EnumToStyleConverter),
            new PropertyMetadata(null));

        public Style NotMatchedStyle { get; set; }

        public object Convert(object value, Type targetType, object parameter, string language)
        {
            string val = value.ToString();
            string param = parameter as string;
            if (!string.IsNullOrEmpty(val) && !string.IsNullOrEmpty(param) && val.Equals(param, StringComparison.OrdinalIgnoreCase))
            {
                return MatchedStyle;
            }
            return NotMatchedStyle;
        }

        public object ConvertBack(object value, Type targetType, object parameter, string language)
        {
            return DependencyProperty.UnsetValue;
        }
    }
}
