﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using pdftron.PDF;
using pdftron.Common;

namespace pdftron.PDF.Tools
{
    public class Eraser : SimpleShapeCreate
    {
        private bool mShapeHasBeenCreated = false;
        Path mShape;

        protected bool mIsErasing = true;
        protected double mEraseHalfThickness = 10.0;
        protected Dictionary<SDF.Obj, Annots.Ink> mInkMap = null;
        protected List<PDFPoint> mEraseStrokeList = null;
        protected List<PDFPoint> mPointList = null;

        // update UI
        private Windows.UI.Color m_CurrentDrawingColor = Windows.UI.Colors.LightGray;
        protected UIPoint mLastPoint;

        private bool should_update = false;

        protected PathFigureCollection mPathPoints;
        protected PathFigure mCurrentPathFigure;
        protected bool mFirstTime = false;

        public Eraser(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_ink_eraser;
            mToolMode = ToolType.e_ink_eraser;

            mInkMap = new Dictionary<SDF.Obj, Annots.Ink>();
            mEraseStrokeList = new List<PDFPoint>();
            mPointList = new List<PDFPoint>();
        }

        public void SetEraserHalfThickness(double thickness)
        {
            mEraseHalfThickness = thickness;
        }

        public double GetEraserHalfThickness()
        {
            return Settings.GetThickness(ToolType.e_ink_eraser);
        }

        internal override bool PointerPressedHandler(object sender, Windows.UI.Xaml.Input.PointerRoutedEventArgs e)
        {
            PointerPoint pointerPoint = e.GetCurrentPoint(mPDFView);
            UIPoint screenPoint = pointerPoint.Position;

            base.PointerPressedHandler(sender, e);

            if (mSimpleShapeCanceledAtPress)
            {
                return false;
            }

            double halfThickness = GetEraserHalfThickness() * mPDFView.GetZoom();
            if (mPageCropOnClient != null)
            {
                if (m_cnPt1.x < mPageCropOnClient.x1 + halfThickness)
                {
                    m_cnPt1.x = mPageCropOnClient.x1 + halfThickness;
                }
                if (m_cnPt1.x > mPageCropOnClient.x2 - halfThickness)
                {
                    m_cnPt1.x = mPageCropOnClient.x2 - halfThickness;
                }
                if (m_cnPt1.y < mPageCropOnClient.y1 + halfThickness)
                {
                    m_cnPt1.y = mPageCropOnClient.y1 + halfThickness;
                }
                if (m_cnPt1.y > mPageCropOnClient.y2 - halfThickness)
                {
                    m_cnPt1.y = mPageCropOnClient.y2 - halfThickness;
                }
            }

            mEraseStrokeList.Add(new Point(m_cnPt1.x, m_cnPt1.y));
            mPointList.Add(new Point(m_cnPt1.x, m_cnPt1.y));

            // update UI
            if (!mFirstTime)
            {
                mFirstTime = true;
                mPathPoints = new PathFigureCollection();
            }
            mCurrentPathFigure = new PathFigure();
            mCurrentPathFigure.StartPoint = new UIPoint(m_cnPt1.x - mPageCropOnClient.x1, m_cnPt1.y - mPageCropOnClient.y1);
            mPathPoints.Add(mCurrentPathFigure);

            mLastPoint = new UIPoint(m_cnPt1.x, m_cnPt1.y);

            return true;
        }

        /// <summary>
        /// Called when pointer is pressed and moved, so it can collect all point and annots to be erased.
        /// </summary>
        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (mPointerDown)
            {
                if (e.Pointer.PointerId != mPointerID)
                    return true;

                // update PDF

                IList<PointerPoint> intermediatePoints = e.GetIntermediatePoints(mViewerCanvas);
                for (int i = intermediatePoints.Count - 1; i >= 0; i--)
                {
                    UIPoint currentPoint = intermediatePoints[i].Position;

                    // want to make sure we're within our bounding box.
                    m_cnPt2.x = currentPoint.X;
                    m_cnPt2.y = currentPoint.Y;
                    double halfThickness = GetEraserHalfThickness() * mPDFView.GetZoom();
                    if (mPageCropOnClient != null)
                    {
                        m_cnPt2.x = Math.Max(m_cnPt2.x, mPageCropOnClient.x1 + halfThickness);
                        m_cnPt2.x = Math.Min(m_cnPt2.x, mPageCropOnClient.x2 - halfThickness);
                        m_cnPt2.y = Math.Max(m_cnPt2.y, mPageCropOnClient.y1 + halfThickness);
                        m_cnPt2.y = Math.Min(m_cnPt2.y, mPageCropOnClient.y2 - halfThickness);
                    }

                    // Sometimes, we will get 2 point in the same, or at least a very similar position. Also, if we're out of bounds, this can happen.
                    if (Math.Abs(m_cnPt2.x - mLastPoint.X) < 0.1 && Math.Abs(m_cnPt2.y - mLastPoint.Y) < 0.1)
                        continue;

                    LineSegment segment = new LineSegment
                    {
                        Point = new UIPoint(m_cnPt2.x - mPageCropOnClient.x1, m_cnPt2.y - mPageCropOnClient.y1)
                    };

                    mCurrentPathFigure.Segments.Add(segment);

                    mEraseStrokeList.Add(new PDFPoint(m_cnPt2.x, m_cnPt2.y));
                    mPointList.Add(new PDFPoint(m_cnPt2.x, m_cnPt2.y));

                    mLastPoint = new UIPoint(m_cnPt2.x, m_cnPt2.y);
                }

                try
                {
                    mPDFView.DocLock(true);

                    for (int i = 0; i < mEraseStrokeList.Count - 1; ++i)
                    {
                        PDFPoint ppt1 = mEraseStrokeList[i];
                        PDFPoint ppt2 = mEraseStrokeList[i + 1];

                        PDFPoint pdfp1 = new PDFPoint();
                        PDFPoint pdfp2 = new PDFPoint();
                        PDFDouble xpos = new PDFDouble(0);
                        PDFDouble ypos = new PDFDouble(0);
                        xpos.Value = ppt1.x;
                        ypos.Value = ppt1.y;
                        //mPDFView.ConvScreenPtToPagePt(xpos, ypos, mDownPageNumber);
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(xpos, ypos);
                        mPDFView.ConvScreenPtToPagePt(xpos, ypos, mDownPageNumber);
                        pdfp1.x = xpos.Value;
                        pdfp1.y = ypos.Value;
                        xpos = new PDFDouble(0);
                        ypos = new PDFDouble(0);
                        xpos.Value = ppt2.x;
                        ypos.Value = ppt2.y;
                        //mPDFView.ConvScreenPtToPagePt(xpos, ypos, mDownPageNumber);
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(xpos, ypos);
                        mPDFView.ConvScreenPtToPagePt(xpos, ypos, mDownPageNumber);
                        pdfp2.x = xpos.Value;
                        pdfp2.y = ypos.Value;

                        PDFPage page = mPDFView.GetDoc().GetPage(mDownPageNumber);
                        int annot_num = page.GetNumAnnots();
                        for (int j = annot_num - 1; j >= 0; --j)
                        {
                            IAnnot annot = page.GetAnnot(j);
                            if (annot == null || !annot.IsValid()) continue;
                            if (annot.GetAnnotType() == AnnotType.e_Ink)
                            {
                                Annots.Ink ink = new Annots.Ink(annot.GetSDFObj());

                                if (ink.Erase(pdfp1, pdfp2, GetEraserHalfThickness()))
                                {
                                    var comparer = new ObjSpecialComparer();
                                    if (!mInkMap.Keys.Contains(ink.GetSDFObj(), comparer))
                                    {
                                        mInkMap.Add(ink.GetSDFObj(), ink);
                                    }
                                }
                            }
                        }

                    }

                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlock();
                }

                PDFPoint prevPoint = mEraseStrokeList[mEraseStrokeList.Count - 1];
                mEraseStrokeList.Clear();
                mEraseStrokeList.Add(prevPoint);

                // update UI
                if (!mStartDrawing) // Simple Shape is ready to draw
                {
                    // wait until you're at least START_DRAWING_THRESHOLD distance from the start
                    if ((Math.Abs(m_cnPt2.x - m_cnPt1.x) > START_DRAWING_THRESHOLD)
                        || (Math.Abs(m_cnPt2.y - m_cnPt1.y) > START_DRAWING_THRESHOLD))
                    {
                        mStartDrawing = true;

                        var color = m_CurrentDrawingColor;
                        var size = GetEraserHalfThickness() * 2 * mPDFView.GetZoom();

                        if (!mShapeHasBeenCreated)
                        {
                            mShapeHasBeenCreated = true;

                            mViewerCanvas.Children.Add(this);
                            this.SetValue(Canvas.LeftProperty, mPageCropOnClient.x1);
                            this.SetValue(Canvas.TopProperty, mPageCropOnClient.y1);
                            this.Width = Math.Abs(mPageCropOnClient.Width());
                            this.Height = Math.Abs(mPageCropOnClient.Height());

                            PathGeometry geom = new PathGeometry();
                            mShape = new Path();
                            mShape.StrokeLineJoin = PenLineJoin.Round;
                            mShape.StrokeEndLineCap = PenLineCap.Round;
                            mShape.StrokeStartLineCap = PenLineCap.Round;
                            mShape.Data = geom;
                            mShape.StrokeThickness = size;
                            mShape.Stroke = new SolidColorBrush(color);
                            mShape.Opacity = 0.7;
                            this.Children.Add(mShape);
                        }
                    }
                }

                if (mStartDrawing)
                {
                    (mShape.Data as PathGeometry).Figures = mPathPoints;
                }
            }

            return false;
        }

        /// <summary>
        /// Once the eraser marker is completed, the actual erasing is performed
        /// </summary>
        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (mPointerID < 0 || e.Pointer.PointerId != mPointerID)
            {
                return false; // ignore any secondary pointer
            }

            Dictionary<IAnnot, int> modifiedAnnots = new Dictionary<IAnnot, int>();
            Dictionary<IAnnot, int> removedAnnots = new Dictionary<IAnnot, int>();

            bool hasAnnotRemoved = false;
            bool hasAnnotEdited = false;

            if (e.Pointer.PointerId == mPointerID)
            {
                try
                {
                    mPDFView.DocLock(true);
                    Page page;
                    if (mPointList.Count == 1 && mEraseStrokeList.Count == 1)
                    {

                        PDFPoint ppt1 = mEraseStrokeList[0];

                        Point pdfp1 = new Point();
                        pdftron.Common.DoubleRef xpos = new pdftron.Common.DoubleRef(0);
                        pdftron.Common.DoubleRef ypos = new pdftron.Common.DoubleRef(0);
                        xpos.Value = ppt1.x;
                        ypos.Value = ppt1.y;
                        //mPDFView.ConvScreenPtToPagePt(xpos, ypos, mDownPageNumber);
                        mPDFView.ConvAnnotationCanvasPtToScreenPt(xpos, ypos);
                        mPDFView.ConvScreenPtToPagePt(xpos, ypos, mDownPageNumber);
                        pdfp1.x = xpos.Value;
                        pdfp1.y = ypos.Value;

                        page = mPDFView.GetDoc().GetPage(mDownPageNumber);
                        int annot_num = page.GetNumAnnots();
                        for (int j = annot_num - 1; j >= 0; --j)
                        {
                            IAnnot annot = page.GetAnnot(j);
                            if (annot == null || !annot.IsValid()) continue;
                            if (annot.GetAnnotType() == AnnotType.e_Ink)
                            {
                                pdftron.PDF.Annots.Ink ink = new pdftron.PDF.Annots.Ink(annot.GetSDFObj());

                                if (ink.Erase(pdfp1, pdfp1, GetEraserHalfThickness()))
                                {
                                    var comparer = new ObjSpecialComparer();
                                    if (!mInkMap.Keys.Contains(ink.GetSDFObj(), comparer))
                                    {
                                        mInkMap.Add(ink.GetSDFObj(), ink);
                                    }
                                }
                            }
                        }
                    }

                    // update PDF
                    page = mPDFView.GetDoc().GetPage(mDownPageNumber);
                    foreach (KeyValuePair<SDF.Obj, Annots.Ink> entry in mInkMap)
                    {

                        Annots.Ink ink = entry.Value;
                        Rect bbox = ink.GetRect();

                        Rect bbox_new = ConvertFromPageRectToScreenRect(bbox, mDownPageNumber);

                        bool annotRemoved = true;
                        for (int i = 0; i < ink.GetPathCount(); i++)
                        {
                            if (ink.GetPointCount(i) > 0)
                            {
                                annotRemoved = false;
                                break;
                            }
                        }
                        if (annotRemoved)
                        {
                            mToolManager.RaiseAnnotationPreRemovedEvent(ink, mDownPageNumber);
                            page.AnnotRemove(ink);

                            removedAnnots.Add(ink, 0);
                            hasAnnotRemoved = true;
                        }
                        else
                        {
                            try
                            {
                                ink.RefreshAppearance();

                                modifiedAnnots.Add(ink, 0);
                                hasAnnotEdited = true;
                            }
                            catch (Exception ex)
                            {
                                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                            }
                        }

                        mPDFView.Update(bbox_new);
                        should_update = true;

                    }

                    mEraseStrokeList.Clear();
                    mInkMap.Clear();
                    mPointList.Clear();

                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlock();
                }
            }

            // NOTE: in order to maximize performance, we check if more than one ink annot has been removed/modified and raise a
            // group annot removal notification. It will also help with Undo and Redo Manager considering it listens to
            // Annot Edited/Removed/Added events
            if (hasAnnotRemoved)
            {
                if (removedAnnots.Count > 1)
                    mToolManager.RaiseAnnotationGroupRemovedEvent(removedAnnots);
                else if (removedAnnots.Count > 0)
                    mToolManager.RaiseAnnotationRemovedEvent(removedAnnots.First().Key, mDownPageNumber);
            } else if (hasAnnotEdited)
            {
                mToolManager.RaiseAnnotationEditedEvent(modifiedAnnots.First().Key, mDownPageNumber);
            }

            mPointerDown = false;

            if (mStartDrawing) // don't want to add any annotations resulting from a tap
            {
                EnableScrolling();

                if (should_update)
                {
                    mToolManager.mDelayRemoveTimers.Add(new DelayRemoveTimer(mViewerCanvas, this, this, mPDFView, mToolManager));
                    mDelayedRemovalSet = true;
                    should_update = false;
                }

                EndCurrentTool(ToolType.e_pan);
            }

            mPointerID = -1;
            mViewerCanvas.ReleasePointerCaptures();
            return false;
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            return true;
        }

        #region HelperFunctions

        public class ObjSpecialComparer : IEqualityComparer<SDF.Obj>
        {
            public bool Equals(SDF.Obj x, SDF.Obj y)
            {
                return x.IsEqual(y);
            }
            public int GetHashCode(SDF.Obj x)
            {
                return x.GetHashCode();
            }
        }

        private double Distance(double x1, double y1, double x2, double y2)
        {
            double d = 0;
            d = Math.Sqrt(Math.Pow((x2 - x1), 2) + Math.Pow((y2 - y1), 2));
            return d;
        }

        private Rect GetRectFromPoints(List<PDFPoint> pointsList)
        {
            double x1 = pointsList[0].x;
            double y1 = pointsList[0].y;
            double x2 = pointsList[0].x;
            double y2 = pointsList[0].y;

            foreach (PDFPoint pt in pointsList)
            {
                if (pt.x < x1 && pt.x > 0)
                {
                    x1 = pt.x;
                }
                if (pt.y < y1 && pt.y > 0)
                {
                    y1 = pt.y;
                }
                if (pt.x > x2 && pt.x > 0)
                {
                    x2 = pt.x;
                }
                if (pt.y > y2 && pt.y > 0)
                {
                    y2 = pt.y;
                }
            }
            Rect r = new Rect(x1, y1, x2, y2);
            return r;
        }
        #endregion

    }
}
