﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;


using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.Devices.Input;

using UIPoint = Windows.Foundation.Point;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef; 
using pdftron.PDF;
using pdftron.Common;
using Windows.UI.Input;
using pdftron.PDF.Tools.Utilities;

namespace pdftron.PDF.Tools
{
    internal class Pan : Tool
    {
        internal bool mSwitchToTextSelect = false;
        bool mShouldSmartZoom = false;

        protected UIPoint mPopupTargetPoint;

        // For grabbing and scrolling
        protected int mDragPointerID = -1;
        protected UIPoint mLastPoint;
        public Pan(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_pan;
            mToolMode = ToolType.e_pan;
        }

        internal override bool OnScale()
        {
            if (mIsShowingCommandMenu)
            {
                HideContextMenu(true);
                mContextMenu = null;
            }
            mJustSwitchedFromAnotherTool = false;
            base.OnScale();
            return false;
        }

        internal override bool OnSize()
        {
            if (mIsShowingCommandMenu)
            {
                HideContextMenu(true);
                mContextMenu = null;
            }
            mJustSwitchedFromAnotherTool = false;
            base.OnSize();
            EnableScrolling();
            return false;
        }

        internal override bool OnViewChanged(Object sender, Windows.UI.Xaml.Controls.ScrollViewerViewChangedEventArgs e)
        {
            if (mIsShowingCommandMenu)
            {
                HideContextMenu(true);
                mContextMenu = null;
            }
            base.OnViewChanged(sender, e);
            return false;
        }

        internal override bool OnPageNumberChanged(int current_page, int num_pages)
        {
            if (mIsShowingCommandMenu)
            {
                HideContextMenu(true);
                mContextMenu = null;
            }
            base.OnPageNumberChanged(current_page, num_pages);
            return false;
        }
        
        internal override bool HoldingHandler(object sender, HoldingRoutedEventArgs e)
        {
            if (e.HoldingState == Windows.UI.Input.HoldingState.Started && mToolManager.EnablePopupMenuOnLongPress && ToolMode == ToolType.e_pan && !mToolManager.ReadOnlyMode)
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Context Menu", new Dictionary<string, string>() { ["Opened"] = "Holding" });

                UIPoint holdPoint = e.GetPosition(mPDFView);
                CreateCommandMenu(holdPoint);
            }
            return false;
        }

        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            base.PointerPressedHandler(sender, e);

            mShouldSmartZoom = false;
            mDragPointerID = -1;

            if (mPDFView.GetDoc() == null)
            {
                return false;
            }
            if (mIsShowingCommandMenu && !mJustSwitchedFromAnotherTool)
            {
                HideContextMenu(true);
                mContextMenu = null;
                mToolManager.SuppressTap = true;
            }

            PointerPoint sc_pointerPoint = e.GetCurrentPoint(mPDFView);
            UIPoint sc_downPoint = sc_pointerPoint.Position;
            if ((e.Pointer.PointerDeviceType == PointerDeviceType.Mouse && e.GetCurrentPoint(null).Properties.IsLeftButtonPressed))
            {
                bool shouldPan = false;
                if (mToolManager.PanToolTextSelectionMode == ToolManager.TextSelectionBehaviour.AlwaysSelect)
                {
                    mIsUsingMouse = true;
                    mNextToolMode = ToolType.e_text_select;
                }
                else if (mToolManager.PanToolTextSelectionMode == ToolManager.TextSelectionBehaviour.AlwaysPan)
                {
                    shouldPan = true;
                }
                else // mixed mode
                {
                    if (SelectTextAtPoint(sc_downPoint, 5))
                    {
                        mIsUsingMouse = true;
                        mNextToolMode = ToolType.e_text_select;
                    }
                    else
                    {
                        shouldPan = true;
                    }
                }

                if (shouldPan)
                {
                    mDragPointerID = (int)e.Pointer.PointerId;
                    mLastPoint = e.GetCurrentPoint(mPDFView).Position;
                    mPDFView.GetAnnotationCanvas().CapturePointer(e.Pointer);
                }
            }
            else if (e.Pointer.PointerDeviceType == PointerDeviceType.Pen)
            {
                int pgNum = mPDFView.GetPageNumberFromScreenPoint(sc_downPoint.X, sc_downPoint.Y);
                if (pgNum > 0)
                {
                    if (sc_pointerPoint.Properties.IsEraser) // we want to erase ink
                    {
                        mNextToolMode = ToolType.e_ink_eraser;
                        mIsUsingMouse = false;
                    }
                    else if (sc_pointerPoint.Properties.IsBarrelButtonPressed)
                    {
                        try
                        {
                            mPDFView.DocLockRead();

                            SelectAnnot((int)sc_downPoint.X, (int)sc_downPoint.Y, 11, 7);
                            if (mAnnot != null)
                            {
                                mAnnotPageNum = mPDFView.GetPageNumberFromScreenPoint(sc_downPoint.X, sc_downPoint.Y);
                                mNextToolMode = ToolType.e_annot_edit;
                                if (mAnnot.GetAnnotType() == AnnotType.e_Line)
                                {
                                    mNextToolMode = ToolType.e_line_edit;
                                }
                                if (mAnnot is Annots.ITextMarkup)
                                {
                                    mNextToolMode = ToolType.e_annot_edit_text_markup;
                                }
                            }
                        }
                        catch (Exception ex)
                        {
                            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                        }
                        finally
                        {
                            mPDFView.DocUnlockRead();
                        }
                    }
                    else if(mToolManager.StylusAsPen) // we are not erasing or selecting, so we draw ink if we're on a particular page.
                    {
                        mNextToolMode = ToolType.e_ink_create;
                        mIsUsingMouse = false;
                    }
                    // Use stylus to pan otherwise
                    else
                    {
                        mDragPointerID = (int)e.Pointer.PointerId;
                        mLastPoint = e.GetCurrentPoint(mPDFView).Position;
                        mPDFView.GetAnnotationCanvas().CapturePointer(e.Pointer);
                    }
                    
                }
            }
            mJustSwitchedFromAnotherTool = false;
            return false;
        }


        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerId == mDragPointerID)
            {
                UIPoint dragPoint = e.GetCurrentPoint(mPDFView).Position;

                mPDFView.SetHScrollPos(mPDFView.GetHScrollPos() - dragPoint.X + mLastPoint.X);
                mPDFView.SetVScrollPos(mPDFView.GetVScrollPos() - dragPoint.Y + mLastPoint.Y);
                mLastPoint = dragPoint;
                return true;
            }
            return false;
        }


        internal override bool PointerCanceledHandler(object sender, PointerRoutedEventArgs e)
        {
            base.PointerCanceledHandler(sender, e);
            if (e.Pointer.PointerId == mDragPointerID)
            {
                mDragPointerID = -1;
                mPDFView.GetAnnotationCanvas().ReleasePointerCaptures();
            }
            return true;
        }

        internal override bool PointerCaptureLostHandler(object sender, PointerRoutedEventArgs e)
        {
            base.PointerCaptureLostHandler(sender, e);
            if (e.Pointer.PointerId == mDragPointerID)
            {
                mDragPointerID = -1;
            }
            return true;
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            base.PointerReleasedHandler(sender, e);
            // We want this to happen when the finger is released, as opposed to pressed.
            if (mShouldSmartZoom)
            {
                UIPoint tapPoint = e.GetCurrentPoint(mPDFView).Position;

                // We need to marshal this call. At this point in time, the scroll viewer is still being "manipulated" and will behave oddly if we 
                // resize it's contents here. So we create a task that the UI thread can execute after it's done finalizing manipulation.
                CheckForSmartZoom(tapPoint);
                return false;
            }
            else if (mDragPointerID == e.Pointer.PointerId)
            {
                mDragPointerID = -1;
                mPDFView.GetAnnotationCanvas().ReleasePointerCaptures();
            }

            EnableScrolling();
            mJustSwitchedFromAnotherTool = false;
            return false;
        }

        private async void CheckForSmartZoom(UIPoint tapPoint)
        {
            try
            {
                // We need to marshal this call. At this point in time, the scroll viewer is still being "manipulated" and will behave oddly if we 
                // resize it's contents here. So we create a task that the UI thread can execute after it's done finalizing manipulation.
                await Windows.UI.Core.CoreWindow.GetForCurrentThread().Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High, () =>
                {
                    mShouldSmartZoom = false;
                    {
                        PDFViewCtrlPageViewMode mode = mPDFView.GetPageViewMode();
                        PDFViewCtrlPageViewMode refViewMode = mPDFView.GetPageRefViewMode();
                        if (refViewMode == PDFViewCtrlPageViewMode.e_zoom)
                        {
                            refViewMode = PDFViewCtrlPageViewMode.e_fit_page;
                        }
                        double zoom = mPDFView.GetZoom();
                        double refZoom = mPDFView.GetZoomForViewMode(refViewMode);
                        double factor = zoom / refZoom;
                        double whatup = Math.Abs(factor - 1);
                        System.Diagnostics.Debug.WriteLine("Zoom: {0}, RefZoom: {1}, factor: {2}, closeness to 0: {3}", zoom, refZoom, factor, whatup);

                        if (zoom > 0 && refZoom > 0 && Math.Abs((zoom / refZoom) - 1) < 0.05) 
                        {
                            if (!mPDFView.SmartZoom((int)tapPoint.X, (int)tapPoint.Y, true))
                            {
                                mPDFView.SetZoom((int)tapPoint.X, (int)tapPoint.Y, 2.5 * mPDFView.GetZoom(), true);
                            }
                        }
                        else
                        {
                            mPDFView.SetPageViewMode(refViewMode, tapPoint.X, tapPoint.Y, true);
                        }
                    }
                });
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        internal override bool DoubleTappedHandler(object sender, DoubleTappedRoutedEventArgs e)
        {
            if (e.PointerDeviceType != PointerDeviceType.Mouse)
            {
                mJustSwitchedFromAnotherTool = false;
                mSwitchToTextSelect = false;
                mShouldSmartZoom = true;
            }
            return false;
        }

        internal override bool DoubleClickedHandler(object sender, PointerRoutedEventArgs e)
        {
            // We're still in pan mode -> we did not select any annotation (we would have been in annot edit mode
            mNextToolMode = ToolType.e_text_select;

            return true;
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            if (mJustSwitchedFromAnotherTool && !mIsInHelperMode)
            {
                mJustSwitchedFromAnotherTool = false;
                return true;
            }
            if (mPDFView.GetDoc() == null)
            {
                return false;
            }
            if (!mPDFView.IsEnabled)
            {
                return false;
            }
            
            int x = (int)(e.GetPosition(mPDFView).X + 0.5);
            int y = (int)(e.GetPosition(mPDFView).Y + 0.5);

            mAnnot = null;
            bool foundSomething = false;
            mPDFView.DocLockRead();
            try
            {
                if (!mIsInSnappedView)
                {
                    if (e.PointerDeviceType == PointerDeviceType.Mouse)
                    {
                        SelectAnnot(x, y, 6, 3);
                    }
                    else if (e.PointerDeviceType == PointerDeviceType.Pen)
                    {
                        SelectAnnot(x, y, 11, 7);
                    }
                    else
                    {
                        SelectAnnot(x, y, 17, 11);
                    }

                    bool isRtReply = false;
                    try
                    {
                        isRtReply = UtilityFunctions.IsReplyTypeReply(mAnnot);
                    }
                    catch (Exception ex)
                    {
                        Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                    }

                    if ((mAnnot != null) && (!isRtReply))
                    {
                        foundSomething = true;
                        try
                        {
                            if (mAnnot.GetAnnotType() == AnnotType.e_Link)
                            {
                                InterceptAnnotationEventArgs interceptAnnotationEventArgs = new InterceptAnnotationEventArgs(mAnnot, mAnnotPageNum);
                                mToolManager.RaiseShouldHandleLinkAnnotationEvent(this, interceptAnnotationEventArgs);
                                if (!interceptAnnotationEventArgs.Handled)
                                {
                                    if (CanExectueAction(mAnnot))
                                    {
                                        //link navigation
                                        mNextToolMode = ToolType.e_link_action;
                                    }
                                }
                            }
                            else if (mAnnot.GetAnnotType() == AnnotType.e_Widget)
                            {
                                //form filling
                                mNextToolMode = ToolType.e_form_fill;
                            }
                            else if (mAnnot.GetAnnotType() == AnnotType.e_RichMedia)
                            {
                                // Movie annotation
                                mNextToolMode = ToolType.e_rich_media;
                            }
                            else
                            {
                                if (mAnnot is Annots.IMarkup)
                                {
                                    //Annotation editing
                                    mNextToolMode = ToolType.e_annot_edit;
                                    if (mAnnot.GetAnnotType() == AnnotType.e_Line)
                                    {
                                        mNextToolMode = ToolType.e_line_edit;
                                    }
                                    else if ((mAnnot.GetAnnotType() == AnnotType.e_Polyline)
                                        || (mAnnot.GetAnnotType() == AnnotType.e_Polygon))
                                    {
                                        mNextToolMode = ToolType.e_advanced_shape_edit;
                                    }

                                    if (mAnnot is Annots.ITextMarkup)
                                    {
                                        mNextToolMode = ToolType.e_annot_edit_text_markup;
                                    }

                                    KeyValuePair<ToolType, List<IAnnot>> pair = CanSelectGroupAnnotation(mPDFView, mAnnot, mAnnotPageNum);

                                    if (pair.Key != ToolType.e_none)
                                    {
                                        mNextToolMode = pair.Key;
                                        mAnnot = null;
                                        mGroupAnnotations = pair.Value;
                                    }
                                }
                                else
                                {
                                    foundSomething = false;
                                    mAnnot = null;
                                }
                            }
                            mAnnotPageNum = mPDFView.GetPageNumberFromScreenPoint(x, y);
                        }
                        catch (Exception ex)
                        {
                            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                        }
                        finally
                        {
                        }

                        if (mAnnot != null && 
                            mToolManager.ReadOnlyMode &&
                            (mNextToolMode == ToolType.e_form_fill || mNextToolMode == ToolType.e_annot_edit) )
                        {
                            foundSomething = false;
                            mNextToolMode = ToolType.e_pan;
                        }
                        else if (mAnnot != null)
                        {
                            return true;
                        }
                    }

                    if (!foundSomething && mIsInHelperMode)
                    {
                        return true;
                    }
                    else if (!foundSomething)
                    {
                        PDFViewCtrlLinkInfo linkInfo = mPDFView.GetLinkAt(x, y);
                        if (linkInfo != null)
                        {
                            foundSomething = true;
                            string uriString = linkInfo.GetUrl();
                            LaunchBrowser(uriString);
                        }
                    }
                }


                if (!foundSomething && e.PointerDeviceType != PointerDeviceType.Mouse)
                {
                    //if hit text, run text select tool
                    double zoomFactor = 1;
                    double x1 = e.GetPosition(mPDFView).X;
                    double y1 = e.GetPosition(mPDFView).Y;
                    double diff = 0.05 * zoomFactor;
                    x1 = Math.Max(x1 - diff, 0);
                    y1 = Math.Max(y1 - diff, 0);
                    diff *= 2;
                    double x2 = x1 + diff;
                    double y2 = y1 + diff;
                    mPDFView.ClearSelection();
                    mPDFView.SetTextSelectionMode(TextSelectionMode.e_rectangular);
                    if (mPDFView.Select(x1, y1, x2, y2) && !mIsUsingMouse)
                    {
                        mPDFView.ClearSelection();
                        DelayTextSelect(e);
                        return true;
                    }
                    else
                    {
                        mToolManager.PageNumberIndicator.Show();
                    }
                }
                else
                {
                    mToolManager.PageNumberIndicator.Show();
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlockRead();
            }

            return false;
        }

        internal async void DelayTextSelect(TappedRoutedEventArgs e)
        {
            mSwitchToTextSelect = true;
            await Task.Delay(ToolManager._SingleTapDelayMS);
            if (mSwitchToTextSelect)
            {
                mSwitchToTextSelect = false;
                mToolManager.CreateTool(ToolType.e_text_select, this);
                mToolManager.CurrentTool.TappedHandler(this, e);
            }
        }

        /// <summary>
        /// Listen to right-click of mouse events
        /// </summary>
        /// <remarks>Note: it can also listen to touch holding down, which in some system mean right-click of the mouse</remarks>
        internal override bool RightTappedHandler(object sender, RightTappedRoutedEventArgs e)
        {
            if (mToolManager.ReadOnlyMode)
                return false;

            // Shift+F10 triggers right tap, which should not open the context menu
            // https://docs.microsoft.com/en-us/uwp/api/windows.ui.xaml.uielement.righttapped?view=winrt-20348
            var shiftState = Window.Current.CoreWindow.GetKeyState(Windows.System.VirtualKey.Shift);
            var f10State = Window.Current.CoreWindow.GetKeyState(Windows.System.VirtualKey.F10);

            if ((shiftState & Windows.UI.Core.CoreVirtualKeyStates.Down) == Windows.UI.Core.CoreVirtualKeyStates.Down
               && (f10State & Windows.UI.Core.CoreVirtualKeyStates.Down) == Windows.UI.Core.CoreVirtualKeyStates.Down)
            {
                return false;
            }

            if (e.PointerDeviceType != PointerDeviceType.Touch)
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Context Menu", new Dictionary<string, string>() { ["Opened"] = "Right Click/Tap" });
                CreateCommandMenu(e.GetPosition(mPDFView));
            }

            return false;
        }

        internal override bool KeyDownAction(object sender, KeyRoutedEventArgs e)

        {
            mJustSwitchedFromAnotherTool = false;
            if (!mIsModifierKeyDown)
            {
                switch (e.Key)
                {
                        
                    case Windows.System.VirtualKey.Escape:
                        if (mIsShowingCommandMenu)
                        {
                            HideContextMenu(true);
                            mContextMenu = null;
                        }
                        return true;
                }
            }
            return base.KeyDownAction(sender, e);
        }

        #region Handlers For Popup Menu

        internal bool CreateCommandMenu(UIPoint target)
        {
            if (mPDFView.GetPageNumberFromScreenPoint(target.X, target.Y) > 0)
            {
                var creationOptionsDict = new Dictionary<string, string>();
                var flyoutOptions = new Dictionary<string, string>();

                creationOptionsDict["note"] = ResourceHandler.GetString("CreateOption_Note");
                creationOptionsDict["signature"] = ResourceHandler.GetString("CreateOption_Signature");
                creationOptionsDict["ink"] = ResourceHandler.GetString("CreateOption_Ink");
                creationOptionsDict["marker"] = ResourceHandler.GetString("CreateOption_Marker");
                creationOptionsDict["text"] = ResourceHandler.GetString("CreateOption_Text");
                creationOptionsDict["arrow"] = ResourceHandler.GetString("CreateOption_Arrow");
                creationOptionsDict["line"] = ResourceHandler.GetString("CreateOption_Line");
                creationOptionsDict["rectangle"] = ResourceHandler.GetString("CreateOption_Rectangle");
                creationOptionsDict["ellipse"] = ResourceHandler.GetString("CreateOption_Ellipse");
                creationOptionsDict["image"] = ResourceHandler.GetString("CreateOption_Image");

                if (mToolManager.IsSoundAnnotEnabled)
                    creationOptionsDict["mic"] = ResourceHandler.GetString("CreateOption_Sound");

                flyoutOptions["reset zoom"] = ResourceHandler.GetString("Reset_Zoom");
                flyoutOptions["select page text"] = ResourceHandler.GetString("SelectTextPage");
                flyoutOptions["select all text"] = ResourceHandler.GetString("SelectAllText");

                if (AnnotationClipboardHelper.IsAnnotToPaste)
                {
                    flyoutOptions["paste annot from clipboard"] = ResourceHandler.GetString("CreateOption_PasteAnnot");
                }

                mContextMenu = new Controls.ContextMenuControl(mPDFView, creationOptionsDict);
                mContextMenu.CreateOverflowFromDictionary(flyoutOptions);
                mContextMenu.ItemClicked += OnCommandMenuClicked;
            }
            else
            {
                return false;
            }

            mPopupTargetPoint = target;
            mContextMenu.TargetPoint(mPopupTargetPoint);

            var eventArgs = new ShowingContextMenuEventArgs
            {
                ToolType = ToolType.e_pan,
                AnnotType = mAnnot == null ? AnnotType.e_Unknown : mAnnot.GetAnnotType(),
                Annotation = mAnnot,
                PageNumber = mPDFView.GetCurrentPage()
            };
            mToolManager.RaiseOnShowContextMenu(mContextMenu, eventArgs);

            mContextMenu.Show();

            return true;
        }

        internal override async void OnCommandMenuClicked(string title)
        {
            Utilities.AnalyticsHandlerBase.CURRENT.SendEvent("QuickMenu Tool", title + " selected");

            if (title.Equals("line", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_line_create;
            }
            else if (title.Equals("arrow", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_arrow_create;
            }
            else if (title.Equals("rectangle", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_rect_create;
            }
            else if (title.Equals("ellipse", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_oval_create;
            }
            else if (title.Equals("cloud", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_cloud_create;
            }
            else if (title.Equals("polygon", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_polygon_create;
            }
            else if (title.Equals("polyline", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_polyline_create;
            }
            else if (title.Equals("note", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_sticky_note_create;
                mToolManager.CreateTool(ToolType.e_sticky_note_create, this);
                StickyNoteCreate stickyTool = mToolManager.CurrentTool as StickyNoteCreate;
                stickyTool.SetTargetPoint(mPopupTargetPoint);
            }
            else if (title.Equals("ink", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_ink_create;
                mToolManager.CreateTool(ToolType.e_ink_create, this);
            }
            else if (title.Equals("marker", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_ink_marker;
                mToolManager.CreateTool(ToolType.e_ink_marker, this);
            }
            else if (title.Equals("text", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_text_annot_create;
                mToolManager.CreateTool(ToolType.e_text_annot_create, this);
                FreeTextCreate freeTextTool = mToolManager.CurrentTool as FreeTextCreate;
                freeTextTool.SetTargetPoint(mPopupTargetPoint, false);
            }
            else if (title.Equals("signature", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_signature;
                mAnnot = null;
                mToolManager.CreateTool(ToolType.e_signature, this);
                Signature signatureTool = mToolManager.CurrentTool as Signature;
                signatureTool.SetTargetPoint(mPopupTargetPoint);
            }
            else if (title.Equals("image", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_stamper;
                mAnnot = null;
                mToolManager.CreateTool(ToolType.e_stamper, this);
                ImageStamper stamperTool = mToolManager.CurrentTool as ImageStamper;
                stamperTool.SetTargetPoint(mPopupTargetPoint);
            }
            else if (title.Equals("annotation group", StringComparison.OrdinalIgnoreCase))
            {
                mNextToolMode = ToolType.e_annot_edit_rect_group;
            }
            else if (title.Equals("paste annot from clipboard", StringComparison.OrdinalIgnoreCase))
            {
                await HandlePasteAsync();
            }
            else if (title.Equals("mic", StringComparison.OrdinalIgnoreCase))
            {
                HandleSoundAnnotationAsync();
            }
            else if (title.Equals("reset zoom", StringComparison.OrdinalIgnoreCase))
            {
                if (mPDFView != null)
                {
                    AppCenterHandlerBase.CURRENT.SendEvent("Context Menu", new Dictionary<string, string>() { ["Selection"] = "Reset Zoom" });

                    mPDFView.SetZoom(0.8); // Default value to a known safe one for better view
                }
            }
            else if (title.Equals("select page text", StringComparison.OrdinalIgnoreCase))
            {
                var textSelect = mToolManager.CreateTool(ToolType.e_text_select, this) as TextSelect;
                if (textSelect == null)
                    return;

                AppCenterHandlerBase.CURRENT.SendEvent("Context Menu", new Dictionary<string, string>() { ["Selection"] = "page text" });

                textSelect.SelectPageText();
            }
            else if (title.Equals("select all text", StringComparison.OrdinalIgnoreCase))
            {
                var textSelectAll = mToolManager.CreateTool(ToolType.e_text_select, this) as TextSelect;
                if (textSelectAll == null)
                {
                    return;
                }

                AppCenterHandlerBase.CURRENT.SendEvent("Context Menu", new Dictionary<string, string>() { ["Selection"] = "Select All Text" });

                textSelectAll.SelectAllText();
            }
            else
            {
                throw new Exception("The option doesn't exist");
            }

            if (mContextMenu != null)
            {
                HideContextMenu(true);
                mContextMenu = null;
            }
        }

        private async Task HandlePasteAsync()
        {
            if (AnnotationClipboardHelper.IsAnnotToPaste)
            {
                Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Clipboard", new Dictionary<string, string>() { ["Command"] = "Paste Annotation - Popup Menu" });

                var currentPage = mPDFView.GetCurrentPage();

                bool hasPasted = await AnnotationClipboardHelper.PasteAnnotFromClipboardAsync(mPDFView, currentPage);
                if (hasPasted == false)
                    return;

                try
                {
                    var doc = mPDFView.GetDoc();
                    if (doc.GetPageCount() >= currentPage)
                    {
                        var page = doc.GetPage(currentPage);
                        var annot = page.GetAnnot(page.GetNumAnnots() - 1);
                        mToolManager.RaiseAnnotationAddedEvent(annot, currentPage);
                        mToolManager.SelectAnnotation(annot, currentPage);
                    }
                }
                catch
                {
                    //Swallow exception
                }
            }
        }

        private void HandleSoundAnnotationAsync()
        {
            Utilities.AppCenterHandlerBase.CURRENT.SendEvent("Context Menu", new Dictionary<string, string>() { ["Selection"] = "Sound Annotation" });

            mToolManager.RaiseOnSoundAnnotRequestedEvent(mPopupTargetPoint, mPDFView.GetCurrentPage());
        }
        #endregion Handlers For Popup Menu

        private void HideContextMenu(bool shouldDispose = false)
        {
            if(mContextMenu != null)
            {
                if(shouldDispose)
                    ResetContextMenuEvents();
                mContextMenu.Hide(shouldDispose);
            }
        }

        private void ResetContextMenuEvents()
        {
            if(mContextMenu != null)
            {
                mContextMenu.ItemClicked -= OnCommandMenuClicked;
            }
        }
    }
}
