﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using pdftron.PDF;
using pdftron.Common;
using pdftron.PDF.Tools.Utilities;

namespace pdftron.PDF.Tools
{
    public class PerimeterMeasureCreate : PolylineCreate
    {
        /// <summary>
        /// Class constructor
        /// </summary>
        public PerimeterMeasureCreate(PDFViewCtrl pdfViewControl, ToolManager toolManager)
            : base(pdfViewControl, toolManager)
        {
            mNextToolMode = ToolType.e_perimeter_create;
            mToolMode = ToolType.e_perimeter_create;
        }

        internal override void Finish()
        {
            //creates a polyline annotation and adds it to the PDF page
            try
            {
                mPDFView.DocLock(true);

                PDFRect rect = GetAdvancedShapeBBox(mPdfPointList);
                if (rect != null)
                {
                    rect.Inflate(mDrawThickness);
                    Annots.PolyLine pdfPolyline = Annots.PolyLine.Create(mPDFView.GetDoc().GetSDFDoc(), rect);

                    int vertexIdx = 0;
                    foreach (PDFPoint point in mPdfPointList)
                    {
                        // The Rect is normalized, so we need to recalculate the points if we don't want it to start from 0,0
                        PDFPoint pdfPoint = ConvertFromCanvasPointToPagePoint(point, mDownPageNumber);
                        pdfPolyline.SetVertex(vertexIdx++, pdfPoint);
                    }

                    // Load Settings data
                    var precision = Settings.GetMeasure(this.ToolMode, MeasureInfoType.Precision, MeasureUtils.DEFAULT_PRECISION);
                    var baseScale = Settings.GetMeasure(this.ToolMode, MeasureInfoType.BaseScale, MeasureUtils.DEFAULT_BASE_SCALE);
                    var baseUnit = Settings.GetMeasure(this.ToolMode, MeasureInfoType.BaseUnit, MeasureUtils.DEFAULT_BASE_UNIT);
                    var translateScale = Settings.GetMeasure(this.ToolMode, MeasureInfoType.TranslateScale, MeasureUtils.DEFAULT_TRANSLATE_SCALE);
                    var translateUnit = Settings.GetMeasure(this.ToolMode, MeasureInfoType.TranslateUnit, MeasureUtils.DEFAULT_TRANSLATE_UNIT);

                    // Load settings data to Measure info
                    MeasureInfo measureInfo = new MeasureInfo();
                    measureInfo.Precision = MeasureUtils.ConvStrToPDFPrecision(precision);
                    measureInfo.BaseUnit = baseUnit;
                    measureInfo.BaseFactor = MeasureUtils.ConvScaleToPDFFactor(MeasureUtils.GetPageUnitFromString(baseUnit), mToolMode, baseScale, translateScale);
                    measureInfo.Unit = translateUnit;
                    measureInfo.Scale = baseScale + " " + baseUnit + " = " + translateScale + " " + translateUnit;

                    // Add all PDF spec info to the annot
                    MeasureUtils.SetMeasurementInfo(pdfPolyline, measureInfo);

                    // Proper convert Canvas pts to Page pts
                    List<Point> pgPoints = new List<PDFPoint>();
                    for (int i = 0; i < mPdfPointList.Count; i++)
                        pgPoints.Add(ConvertFromCanvasPointToPagePoint(mPdfPointList[i], CurrentAnnotPageNumber));
                    
                    // Calc and set perimeter to annot's content
                    double perimeter = MeasureUtils.CalcPerimeter(pgPoints);
                    var rsl_val = perimeter * measureInfo.BaseFactor;
                    var strVal = MeasureUtils.FormatDecimals(rsl_val, precision) + " " + translateUnit;

                    pdfPolyline.SetContents(strVal);

                    SetStyle(pdfPolyline);
                    pdfPolyline.RefreshAppearance();

                    UtilityFunctions.SetUniqueId(pdfPolyline, Guid.NewGuid().ToString());

                    PDFPage page = mPDFView.GetDoc().GetPage(mDownPageNumber);
                    page.AnnotPushBack(pdfPolyline);

                    mAnnotPageNum = mDownPageNumber;
                    mAnnot = pdfPolyline;
                    mPDFView.UpdateWithAnnot(pdfPolyline, mAnnotPageNum);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }
        }
    }
}
