﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using pdftron.PDF;
using pdftron.Common;

namespace pdftron.PDF.Tools
{
    class RectRedactionCreate : RectCreate
    {
        public RectRedactionCreate(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_rect_redaction_create;
            mToolMode = ToolType.e_rect_redaction_create;
        }

        internal override void Finish()
        {
            try
            {
                mPDFView.DocLock(true);

                Rect rect = GetShapeBBox();
                if (rect != null)
                {
                    Annots.Redaction redaction = Annots.Redaction.Create(mPDFView.GetDoc().GetSDFDoc(), rect);

                    var textOverlay = Settings.GetRedactionText();
                    redaction.SetOverlayText(textOverlay);

                    SetStyle(redaction, true);
                    redaction.RefreshAppearance();

                    UtilityFunctions.SetUniqueId(redaction, Guid.NewGuid().ToString());

                    pdftron.PDF.Page page = mPDFView.GetDoc().GetPage(mDownPageNumber);
                    page.AnnotPushBack(redaction);

                    mAnnotPageNum = mDownPageNumber;
                    mAnnot = redaction;
                    mPDFView.UpdateWithAnnot(redaction, mAnnotPageNum);                    
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }
        }
    }
}
