﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using pdftron.PDF.Tools.Utilities;
using pdftron.PDF;
using pdftron.Common;

/*
 * Ruler Annotation
 * Draw a line with distance information using a TextBlock to give realtime feedback
 */
namespace pdftron.PDF.Tools
{
    public class RulerCreate : SimpleShapeCreate
    {
        // Aux variables
        Path mShape;
        private bool mShapeHasBeenCreated = false;
        string mPrecision;
        string mBaseUnit;
        string mBaseScale;
        string mTranslateUnit;
        string mTranslateScale;
        const double DEFAULT_ZOOM_LEVEL = 0.8;
        const int DEFAULT_FONT_SIZE = 12;
        double angle;

        // Text information to be draw in the middle of the line
        TextBlock mTextInfo;
        double mTextOffsetY = -17; // Offset to position the text above the line draw

        public RulerCreate(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_ruler_create;
            mToolMode = ToolType.e_ruler_create;
        }

        #region Interaction Events

        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            // Load measure settings
            if (!mSimpleShapeCreateSubclassHandlesProperties)
            {
                mPrecision = Settings.GetMeasure(this.ToolMode, MeasureInfoType.Precision, MeasureUtils.DEFAULT_PRECISION);
                mBaseScale = Settings.GetMeasure(this.ToolMode, MeasureInfoType.BaseScale, MeasureUtils.DEFAULT_BASE_SCALE);
                mBaseUnit = Settings.GetMeasure(this.ToolMode, MeasureInfoType.BaseUnit, MeasureUtils.DEFAULT_BASE_UNIT);
                mTranslateScale = Settings.GetMeasure(this.ToolMode, MeasureInfoType.TranslateScale, MeasureUtils.DEFAULT_TRANSLATE_SCALE);
                mTranslateUnit = Settings.GetMeasure(this.ToolMode, MeasureInfoType.TranslateUnit, MeasureUtils.DEFAULT_TRANSLATE_UNIT);                
            }

            // Go ahead and perform all base routines
            base.PointerPressedHandler(sender, e);

            if (mSimpleShapeCanceledAtPress)
                return false;

            mShapeHasBeenCreated = false;
            double halfThickness = mDrawThickness / 2;
            // we need to restrict the border further, since the line can stick out by halfthickness
            if (mPageCropOnClient != null)
            {
                if (m_cnPt1.x < mPageCropOnClient.x1 + halfThickness)
                {
                    m_cnPt1.x = mPageCropOnClient.x1 + halfThickness;
                }
                if (m_cnPt1.x > mPageCropOnClient.x2 - halfThickness)
                {
                    m_cnPt1.x = mPageCropOnClient.x2 - halfThickness;
                }
                if (m_cnPt1.y < mPageCropOnClient.y1 + halfThickness)
                {
                    m_cnPt1.y = mPageCropOnClient.y1 + halfThickness;
                }
                if (m_cnPt1.y > mPageCropOnClient.y2 - halfThickness)
                {
                    m_cnPt1.y = mPageCropOnClient.y2 - halfThickness;
                }
            }

            return false;
        }

        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            // Canvas takes care of mirroring...
            base.PointerMovedHandler(sender, e);
            
            if (mStartDrawing) // Simple Shape is ready to draw
            {
                if (!mShapeHasBeenCreated)
                {
                    mShapeHasBeenCreated = true;
                    
                    // Start drawing the line
                    LineGeometry geom = new LineGeometry();
                    geom.StartPoint = new UIPoint(0, 0);
                    mShape = new Path();
                    mShape.Data = geom;
                    mShape.StrokeThickness = mDrawThickness;
                    mShape.Stroke = mStrokeBrush;
                    this.Children.Add(mShape);

                    var FontSize = DEFAULT_FONT_SIZE / DEFAULT_ZOOM_LEVEL;
                    mTextOffsetY = mTextOffsetY / DEFAULT_ZOOM_LEVEL;
                    mTextOffsetY = mTextOffsetY * mZoomLevel;

                    mTextInfo = new TextBlock();
                    mTextInfo.FontSize = FontSize * mZoomLevel;
                    mTextInfo.Foreground = mStrokeBrush;
                    this.Children.Add(mTextInfo);
                }

                // This ensures that the line stays within page bounds, since the line can stick out by halfthickness
                double halfThickness = mDrawThickness / 2;
                if (mPageCropOnClient != null)
                {
                    if (m_cnPt2.x < mPageCropOnClient.x1 + halfThickness)
                    {
                        m_cnPt2.x = mPageCropOnClient.x1 + halfThickness;
                    }
                    if (m_cnPt2.x > mPageCropOnClient.x2 - halfThickness)
                    {
                        m_cnPt2.x = mPageCropOnClient.x2 - halfThickness;
                    }
                    if (m_cnPt2.y < mPageCropOnClient.y1 + halfThickness)
                    {
                        m_cnPt2.y = mPageCropOnClient.y1 + halfThickness;
                    }
                    if (m_cnPt2.y > mPageCropOnClient.y2 - halfThickness)
                    {
                        m_cnPt2.y = mPageCropOnClient.y2 - halfThickness;
                    }
                }
                this.Width = Math.Abs(m_cnPt2.x - m_cnPt1.x);
                this.Height = Math.Abs(m_cnPt2.y - m_cnPt1.y);

                // Adjust the Line size
                var lineShape = mShape.Data as LineGeometry;
                lineShape.EndPoint = new UIPoint(this.Width, this.Height);

                /* 
                 * Calculate the position to draw the text info using 
                 * Pythagorean theroem: d^2 = h^2 + w^2
                 */
                var line_length = Math.Sqrt(this.Width * this.Width + this.Height * this.Height);                
                var line_length_midPoint = line_length / 2;

                // Proper convert Canvas pts to Page pts
                var pt1 = ConvertFromCanvasPointToPagePoint(m_cnPt1, CurrentAnnotPageNumber);
                var pt2 = ConvertFromCanvasPointToPagePoint(m_cnPt2, CurrentAnnotPageNumber);
                var width = pt2.x - pt1.x;
                var height = pt2.y - pt1.y;

                // Convert Canvas pts to Page pts
                var x_pg_pt = new PDFDouble(width);
                var y_pg_pt = new PDFDouble(height);                
                var line_length_conv = Math.Sqrt(x_pg_pt.Value * x_pg_pt.Value + y_pg_pt.Value * y_pg_pt.Value);

                // Calculate final conversion and add translated unit
                var val = MeasureUtils.ConvertToFinalVal(line_length_conv, mBaseScale, mBaseUnit, mTranslateScale, mTranslateUnit, mPrecision);
                mTextInfo.Text = val;

                // Calculate the angle to properly rotate the text block
                var xDiff = (float)(m_cnPt2.x - m_cnPt1.x);
                var yDiff = (float)(m_cnPt2.y - m_cnPt1.y);
                angle = Math.Atan2(yDiff, xDiff) * 180 / Math.PI;

                // Use Tranform to proper position text on canvas and ensure angle and scale are proper calculated
                TransformGroup tGroup = new TransformGroup();
                ScaleTransform tScale;
                TranslateTransform tTranslate;
                RotateTransform tRotate = new RotateTransform { Angle = angle };

                if ((angle < 0) && (angle > -90))
                {
                    tScale = new ScaleTransform { ScaleY = -1 };
                    tTranslate = new TranslateTransform { X = line_length_midPoint, Y = mTextOffsetY };
                }
                else if (angle < -90)
                {
                    tScale = new ScaleTransform { ScaleY = -1, ScaleX = -1 };
                    tTranslate = new TranslateTransform { X = line_length_midPoint, Y = mTextOffsetY };
                }
                else if (angle > 90)
                {
                    tScale = new ScaleTransform { ScaleX = -1 };
                    tTranslate = new TranslateTransform { X = line_length_midPoint, Y = mTextOffsetY };
                }
                else
                {
                    tScale = new ScaleTransform { ScaleX = 1 };
                    tTranslate = new TranslateTransform { X = line_length_midPoint, Y = mTextOffsetY };
                }

                tGroup.Children.Add(tTranslate);
                tGroup.Children.Add(tRotate);
                tGroup.Children.Add(tScale);

                mTextInfo.RenderTransform = tGroup;
            }

            return false;
        }

        internal override void Finish()
        {
            try
            {
                mPDFView.DocLock(true);

                Rect rect = GetShapeBBox();
                if (rect != null)
                {
                    // Prepare annotation data
                    MeasureInfo measureInfo = new MeasureInfo();
                    measureInfo.Precision = MeasureUtils.ConvStrToPDFPrecision(mPrecision);
                    measureInfo.BaseFactor = MeasureUtils.ConvScaleToPDFFactor(MeasureUtils.GetPageUnitFromString(mBaseUnit), mToolMode, mBaseScale, mTranslateScale);
                    measureInfo.TranslateFactor = double.Parse(mTranslateScale);
                    measureInfo.BaseUnit = mBaseUnit;
                    measureInfo.Unit = mTranslateUnit;
                    measureInfo.Scale = mBaseScale + " " + mBaseUnit + " = " + mTranslateScale + " " + mTranslateUnit;

                    pdftron.PDF.Annots.Line line = pdftron.PDF.Annots.Line.Create(mPDFView.GetDoc().GetSDFDoc(), rect);

                    // The Rect is normalized, so we need to recalculate the points if we don't want it to start from 0,0
                    PDFPoint point1 = ConvertFromCanvasPointToPagePoint(m_cnPt1, mDownPageNumber);
                    PDFPoint point2 = ConvertFromCanvasPointToPagePoint(m_cnPt2, mDownPageNumber);

                    line.SetStartPoint(point1);
                    line.SetEndPoint(point2);
                    line.SetShowCaption(true);
                    line.SetContents(mTextInfo.Text);
                    line.SetCaptionPosition(Annots.LineCapPos.e_Top);

                    MeasureUtils.SetMeasurementInfo(line, measureInfo);

                    SetStyle(line);
                    line.RefreshAppearance();

                    UtilityFunctions.SetUniqueId(line, Guid.NewGuid().ToString());

                    pdftron.PDF.Page page = mPDFView.GetDoc().GetPage(mDownPageNumber);
                    page.AnnotPushBack(line);

                    mAnnotPageNum = mDownPageNumber;
                    mAnnot = line;
                    mPDFView.UpdateWithAnnot(line, mAnnotPageNum);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }
        }

        #endregion
    }
}
