using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.UI;
using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Media.Imaging;
using Windows.UI.Xaml.Shapes;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;
using UIImage = Windows.UI.Xaml.Controls.Image;

using pdftron.PDF;
using PDFDouble = pdftron.Common.DoubleRef;
using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using pdftron.PDF.Annots;
using System.IO;
using Windows.Storage;
using pdftron.SDF;
using pdftron.PDF.Tools.Utilities;

namespace pdftron.PDF.Tools
{
    class StickyNoteCreate : SimpleShapeCreate
    {
        protected FixedSizedBoxPopup mBoxPopup; // Holds fixed size rectangular elements

        protected bool mCancelledByTap = false;

        protected FrameworkElement mStickyIcon;

        // Some magic number that transform ui widgets space to PDF space
        protected const double mStickyIconWidthInPageSpace = 21;
        protected const double mStickyIconHeightInPageSpace = 21;
        protected const double mStickyIconOffsetFactor = 1.1;
        protected const double mStickyIconFontSize = mStickyIconHeightInPageSpace * 1.35;

        // The following numbers would work at 100% zoom
        protected double mStickyIconWidthInScreenSpace = mStickyIconWidthInPageSpace;
        protected double mStickyIconHeightInScreenSpace = mStickyIconHeightInPageSpace;
        protected double mStickyIconVerticalOffset = mStickyIconWidthInPageSpace * mStickyIconOffsetFactor;
        protected double mScaledStickyIconFontSize = mStickyIconFontSize;

        protected UIPoint mStickyIconOffset;

        public StickyNoteCreate(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_sticky_note_create;
            mToolMode = ToolType.e_sticky_note_create;
        }

        internal override void OnClose()
        {
            if (!mDelayedRemovalSet && mViewerCanvas != null && mViewerCanvas.Children.Contains(this)) 
            {
                mViewerCanvas.Children.Remove(this);
            }
            if (mBoxPopup != null)
            {
                mBoxPopup.Hide();
                mBoxPopup = null;
            }
            EnableScrolling();
            base.OnClose();
        }

        internal override bool OnSize()
        {
            EndCurrentTool(ToolType.e_pan);
            return true;
        }

        /// <summary>
        /// This will make the popup menu pop up at the targeted point
        /// </summary>
        /// <param name="point">The point in screen space.</param>
        public void SetTargetPoint(UIPoint point)
        {
            mDownPageNumber = mPDFView.GetPageNumberFromScreenPoint(point.X, point.Y);

            PDFDouble x = new PDFDouble(point.X);
            PDFDouble y = new PDFDouble(point.Y);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(x, y);
            m_cnPt1 = new PDFPoint(x.Value, y.Value);

            mPageCropOnClient = BuildPageBoundBoxOnClient(mDownPageNumber);

            mStickyIcon = GetStickyNoteIcon(Settings.GetIcon(ToolType.e_sticky_note_create));
            this.Children.Add(mStickyIcon);

            PositionStickyIcon();

            //Finish();
            CreateTypingArea();
        }

        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            AddPointer(e.Pointer);
            if (mToolManager.ContactPoints.Count > 1)
            {
                mSimpleShapeCanceledAtPress = true;
                mCancelledDrawing = true;
                mPointerID = -1;
                if (mViewerCanvas != null && mViewerCanvas.Children.Contains(this))
                {
                    mViewerCanvas.Children.Remove(this);
                }
                if (mStickyIcon != null && this.Children.Contains(mStickyIcon))
                {
                    this.Children.Remove(mStickyIcon);
                }
                mToolManager.EnableScrollByManipulation = true;
                return false; // ignore additional pointer presses
            }
            if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Mouse && !e.GetCurrentPoint(null).Properties.IsLeftButtonPressed)
            {
                mSimpleShapeCanceledAtPress = true;
                return false; // if mouse, only allow left button
            }

            PDFDoc doc = mPDFView.GetDoc();
            if (doc == null)
            {
                mSimpleShapeCanceledAtPress = true;
                return false;
            }

            mViewerCanvas = mPDFView.GetAnnotationCanvas();
            UIPoint screenPoint = e.GetCurrentPoint(mPDFView).Position;

            DisableScrolling();

            // Ensure we're on a page
            mDownPageNumber = mPDFView.GetPageNumberFromScreenPoint(screenPoint.X, screenPoint.Y);
            if (mDownPageNumber < 1)
            {
                mSimpleShapeCanceledAtPress = true;
                mPointerID = -1;
                mToolManager.EnableOneFingerScroll = true;
                mToolManager.EnableScrollByManipulation = true;
                return false;
            }

            UIPoint canvasPoint = e.GetCurrentPoint(mViewerCanvas).Position;
            mPointerID = (int)e.Pointer.PointerId;
            mViewerCanvas.CapturePointer(e.Pointer);

            // set up tool's canvas
            mPageCropOnClient = BuildPageBoundBoxOnClient(mDownPageNumber);
            this.Width = mPageCropOnClient.x2 - mPageCropOnClient.x1;
            this.Height = mPageCropOnClient.y2 - mPageCropOnClient.y1;

            mViewerCanvas.Children.Add(this);

            m_cnPt1 = new PDFPoint(canvasPoint.X, canvasPoint.Y);

            double zoom = mPDFView.GetZoom();
            mStickyIconWidthInScreenSpace = mStickyIconWidthInPageSpace * zoom;
            mStickyIconHeightInScreenSpace = mStickyIconHeightInPageSpace * zoom;
            mScaledStickyIconFontSize = mStickyIconFontSize;
            mStickyIconVerticalOffset = mStickyIconHeightInPageSpace * mStickyIconOffsetFactor;
            if (zoom < 1.0)
            {
                mStickyIconVerticalOffset *= zoom;
                mScaledStickyIconFontSize *= zoom;
            }
            mStickyIcon = GetStickyNoteIcon(Settings.GetIcon(ToolType.e_sticky_note_create));

            PositionStickyIcon();

            return false;
        }

        internal override bool PointerMovedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerId != mPointerID)
            {
                return false;
            }
            UIPoint canvasPoint = e.GetCurrentPoint(mViewerCanvas).Position;
            m_cnPt1.x = canvasPoint.X;
            m_cnPt1.y = canvasPoint.Y;
            PositionStickyIcon();

            return false;
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            // Remove contact from dictionary.
            RemovePointer(e.Pointer);

            if (mPointerID < 0 || e.Pointer.PointerId != mPointerID)
            {
                EnableScrolling();
                return false; // ignore any secondary pointer
            }

            mCancelledByTap = false;

            AddNotePendingTapped();

            if (!this.Children.Contains(mStickyIcon))
            {
                UIPoint canvasPoint = e.GetCurrentPoint(mViewerCanvas).Position;
                m_cnPt1.x = canvasPoint.X;
                m_cnPt1.y = canvasPoint.Y;
                this.Children.Add(mStickyIcon);
                PositionStickyIcon();
            }

            mPointerID = -1;
            mViewerCanvas.ReleasePointerCaptures();
            return false;
        }

        private async void AddNotePendingTapped()
        {
            await Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.Normal, () =>
            {
                if (!mCancelledByTap)
                {
                    CreateTypingArea();
                }
            });
        }

        internal override bool ManipulationStartedEventHandler(object sender, ManipulationStartedRoutedEventArgs e)
        {
            if (mToolManager.ContactPoints.Count == 1 && mDownPageNumber > 0)
            {
                this.Children.Add(mStickyIcon);
                PositionStickyIcon();
            }
            return base.ManipulationStartedEventHandler(sender, e);
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            Pan pan = new Pan(mPDFView, mToolManager);
            pan.mIsInHelperMode = true;
            pan.TappedHandler(sender, e);
            if (pan.NextToolMode == ToolType.e_annot_edit && pan.CurrentAnnot != null && pan.CurrentAnnot.GetAnnotType() == AnnotType.e_Text)
            {
                mToolManager.ToolModeAfterEditing = mToolMode;
                mNextToolMode = pan.NextToolMode;
                mAnnot = pan.CurrentAnnot;
                mAnnotPageNum = pan.CurrentAnnotPageNumber;
                mAnnotBBox = pan.CurrentAnnotBBox;
                mCancelledByTap = true;
            }

            return true;
        }

        internal override bool KeyDownAction(object sender, KeyRoutedEventArgs e)
        {
            if (mPointerID >= 0) // if we're currently moving the widget, don't allow key presses
            {
                return false;
            }
            if (e.Key == Windows.System.VirtualKey.Escape)
            {
                if (mBoxPopup != null)
                {
                    mBoxPopup.Hide();
                    mBoxPopup = null;
                    mPDFView.Focus(Windows.UI.Xaml.FocusState.Pointer);
                    EndCurrentTool(ToolType.e_pan);
                    return false;
                }
            }
            if (mBoxPopup != null) // take no action
            {
                return false;
            }
            return base.KeyDownAction(sender, e);
        }


        private string _TextToAdd = string.Empty;
        // This is expected by SimpleShapeCreate, but it's not needed by StickyNoteCreate
        internal async override void Finish()
        {
            pdftron.PDF.Annots.IMarkup markup = null;
            try
            {
                mPDFView.DocLock(true);

                PDFRect rect = new PDFRect(m_cnPt1.x, m_cnPt1.y - mStickyIconVerticalOffset, 
                    m_cnPt1.x + mStickyIconWidthInScreenSpace, m_cnPt1.y - mStickyIconVerticalOffset + mStickyIconHeightInScreenSpace);

                rect = ConvertFromCanvasRectToPageRect(rect, mDownPageNumber);
                rect.Normalize();

                double pageSpaceWidth = rect.Width();
                double pageSpaceHeight = rect.Height();

                //set the size of the rect in page space, and take rotation into account
                PageRotate pr = mPDFView.GetDoc().GetPage(mDownPageNumber).GetRotation();
                switch (pr)
                {
                    case PageRotate.e_90:
                        rect.Set(rect.x1, rect.y1, rect.x2, rect.y2 - pageSpaceHeight - pageSpaceHeight);
                        break;
                    case PageRotate.e_180:
                        rect.Set(rect.x2, rect.y1, rect.x1 + pageSpaceWidth + pageSpaceWidth, rect.y2 - pageSpaceHeight - pageSpaceHeight);
                        break;
                    case PageRotate.e_270:
                        rect.Set(rect.x1 + pageSpaceWidth + pageSpaceWidth, rect.y1, rect.x2, rect.y2);
                        break;
                    default:
                        break;
                }
                rect.Normalize();

                pdftron.PDF.Annots.Text text = pdftron.PDF.Annots.Text.Create(mPDFView.GetDoc().GetSDFDoc(), rect);
                text.SetIcon(Settings.GetIcon(ToolType.e_sticky_note_create));

                SettingsColor settingsColor = Settings.GetShapeColor(ToolType.e_sticky_note_create, false);
                double red = settingsColor.R / 255.0;
                double green = settingsColor.G / 255.0;
                double blue = settingsColor.B / 255.0;
                ColorPt color = new ColorPt(red, green, blue);
                text.SetColor(color); 

                Annots.Popup pop = UtilityFunctions.AddPopupToAnnot(mPDFView.GetDoc(), text, _TextToAdd);

                SetAuthor(text);

                UtilityFunctions.SetUniqueId(text, Guid.NewGuid().ToString());
                UtilityFunctions.SetUniqueId(pop, Guid.NewGuid().ToString());

                text.RefreshAppearance();
                mPage = mPDFView.GetDoc().GetPage(mDownPageNumber);
                mPage.AnnotPushBack(text);
                mPage.AnnotPushBack(pop);

                // required to make it appear upright in rotated documents
                text.RefreshAppearance();

                mAnnot = text;

                // required to make it appear upright in rotated documents
                bool result = await RefreshCustomStickyNoteAppearance(mAnnot, mPDFView);
                if (!result)
                    text.RefreshAppearance();

                mPDFView.UpdateWithAnnot(text, mDownPageNumber);

                markup = text;

                // Make sure to fire the Add event after annot has been added and PDFViewCtrl updated
                mToolManager.RaiseAnnotationAddedEvent(mAnnot, mDownPageNumber);


            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlock();
            }

            // back to pan
            if (mReturnToPanModeWhenFinished)
            {
                mToolManager.CreateTool(ToolType.e_pan, this);
            }
            else
            {
                mToolManager.CreateTool(ToolType.e_sticky_note_create, this, true);
            }
        }

        private void CreateTypingArea()
        {
            PDFPoint screenPoint = ConvertFromCanvasPointToScreenPoint(m_cnPt1);
            NoteDialogBase noteDialog = null;
            if (mToolManager.IsOnPhone)
            {
                noteDialog = new NoteDialogPhone(null, false, true, mPDFView, mToolManager);
            }
            else
            {
               noteDialog = new NoteDialogWindows(null, false, true, mPDFView, mToolManager, new UIPoint(screenPoint.x + mStickyIconFontSize / 2, screenPoint.y - mStickyIconFontSize / 2));
                
            }
            noteDialog.NoteClosed += NoteDialog_NoteClosed;
        }

        private void NoteDialog_NoteClosed(NoteDialogBase.NotedialogResults result, string newText)
        {
            if (result == NoteDialogBase.NotedialogResults.Changed)
            {
                _TextToAdd = newText;
                Finish();
                mToolManager.mDelayRemoveTimers.Add(new DelayRemoveTimer(mViewerCanvas, this, this, mPDFView, mToolManager));
                mDelayedRemovalSet = true;
            }

            // back to pan
            if (mReturnToPanModeWhenFinished)
            {
                mToolManager.CreateTool(ToolType.e_pan, this);
            }
            else
            {
                mToolManager.CreateTool(ToolType.e_sticky_note_create, this, true);
            }
        }
     
        private UIRect GetElementRect(Windows.UI.Xaml.FrameworkElement element)
        {
            GeneralTransform elementtransform = element.TransformToVisual(null);
            UIRect rect = elementtransform.TransformBounds(new UIRect(new UIPoint(0, 0), new Windows.Foundation.Size(element.ActualWidth, element.ActualHeight)));
            return rect;
        }

        public static async Task<bool> RefreshCustomStickyNoteAppearance(IAnnot annot, PDFViewCtrl pdfView)
        {
            try
            {
                Text text = new Text(annot.GetSDFObj());
                string iconName = text.GetIconName();

                StorageFile iconsFile = await StorageFile.GetFileFromApplicationUriAsync(new Uri("ms-appx:///pdftron.PDF.Tools/Icons/StickyIcons.pdf"));
                PDFDoc doc = new PDFDoc(iconsFile);

                Page iconPage = null;
                int pageCount = doc.GetPageCount();
                for(int pageNum = 1; pageNum <= pageCount; pageNum++)
                {
                    string label = doc.GetPageLabel(pageNum).GetPrefix();
                    if (iconName.Equals(doc.GetPageLabel(pageNum).GetPrefix(), StringComparison.OrdinalIgnoreCase))
                    {
                        iconPage = doc.GetPage(pageNum);
                        Obj contents = iconPage.GetContents();
                        Obj importedContents = annot.GetSDFObj().GetDoc().ImportObj(contents, true);
                        Rect bbox = iconPage.GetMediaBox();
                        importedContents.PutRect("BBox", bbox.x1, bbox.y1, bbox.x2, bbox.y2);
                        importedContents.PutName("Subtype", "Form");
                        importedContents.PutName("Type", "XObject");

                        ElementReader reader = new ElementReader();
                        ElementWriter writer = new ElementWriter();
                        reader.Begin(importedContents);
                        writer.Begin(importedContents, true);
                        ColorPt rgb_color = text.GetColorAsRGB();

                        for (Element element = reader.Next(); element != null; element = reader.Next())
                        {
                            if (element.GetType() == ElementType.e_path && !element.IsClippingPath())
                            {
                                element.GetGState().SetFillColorSpace(ColorSpace.CreateDeviceRGB());
                                element.GetGState().SetFillColor(rgb_color);
                                element.SetPathStroke(true);
                                element.SetPathFill(true);
                            }
                            writer.WriteElement(element);
                        }
                        reader.End();
                        writer.End();
                        reader.Dispose();
                        writer.Dispose();

                        text.SetAppearance(importedContents);

                        return true;
                    }

                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return false;
        }

        private FrameworkElement GetStickyNoteIcon(string iconName)
        {
            double zoom = mPDFView.GetZoom();


            Grid grid = new Grid();
            //grid.Background = new SolidColorBrush(Color.FromArgb(55, 150, 0, 150));

            double resolutionScaleFactor = 1.0 / Windows.Graphics.Display.DisplayInformation.GetForCurrentView().RawPixelsPerViewPixel;

            TextBlock outline = new TextBlock();
            outline.Foreground = new SolidColorBrush(Colors.Black);
            outline.FontFamily = new FontFamily("ms-appx:///pdftron.PDF.Tools/Controls/Resources/custom.ttf#PDFTron Glyphicons");
            outline.FontSize = mScaledStickyIconFontSize * resolutionScaleFactor * UtilityFunctions.IconNameToFontSizeMultiplier(iconName);
            outline.Text = UtilityFunctions.IconNameToFontIconOutline(iconName);
            outline.HorizontalAlignment = HorizontalAlignment.Center;

            TextBlock fill = new TextBlock();
            SettingsColor color = Settings.GetShapeColor(mToolMode, false);
            byte r = color.R;
            byte g = color.G;
            byte b = color.B;
            mUseStroke = true;
            fill.Foreground = new SolidColorBrush(Color.FromArgb(255, r, g, b));
            fill.FontFamily = new FontFamily("ms-appx:///pdftron.PDF.Tools/Controls/Resources/custom.ttf#PDFTron Glyphicons");
            fill.FontSize = mScaledStickyIconFontSize * resolutionScaleFactor * UtilityFunctions.IconNameToFontSizeMultiplier(iconName);
            fill.Text = UtilityFunctions.IconNameToFontIconFill(iconName);
            fill.HorizontalAlignment = HorizontalAlignment.Center;

            TextBlock filler = new TextBlock();
            filler.Opacity = 0;
            filler.FontFamily = new FontFamily("ms-appx:///pdftron.PDF.Tools/Controls/Resources/custom.ttf#PDFTron Glyphicons");
            filler.FontSize = mScaledStickyIconFontSize * resolutionScaleFactor * UtilityFunctions.IconNameToFontSizeMultiplier("Comment"); ;
            filler.Text = UtilityFunctions.IconNameToFontIconFill("Comment");
            filler.HorizontalAlignment = HorizontalAlignment.Center;

            grid.Children.Add(filler);
            grid.Children.Add(fill);
            grid.Children.Add(outline);
            grid.Margin = new Thickness(0, 3, 0, 0);
            grid.SizeChanged += Grid_SizeChanged;

            return grid;
        }

        private void Grid_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (e.NewSize.Width > 0 && e.NewSize.Height > 0)
            {
                mStickyIconOffset.X = 0;
                mStickyIconOffset.Y = - (e.NewSize.Height + mScaledStickyIconFontSize) / 2;
                PositionStickyIcon();
            }
        }

        private void SetStickyIconBounds()
        {
            if (mPageCropOnClient != null)
            {
                if (m_cnPt1.x < mPageCropOnClient.x1)
                {
                    m_cnPt1.x = mPageCropOnClient.x1;
                }
                if (m_cnPt1.x > mPageCropOnClient.x2 - mScaledStickyIconFontSize)
                {
                    m_cnPt1.x = mPageCropOnClient.x2 - mScaledStickyIconFontSize;
                }
                if (m_cnPt1.y < mPageCropOnClient.y1 + mScaledStickyIconFontSize)
                {
                    m_cnPt1.y = mPageCropOnClient.y1 + mScaledStickyIconFontSize;
                }
                if (m_cnPt1.y > mPageCropOnClient.y2)
                {
                    m_cnPt1.y = mPageCropOnClient.y2;
                }
            }
        }

        private void PositionStickyIcon()
        {
            SetStickyIconBounds();
            mStickyIcon.SetValue(Canvas.LeftProperty, m_cnPt1.x + mStickyIconOffset.X);
            mStickyIcon.SetValue(Canvas.TopProperty, m_cnPt1.y + mStickyIconOffset.Y);
        }
    }
}
