﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.ApplicationModel.DataTransfer;

using PointerDeviceType = Windows.Devices.Input.PointerDeviceType;

using Windows.UI;
using Windows.UI.Input;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Animation;
using Windows.UI.Xaml.Shapes;

using pdftron.PDF;
using pdftron.Common;
using pdftron.PDF.Annots;

using UIPoint = Windows.Foundation.Point;
using UIRect = Windows.Foundation.Rect;

using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFPage = pdftron.PDF.Page;
using PDFDouble = pdftron.Common.DoubleRef;
using pdftron.PDF.Tools.Controls;
using static pdftron.PDF.Tools.UtilityFunctions;


namespace pdftron.PDF.Tools
{
    class TextMarkupEdit : TextSelect
    {
        protected ITextMarkup mSelectedTextMarkup;
        protected AnnotEdit mAnnotEdit;
        protected string mOriginalHighlightedText = "";

        public TextMarkupEdit(PDFViewCtrl ctrl, ToolManager tMan)
            : base(ctrl, tMan)
        {
            mNextToolMode = ToolType.e_annot_edit_text_markup;
            mToolMode = ToolType.e_annot_edit_text_markup;

            mSurpressTextSelectPopupMenu = true;
        }

        internal override void OnClose()
        {
            base.OnClose();
            mAnnotEdit.AnnotPropertyChangedInHelperMode -= mAnnotEdit_AnnotPropertyChangedInHelperMode;
            mAnnotEdit.HelperModeHideMenu();
            mAnnotEdit.OnClose();
        }

        internal override void OnCreate()
        {
            mSelectedTextMarkup = (ITextMarkup)mAnnot;
            base.OnCreate();

            PageRotate rotation = PageRotate.e_0;
            int qpCount = 0;
            QuadPoint pg_qp1 = null;
            QuadPoint pg_qp2 = null;

            try
            {
                mPDFView.DocLockRead();
                rotation = mPDFView.GetDoc().GetPage(mAnnotPageNum).GetRotation();

                qpCount = mSelectedTextMarkup.GetQuadPointCount();
                pg_qp1 = mSelectedTextMarkup.GetQuadPoint(0);
                pg_qp2 = mSelectedTextMarkup.GetQuadPoint(qpCount - 1);
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                mPDFView.DocUnlockRead();
            }
            
            bool rightToLeft = mPDFView.GetRightToLeftLanguage();

            DoubleRef screen1X = new PDFDouble(pg_qp1.p1.x);
            DoubleRef screen1Y = new PDFDouble(pg_qp1.p1.y);
            DoubleRef screen2X = new PDFDouble(pg_qp1.p2.x);
            DoubleRef screen2Y = new PDFDouble(pg_qp1.p2.y);
            DoubleRef screen3X = new PDFDouble(pg_qp1.p3.x);
            DoubleRef screen3Y = new PDFDouble(pg_qp1.p3.y);
            DoubleRef screen4X = new PDFDouble(pg_qp1.p4.x);
            DoubleRef screen4Y = new PDFDouble(pg_qp1.p4.y);

            mPDFView.ConvPagePtToScreenPt(screen1X, screen1Y, mAnnotPageNum);
            mPDFView.ConvPagePtToScreenPt(screen2X, screen2Y, mAnnotPageNum);
            mPDFView.ConvPagePtToScreenPt(screen3X, screen3Y, mAnnotPageNum);
            mPDFView.ConvPagePtToScreenPt(screen4X, screen4Y, mAnnotPageNum);

            mPDFView.ConvScreenPtToAnnotationCanvasPt(screen1X, screen1Y);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(screen2X, screen2Y);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(screen3X, screen3Y);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(screen4X, screen4Y);

            double minX = Math.Min(Math.Min(screen1X.Value, screen2X.Value), Math.Min(screen3X.Value, screen4X.Value));
            double maxX = Math.Max(Math.Max(screen1X.Value, screen2X.Value), Math.Max(screen3X.Value, screen4X.Value));
            double minY = Math.Min(Math.Min(screen1Y.Value, screen2Y.Value), Math.Min(screen3Y.Value, screen4Y.Value));
            double maxY = Math.Max(Math.Max(screen1Y.Value, screen2Y.Value), Math.Max(screen3Y.Value, screen4Y.Value));
            double newX1 = minX;
            double newY1 = (0.4 * minY) + (0.6 * maxY);
            if (rightToLeft)
            {
                newX1 = maxX;
            }

            DoubleRef x1 = new PDFDouble(newX1);
            DoubleRef y1 = new PDFDouble(newY1);

            screen1X = new PDFDouble(pg_qp2.p1.x);
            screen1Y = new PDFDouble(pg_qp2.p1.y);
            screen2X = new PDFDouble(pg_qp2.p2.x);
            screen2Y = new PDFDouble(pg_qp2.p2.y);
            screen3X = new PDFDouble(pg_qp2.p3.x);
            screen3Y = new PDFDouble(pg_qp2.p3.y);
            screen4X = new PDFDouble(pg_qp2.p4.x);
            screen4Y = new PDFDouble(pg_qp2.p4.y);

            mPDFView.ConvPagePtToScreenPt(screen1X, screen1Y, mAnnotPageNum);
            mPDFView.ConvPagePtToScreenPt(screen2X, screen2Y, mAnnotPageNum);
            mPDFView.ConvPagePtToScreenPt(screen3X, screen3Y, mAnnotPageNum);
            mPDFView.ConvPagePtToScreenPt(screen4X, screen4Y, mAnnotPageNum);

            mPDFView.ConvScreenPtToAnnotationCanvasPt(screen1X, screen1Y);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(screen2X, screen2Y);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(screen3X, screen3Y);
            mPDFView.ConvScreenPtToAnnotationCanvasPt(screen4X, screen4Y);

            minX = Math.Min(Math.Min(screen1X.Value, screen2X.Value), Math.Min(screen3X.Value, screen4X.Value));
            maxX = Math.Max(Math.Max(screen1X.Value, screen2X.Value), Math.Max(screen3X.Value, screen4X.Value));
            minY = Math.Min(Math.Min(screen1Y.Value, screen2Y.Value), Math.Min(screen3Y.Value, screen4Y.Value));
            maxY = Math.Max(Math.Max(screen1Y.Value, screen2Y.Value), Math.Max(screen3Y.Value, screen4Y.Value));

            double newX2 = maxX;
            double newY2 = (0.6 * minY + 0.4 * maxY);
            if (rightToLeft)
            {
                newX2 = minX;
            }

            DoubleRef x2 = new PDFDouble(newX2);
            DoubleRef y2 = new PDFDouble(newY2);

            SetTextSelectionColor();

            mPDFView.SetTextSelectionMode(TextSelectionMode.e_structural);
            SelectAndDraw(new UIPoint(x1.Value, y1.Value), new UIPoint(x2.Value, y2.Value), false);
            PositionSelectionWidgets();

            mPageBoundingBoxRect = BuildPageBoundBoxOnClient(mAnnotPageNum);
            mLimitToPage = mAnnotPageNum;

            mForceControlPoints = true;

            mAnnotEdit = new AnnotEdit(mPDFView, mToolManager);
            mAnnotEdit.SetAnnotEditAsHelperTool(mAnnot, mAnnotPageNum);
            mAnnotEdit.AnnotPropertyChangedInHelperMode += mAnnotEdit_AnnotPropertyChangedInHelperMode;
            mToolManager.RaiseSelectionChangedEvent(mAnnot, mAnnotPageNum);
        }

        void mAnnotEdit_AnnotPropertyChangedInHelperMode()
        {
            SetTextSelectionColor();
            DetachAllTextSelection();
            mPagesOnScreen.Clear();
            mSelectionCanvases.Clear();
            DrawSelection();
        }

        internal override bool TappedHandler(object sender, TappedRoutedEventArgs e)
        {
            if (mJustSwitchedFromAnotherTool)
            {
                mJustSwitchedFromAnotherTool = false;

                ShowCommandMenu();
                return false;
            }

            if (IsPointInSelection(e.GetPosition(mPDFView), true))
            {
                return true;
            }
            else
            {
                mNextToolMode = ToolType.e_pan;
            }


            return false;
        }

        internal override bool RightTappedHandler(object sender, RightTappedRoutedEventArgs e)
        {
            e.Handled = true;
            return false;
        }

        internal override bool DoubleClickedHandler(object sender, PointerRoutedEventArgs e)
        {
            e.Handled = true;
            return false;
        }

        internal override bool PointerPressedHandler(object sender, PointerRoutedEventArgs e)
        {
            mAnnotEdit.HelperModeHideMenu();
            if (mJustSwitchedFromAnotherTool)
            {
                if (e.Pointer.PointerDeviceType == Windows.Devices.Input.PointerDeviceType.Pen)
                {
                    PointerPoint cn_downPoint = e.GetCurrentPoint(mViewerCanvas);
                    if (cn_downPoint.Properties.IsBarrelButtonPressed)
                    {
                        mPointerID = (int)e.Pointer.PointerId;
                        return false;
                    }
                }   
            }
            bool returnVal = base.PointerPressedHandler(sender, e);
            if (mToolManager.CopyAnnotatedTextToNote)
            {
                mOriginalHighlightedText = UtilityFunctions.GetSelectedTextOnPage(mPDFView, mAnnotPageNum);
            }
            return returnVal;
        }

        internal override bool PointerReleasedHandler(object sender, PointerRoutedEventArgs e)
        {
            if (mPointerID < 0 || e.Pointer.PointerId != mPointerID)
            {
                // do nothing
                return false;
            }

            int ctrlPoint = mCurrentWidget;

            base.PointerReleasedHandler(sender, e);

            if (ctrlPoint >= 0)
            {
                if (!mPDFView.HasSelectionOnPage(mAnnotPageNum))
                {
                    return true;
                }

                PDFRect sc_OldRect = null;

                try
                {
                    mPDFView.DocLock(true);
                    double[] quads = mPDFView.GetSelection(mAnnotPageNum).GetQuads();
                    if (quads == null)
                    {
                        return true;
                    }
                    int sz = quads.Length / 8;
                    if (sz == 0)
                    {
                        return true;
                    }

                    PDFRect pg_oldRect = mSelectedTextMarkup.GetRect();
                    sc_OldRect = ConvertFromPageRectToScreenRect(pg_oldRect, mAnnotPageNum);

                    // delete the old quad points and rect
                    mSelectedTextMarkup.GetSDFObj().Erase("QuadPoints");
                    mSelectedTextMarkup.GetSDFObj().Erase("Rect");

                    mSelectedTextMarkup.SetRect(new PDFRect(quads[0], quads[1], quads[4], quads[5]));

                    SetQuadPoints(mSelectedTextMarkup, quads);
                    mSelectedTextMarkup.RefreshAppearance();
                    
                    if (mToolManager.CopyAnnotatedTextToNote)
                    {
                        string currentNote = mSelectedTextMarkup.GetContents();
                        if (!string.IsNullOrEmpty(currentNote) && currentNote.Equals(mOriginalHighlightedText))
                        {
                            Popup p = mSelectedTextMarkup.GetPopup();
                            if (p == null || !p.IsValid())
                            {
                                p = Popup.Create(mPDFView.GetDoc().GetSDFDoc(), mSelectedTextMarkup.GetRect());
                                p.SetParent(mSelectedTextMarkup);
                                mSelectedTextMarkup.SetPopup(p);
                            }
                            p.SetContents(UtilityFunctions.GetSelectedTextOnPage(mPDFView, mAnnotPageNum));
                        }
                    }

                    mSelectedTextMarkup.SetDateToNow();

                    mPDFView.UpdateWithAnnot(mSelectedTextMarkup, mAnnotPageNum);
                    mToolManager.RaiseAnnotationEditedEvent(mSelectedTextMarkup, mAnnotPageNum);

                    if (sc_OldRect != null)
                    {
                        mPDFView.Update(sc_OldRect);
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlock();
                }
            }

            if (ctrlPoint >= 0 || IsPointInSelection(e.GetCurrentPoint(mPDFView).Position, true))
            { 
                ShowCommandMenu();
            }

            return false;
        }

        protected void ShowCommandMenu()
        {
            PDFRect rect = ConvertFromPageRectToScreenRect(mAnnot.GetRect(), mAnnotPageNum);
            rect.Normalize();
            mAnnotEdit.HelperModeShowMenu(rect);
        }
         
        [Obsolete("HideMenu is deprecated, please use HideContextMenu with ContextMenuControl instead.")]
        internal void HideMenu(PopupCommandMenu m)
        {
            if (m != null)
            {
                m.Hide();
            }
        }

        internal void HideContextMenu(ContextMenuControl m)
        {
            if (m != null)
            {
                m.Hide();
            }
        }

        internal override bool CloseOpenDialog()
        {
            return mAnnotEdit.CloseOpenDialog(); 
        }

        protected void SetTextSelectionColor()
        {
            if (mAnnot.GetAnnotType() == AnnotType.e_Highlight)
            {
                mTextSelectionBrush = DEFAULT_TEXT_SELECTION_COLOR;
            }
            else
            {
                ColorPt color = mAnnot.GetColorAsRGB();
                double opacity = mSelectedTextMarkup.GetOpacity();
                Color col = new Color();
                double r = (int)(color.Get(0) * 255 + 0.5);
                double g = (int)(color.Get(1) * 255 + 0.5);
                double b = (int)(color.Get(2) * 255 + 0.5);
                col.R = (byte)r;
                col.G = (byte)g;
                col.B = (byte)b;
                col.A = (byte)(20 + (40 * opacity));
                mTextSelectionBrush = new SolidColorBrush(col);
            }
        }

        internal override bool KeyDownAction(object sender, KeyRoutedEventArgs e)
        {
            if (mPointerID >= 0) // if we're currently moving the widget, don't allow key presses
            {
                return false;
            }

            if (e.Key == Windows.System.VirtualKey.Escape)
            {
                if (mIsShowingCommandMenu)
                {
                    HideContextMenu(mContextMenu);
                }
                else
                {
                    mNextToolMode = ToolType.e_pan;
                }
                return false;
            }

            if (e.Key == Windows.System.VirtualKey.Delete) // delete annotation
            {
                HideContextMenu(mContextMenu);
                try
                {
                    mPDFView.DocLock(true);

                    mToolManager.RaiseAnnotationPreRemovedEvent(mAnnot, mAnnotPageNum);

					// Perform the Annotation removal and update PDFViewer
                    mPage = mPDFView.GetDoc().GetPage(mAnnotPageNum);
                    mAnnot = AnnotUtils.SafeDeleteAnnotAndUpdate(mPDFView, mPage, mAnnot, mAnnotPageNum);

					// make sure to raise removed event after mPDFView.UpdateWithAnnot()
                    mToolManager.RaiseAnnotationRemovedEvent(mAnnot, mAnnotPageNum);                    
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    mPDFView.DocUnlock();
                    mAnnot = null;
                }
                mNextToolMode = ToolType.e_pan;
                return true;
            }

            return base.KeyDownAction(sender, e);
        }
    }
}
