﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Windows.UI.Text;
using Windows.UI.Xaml.Controls;
using Windows.UI;

namespace pdftron.PDF.Tools.Utilities
{
    internal static class FormatRichText
    {
        /// <summary>
        /// Selected text to bold
        /// </summary>
        /// <param name="selectedText">The selected text from textbox</param>
        public static void BoldSelectedText(ITextSelection selectedText)
        {
            if (selectedText != null)
            {
                ITextCharacterFormat charFormatting = selectedText.CharacterFormat;
                charFormatting.Bold = FormatEffect.Toggle;
                selectedText.CharacterFormat = charFormatting;
            }
        }

        /// <summary>
        /// Selected text to italic
        /// </summary>
        /// <param name="selectedText">The selected text from textbox</param>
        public static void ItalicSelectedText(ITextSelection selectedText)
        {
            if (selectedText != null)
            {
                ITextCharacterFormat charFormatting = selectedText.CharacterFormat;
                charFormatting.Italic = FormatEffect.Toggle;
                selectedText.CharacterFormat = charFormatting;
            }
        }

        /// <summary>
        /// Underline selected text
        /// </summary>
        /// <param name="selectedText">The selected text from textbox</param>
        public static void UnderlineSelectedText(ITextSelection selectedText)
        {
            if (selectedText != null)
            {
                ITextCharacterFormat charFormatting = selectedText.CharacterFormat;
                if (selectedText.CharacterFormat.Underline != UnderlineType.Single)
                {
                    charFormatting.Underline = UnderlineType.Single;
                }
                else
                {
                    charFormatting.Underline = UnderlineType.None;
                }

                selectedText.CharacterFormat = charFormatting;
            }
        }

        /// <summary>
        /// Strikethrough selected text
        /// </summary>
        /// <param name="selectedText">The selected text from textbox</param>
        public static void StrikethroughSelectedText(ITextSelection selectedText)
        {
            if (selectedText != null)
            {
                ITextCharacterFormat charFormatting = selectedText.CharacterFormat;
                charFormatting.Strikethrough = FormatEffect.Toggle;
                selectedText.CharacterFormat = charFormatting;
            }
        }

        /// <summary>
        /// Apply color to selected text
        /// </summary>
        /// <param name="selectedText">The selected text from textbox</param>
        /// <param name="color">The color to apply to the text</param>
        public static void ColorSelectedText(ITextSelection selectedText, Color color)
        {
            if (selectedText != null)
            {
                ITextCharacterFormat charFormatting = selectedText.CharacterFormat;
                charFormatting.ForegroundColor = color;
                selectedText.CharacterFormat = charFormatting;
            }
        }

        /// <summary>
        /// Check if RichEditBox is null or if document is null
        /// </summary>
        /// <param name="richEditBox">The rich edit box to check</param>
        /// <returns>True if null</returns>
        public static bool IsEditTextBoxNull(RichEditBox richEditBox)
        {
            if (richEditBox == null || richEditBox.Document == null)
            {
                return true;
            }
            return false;
        }

        /// <summary>
        /// Check if text selection is null or empty
        /// </summary>
        /// <param name="selectedText">The selected text from textbox</param>
        /// <returns>True if null or empty text</returns>
        public static bool IsSelectionNullOrEmpty(ITextSelection selectedText)
        {
            if (selectedText == null || selectedText.Text == null || selectedText.Text.Length == 0)
            {
                return true;
            }
            return false;
        }
    }
}
