﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Foundation;
using Windows.System;
using Windows.UI.Core;
using Windows.UI.Xaml;

namespace pdftron.PDF.Tools.Utilities
{
    public sealed class HotKeyHandler
    {
        private static HotKeyHandler _current;

        public event TypedEventHandler<CoreWindow, KeyEventArgs> KeyPressedEvent;
        public event TypedEventHandler<CoreWindow, KeyEventArgs> HotKeyPressedEvent;
        public event TypedEventHandler<CoreWindow, KeyEventArgs> AltHotKeyPressedEvent;

        public static HotKeyHandler Current => _current ?? (_current = new HotKeyHandler());

        private HotKeyHandler()
        {
            Window.Current.CoreWindow.KeyDown += CoreWindow_KeyDown;
        }

        private void CoreWindow_KeyDown(CoreWindow sender, KeyEventArgs args)
        {
            if (args.KeyStatus.WasKeyDown || args.VirtualKey == VirtualKey.Menu || args.VirtualKey == VirtualKey.Control)
            {
                return;
            }
            CoreVirtualKeyStates ctrlState = CoreWindow.GetForCurrentThread().GetKeyState(VirtualKey.Control);
            if ((ctrlState & CoreVirtualKeyStates.Down) == CoreVirtualKeyStates.Down)
            {
                CoreVirtualKeyStates altState = CoreWindow.GetForCurrentThread().GetKeyState(VirtualKey.Menu);
                if ((altState & CoreVirtualKeyStates.Down) == CoreVirtualKeyStates.Down)
                {
                    AltHotKeyPressedEvent?.Invoke(sender, args);
                }
                else
                {
                    HotKeyPressedEvent?.Invoke(sender, args);
                    CheckforEasterEgg(args.VirtualKey);
                }
            }
            else
            {
                KeyPressedEvent?.Invoke(sender, args);
            }
        }

        public bool IsShiftDown
        {
            get
            {
                CoreVirtualKeyStates shiftState = CoreWindow.GetForCurrentThread().GetKeyState(VirtualKey.Shift);
                return (shiftState & CoreVirtualKeyStates.Down) == CoreVirtualKeyStates.Down;
            }
        }

        private const string EASTER_EGG_STRING = "CRASHMENOW";
        private static int _currentEasterEggPosition = -1;
        private static string _EasterEggCrashString;

        private void CheckforEasterEgg(VirtualKey key)
        {
            int curr_index = _currentEasterEggPosition;
            if (curr_index >= 0 && curr_index < EASTER_EGG_STRING.Length)
            {
                curr_index++;
                string nextPosition = EASTER_EGG_STRING.Substring(curr_index, 1);
                string keyString = key.ToString();
                if (nextPosition.Equals(keyString, StringComparison.OrdinalIgnoreCase))
                {
                    _currentEasterEggPosition = curr_index;
                    if (keyString == "_")
                    {
                        _EasterEggCrashString = "LALA";
                    }
                    if (_currentEasterEggPosition == EASTER_EGG_STRING.Length - 1)
                    {
                        curr_index = _EasterEggCrashString.IndexOf("_");
                    }
                }
                else
                {
                    _currentEasterEggPosition = -1;
                }
            }

            if (key == VirtualKey.C)
            {
                _currentEasterEggPosition = 0;
            }

            Debug.WriteLineIf(_currentEasterEggPosition >= 0 && _currentEasterEggPosition < EASTER_EGG_STRING.Length, _currentEasterEggPosition);
        }
    }
}
