﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Storage.Streams;
using System.Runtime.InteropServices.WindowsRuntime;

using UIPoint = Windows.Foundation.Point;
using PDFPoint = pdftron.PDF.Point;
using PDFRect = pdftron.PDF.Rect;
using PDFDouble = pdftron.Common.DoubleRef;
using UIRect = Windows.Foundation.Rect;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Media;
using Windows.UI;
using pdftron.Common;
using Windows.Data.Json;
using pdftron.SDF;
using Windows.UI.Text;
using Windows.Storage;
using Windows.System.Profile;
using Windows.UI.Xaml.Controls;
using System.IO;
using System.Text.RegularExpressions;
using Windows.UI.ViewManagement;
using System.Diagnostics;
using pdftron.PDF.Tools.Controls.Viewer;
using pdftron.PDF.Tools.Controls.ViewModels.Viewer.Tab;
using Windows.Storage.FileProperties;
using Windows.ApplicationModel.DataTransfer;
using pdftron.PDF.Tools.Utilities;
using Windows.UI.Notifications;
using Windows.UI.Core;
using pdftron.PDF.Tools.Extensions;

namespace pdftron.PDF.Tools
{
    public class UtilityLogger
    {
        public class LogItem
        {
            public string LogText { get; private set; }
            public LogItem(string txt)
            {
                LogText = txt;
            }
        }

        private static UtilityLogger _Current;
        public static UtilityLogger Current
        {
            get
            {
                if (_Current == null)
                {
                    _Current = new UtilityLogger();
                }
                return _Current;
            }
        }

        public UtilityLogger()
        {
            _LogList = new System.Collections.ObjectModel.ObservableCollection<LogItem>();
            _Timings = new Dictionary<string, DateTime>();
            _Current = this;
        }

        private System.Collections.ObjectModel.ObservableCollection<LogItem> _LogList;

        public System.Collections.ObjectModel.ObservableCollection<LogItem> LogList
        {
            get { return _LogList; }
        }

        public void LogString(string log)
        {
            DateTime dt = DateTime.Now;
            string logString = string.Format("{0}:{1}:{2}.{3:000} - {4}", dt.Hour, dt.Minute, dt.Second, dt.Millisecond, log);
            _LogList.Insert(0, new LogItem(logString));
        }

        private Dictionary<string, DateTime> _Timings;
        public bool LogString(string log, string key)
        {
            if (_Timings.ContainsKey(key))
            {
                DateTime dt = DateTime.Now;
                TimeSpan timeTaken = dt - _Timings[key];
                _Timings.Remove(key);

                string logString = string.Format("{0}:{1}:{2}.{3:000} ({5}ms)- {4}", dt.Hour, dt.Minute, dt.Second, dt.Millisecond,
                    log, (int)timeTaken.TotalMilliseconds);
                _LogList.Insert(0, new LogItem(logString));

                if (timeTaken.TotalMilliseconds > 100)
                {
                    return true;
                }
            }
            else
            {
                _Timings[key] = DateTime.Now;
                LogString(log);
            }

            return false;
        }

        public FrameworkElement GetScrollableLogStack()
        {
            Windows.UI.Xaml.Controls.ScrollViewer sv = new Windows.UI.Xaml.Controls.ScrollViewer();
            sv.ZoomMode = Windows.UI.Xaml.Controls.ZoomMode.Disabled;
            sv.HorizontalScrollMode = Windows.UI.Xaml.Controls.ScrollMode.Disabled;
            sv.VerticalScrollMode = Windows.UI.Xaml.Controls.ScrollMode.Auto;

            Windows.UI.Xaml.Controls.StackPanel sp = new Windows.UI.Xaml.Controls.StackPanel();
            sp.Orientation = Windows.UI.Xaml.Controls.Orientation.Vertical;
            sp.Width = 500;
            sv.Content = sp;
            sp.Background = new SolidColorBrush(Colors.Black);

            foreach (LogItem item in LogList)
            {
                Windows.UI.Xaml.Controls.TextBox tb = new Windows.UI.Xaml.Controls.TextBox();
                tb.IsReadOnly = true;
                tb.AcceptsReturn = true;
                tb.HorizontalAlignment = HorizontalAlignment.Stretch;
                tb.Text = item.LogText;
                tb.Foreground = new SolidColorBrush(Colors.White);

                sp.Children.Add(tb);
            }

            return sv;
        }

        public void ShowPopup()
        {
            Windows.UI.Xaml.Controls.Primitives.Popup pup = new Windows.UI.Xaml.Controls.Primitives.Popup();
            pup.IsLightDismissEnabled = true;

            pup.Child = GetScrollableLogStack();
            (pup.Child as FrameworkElement).Height = Window.Current.Bounds.Height - 80;
            pup.Height = Window.Current.Bounds.Height - 80;
            pup.HorizontalOffset = 200;
            pup.VerticalOffset = 80;
            pup.IsOpen = true;
        }
    }


    public static class UtilityFunctions
    {
        public static string KeyIRT = "IRT";
        public static string KeyRT = "RT";
        public static string KeyNM = "NM";
        public static string ValueGroup = "Group";
        public static string PDFTronId = "pdftron";
        /// <summary>
        /// The key value for the dictionary that stores HTML for supporting Rich Text Content for the Free Text Annotation
        /// Note: It is not complete HTML, only a subset of it and CSS
        /// </summary>
        public static string KeyRichContent = "RC";

        public static uint PCM_AUDIO_SAMPLE_RATE = 22050;
        public static uint PCM_AUDIO_BITS_PER_SAMPLE = 16;
        public static uint PCM_AUDIO_NUM_CHANNELS = 1;

        static uint MAX_TIME_TOAST_NOTIFICATION = 10;
        static uint MIN_TIME_TOAST_NOTIFICATION = 4;

        private static Regex _SkyDriveRegex = new Regex(@"\\SkyDrive\\");
        private static Regex _OneDriveRegex = new Regex(@"\\OneDrive\\");
        private static Regex _BoxRegex = new Regex(@"\\Packages\\[A-Za-z0-9_-]*\.Box[A-Za-z0-9_-]*\\LocalState\\");
        private static Regex _DropBoxRegex = new Regex(@"\\Packages\\[A-Za-z0-9_-]*\.Dropbox[A-Za-z0-9_-]*\\");

        private static Regex _SkyDriveTempFolder = new Regex(@"\\Packages\\.*skydrive.*\\LocalState\\OpenWithTempFolder");
        private static Regex _OneDriveTempFolder = new Regex(@"\\Packages\\.*onedrive.*\\LocalState\\OpenWithTempFolder");

        public static double PointToLineDistance(PDFPoint basePoint, PDFPoint tipPoint, PDFPoint target)
        {
            double dx;
            double dy;

            if (basePoint == null || tipPoint == null || target == null)
                return 0;

            double lineXDist = tipPoint.x - basePoint.x;
            double lineYDist = tipPoint.y - basePoint.y;

            double squaredDist = (lineXDist * lineXDist) + (lineYDist * lineYDist);

            if (double.IsNaN(squaredDist) || Math.Abs(squaredDist) < 0.001)
            {
                dx = target.x - basePoint.x;
                dy = target.y - basePoint.y;
                return (dx * dx) + (dy * dy);
            }

            double distRatio = ((target.x - basePoint.x) * lineXDist + (target.y - basePoint.y) * lineYDist) / squaredDist;

            if (distRatio < 0)
            {
                distRatio = 0; // This way, we will compare against mBasePoint
            }
            if (distRatio > 1)
            {
                distRatio = 0; // This way, we will compare against mTipPoint
            }

            dx = basePoint.x - target.x + distRatio * lineXDist;
            dy = basePoint.y - target.y + distRatio * lineYDist;
            return (dx * dx) + (dy * dy);
        }


        /// <summary>
        /// Checks if point is inside quad,
        /// Assumes the quad is convex
        /// </summary>
        /// <param name="quad">A convex quad</param>
        /// <param name="point"></param>
        /// <param name="inflate"></param>
        /// <returns></returns>
        public static bool IsPointInSideQuad(QuadPoint quad, PDFPoint point, double inflate)
        {
            bool l0 = IsLeftTurn(quad.p1, quad.p2, point);
            bool l1 = IsLeftTurn(quad.p2, quad.p3, point);
            bool l2 = IsLeftTurn(quad.p3, quad.p4, point);
            bool l3 = IsLeftTurn(quad.p4, quad.p1, point);
            if (inflate == 0)
            {
                return (l0 && l1 && l2 && l3) == (l0 || l1 || l2 || l3);
            }
            double thresh = inflate * inflate;
            return PointToLineDistance(quad.p1, quad.p2, point) < thresh
                || PointToLineDistance(quad.p2, quad.p3, point) < thresh
                || PointToLineDistance(quad.p3, quad.p4, point) < thresh
                || PointToLineDistance(quad.p4, quad.p1, point) < thresh;
        }

        public static bool IsLeftTurn(PDFPoint p0, PDFPoint p1, PDFPoint p2)
        {
            return ((p1.x - p0.x) * (p2.y - p0.y)
            - (p2.x - p0.x) * (p1.y - p0.y)) >= 0;
        }

        public static UIPoint GetPointCloserToLine(UIPoint basePoint, UIPoint tipPoint, UIPoint target)
        {
            UIPoint returnPoint = new UIPoint(target.X, target.Y);

            double lineXDist = tipPoint.X - basePoint.X;
            double lineYDist = tipPoint.Y - basePoint.Y;

            double squaredDist = (lineXDist * lineXDist) + (lineYDist * lineYDist);

            double distRatio = ((target.X - basePoint.X) * lineXDist + (target.Y - basePoint.Y) * lineYDist) / squaredDist;

            if (double.IsNaN(squaredDist) || Math.Abs(squaredDist) < 0.001)
            {
                returnPoint.X = (target.X + basePoint.X) / 2;
                returnPoint.Y = (target.Y + basePoint.Y) / 2;
                return returnPoint;
            }

            if (distRatio < 0)
            {
                distRatio = 0; // This way, we will compare against mBasePoint
            }
            if (distRatio > 1)
            {
                distRatio = 0; // This way, we will compare against mTipPoint
            }

            double dx = basePoint.X - target.X + distRatio * lineXDist;
            double dy = basePoint.Y - target.Y + distRatio * lineYDist;

            returnPoint = new UIPoint(target.X + (dx / 2), target.Y + (dy / 2));

            return returnPoint;
        }

        public static UIRect GetElementRect(Windows.UI.Xaml.FrameworkElement element, Windows.UI.Xaml.FrameworkElement target = null)
        {
            Windows.UI.Xaml.Media.GeneralTransform elementtransform = element.TransformToVisual(target);
            UIRect rect = elementtransform.TransformBounds(new UIRect(new UIPoint(0, 0), new Windows.Foundation.Size(element.ActualWidth, element.ActualHeight)));
            return rect;
        }

        public static T FindVisualChild<T>(DependencyObject depObj) where T : DependencyObject
        {
            if (depObj != null)
            {
                for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
                {
                    DependencyObject child = VisualTreeHelper.GetChild(depObj, i);
                    if (child != null && child is T)
                    {
                        return (T)child;
                    }

                    T childItem = FindVisualChild<T>(child);
                    if (childItem != null) return childItem;
                }
            }
            return null;
        }
        public static IEnumerable<T> FindAllVisualChildren<T>(DependencyObject depObj) where T : DependencyObject
        {
            if (depObj != null)
            {
                for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
                {
                    DependencyObject child = VisualTreeHelper.GetChild(depObj, i);
                    if (child != null && child is T)
                    {
                        yield return (T)child;
                    }

                    foreach (T childOfChild in FindAllVisualChildren<T>(child))
                    {
                        yield return childOfChild;
                    }
                }
            }
        }

        public static void CopySelectedTextToClipBoard(string text)
        {
            DataPackage dataPackage = new DataPackage
            {
                RequestedOperation = DataPackageOperation.Copy
            };

            dataPackage.SetText(text);

            try
            {
                Clipboard.SetContent(dataPackage);
            }
            catch(Exception)
            {
                // NOTE: Assuming the process of copying data to Clipboard might fail if another app has it opened
            }
        }

        public static bool CopyToClipBoard(DataPackage dataPackage)
        {
            try
            {
                Clipboard.SetContent(dataPackage);
            }
            catch (Exception)
            {
                // NOTE: Assuming the process of copying data to Clipboard might fail if another app has it opened
                return false;
            }

            return true;
        }

        public static void SetCursor(Windows.UI.Core.CoreCursorType cursorType)
        {
#if !WINDOWS_PHONE_APP
            Windows.UI.Xaml.Window.Current.CoreWindow.PointerCursor =
                new Windows.UI.Core.CoreCursor(cursorType, 1);
#endif
        }

        public static SettingsColor ConvertToSettingsColor(Color color)
        {
            return new SettingsColor(color.R, color.G, color.B, color.A != 0);
        }

        public static Color ConvertToColor(SettingsColor color)
        {
            byte alpha = 0;
            if (color.Use)
            {
                alpha = 255;
            }
            return Color.FromArgb(alpha, color.R, color.G, color.B);
        }

        /// <summary>
        /// Convert a <see cref="SettingsColor"/> to a color hex string
        /// </summary>
        /// <param name="color">Color to be converted</param>
        /// <param name="isOffset">If true "RRGGBB", otherwise #RRGGBB</param>
        /// <returns>Color in hex - #RRGGBB or RRGGBB</returns>
        public static string ConvertRgbToHex(SettingsColor color, bool isOffset = false)
        {
            Color myColor = Color.FromArgb(255, color.R, color.G, color.B);

            string colorHex = string.Empty;

            if (!isOffset)
                colorHex = "#";

            colorHex += myColor.R.ToString("X2") + myColor.G.ToString("X2") + myColor.B.ToString("X2");

            return colorHex;
        }

        /// <summary>
        /// Convert hex string color (#RRGGBB) to <see cref="SolidColorBrush"/>
        /// </summary>
        /// <param name="colorStr">The hex color in string format</param>
        /// <returns>The result color</returns>
        public static SolidColorBrush ConvertHexToSolidColor(string colorStr)
        {
            colorStr = colorStr.Replace("#", string.Empty);

            var r = (byte)System.Convert.ToUInt32(colorStr.Substring(0, 2), 16);
            var g = (byte)System.Convert.ToUInt32(colorStr.Substring(2, 2), 16);
            var b = (byte)System.Convert.ToUInt32(colorStr.Substring(4, 2), 16);

            var myBrush = new SolidColorBrush(Color.FromArgb(255, r, g, b));

            return myBrush;
        }

        /// <summary>
        /// Convert hex string color (#RRGGBB) to <see cref="SettingsColor"/>
        /// </summary>
        /// <param name="colorStr">The hex color in string format</param>
        /// <returns>The result color</returns>
        public static SettingsColor ConvertHexToSettingsColor(string colorStr)
        {
            colorStr = colorStr.Replace("#", string.Empty);

            var r = (byte)System.Convert.ToUInt32(colorStr.Substring(0, 2), 16);
            var g = (byte)System.Convert.ToUInt32(colorStr.Substring(2, 2), 16);
            var b = (byte)System.Convert.ToUInt32(colorStr.Substring(4, 2), 16);

            var myBrush = new SettingsColor(r, g, b, false);

            return myBrush;
        }

        public static ToolType GetToolTypeFromAnnotType(IAnnot annot)
        {
            switch (annot.GetAnnotType())
            {
                case AnnotType.e_Line:
                    Annots.Line line = new Annots.Line(annot.GetSDFObj());
                    if (IsArrow(line))
                    {
                        return ToolType.e_arrow_create;
                    }
                    return ToolType.e_line_create;
                case AnnotType.e_Square:
                    return ToolType.e_rect_create;
                case AnnotType.e_Circle:
                    return ToolType.e_oval_create;
                case AnnotType.e_Ink:
                    Annots.Ink ink = new Annots.Ink(annot.GetSDFObj());
                    if (ink.GetBlendMode() == GStateBlendMode.e_bl_multiply)
                    {
                        return ToolType.e_ink_marker;
                    }
                    return ToolType.e_ink_create;
                case AnnotType.e_Polyline:
                    return ToolType.e_polyline_create;
                case AnnotType.e_Polygon:
                    Annots.Polygon polygon = new Annots.Polygon(annot.GetSDFObj());
                    if (polygon.GetBorderEffect() == Annots.MarkupBorderEffect.e_Cloudy)
                    {
                        return ToolType.e_cloud_create;
                    }
                    return ToolType.e_polygon_create;
                case AnnotType.e_Text:
                    return ToolType.e_sticky_note_create;
                case AnnotType.e_FreeText:
                    return ToolType.e_text_annot_create;
                case AnnotType.e_Highlight:
                    return ToolType.e_text_highlight;
                case AnnotType.e_Underline:
                    return ToolType.e_text_underline;
                case AnnotType.e_StrikeOut:
                    return ToolType.e_text_strikeout;
                case AnnotType.e_Squiggly:
                    return ToolType.e_text_squiggly;
                case AnnotType.e_Redact:
                    return ToolType.e_rect_redaction_create;
            }
            return ToolType.e_none;
        }

        public static bool IsArrow(pdftron.PDF.Annots.Line line)
        {
            Annots.LineEndingStyle endingStyle = line.GetStartStyle();
            if (endingStyle == Annots.LineEndingStyle.e_ClosedArrow || endingStyle == Annots.LineEndingStyle.e_OpenArrow
                || endingStyle == Annots.LineEndingStyle.e_RClosedArrow || endingStyle == Annots.LineEndingStyle.e_ROpenArrow)
            {
                return true;
            }
            endingStyle = line.GetEndStyle();
            if (endingStyle == Annots.LineEndingStyle.e_ClosedArrow || endingStyle == Annots.LineEndingStyle.e_OpenArrow
                || endingStyle == Annots.LineEndingStyle.e_RClosedArrow || endingStyle == Annots.LineEndingStyle.e_ROpenArrow)
            {
                return true;
            }
            return false;
        }

        public static bool AreFilesEqual(StorageFile file1, StorageFile file2)
        {
            if (file1 == null && file2 == null)
            {
                return true;
            }
            else if (file1 != null && file2 != null)
            {
#if WINDOWS_PHONE_APP
                string path1 = file1.Path;
                string path2 = file2.Path;
                if (!string.IsNullOrEmpty(path1) && !string.IsNullOrEmpty(path2))
                {
                    return path1.Equals(path2);
                }
#else
                return file1.IsEqual(file2);
#endif
            }
            return false;
        }

        public static string GetMaximalAvailablePath(StorageFile file)
        {
            string path = file.Path;
            if (string.IsNullOrEmpty(path))
                path = file.Name;

            return path;
        }

        public static Windows.Foundation.Size GetUniformStretch(double maxWidth, double maxHeight, double width, double height)
        {
            double scale = Math.Max(width / maxWidth, height / maxHeight);
            return new Windows.Foundation.Size(width / scale, height / scale);
        }

        public static void HideElement(FrameworkElement element)
        {
            element.Opacity = 0;
            element.IsHitTestVisible = false;
            element.Width = 0;
        }

        public static void ShowElement(FrameworkElement element, double width)
        {
            element.Opacity = 1;
            element.IsHitTestVisible = true;
            element.Width = width;
        }

        public static Color GetPostProcessedColor(Color inColor, PDFViewCtrl ctrl)
        {
            double r = inColor.R / 255;
            double g = inColor.G / 255;
            double b = inColor.B / 255;
            ColorPt cp = new ColorPt(r, g, b);
            cp = ctrl.GetPostProcessedColor(cp);
            return Color.FromArgb(inColor.A, (byte)(cp.Get(0) * 255), (byte)(cp.Get(1) * 255), (byte)(cp.Get(2) * 255));
        }

        public static int GetAnnotIdexForAddingTextMarkup(Page page)
        {
            int insertionIndex = page.GetNumAnnots() - 1;
            for (; insertionIndex >= 0; --insertionIndex)
            {
                try
                {
                    var annot = page.GetAnnot(insertionIndex);
                    if (annot == null || !annot.IsValid()) continue;

                    AnnotType annotType = annot.GetAnnotType();
                    if (annotType == AnnotType.e_Highlight ||
                        annotType == AnnotType.e_Underline ||
                        annotType == AnnotType.e_StrikeOut ||
                        annotType == AnnotType.e_Squiggly ||
                        annotType == AnnotType.e_Link ||
                        annotType == AnnotType.e_Widget)
                    {
                        break;
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
            }
            return insertionIndex + 1;
        }

        /// <summary>
        /// Adds a popup to a markup annotation.
        /// Assumes we already have a write lock, and that the annot does not have a popup already.
        /// </summary>
        /// <param name="doc">the document the annotation is in</param>
        /// <param name="markup">the annotation</param>
        /// <param name="text">the text to add</param>
        public static Annots.Popup AddPopupToAnnot(PDFDoc doc, Annots.IMarkup markup, string text)
        {
            Rect rect = markup.GetRect();
            // This rect is offset a bit. We do this because Adobe uses this to position the popup when you open it.
            Annots.Popup p = Annots.Popup.Create(doc.GetSDFDoc(), new Rect(rect.x2 + 50, rect.y2 + 50, rect.x2 + 200, rect.y2 + 200));
            p.SetParent(markup);
            markup.SetPopup(p);
            p.SetContents(text);
            return p;
        }

        public static string GetAnnotContent(Annots.IMarkup markup)
        {
            Annots.Popup pop = markup.GetPopup();
            string content = string.Empty;
            if (pop != null && pop.IsValid())
            {
                content = pop.GetContents();
            }
            else
            {
                content = markup.GetContents();
            }
            return content;
        }

        public static bool DoesAnnotHaveContent(Annots.IMarkup markup)
        {
            return !string.IsNullOrEmpty(GetAnnotContent(markup));
        }

        public static bool DoesAnnotHavePopup(Annots.IMarkup markup)
        {
            Annots.Popup pop = markup.GetPopup();
            return pop != null && pop.IsValid();
        }

        public static String GetSelectedTextOnPage(PDFViewCtrl ctrl, int pageNum)
        {
            String text = "";
            if (ctrl.HasSelectionOnPage(pageNum))
            {
                try
                {
                    ctrl.DocLockRead();
                    PDFViewCtrlSelection sel = ctrl.GetSelection(pageNum);
                    text = sel.GetAsUnicode();
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }
                finally
                {
                    ctrl.DocUnlockRead();
                }
            }
            return text;
        }

        private static string _HelpIconString = new string((char)0x00a5, 1);
        private static string _HelpIconBGString = new string((char)0x017a, 1);
        public static string IconNameToFontIconOutline(string name)
        {
            switch (name)
            {
                case "Comment":
                    return "ś";
                case "Help":
                    return _HelpIconString;
                case "InsertCaret":
                    return "˘";
                case "Star":
                    return "Ş";
                case "LightBulb":
                    return "¤";
                case "NewParagraph":
                    return "¨";
                case "CrossMark":
                    return "ž";
                case "CheckMark":
                    return "š";
                case "LocationPin":
                    return "¦";
                case "ThumbsUp":
                    return "¬";
                default:
                    return name;
            }
        }

        public static string IconNameToFontIconFill(string name)
        {
            switch (name)
            {
                case "Comment":
                    return "›";
                case "Help":
                    return _HelpIconBGString;
                case "InsertCaret":
                    return "ˇ";
                case "Star":
                    return "©";
                case "LightBulb":
                    return "Ł";
                case "NewParagraph":
                    return "§";
                case "CrossMark":
                    return "ť";
                case "CheckMark":
                    return "™";
                case "LocationPin":
                    return "Ą";
                case "ThumbsUp":
                    return "«";
                default:
                    return name;
            }
        }

        /// <summary>
        /// This returns a magic number to multiply the font size of an icon in order to
        /// make it match the icon for the sticky note pushed into the document.
        /// </summary>
        /// <param name="name"></param>
        /// <returns></returns>
        public static double IconNameToFontSizeMultiplier(string name)
        {
            switch (name)
            {
                case "Comment":
                    return 1.0;
                case "Help":
                    return 1.04;
                case "InsertCaret":
                    return 0.92;
                case "Star":
                    return 1.04;
                case "LightBulb":
                    return 1.0;
                case "NewParagraph":
                    return 1.08;
                case "CrossMark":
                    return 1.08;
                case "CheckMark":
                    return 1.08;
                case "LocationPin":
                    return 1.04;
                case "ThumbsUp":
                    return 1.0;
                default:
                    return 1.0;
            }
        }

        public static Rect ConvertFromScreenRectToAnnotationCanvasRect(PDFViewCtrl ctrl, Rect rect)
        {
            DoubleRef x1 = new DoubleRef(rect.x1);
            DoubleRef y1 = new DoubleRef(rect.y1);
            DoubleRef x2 = new DoubleRef(rect.x2);
            DoubleRef y2 = new DoubleRef(rect.y2);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x1, y1);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x2, y2);
            return new Rect(x1.Value, y1.Value, x2.Value, y2.Value);
        }

        // 1 x 1 png
        static readonly byte[] PngBytes = {
            137, 80, 78, 71, 13, 10, 26, 10, 0, 0, 0, 13, 73, 72, 68, 82, 0, 0, 0, 1, 0, 0, 0, 1,
            8, 6, 0, 0, 0, 31, 21, 196, 137, 0, 0, 0, 1, 115, 82, 71, 66, 0, 174, 206, 28, 233, 0,
            0, 0, 4, 103, 65, 77, 65, 0, 0, 177, 143, 11, 252, 97, 5, 0, 0, 0, 9, 112, 72, 89, 115,
            0, 0, 14, 195, 0, 0, 14, 195, 1, 199, 111, 168, 100, 0, 0, 0, 24, 116, 69, 88, 116, 83,
            111, 102, 116, 119, 97, 114, 101, 0, 112, 97, 105, 110, 116, 46, 110, 101, 116, 32, 52,
            46, 48, 46, 51, 140, 230, 151, 80, 0, 0, 0, 13, 73, 68, 65, 84, 24, 87, 99, 248, 255,
            255, 255, 127, 0, 9, 251, 3, 253, 5, 67, 69, 202, 0, 0, 0, 0, 73, 69, 78, 68, 174, 66,
            96, 130
        };

        public static async Task<InMemoryRandomAccessStream> ConvertToPngStreamAsync(byte[] buffer, int w, int h)
        {
            InMemoryRandomAccessStream memoryStream = new InMemoryRandomAccessStream();
            await memoryStream.WriteAsync(PngBytes.AsBuffer()).AsTask().ConfigureAwait(false);
            memoryStream.Seek(0);

            Windows.Graphics.Imaging.BitmapEncoder encode = await Windows.Graphics.Imaging.BitmapEncoder.CreateAsync(Windows.Graphics.Imaging.BitmapEncoder.PngEncoderId,
                memoryStream).AsTask().ConfigureAwait(false);
            encode.SetPixelData(Windows.Graphics.Imaging.BitmapPixelFormat.Bgra8, Windows.Graphics.Imaging.BitmapAlphaMode.Straight,
                (uint)w, (uint)h, 96.0, 96.0, buffer);
            await encode.FlushAsync().AsTask().ConfigureAwait(false);
            memoryStream.Seek(0);
            return memoryStream;
        }

        public static void CloseAllOpenPopups()
        {
            Controls.ViewModels.AnnotationPropertiesFontNameSelectorViewModel.ClosePopups();
            Controls.ViewModels.AnnotationPropertiesDisplay.CloseCurrentPopup();
        }

        public static bool IsContinuous(PDFViewCtrlPagePresentationMode presentationMode)
        {
            return presentationMode == PDFViewCtrlPagePresentationMode.e_facing_continuous
                || presentationMode == PDFViewCtrlPagePresentationMode.e_facing_continuous_cover
                || presentationMode == PDFViewCtrlPagePresentationMode.e_single_continuous;
        }

        /// <summary>
        /// Checks if the annotation is an annotation reply
        /// </summary>
        /// <param name="annot">the annotation</param>
        /// <returns>true if it is an annotation reply, false otherwise</returns>
        public static bool IsReplyTypeReply(IAnnot annot)
        {
            try
            {
                if (annot == null)
                {
                    return false;
                }

                Obj annotSDFObj = annot.GetSDFObj();
                if (annotSDFObj != null)
                {
                    Obj irt = annotSDFObj.FindObj(KeyIRT);
                    // if annotation contains IRT and does not contain RT of "Group"
                    // then it is a reply
                    return ((irt != null) && (!IsReplyTypeGroup(annot)));
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            return false;
        }

        /// <summary>
        /// Gets the list of annotations on a given page.
        /// </summary>
        /// <param name="pdfViewCtrl">PDFViewCtrl object which contains the current document.</param>
        /// <param name="pageNumber">Page number on which to search for annotations.</param>
        /// <returns>List of annotations on a given page.</returns>
        public static List<IAnnot> GetAnnotationsOnPage(PDFViewCtrl pdfViewCtrl, int pageNumber)
        {
            List<IAnnot> annotationList = new List<IAnnot>();
            try
            {
                PDFDoc document = pdfViewCtrl.GetDoc();
                if (document != null)
                {
                    Page page = document.GetPage(pageNumber);

                    if (page != null)
                    {
                        int annotationCount = page.GetNumAnnots();
                        for (int i = 0; i < annotationCount; i++)
                        {
                            var annot = page.GetAnnot(i);
                            if (annot == null || !annot.IsValid()) continue;

                            annotationList.Add(annot);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return annotationList;
        }

        /// <summary>
        /// Gets the unique Id of the IRT parent annotation of this annotation
        /// </summary>
        /// <param name="annotation">The annotation</param>
        /// <returns>The unique id of the IRT parent annotation of this annotation</returns>
        public static string GetIRTAsString(IAnnot annotation)
        {
            if (annotation == null)
            {
                return null;
            }

            Obj annotSDFObj = annotation.GetSDFObj();
            if (annotSDFObj != null)
            {
                Obj irt = annotSDFObj.FindObj(KeyIRT);
                if (irt != null)
                {
                    // first check string, for backwards compatibility
                    if (irt.IsString())
                    {
                        return irt.GetAsPDFText();
                    }
                    // IRT is dict link
                    if (irt.IsDict())
                    {
                        Obj nm = irt.FindObj(KeyNM);
                        if ((nm != null) && nm.IsString())
                        {
                            return nm.GetAsPDFText();
                        }
                    }
                }
            }
            return null;
        }

        /// <summary>
        /// Checks if the annotation is part of a group annotation.
        /// </summary>
        /// <param name="annotation">The annotation.</param>
        /// <returns>True if it is part of a group annotation, false otherwise.</returns>
        public static bool IsReplyTypeGroup(IAnnot annotation)
        {
            if (annotation == null)
            {
                return false;
            }

            // group annotation
            // RT is only meaningful if IRT is present
            // valid values are "R" and "Group"
            Obj annotSDFObj = annotation.GetSDFObj();
            if (annotSDFObj != null)
            {
                Obj irt = annotSDFObj.FindObj(KeyIRT);
                Obj rt = annotSDFObj.FindObj(KeyRT);
                if ((irt != null) && (rt != null) && rt.IsName())
                {
                    string rtVal = rt.GetName();
                    return ValueGroup.Equals(rtVal);
                }
            }
            return false;
        }

        /// <summary>
        /// Returns all annotations in this group if the annotation passed in is in an annotation group
        /// </summary>
        /// <param name="pdfViewCtrl">PDFViewCtrl object which contains the current document</param>
        /// <param name="annotation">the annotation.</param>
        /// <param name="pageNumber">page number of page that contains the annotations</param>
        /// <returns>returns all annotations in this group if the annotation passed in is in an annotation group</returns>
        public static List<IAnnot> GetAnnotationsInGroup(PDFViewCtrl pdfViewCtrl, IAnnot annotation, int pageNumber)
        {
            if ((pdfViewCtrl == null) || (annotation == null))
            {
                return null;
            }

            if (!annotation.IsValid())
            {
                return null;
            }

            bool shouldUnlockRead = false;
            try
            {
                List<IAnnot> annotationsInGroup = new List<IAnnot>();
                pdfViewCtrl.DocLockRead();
                shouldUnlockRead = true;

                string mainAnnotationId = null;

                if (IsReplyTypeGroup(annotation))
                {
                    // this is a sub annot in a group
                    // its IRT will be the primary annot in the group
                    mainAnnotationId = GetIRTAsString(annotation);
                }
                else if (annotation.GetUniqueID() != null)
                {
                    // this is a primary annot
                    mainAnnotationId = annotation.GetUniqueID().GetAsPDFText();
                }

                if (String.IsNullOrEmpty(mainAnnotationId))
                {
                    return null;
                }

                List<IAnnot> annotationsInPage = GetAnnotationsOnPage(pdfViewCtrl, pageNumber);
                foreach (IAnnot annotationItem in annotationsInPage)
                {
                    if ((annotationItem != null) && annotationItem.IsValid() && (annotationItem.GetUniqueID() != null))
                    {
                        // find all annotations in the group
                        string id = annotationItem.GetUniqueID().GetAsPDFText();
                        if ((id != null) && id.Equals(mainAnnotationId))
                        {
                            // add primary annot
                            annotationsInGroup.Add(annotationItem);
                        }
                        string irt = GetIRTAsString(annotationItem);
                        if (IsReplyTypeGroup(annotationItem) && (irt != null) && irt.Equals(mainAnnotationId))
                        {
                            // add subordinate annot
                            annotationsInGroup.Add(annotationItem);
                        }
                    }
                }

                return annotationsInGroup;
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (shouldUnlockRead)
                {
                    pdfViewCtrl.DocUnlockRead();
                }
            }

            return null;
        }

        /// <summary>
        /// Snaps a point to the specified rectangle
        /// </summary>
        /// <param name="point">the resulted point snapped to the rectangle</param>
        /// <param name="rect">the rectangle</param>
        public static void SnapPointToRect(PDFPoint point, PDFRect rect)
        {
            if (rect != null)
            {
                if (point.x < rect.x1)
                {
                    point.x = rect.x1;
                }
                else if (point.x > rect.x2)
                {
                    point.x = rect.x2;
                }
                if (point.y < rect.y2)
                {
                    point.y = rect.y2;
                }
                else if (point.y > rect.y1)
                {
                    point.y = rect.y1;
                }
            }
        }

        /// <summary>
        /// Gets the annotation creation date in local date time; if it fails then UTC time is returned
        /// </summary>
        /// <param name="annotation">the annotation</param>
        /// <returns>the date</returns>
        public static DateTime GetAnnotationLocalCreationDate(IAnnot annotation)
        {
            pdftron.PDF.Date date = annotation.GetDate();
            if (annotation.IsMarkup())
            {
                date = (annotation as Annots.Markup).GetCreationDates();
            }

            DateTime annotTime = new DateTime(date.year, date.month, date.day, date.hour, date.minute, date.second);
            DateTime timeUtc = annotTime.ToUniversalTime();

            try
            {
                TimeZoneInfo localZone = TimeZoneInfo.Local;
                DateTime localTime = TimeZoneInfo.ConvertTime(timeUtc, localZone);
                return localTime;
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }

            return timeUtc;
        }

        /// <summary>
        /// Returns true if the annotation was created with PDFTron SDK
        /// </summary>
        /// <param name="annotation">the annotation</param>
        /// <returns>true if created by PDFTron SDK; false otherwise</returns>
        public static bool IsMadeByPDFTron(IAnnot annotation)
        {
            string[] supportedTag = new string[] { PDFTronId, "pdftronlink" };
            if (annotation != null && annotation.GetSDFObj() != null)
            {
                Obj sdfObj = annotation.GetSDFObj();
                foreach (String tag in supportedTag)
                {
                    Object selfMadeObj = sdfObj.FindObj(tag);
                    if (selfMadeObj != null)
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        /// <summary>
        /// Ungroups the annotations
        /// </summary>
        /// <param name="pdfViewCtrl">PDFViewCtrl object which contains the current document</param>
        /// <param name="allAnnotationsInGroup">the annotations in the group</param>
        public static void UngroupAnnotations(PDFViewCtrl pdfViewCtrl, List<IAnnot> allAnnotationsInGroup)
        {
            if ((pdfViewCtrl == null) || (allAnnotationsInGroup == null))
            {
                return;
            }
            bool shouldUnlock = false;
            try
            {
                pdfViewCtrl.DocLock(true);
                shouldUnlock = true;
                foreach (IAnnot annotation in allAnnotationsInGroup)
                {
                    // remove all related keys
                    annotation.GetSDFObj().Erase(KeyRT);
                    annotation.GetSDFObj().Erase(KeyIRT);

                    annotation.SetDateToNow(); // Update Date in case of using ExternalAnnotManager
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (shouldUnlock)
                {
                    pdfViewCtrl.DocUnlock();
                }
            }
        }

        /// <summary>
        /// Creates a group of annotations
        /// </summary>
        /// <param name="pdfViewCtrl">PDFViewCtrl object which contains the current document</param>
        /// <param name="primaryAnnotation">the primary annotation</param>
        /// <param name="allAnnotationsInGroup">all annotations to be added to this group</param>
        public static void CreateAnnotationGroup(PDFViewCtrl pdfViewCtrl, IAnnot primaryAnnotation, List<IAnnot> allAnnotationsInGroup)
        {
            if ((pdfViewCtrl == null) || (primaryAnnotation == null) || (allAnnotationsInGroup == null))
            {
                return;
            }

            bool shouldUnlock = false;

            try
            {
                pdfViewCtrl.DocLock(true);
                shouldUnlock = true;
                foreach (IAnnot annotation in allAnnotationsInGroup)
                {
                    // first set Guid if not exist yet
                    if (annotation.GetUniqueID() == null)
                    {
                        SetUniqueId(annotation, Guid.NewGuid().ToString());
                    }
                    if (!annotation.Equals(primaryAnnotation))
                    {
                        annotation.GetSDFObj().PutName(KeyRT, ValueGroup);
                        annotation.GetSDFObj().Put(KeyIRT, primaryAnnotation.GetSDFObj());
                    }
                    annotation.SetDateToNow(); // Update Date in case of using ExternalAnnotManager
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(
                    Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                    Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            finally
            {
                if (shouldUnlock)
                {
                    pdfViewCtrl.DocUnlock();
                }
            }
        }

        /// <summary>
        /// Sets the unique identifier for the specified annotation, if possible.
        /// This method does not lock the document, so a write lock should be acquired outside.
        /// </summary>
        /// <param name="annotation">the annotation</param>
        /// <param name="id">the unique identifier</param>
        public static void SetUniqueId(IAnnot annotation, string id)
        {
            try
            {
                if (annotation != null)
                {
                    annotation.SetUniqueID(id);
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
        }

        public class AnnotUtils
        {
            /// <summary>
            /// Get the annotation screen position
            /// </summary>
            /// <returns>The annotation screen position</returns>
            public static Rect GetAnnotScreenPosition(PDFViewCtrl pDFViewCtrl, IAnnot annot, int annotPageNum)
            {
                if (annot == null)
                    return null;

                Rect oldUpdateRect = pDFViewCtrl.GetScreenRectForAnnot(annot, annotPageNum);
                oldUpdateRect.Normalize();

                return oldUpdateRect;
            }

            /// <summary>
            /// Used in replacement of <see cref="Page.AnnotRemove(IAnnot)"/> to handle direct object and updates the viewer
            /// </summary>
            /// <remarks>
            /// The annot parameter should no longer be used after this call
            /// </remarks>
            /// <returns>The returned annot object should be used after this function</returns>
            public static IAnnot SafeDeleteAnnotAndUpdate(PDFViewCtrl pDFViewCtrl, Page page, IAnnot annot, int annotPageNum)
            {
                bool isIndirect = annot.GetSDFObj().IsIndirect();
                Rect oldScreenRect = null;

                if (!isIndirect)
                {
                    oldScreenRect = GetAnnotScreenPosition(pDFViewCtrl, annot, annotPageNum);
                }

                annot = SafeDeleteAnnot(pDFViewCtrl.GetDoc(), page, annot);
                if (oldScreenRect != null)
                    pDFViewCtrl.Update(oldScreenRect);
                else
                    pDFViewCtrl.UpdateWithAnnot(annot, annotPageNum);

                return annot;
            }

            /// <summary>
            /// Used in replacement of<see cref= "Page.AnnotRemove(IAnnot)" /> to handle direct object and updates the viewer
            /// </summary>
            /// <remarks>
            /// The annot parameter should no longer be used after this call
            /// </remarks>
            /// <returns>The returned annot object should be used after this function</returns>
            public static IAnnot SafeDeleteAnnot(PDFDoc doc, Page page, IAnnot annot)
            {
                Obj annotObj = annot.GetSDFObj();
                Obj indirectCopy = annotObj.IsIndirect() ? annotObj : doc.GetSDFDoc().ImportObj(annotObj, false);

                page.AnnotRemove(annot);

                return new Annot(indirectCopy);
            }
        }

        #region FreeText fonts
        public class FontBundle
        {
            /// <summary>
            /// Returns a string intended for use by system UI widgets like TextBoxes.
            /// </summary>
            public string FontUri
            {
                get
                {
                    // TODO - Figure out how this works with imported fonts.
                    string localFolder = Windows.Storage.ApplicationData.Current.LocalFolder.Path;
                    localFolder = localFolder.Replace("\\", "/");
                    if (FontPath.Contains(localFolder))
                    {
                        string fontUri = FontPath.Replace(localFolder, "ms-appdata:///local") + "#" + DisplayName;
                        return fontUri;
                    }
                    return FontFamily;
                }
            }

            public string FontPath { get; private set; }
            public string DisplayName { get; private set; }
            public string PDFTronName { get; private set; }

            public string FontFamily { get; private set; }

            public FontWeight FontWeight { get; private set; }

            public FontStyle FontStyle { get; private set; }

            public static string FontListJson { get; private set; }

            public FontBundle(string fontPath, string displayName, string pdfTronName, string family, FontWeight weight, FontStyle style)
            {
                FontPath = fontPath;
                DisplayName = displayName;
                PDFTronName = pdfTronName;
                FontFamily = family;
                FontWeight = weight;
                FontStyle = style;
            }

            private enum CompareProperties
            {
                DisplayName,
                PDFTronName,
            }

            public static FontBundle CreateNullFont()
            {
                return new FontBundle("", "", "", "", FontWeights.Normal, FontStyle.Normal);
            }

            public static bool IsNullFont(FontBundle font)
            {
                return string.IsNullOrWhiteSpace(font.DisplayName) || string.IsNullOrWhiteSpace(font.FontPath) || string.IsNullOrWhiteSpace(font.PDFTronName);
            }

            public static FontBundle Create(ToolType toolType)
            {
                return Create(Settings.GetFontName(toolType), CompareProperties.DisplayName);
            }

            public static FontBundle CreateFromDisplayName(string displayName)
            {
                return FontBundle.Create(displayName, CompareProperties.DisplayName);
            }

            public static FontBundle CreateFromPDFTronName(string pdfTronName)
            {
                return FontBundle.Create(pdfTronName, CompareProperties.PDFTronName);
            }

            private static FontBundle Create(string fontName, CompareProperties comparer)
            {
                return Create(fontName, comparer, false);
            }

            public static FontBundle CreateDefaultFontBundle()
            {
                return Create(ToolSettings.Constants.FONT_DEFAULT_NAME, CompareProperties.DisplayName, true);
            }

            public static async Task<string> GetSystemFontListAsync()
            {
                try
                {
                    FontListJson = await Task.Run(() => pdftron.PDFNet.GetSystemFontListJSON());
                }
                catch (Exception exception)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY,
                            Utilities.AnalyticsHandlerBase.GetExceptionMessage(exception));
                }
                return FontListJson;
            }

            private static FontBundle Create(string fontName, CompareProperties comparer, bool pickFistIfFails)
            {
                if (String.IsNullOrEmpty(FontListJson))
                {
                    return null;
                }

                FontBundle firstFontbundle = null;
                try
                {
                    string fontListJson = FontListJson;

                    JsonObject jobj = JsonObject.Parse(fontListJson);

                    JsonArray fontArr = jobj.GetNamedArray(ToolSettings.Constants.FONT_JSON_NAME);

                    for (uint i = 0; i < fontArr.Count; ++i)
                    {
                        JsonObject fontObj = fontArr.GetObjectAt(i);
                        string displayName = fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_DISPLAY_NAME);
                        string pdfTronName = fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_PDFTRON_NAME);
                        if ((comparer == CompareProperties.DisplayName && displayName.Equals(fontName)) ||
                            (comparer == CompareProperties.PDFTronName && pdfTronName.Equals(fontName)))
                        {
                            FontBundle fontBundle = new FontBundle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FILE_PATH), displayName, pdfTronName,
                                fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FAMILY_NAME),
                                StringToFontWeight(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_WEIGTH)),
                                StringToFontStyle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_STYLE)));
                            return fontBundle;
                        }
                        if (pickFistIfFails)
                        {
                            firstFontbundle = new FontBundle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FILE_PATH), displayName, pdfTronName,
                                fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FAMILY_NAME),
                                StringToFontWeight(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_WEIGTH)),
                                StringToFontStyle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_STYLE)));
                            pickFistIfFails = false;
                        }
                    }
                }
                catch (Exception ex)
                {
                    Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
                }

                return firstFontbundle;
            }

            public override bool Equals(object obj)
            {
                FontBundle other = obj as FontBundle;
                if (other != null)
                {
                    return this.DisplayName.Equals(other.DisplayName);
                }
                return base.Equals(obj);
            }

            public override int GetHashCode()
            {
                return this.DisplayName.GetHashCode();
            }

            public static FontWeight StringToFontWeight(string weightString)
            {
                FontWeight weight = FontWeights.Normal;
                if (!string.IsNullOrWhiteSpace(weightString))
                {
                    if (weightString.Equals("Thin", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Thin;
                    }
                    else if (weightString.Equals("ExtraLight", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.ExtraLight;
                    }
                    else if (weightString.Equals("Light", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Light;
                    }
                    else if (weightString.Equals("Medium", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Medium;
                    }
                    else if (weightString.Equals("SemiBold", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.SemiBold;
                    }
                    else if (weightString.Equals("Bold", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Bold;
                    }
                    else if (weightString.Equals("Heavy", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.ExtraBold;
                    }
                    else if (weightString.Equals("Black", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.Black;
                    }
                    else if (weightString.Equals("UltraBlack", StringComparison.OrdinalIgnoreCase))
                    {
                        weight = FontWeights.ExtraBlack;
                    }
                }
                return weight;
            }

            public static FontStyle StringToFontStyle(string styleString)
            {
                FontStyle style = FontStyle.Normal;
                if (!string.IsNullOrWhiteSpace(styleString))
                {
                    if (styleString.Equals("Italic", StringComparison.OrdinalIgnoreCase))
                    {
                        style = FontStyle.Italic;
                    }
                    else if (styleString.Equals("Oblique", StringComparison.OrdinalIgnoreCase))
                    {
                        style = FontStyle.Oblique;
                    }
                }
                return style;
            }
        }

        public static List<FontBundle> GetListOfSystemFonts()
        {
            List<FontBundle> fontList = new List<FontBundle>();

            if (String.IsNullOrEmpty(FontBundle.FontListJson))
            {
                return fontList;
            }

            try
            {
                string fontListJson = FontBundle.FontListJson;

                JsonObject jobj = JsonObject.Parse(fontListJson);

                JsonArray fontArr = jobj.GetNamedArray(ToolSettings.Constants.FONT_JSON_NAME);
                string lastDisplayName = "";

                for (uint i = 0; i < fontArr.Count; ++i)
                {
                    JsonObject fontObj = fontArr.GetObjectAt(i);
                    string displayName = fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_DISPLAY_NAME);
                    if (!string.IsNullOrWhiteSpace(displayName) && !displayName.Equals(lastDisplayName))
                    {
                        fontList.Add(new FontBundle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FILE_PATH), displayName,
                            fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_PDFTRON_NAME),
                                fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FAMILY_NAME),
                                FontBundle.StringToFontWeight(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_WEIGTH)),
                                FontBundle.StringToFontStyle(fontObj.GetNamedString(ToolSettings.Constants.FONT_JSON_FONT_STYLE))));

                        lastDisplayName = displayName;
                    }
                }
            }
            catch (Exception ex)
            {
                Utilities.AnalyticsHandlerBase.CURRENT.SendEvent(Utilities.AnalyticsHandlerBase.EXCEPTION_EVENT_CATEGORY, Utilities.AnalyticsHandlerBase.GetExceptionMessage(ex));
            }
            return fontList;
        }

        public static void SetFontFromBundle(Annots.FreeText freeText, PDFDoc doc, FontBundle bundle)
        {
            string pdftronFontName = bundle.PDFTronName;

            // Create a DR entry for embedding the font
            Obj annotObj = freeText.GetSDFObj();
            Obj drDict = annotObj.PutDict("DR");

            // Embed the font
            Obj fontDict = drDict.PutDict("Font");
            Font font = Font.Create(doc, pdftronFontName, freeText.GetContents());
            fontDict.Put(ToolSettings.Constants.FONT_DR_NAME, font.GetSDFObj());
        }

        public static void SetFreeTextDefaultAppearanceWithoutFont(Annots.FreeText freeText)
        {
            SetFreeTextDefaultAppearance(freeText, ToolSettings.Constants.FONT_FREETEXT_DA_DEFAULT);
        }

        public static void SetFreeTextDefaultAppearanceWithFont(Annots.FreeText freeText)
        {
            SetFreeTextDefaultAppearance(freeText, ToolSettings.Constants.FONT_DR_NAME);
        }

        private static void SetFreeTextDefaultAppearance(Annots.FreeText freeText, string fontName)
        {
            // Set DA string
            String DA = freeText.GetDefaultAppearance();
            int slashPosition = DA.IndexOf("/", 0);

            // if DR string contains '/' which it always should.
            if (slashPosition > 0)
            {
                String beforeSlash = DA.Substring(0, slashPosition);
                String afterSlash = DA.Substring(slashPosition);
                String afterFont = afterSlash.Substring(afterSlash.IndexOf(" "));
                String updatedDA = beforeSlash + "/" + fontName + afterFont;

                freeText.SetDefaultAppearance(updatedDA);
            }
        }

        public static FontBundle GetFontFromAnnot(Annots.FreeText freeText)
        {
            string fontName = GetFontNameFromAnnot(freeText);
            if (fontName != null)
            {
                return UtilityFunctions.FontBundle.CreateFromPDFTronName(fontName);
            }
            return null;
        }

        public static string GetFontNameFromAnnot(Annots.FreeText freeText)
        {
            Obj ftObj = freeText.GetSDFObj();
            Obj drDict = ftObj.FindObj("DR");
            if (drDict != null && drDict.IsDict())
            {
                Obj fontDict = drDict.FindObj("Font");
                if (fontDict != null && fontDict.IsDict())
                {
                    DictIterator fItr = fontDict.GetDictIterator();
                    if (fItr.HasNext())
                    {
                        Font f = new Font(fItr.Value());
                        Obj fontDescriptor = f.GetDescriptor();
                        if (fontDescriptor != null && fontDescriptor.IsDict())
                        {
                            Obj fontNameObj = fontDescriptor.FindObj("FontName");
                            if (fontNameObj != null && fontNameObj.IsName())
                            {
                                return fontNameObj.GetName();
                            }
                        }
                        string fontName = f.GetName();
                        return fontName;
                    }
                }
            }
            return null;
        }

        public static void GenerateDSEntry(Annots.FreeText freeText, FontBundle fontBundle = null)
        {
            if (fontBundle == null)
            {
                fontBundle = GetFontFromAnnot(freeText);
            }
            ColorPt textColor = freeText.GetTextColor();
            int r = (int)(textColor.Get(0) * 255 + 0.5);
            int g = (int)(textColor.Get(1) * 255 + 0.5);
            int b = (int)(textColor.Get(2) * 255 + 0.5);
            string colorString = Windows.UI.Color.FromArgb(255, (byte)r, (byte)g, (byte)b).ToString();
            colorString = "#" + colorString.Substring(3);

            double fs = freeText.GetFontSize();

            FontStyle style = fontBundle.FontStyle;
            if (style == FontStyle.Oblique) // oblique is not a valid value
            {
                style = FontStyle.Italic;
            }

            string formatString = string.Format("font-family: {0}; font-style: {1}; font-weight: {2}; font-size: {3}pt; color: {4}",
                fontBundle.FontFamily, style, fontBundle.FontWeight.Weight, fs, colorString);
            freeText.GetSDFObj().PutString("DS", formatString);
        }

        /// <summary>
        /// Get value from RC, the value is in HTML
        /// </summary>
        /// <returns></returns>
        public static string GetRichFreeText(pdftron.PDF.Annots.FreeText freeTextAnnot)
        {
            if (!IsRichFreeText(freeTextAnnot))
            {
                return "";
            }
            DictIterator dictIter = freeTextAnnot.GetSDFObj().Get(KeyRichContent);
            string richText = "";
            if (dictIter.Value().IsString())
            {
                richText = dictIter.Value().GetAsPDFText();
            }
            return richText;
        }

        /// <summary>
        /// Check if RC exists
        /// </summary>
        /// <param name="freeTextAnnot"></param>
        /// <returns></returns>
        public static bool IsRichFreeText(pdftron.PDF.Annots.FreeText freeTextAnnot)
        {
            var sdfObj = freeTextAnnot.GetSDFObj();

            DictIterator dictIter = sdfObj.Find(KeyRichContent);

            if (dictIter.HasNext())
            {
                return true;
            }

            return false;
        }

        #endregion FreeText fonts

        #region Rectangle Conversions

        public static PDFPoint ConvertFromScreenPointToCanvasPoint(PDFPoint point, PDFViewCtrl ctrl)
        {
            PDFDouble x = new PDFDouble(point.x);
            PDFDouble y = new PDFDouble(point.y);

            ctrl.ConvScreenPtToAnnotationCanvasPt(x, y);

            return new PDFPoint(x.Value, y.Value);
        }

        public static PDFPoint ConvertFromScreenPointToCanvasPoint(double x, double y, PDFViewCtrl ctrl)
        {
            PDFDouble px = new PDFDouble(x);
            PDFDouble py = new PDFDouble(y);

            ctrl.ConvScreenPtToAnnotationCanvasPt(px, py);

            return new PDFPoint(px.Value, py.Value);
        }

        public static PDFPoint ConvertFromCanvasPointToScreenPoint(PDFPoint point, PDFViewCtrl ctrl)
        {
            PDFDouble x = new PDFDouble(point.x);
            PDFDouble y = new PDFDouble(point.y);

            ctrl.ConvAnnotationCanvasPtToScreenPt(x, y);

            return new PDFPoint(x.Value, y.Value);
        }

        public static PDFPoint ConvertFromCanvasPointToScreenPoint(double x, double y, PDFViewCtrl ctrl)
        {
            PDFDouble px = new PDFDouble(x);
            PDFDouble py = new PDFDouble(y);

            ctrl.ConvAnnotationCanvasPtToScreenPt(px, py);

            return new PDFPoint(px.Value, py.Value);
        }

        public static PDFPoint ConvertFromCanvasPointToPagePoint(PDFPoint point, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x = new PDFDouble(point.x);
            PDFDouble y = new PDFDouble(point.y);

            ctrl.ConvAnnotationCanvasPtToScreenPt(x, y);
            ctrl.ConvScreenPtToPagePt(x, y, pageNum);

            return new PDFPoint(x.Value, y.Value);
        }

        public static PDFPoint ConvertFromPagePointToCanvasPoint(PDFPoint point, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x = new PDFDouble(point.x);
            PDFDouble y = new PDFDouble(point.y);

            ctrl.ConvPagePtToScreenPt(x, y, pageNum);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x, y);

            return new PDFPoint(x.Value, y.Value);
        }

        public static PDFRect ConvertFromScreenRectToPageRect(PDFRect rect, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvScreenPtToPagePt(x1, y1, pageNum);
            ctrl.ConvScreenPtToPagePt(x2, y2, pageNum);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        public static PDFRect ConvertFromPageRectToScreenRect(PDFRect rect, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvPagePtToScreenPt(x1, y1, pageNum);
            ctrl.ConvPagePtToScreenPt(x2, y2, pageNum);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        public static PDFRect ConvertFromCanvasRectToPageRect(PDFRect rect, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvAnnotationCanvasPtToScreenPt(x1, y1);
            ctrl.ConvAnnotationCanvasPtToScreenPt(x2, y2);

            ctrl.ConvScreenPtToPagePt(x1, y1, pageNum);
            ctrl.ConvScreenPtToPagePt(x2, y2, pageNum);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);
            retRect.Normalize();

            return retRect;
        }

        public static PDFRect ConvertFromPageRectToCanvasRect(PDFRect rect, int pageNum, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvPagePtToScreenPt(x1, y1, pageNum);
            ctrl.ConvPagePtToScreenPt(x2, y2, pageNum);

            ctrl.ConvScreenPtToAnnotationCanvasPt(x1, y1);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x2, y2);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        public static PDFRect ConvertFromCanvasRectToScreenRect(PDFRect rect, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvAnnotationCanvasPtToScreenPt(x1, y1);
            ctrl.ConvAnnotationCanvasPtToScreenPt(x2, y2);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        public static PDFRect ConvertFromScreenRectToCanvasRect(PDFRect rect, PDFViewCtrl ctrl)
        {
            PDFDouble x1 = new PDFDouble(rect.x1);
            PDFDouble y1 = new PDFDouble(rect.y1);
            PDFDouble x2 = new PDFDouble(rect.x2);
            PDFDouble y2 = new PDFDouble(rect.y2);

            ctrl.ConvScreenPtToAnnotationCanvasPt(x1, y1);
            ctrl.ConvScreenPtToAnnotationCanvasPt(x2, y2);

            PDFRect retRect = new PDFRect(x1.Value, y1.Value, x2.Value, y2.Value);

            return retRect;
        }

        #endregion Rectangle Conversions


        #region Color Space

        public struct RGB
        {
            public int R; // 0 - 255
            public int G;
            public int B;

            public RGB(int r, int g, int b)
            {
                R = r;
                G = g;
                B = b;
            }
        }

        public struct HSV
        {
            public double H; // 0 - 1
            public double S; // 0 - 1
            public double V; // 0 - 1

            public HSV(double h, double s, double v)
            {
                H = h;
                S = s;
                V = v;
            }
        }

        public static RGB HSVtoRGB(HSV hsv)
        {
            // HSV contains values scaled as in the color wheel:
            // that is, all from 0 to 255.

            // for ( this code to work, HSV.Hue needs
            // to be scaled from 0 to 360 (it//s the angle of the selected
            // point within the circle). HSV.Saturation and HSV.value must be
            // scaled to be between 0 and 1.

            double h;
            double s;
            double v;

            double r = 0;
            double g = 0;
            double b = 0;

            // Scale Hue to be between 0 and 360. Saturation
            // and value scale to be between 0 and 1.
            h = ((double)hsv.H * 360) % 360;
            s = (double)hsv.S;
            v = (double)hsv.V;

            if (s == 0)
            {
                // If s is 0, all colors are the same.
                // This is some flavor of gray.
                r = v;
                g = v;
                b = v;
            }
            else
            {
                double p;
                double q;
                double t;

                double fractionalSector;
                int sectorNumber;
                double sectorPos;

                // The color wheel consists of 6 sectors.
                // Figure out which sector you//re in.
                sectorPos = h / 60;
                sectorNumber = (int)(Math.Floor(sectorPos));

                // get the fractional part of the sector.
                // That is, how many degrees into the sector
                // are you?
                fractionalSector = sectorPos - sectorNumber;

                // Calculate values for the three axes
                // of the color.
                p = v * (1 - s);
                q = v * (1 - (s * fractionalSector));
                t = v * (1 - (s * (1 - fractionalSector)));

                // Assign the fractional colors to r, g, and b
                // based on the sector the angle is in.
                switch (sectorNumber)
                {
                    case 0:
                        r = v;
                        g = t;
                        b = p;
                        break;

                    case 1:
                        r = q;
                        g = v;
                        b = p;
                        break;

                    case 2:
                        r = p;
                        g = v;
                        b = t;
                        break;

                    case 3:
                        r = p;
                        g = q;
                        b = v;
                        break;

                    case 4:
                        r = t;
                        g = p;
                        b = v;
                        break;

                    case 5:
                        r = v;
                        g = p;
                        b = q;
                        break;
                }
            }
            // return an RGB structure, with values scaled
            // to be between 0 and 255.
            return new RGB((int)(r * 255), (int)(g * 255), (int)(b * 255));
        }

        public static HSV RGBtoHSV(RGB rgb)
        {
            // In this function, R, G, and B values must be scaled
            // to be between 0 and 1.
            // HSV.Hue will be a value between 0 and 360, and
            // HSV.Saturation and value are between 0 and 1.
            // The code must scale these to be between 0 and 255 for
            // the purposes of this application.

            double min;
            double max;
            double delta;

            double r = (double)rgb.R / 255;
            double g = (double)rgb.G / 255;
            double b = (double)rgb.B / 255;

            double h;
            double s;
            double v;

            min = Math.Min(Math.Min(r, g), b);
            max = Math.Max(Math.Max(r, g), b);
            v = max;
            delta = max - min;
            if (max == 0 || delta == 0)
            {
                // R, G, and B must be 0, or all the same.
                // In this case, S is 0, and H is undefined.
                // Using H = 0 is as good as any...
                s = 0;
                h = 0;
            }
            else
            {
                s = delta / max;
                if (r == max)
                {
                    // Between Yellow and Magenta
                    h = (g - b) / delta;
                }
                else if (g == max)
                {
                    // Between Cyan and Yellow
                    h = 2 + (b - r) / delta;
                }
                else
                {
                    // Between Magenta and Cyan
                    h = 4 + (r - g) / delta;
                }

            }
            // Scale h to be between 0 and 360.
            // This may require adding 360, if the value
            // is negative.
            h *= 60;
            if (h < 0)
            {
                h += 360;
            }

            // Scale to the requirements of this
            // application. All values are between 0 and 1.
            return new HSV((h / 360), s, v);
        }

        public static Tuple<double, double, double> GetHSVFromRGB(Color color)
        {
            double h = 0;
            double s = 0;
            double v = 0;

            double r = color.R / 255.0;
            double g = color.G / 255.0;
            double b = color.B / 255.0;

            double cMax = Math.Max(b, Math.Max(r, g));
            double cMin = Math.Min(b, Math.Min(r, g));

            double delta = cMax - cMin;

            if (delta == 0)
            {
                h = 0;
            }
            else if (cMax == r)
            {
                h = 60 * (((g - b) / delta) % 6);
            }
            else if (cMax == g)
            {
                h = 60 * (((b - r) / delta) + 2);
            }
            else
            {
                h = 60 * (((r - g) / delta) + 4);
            }

            if (cMax == 0)
            {
                s = 0;
            }
            else
            {
                s = (delta / cMax) * 100;
            }

            if (h < 0)
            {
                h += 360;
            }

            v = cMax * 100;

            return new Tuple<double, double, double>(h, s, v);
        }

        public static Color GetRGBFromHSV(double h, double s, double v)
        {
            Color color = new Color();

            double c = v / 100.0 * s / 100.0;
            double x = c * (1 - Math.Abs(((h / 60) % 2) - 1));
            double m = v / 100.0 - c;

            double r = 0;
            double g = 0;
            double b = 0;

            if (h < 60)
            {
                r = c;
                g = x;
                b = 0;
            }
            else if (h < 120)
            {
                r = x;
                g = c;
                b = 0;
            }
            else if (h < 180)
            {
                r = 0;
                g = c;
                b = x;
            }
            else if (h < 240)
            {
                r = 0;
                g = x;
                b = c;
            }
            else if (h < 300)
            {
                r = x;
                g = 0;
                b = c;
            }
            else
            {
                r = c;
                g = 0;
                b = x;
            }

            color.A = 255;
            color.R = (byte)((r + m) * 255);
            color.G = (byte)((g + m) * 255);
            color.B = (byte)((b + m) * 255);

            return color;
        }

        #endregion Color Space

        #region Enum Helpers

        /// <summary>
        /// Get annotation description from Annotation Enum Type
        /// </summary>
        /// <param name="annotType"> Annotation Type </param>
        /// <returns> Description in case it exists otherwise "Unknown" </returns>
        public static string GetAnnotDescriptionFromEnum(AnnotType annotType)
        {
            switch (annotType)
            {
                case AnnotType.e_3D:
                    return "3D";
                case AnnotType.e_Caret:
                    return "Caret";
                case AnnotType.e_Circle:
                    return "Circle";
                case AnnotType.e_FileAttachment:
                    return "File Attachment";
                case AnnotType.e_FreeText:
                    return "Free Text";
                case AnnotType.e_Highlight:
                    return "Highlight";
                case AnnotType.e_Ink:
                    return "Ink";
                case AnnotType.e_Line:
                    return "Line";
                case AnnotType.e_Movie:
                    return "Movie";
                case AnnotType.e_Polygon:
                    return "Polygon";
                case AnnotType.e_Polyline:
                    return "Polyline";
                case AnnotType.e_Popup:
                    return "Popup";
                case AnnotType.e_PrinterMark:
                    return "Printer Mark";
                case AnnotType.e_Projection:
                    return "Projection";
                case AnnotType.e_Redact:
                    return "Redact";
                case AnnotType.e_RichMedia:
                    return "Rich Media";
                case AnnotType.e_Screen:
                    return "Screen";
                case AnnotType.e_Sound:
                    return "Sound";
                case AnnotType.e_Square:
                    return "Square";
                case AnnotType.e_Squiggly:
                    return "Squiggly";
                case AnnotType.e_Stamp:
                    return "Stamp";
                case AnnotType.e_StrikeOut:
                    return "Strikeout";
                case AnnotType.e_Text:
                    return "Sticky";
                case AnnotType.e_TrapNet:
                    return "Trap Net";
                case AnnotType.e_Underline:
                    return "Underline";
                case AnnotType.e_Watermark:
                    return "Watermark";
                case AnnotType.e_Widget:
                    return "Widget";

                case AnnotType.e_Unknown:
                default:
                    return "Unknown";
            }
        }
        #endregion

        public static DeviceFormFactorType GetDeviceFormFactorType()
        {
            switch (AnalyticsInfo.VersionInfo.DeviceFamily)
            {
                case "Windows.Mobile":
                    return DeviceFormFactorType.Phone;
                case "Windows.Desktop":
                    return DeviceFormFactorType.Desktop;
                case "Windows.Universal":
                    return DeviceFormFactorType.IoT;
                case "Windows.Team":
                    return DeviceFormFactorType.SurfaceHub;
                default:
                    return DeviceFormFactorType.Other;
            }
        }

        // Calculates the color of a point in a rectangle that is filled
        // with a LinearGradientBrush.
        public static Color GetColorAtPoint(Grid grid, UIPoint thePoint)
        {
            // Get properties
            LinearGradientBrush br = (LinearGradientBrush)grid.Background;

            double y3 = thePoint.Y;
            double x3 = thePoint.X;

            double x1 = br.StartPoint.X * grid.ActualWidth;
            double y1 = br.StartPoint.Y * grid.ActualHeight;
            Windows.Foundation.Point p1 = new Windows.Foundation.Point(x1, y1); // Starting point

            double x2 = br.EndPoint.X * grid.ActualWidth;
            double y2 = br.EndPoint.Y * grid.ActualHeight;
            Windows.Foundation.Point p2 = new Windows.Foundation.Point(x2, y2);  // End point

            // Calculate intersecting points
            Windows.Foundation.Point p4 = new Windows.Foundation.Point(); // with tangent

            if (y1 == y2) // Horizontal case
            {
                p4 = new Windows.Foundation.Point(x3, y1);
            }

            else if (x1 == x2) // Vertical case
            {
                p4 = new Windows.Foundation.Point(x1, y3);
            }

            else // Diagnonal case
            {
                double m = (y2 - y1) / (x2 - x1);
                double m2 = -1 / m;
                double b = y1 - m * x1;
                double c = y3 - m2 * x3;

                double x4 = (c - b) / (m - m2);
                double y4 = m * x4 + b;
                p4 = new Windows.Foundation.Point(x4, y4);
            }

            // Calculate distances relative to the vector start
            double d4 = dist(p4, p1, p2);
            double d2 = dist(p2, p1, p2);

            double x = d4 / d2;

            // Clip the input if before or after the max/min offset values
            double max = br.GradientStops.Max(n => n.Offset);
            if (x > max)
            {
                x = max;
            }
            double min = br.GradientStops.Min(n => n.Offset);
            if (x < min)
            {
                x = min;
            }

            // Find gradient stops that surround the input value
            GradientStop gs0 = br.GradientStops.Where(n => n.Offset <= x).OrderBy(n => n.Offset).Last();
            GradientStop gs1 = br.GradientStops.Where(n => n.Offset >= x).OrderBy(n => n.Offset).First();

            float y = 0f;
            if (gs0.Offset != gs1.Offset)
            {
                y = (float)((x - gs0.Offset) / (gs1.Offset - gs0.Offset));
            }

            // Interpolate color channels
            Color cx = new Color();
            if (br.ColorInterpolationMode == ColorInterpolationMode.ScRgbLinearInterpolation)
            {
                byte aVal = (byte)((gs1.Color.A - gs0.Color.A) * y + gs0.Color.A);
                byte rVal = (byte)((gs1.Color.R - gs0.Color.R) * y + gs0.Color.R);
                byte gVal = (byte)((gs1.Color.G - gs0.Color.G) * y + gs0.Color.G);
                byte bVal = (byte)((gs1.Color.B - gs0.Color.B) * y + gs0.Color.B);
                cx = Color.FromArgb(aVal, rVal, gVal, bVal);
            }
            else
            {
                byte aVal = (byte)((gs1.Color.A - gs0.Color.A) * y + gs0.Color.A);
                byte rVal = (byte)((gs1.Color.R - gs0.Color.R) * y + gs0.Color.R);
                byte gVal = (byte)((gs1.Color.G - gs0.Color.G) * y + gs0.Color.G);
                byte bVal = (byte)((gs1.Color.B - gs0.Color.B) * y + gs0.Color.B);
                cx = Color.FromArgb(aVal, rVal, gVal, bVal);
            }
            return cx;
        }

        private static double dist(UIPoint px, UIPoint po, UIPoint pf)
        {
            double d = Math.Sqrt((px.Y - po.Y) * (px.Y - po.Y) + (px.X - po.X) * (px.X - po.X));
            if (((px.Y < po.Y) && (pf.Y > po.Y)) ||
                ((px.Y > po.Y) && (pf.Y < po.Y)) ||
                ((px.Y == po.Y) && (px.X < po.X) && (pf.X > po.X)) ||
                ((px.Y == po.Y) && (px.X > po.X) && (pf.X < po.X)))
            {
                d = -d;
            }
            return d;
        }

        public static int GetLuminance(Color color)
        {
            return (int)Math.Sqrt(0.241 * color.R * color.R + 0.691 * color.G * color.G + 0.068 * color.B * color.B);
        }

        public static string ShortenFileName(string fileName)
        {
            string ext = Path.GetExtension(fileName);
            string fn = Path.GetFileNameWithoutExtension(fileName);
            if (!string.IsNullOrWhiteSpace(fn) && fn.Length > 1)
            {
                fn = fn.Substring(0, fn.Length / 2);
                if (!string.IsNullOrWhiteSpace(ext))
                {
                    fn += ext;
                    return fn;
                }
            }
            return fileName;
        }

        public static bool IsOneDrive(string path)
        {
            return _SkyDriveRegex.IsMatch(path) || _OneDriveRegex.IsMatch(path);
        }

        public static bool IsOneDriveTempFile(string path)
        {
            return _OneDriveTempFolder.IsMatch(path) || _SkyDriveTempFolder.IsMatch(path);
        }

        public static bool IsBox(string path)
        {
            return _BoxRegex.IsMatch(path);
        }

        public static bool IsDropBox(string path)
        {
            return string.IsNullOrEmpty(path) || _DropBoxRegex.IsMatch(path);
        }

        public static string GetPagePresentationModeName(PDFViewCtrlPagePresentationMode mode)
        {
            string modeName = "Unknown Presentation Mode";
            switch (mode)
            {
                case PDFViewCtrlPagePresentationMode.e_single_continuous:
                    modeName = "Continuous";
                    break;
                case PDFViewCtrlPagePresentationMode.e_single_page:
                    modeName = "Single Page";
                    break;
                case PDFViewCtrlPagePresentationMode.e_facing:
                    modeName = "Facing";
                    break;
                case PDFViewCtrlPagePresentationMode.e_facing_cover:
                    modeName = "Facing Cover";
                    break;
            }
            return modeName;
        }

        public static async Task SetFullScreenModeAsync(bool fullScreen)
        {
            if (GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                StatusBar statusBar = StatusBar.GetForCurrentView();
                if (fullScreen)
                {
                    await statusBar.HideAsync();
                }
                else
                {
                    await statusBar.ShowAsync();
                }
            }
            else
            {
                if (fullScreen)
                {
                    if (!ApplicationView.GetForCurrentView().TryEnterFullScreenMode())
                    {
                        Debug.WriteLine("Failed to set fullscreen mode");
                    }
                }
                else
                {
                    ApplicationView.GetForCurrentView().ExitFullScreenMode();
                }
            }
        }

        public static async void SetFullScreenModeDontWait(bool fullscreen) => await SetFullScreenModeAsync(fullscreen);

        public static bool IsFullScreen()
        {
            if (GetDeviceFormFactorType() == DeviceFormFactorType.Phone)
            {
                return ViewerSettings.PhoneFullScreen == 1;
            }

            return ApplicationView.GetForCurrentView().IsFullScreenMode;
        }

        /// <summary>
        /// Removes invalid characters from a file name. Will return temp.ext if there are no valid characters left.
        /// </summary>
        /// <param name="filename"></param>
        /// <returns></returns>
        public static string SanitizeFileName(string filename)
        {
            char[] invalidChars = Path.GetInvalidFileNameChars();
            string invalidString = Regex.Escape(new string(invalidChars));
            string valid = Regex.Replace(filename, "[" + invalidString + "]", "");
            string extension = Path.GetExtension(valid);
            string name = Path.GetFileNameWithoutExtension(valid);
            if (string.IsNullOrWhiteSpace(name))
                name = "temp";

            return name + extension;
        }

        /// <summary>
        /// Create a new temp file in the local app's folder to be accessed with full permission
        /// </summary>
        /// <remarks>
        /// This is usually used when the original/source file is Read Only
        /// </remarks>
        /// <param name="newName">File name</param>
        /// <param name="cleanTempfolder">Choose whether or not we delete temp files that are no longer used by ViewerControl</param>
        /// <returns>Temp file created</returns>
        public static async Task<StorageFile> CreateTemporaryAccessFileAsync(string newName, bool cleanTempfolder = true)
        {
            StorageFolder folder = await GetTemporaryAccessFilefolderAsync(cleanTempfolder);

            if (folder != null)
            {
                StorageFile file = null;
                if (file == null)
                {
                    try
                    {
                        file = await folder.CreateFileAsync(newName, CreationCollisionOption.GenerateUniqueName);
                    }
                    catch
                    {
                        //swallow exception
                    }
                }
                return file;
            }
            return null;
        }

        /// <summary>
        /// Get the temporary folder used to save temporary access files
        /// </summary>
        /// <param name="cleanTempfolder">Choose whether or not we delete temp files that are no longer used by ViewerControl</param>
        /// <returns>Temp folder</returns>
        public static async Task<StorageFolder> GetTemporaryAccessFilefolderAsync(bool cleanTempfolder = true)
        {
            StorageFolder folder = null;
            try
            {
                folder = await ApplicationData.Current.LocalFolder.CreateFolderAsync(
                    "TemporaryPDFS",
                    CreationCollisionOption.OpenIfExists);
            }
            catch (Exception) { }

            if (folder == null)
            {
                try
                {
                    folder = await ApplicationData.Current.LocalFolder.CreateFolderAsync(
                        "TemporaryPDFS",
                        CreationCollisionOption.GenerateUniqueName);
                }
                catch (Exception) { }
            }

            if (folder != null && cleanTempfolder && TabControlViewModel.IsReady)
            {
                try
                {
                    IReadOnlyList<StorageFile> files = await folder.GetFilesAsync();
                    foreach (StorageFile file in files)
                    {
                        if (!TabControlViewModel.Instance.ContainsFile(file))
                        {
                            await file.DeleteAsync();
                        }
                    }
                }
                catch (Exception) { }
            }

            return folder;
        }

        /// <summary>
        /// Remove ReadOnly and Temporary flags from the StorageFile properties
        /// </summary>
        /// <param name="file">The file to be processed</param>
        public static async Task RemoveReadOnlyAndTemporaryFlagsAsync(StorageFile file)
        {
            try
            {
                const string fileAttributeString = "System.FileAttributes";
                List<string> props = new List<string>
                {
                    fileAttributeString
                };

                IDictionary<string, object> propertyDictionary = await file.Properties.RetrievePropertiesAsync(props);
                if (propertyDictionary.ContainsKey(fileAttributeString))
                {
                    uint properties = (uint)propertyDictionary[props[0]];
                    const uint temporaryProp = (uint)Windows.Storage.FileAttributes.Temporary;
                    const uint nonTemporaryMark = 0xFFFFFFFF ^ temporaryProp;
                    const uint readOnlyProp = (uint)Windows.Storage.FileAttributes.ReadOnly;
                    const uint nonReadOnlyMask = 0xFFFFFFFF ^ readOnlyProp;
                    properties &= nonTemporaryMark;
                    properties &= nonReadOnlyMask;
                    propertyDictionary[fileAttributeString] = properties;
                    await file.Properties.SavePropertiesAsync(propertyDictionary);
                }
            }
            catch (Exception ex)
            {
                AnalyticsHandlerBase.CURRENT.SendException(ex);
            }
        }

        public static async Task<Tuple<DateTimeOffset, ulong>> GetDateModifiedAndSizeAsync(StorageFile file)
        {
            if (file == null)
                return new Tuple<DateTimeOffset, ulong>(DateTimeOffset.MinValue, 0);

            BasicProperties props = await file.GetBasicPropertiesAsync();
            return new Tuple<DateTimeOffset, ulong>(props.DateModified, props.Size);
        }

        /// <summary>
        /// Check if file is in the recent list of accessed documents
        /// </summary>
        /// <param name="file">The file to be checked</param>
        /// <returns>Returns true if we can add the file to the recent list safely.</returns>
        public static bool DoesFileBelongInRecentList(StorageFile file)
        {
            if (file == null)
            {
                return false;
            }
            if (string.IsNullOrWhiteSpace(file.Path))
            {
                return false;
            }
            else
            {
                string pathy = file.Path;
                bool isDropBox = IsDropBox(pathy);
                bool isBox = IsBox(pathy);
                bool isOneDrive = IsOneDrive(pathy);
                if (IsRemoteStorage(file.Path) && !IsOneDrive(file.Path))
                {
                    return false;
                }
            }
            return true;
        }

        public static bool IsRemoteStorage(string path)
        {
            return IsOneDrive(path) || IsBox(path) || IsDropBox(path);
        }

        public static bool MatchesColorPostProcessMode(PDFViewCtrl ctrl, PDFRasterizerColorPostProcessMode mode, Color newWhite, Color newblack)
        {
            if (ctrl.GetColorPostProcessMode() != mode)
            {
                return false;
            }
            else if (mode != PDFRasterizerColorPostProcessMode.e_postprocess_gradient_map)
            {
                return true;
            }

            ColorPt white = new ColorPt(1, 1, 1);
            ColorPt processedWhite = ctrl.GetPostProcessedColor(white);
            ColorPt black = new ColorPt(0, 0, 0);
            ColorPt processedBlack = ctrl.GetPostProcessedColor(black);

            Color currentWhite = Color.FromArgb(255, (byte)(255 * processedWhite.Get(0)), (byte)(255 * processedWhite.Get(1)), (byte)(255 * processedWhite.Get(2)));
            Color currentBlack = Color.FromArgb(255, (byte)(255 * processedBlack.Get(0)), (byte)(255 * processedBlack.Get(1)), (byte)(255 * processedBlack.Get(2)));

            int rwDiff = Math.Abs((int)newWhite.R - (int)currentWhite.R);
            int gwDiff = Math.Abs((int)newWhite.G - (int)currentWhite.G);
            int bwDiff = Math.Abs((int)newWhite.B - (int)currentWhite.B);

            int rbDiff = Math.Abs((int)newblack.R - (int)currentBlack.R);
            int gbDiff = Math.Abs((int)newblack.G - (int)currentBlack.G);
            int bbDiff = Math.Abs((int)newblack.B - (int)currentBlack.B);

            return rwDiff < 2 && gwDiff < 2 && bwDiff < 2 && rbDiff < 2 && gbDiff < 2 && bbDiff < 2;
        }

        public static Color GetViewerBackgroundColor(Color pageColor)
        {
            HSV hsv = RGBtoHSV(new RGB(pageColor.R, pageColor.G, pageColor.B));
            double lowEarthHue = 0.05;
            double hightEarthHue = 0.11;
            bool earthTones = hsv.H >= lowEarthHue && hsv.H <= hightEarthHue;
            if (hsv.V > 0.5)
            {
                if (earthTones)
                {
                    hsv.V -= 0.2;
                    hsv.S = Math.Min(hsv.S * 2, Math.Min(hsv.S + 0.05, 1.0));
                }
                else
                {
                    hsv.V *= 0.6;
                }
            }
            else if (hsv.V >= 0.3)
            {
                hsv.V = (hsv.V / 2) + 0.05;
            }
            else if (hsv.V >= 0.1)
            {
                hsv.V -= 0.1;
            }
            else
            {
                hsv.V += 0.1;
            }
            if (!earthTones)
            {
                double dist = Math.Min(0.05, lowEarthHue - hsv.H);
                if (hsv.H > hightEarthHue)
                {
                    dist = Math.Min(0.05, hsv.H - hightEarthHue);
                }
                hsv.S = hsv.S - (hsv.S * (20 * dist) * 0.6);
            }

            RGB rgb = HSVtoRGB(hsv);
            return Color.FromArgb(255, (byte)rgb.R, (byte)rgb.G, (byte)rgb.B);
        }

        /// <summary>
        /// Check if the value of an "enum instance" contains more than one flag (enum value)
        /// </summary>
        /// <typeparam name="TValue">Enum class</typeparam>
        /// <param name="flag">Enum instance value</param>
        /// <returns>True if more than one flag, false otherwise</returns>
        public static bool MoreThanOneFlag<TValue>(TValue flag) where TValue : Enum => (System.Convert.ToInt32(flag) & (System.Convert.ToInt32(flag) - 1)) != 0;

        /// <summary>
        /// Check if a audio capture device (Microphone) is available / installed on the computer
        /// </summary>
        /// <returns>True if microphone is available, false otherwise</returns>
        public static bool IsAudioCaptureAvailable()
        {
            string audioId = string.Empty;
            try
            {
                audioId = Windows.Media.Devices.MediaDevice.GetDefaultAudioCaptureId(Windows.Media.Devices.AudioDeviceRole.Default);
            }
            catch 
            {
                return false;
            }

            if (string.IsNullOrEmpty(audioId))
                return false;
            else
                return true;
        }

        /// <summary>
        /// Display a toast notification on the bottom right to let 
        /// users know about Xodo activities
        /// </summary>
        /// <param name="title">the title</param>
        /// <param name="content">the content text</param>
        /// <param name="expirationTime">how long the toast should be displayed in seconds</param>
        public static void ShowToastNotification(string title, string content, string extraInfo,
            uint expirationTime = 5,
            Windows.Foundation.TypedEventHandler<ToastNotification, object> typedEvent = null)
        {
            if (string.IsNullOrEmpty(title) || string.IsNullOrEmpty(content))
                return;

            if (expirationTime == 0)
                expirationTime = MIN_TIME_TOAST_NOTIFICATION;
            else if (expirationTime > MAX_TIME_TOAST_NOTIFICATION)
                expirationTime = MAX_TIME_TOAST_NOTIFICATION;

            try
            {
                ToastNotifier toastNotifier = ToastNotificationManager.CreateToastNotifier();
                Windows.Data.Xml.Dom.XmlDocument toastXml = ToastNotificationManager.GetTemplateContent(ToastTemplateType.ToastText02);
                Windows.Data.Xml.Dom.XmlNodeList toastNodeList = toastXml.GetElementsByTagName("text");
                toastNodeList.Item(0).AppendChild(toastXml.CreateTextNode(title));
                toastNodeList.Item(1).AppendChild(toastXml.CreateTextNode(content));
                Windows.Data.Xml.Dom.IXmlNode toastNode = toastXml.SelectSingleNode("/toast");
                Windows.Data.Xml.Dom.XmlElement audio = toastXml.CreateElement("audio");
                audio.SetAttribute("src", "ms-winsoundevent:Notification.SMS");

                ToastNotification toast = new ToastNotification(toastXml);
                toast.Tag = extraInfo;

                if (typedEvent != null)
                    toast.Activated += typedEvent;

                toast.ExpirationTime = DateTime.Now.AddSeconds(expirationTime);
                toastNotifier.Show(toast);
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        /// <summary>
        /// Convert a MemoryStream to IRandomAccessStream with cloning support
        /// </summary>
        /// <param name="memoryStream">The memory stream</param>
        /// <returns>IRandomAccessStream with cloning support</returns>
        public static async Task<IRandomAccessStream> ConvertToRandomAccessStreamAsync(MemoryStream memoryStream)
        {
            if (memoryStream == null || memoryStream.Length == 0)
                return null;

            var randomAccessStream = new InMemoryRandomAccessStream();
            var outputStream = randomAccessStream.GetOutputStreamAt(0);

            var dw = new DataWriter(outputStream);
            dw.WriteBytes(memoryStream.ToArray());
            await dw.StoreAsync();

            await outputStream.FlushAsync();

            return randomAccessStream;
        }

        /// <summary>
        /// Helper method to swap the byte array for the audio stream since PDFTron SDK
        /// requires it to be Big Endian
        /// </summary>
        /// <returns></returns>
        public static byte[] SwapByteArray(byte[] a)
        {
            if (a == null)
                return null;

            byte[] ret = new byte[a.Length];
            // if array is odd we set limit to a.length - 1.
            int limit = a.Length - (a.Length % 2);
            if (limit < 1) return null;
            for (int i = 0; i < limit - 1; i = i + 2)
            {
                ret[i] = a[i + 1];
                ret[i + 1] = a[i];
            }
            return ret;
        }

        /// <summary>
        /// Run a Dispatchedhandler async calling the Corewindow Dispatcher
        /// </summary>
        /// <remarks>
        /// This is mainly used in an attempt to avoid <code>CoreApplicationView.get_CoreWindow()</code> exception
        /// </remarks>
        /// <param name="coreWindow">The CoreWindow which will run the Func using its Dispatcher</param>
        /// <param name="agileCallback">DispatchedHandler to be run</param>
        public static async Task RunAsyncSaferAsync(CoreWindow coreWindow, DispatchedHandler agileCallback)
        {
            try
            {
                await coreWindow.Dispatcher.RunAsync(CoreDispatcherPriority.Normal, agileCallback);
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }

        /// <summary>
        /// Run a Func async calling the Corewindow Dispatcher
        /// </summary>
        /// <remarks>
        /// This is mainly used in an attempt to avoid <code>CoreApplicationView.get_CoreWindow()</code> exception
        /// </remarks>
        /// <param name="coreWindow">The CoreWindow which will run the Func using its Dispatcher</param>
        /// <param name="T">Func to be run</param>
        public static async Task RunAsyncSaferAsync(CoreWindow coreWindow, Func<Task> T)
        {
            try
            {
                await coreWindow.Dispatcher.RunTaskAsync(T);
            }
            catch (Exception ex)
            {
                AppCenterHandlerBase.CURRENT.TrackError(ex);
            }
        }
    }
}
